/* $Id: basic_op.c,v 1.7 2002/07/26 11:34:39 peteri Exp $ */

#include <assert.h>
#include "e_lib.h"

#include "basic_op.h"

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   ExchangeVertex 
 * DESCRIPTION
 *   Exchanges Vertex in tetrahedron
 * ARGUMENTS
 *   tet  -- tetrahedron 
 *   oldVertex -- vertex to be changed in tet
 *   newVertex -- new vertex to be set in tet
 * RETURN VALUE
 *   NONE
 ******/
void ExchangeVertex (Tetrah *tet, Vertex *oldVertex, Vertex *newVertex) 
{ 
  int old_in_tet;
  old_in_tet = WhichVertex (tet, oldVertex);
  Tos_MStackRemove (oldVertex->adjElem, (void *)tet);
  TET_VERTEX(tet, old_in_tet) = newVertex;
  Tos_MStackPush (newVertex->adjElem, (void *)tet);  
} /* ExchangeVertex */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   CompareFaces 
 * DESCRIPTION
 *   Check if given faces has the same vertices.
 * ARGUMENTS
 *   f1  - pointer to face
 *   f2  - pointer to face
 * RETURN VALUE
 *   1 if faces has the same vertices 0 otherwise. 
 ******/
int CompareFaces (Face *f1, Face *f2)
{
  assert (f1 != NULL && f2 != NULL);

  if (FACE_VERTEX(f1,0) == FACE_VERTEX(f2,0))
  {
    if (FACE_VERTEX(f1,1) == FACE_VERTEX(f2,1))
    {
       if (FACE_VERTEX(f1,2) == FACE_VERTEX(f2,2))
         return 1;
    }
    else if (FACE_VERTEX(f1,1) == FACE_VERTEX(f2,2))
    {
      if (FACE_VERTEX(f1,2) == FACE_VERTEX(f2,1))
         return 1;
    } 
  }
  else if (FACE_VERTEX(f1,0) == FACE_VERTEX(f2,1))
  {
    if (FACE_VERTEX(f1,1) == FACE_VERTEX(f2,0))
    {
       if (FACE_VERTEX(f1,2) == FACE_VERTEX(f2,2))
         return 1;
    }
    else if (FACE_VERTEX(f1,1) == FACE_VERTEX(f2,2))
    {
      if (FACE_VERTEX(f1,2) == FACE_VERTEX(f2,0))
         return 1;
    } 
  }
  else if (FACE_VERTEX(f1,0) == FACE_VERTEX(f2,2))
  {
    if (FACE_VERTEX(f1,1) == FACE_VERTEX(f2,0))
    {
       if (FACE_VERTEX(f1,2) == FACE_VERTEX(f2,1))
         return 1;
    }
    else if (FACE_VERTEX(f1,1) == FACE_VERTEX(f2,1))
    {
      if (FACE_VERTEX(f1,2) == FACE_VERTEX(f2,0))
         return 1;
    } 
  }   
 return 0;
} /* end of CompareFaces */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   GetFourthVertex 
 * DESCRIPTION
 *   Givet three vertices of tetrhedron returns the fourth one.
 * ARGUMENTS
 *   tet - pointer to tetrahedron
 *   v1, v2, v3 - poiners to tetrahdron vertices
 * RETURN VALUE
 *   Pointer to tetrahedron vertex.
 ******/
Vertex *GetFourthVertex (Tetrah *tet, Vertex *v1, Vertex *v2, Vertex *v3)
{
  int i;
  Vertex *v;
  int found=-1;
  for (i=0; i<4; i++)
  {
    v = TET_VERTEX(tet, i);
    if (v != v1 && v != v2 && v != v3)
    {
      assert (found == -1); /* to make sure that v1,v2,v3 are in tet */
      found = i;
    }
  }
  return TET_VERTEX(tet, found);
} /* end of GetFourthVertex */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   GetNeighbourOpositeVertex 
 * DESCRIPTION
 *   Retruns neighbour of a tetrahedron oposite given vetex.
 * ARGUMENTS
 *   tet  - pointer to tetrahedron
 *   v    - pointer to vertex
 * RETURN VALUE
 *   Pointer to neighbour of tetrahedron
 ******/
Tetrah * GetNeighbourOpositeVertex (Tetrah *tet, Vertex *v)
{
  int n;
  n = WhichVertex (tet, v);
  return TET_NEIGHBOUR(tet, n);
} /* end of GetNeighbourOpositeVertex */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   TetGenericMapping 
 * DESCRIPTION
 *   This mapping allows to find end nodes of edge which is shared
 *   by two faces of tetrahedron or two faces adjacent to edge connecting
 *   two given nodes of the tetrahedron. 
 * ARGUMENTS
 *       i1, i2  - input number pair (IN)
 *       o1, o2  - output number pair (OUT)
 * RETURN VALUE
 *  NONE 
 ******/
void TetGenericMapping (int i1, int i2, int *o1, int *o2)
{
  register int i;
  register int k=-1;
  /* This trick is to have a bit faster loop, equivalent to:
   * for (i=0; i<4; i++)
   */
  for (i=4; i--; )
  {
    if (i1 != i && i2 != i)
    {
      k = i;
      break;
    }
  }
  assert(k > -1);
  *o1 = k;
  *o2 = 6 - i1 - i2 - (*o1);

  return;
} /* end of TetGenericMapping */ 

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   TetHasFace 
 * DESCRIPTION
 *   Check if given face is one of the tetrahedron face.
 * ARGUMENTS
 *       i1, i2  - input number pair (IN)
 *       o1, o2  - output number pair (OUT)
 * RETURN VALUE
 *       If face 'f' is one of tetrahedron face retruns 1 and
 *       set 'facenum' to face generic number. Otherwise retruns 0 
 *       and facenum is not defined.
 ******/
int TetHasFace (Tetrah *tet, Face *f, int *facenum)
{
 register int face; 
 register int count=0; /* counter of vertices shared by face and tetrahedron */
 
 *facenum = 6;
 /* This trick make loop e bit faster, equivalent to:
  * for (face=0; face<4; face++)
  */
 for (face=4; face--; )
 {
   /* Checking face vertex 0 */
   if (TET_VERTEX(tet, face) == FACE_VERTEX (f, 0))
   {
     count++;
     (*facenum) -= face;
   }
   /* Checking face vertex 1 */
   if (TET_VERTEX(tet, face) == FACE_VERTEX (f, 1))
   {
     count++;
     (*facenum) -= face;
   }
   /* Checking face vertex 2 */
   if (TET_VERTEX(tet, face) == FACE_VERTEX (f, 2))
   {
     count++;
     (*facenum) -= face;
   }
 } 

 if (count == 3)   /* face has been found */
 {
   return 1; 
 }

 return 0;
} /* end of TetHasFace */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   TetHasVertex 
 * DESCRIPTION
 *   Check if given vertex is one of the tetrahedron vertices.
 * ARGUMENTS
 *   tet - pointer to tetrah (IN)
 *   v   - pointer to vertex (IN)
 *   vertexnum - vertex number in tetrah (OUT)
 * RETURN VALUE
 *   If vertex 'v' is one of tetrahedron vertices retrun 1 and
 *   set 'vertexnum' to vertex generic number. 
 *   Otherwise retrun 0 and vertexnum is not defined.
 ******/
int TetHasVertex (Tetrah *tet, Vertex *v, int *vertexnum)
{
 int i;
 int result = 0;

 assert(tet != NULL);
 assert(v != NULL);

 for (i=4; i--; )
 {
   if (TET_VERTEX(tet, i) == v)
   {
     (*vertexnum) = i;
     result = 1;
     break;
   }
 } 
 return result;
} /* end of TetHasVertex */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   FaceHasEdge 
 * DESCRIPTION
 *       Check if given edge is one of the edges of the face.
 * ARGUMENTS
 * RETURN VALUE
 *  If edge 'e' belongs to the face return 1 and 
 *  set 'sidenum to side generic number. Otherwise return 0
 *  and 'sidenum' is not defined.
 ******/
int FaceHasEdge (Face *f, Edge *e, int *sidenum)
{
 register int count=0; /* counter of vertices shared by face and edge */
 register int side;
 
 *sidenum = 3;
 /* This is to speed up the loop a litle and is the same as:
  *     for (side=0; side<3; side++)
  */  
 for (side=3; side--; )
 {
   if (FACE_VERTEX(f, side) == EDGE_VERTEX (e, 0))
   {
     count++;
     (*sidenum) -= side;
   }
   if (FACE_VERTEX(f, side) == EDGE_VERTEX (e, 1))
   {
     count++;
     (*sidenum) -= side;
   }
 } 
 if (count == 2) return 1; 

 return 0;
} /* end of FaceHasEdge */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   TetHasNeighbour 
 * DESCRIPTION
 *   Check if a neighbour is realy an neighbour of tet.
 * ARGUMENTS
 *   tet - pointer to tetrahedron (IN)
 *   neighbour - pointer to tetrahedron neighbour (IN)
 *   neighbnum - number on neighbour in tet (OUT)
 * RETURN VALUE
 *   If 'tet' and 'neighbour' are realy neighbours return 1
 *   and set neighbnum to generic number of neigbour in tet
 *   otherwise return 0 and neighbnum is set to -999
 ******/
int TetHasNeighbour (Tetrah *tet, Tetrah *neighbour, int *neighbnum)
{
  int i;
  for (i=0; i<4; i++)
  {
    if (TET_NEIGHBOUR (tet, i) == neighbour) 
    {
      *neighbnum = i; 
      return 1;
    }
  }
  *neighbnum = -999;
  return 0;
} /* end of TetHasNeighbour */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   WhichNeighbour 
 * DESCRIPTION
 *   Find generic number of 'neighbour' in 'tet'
 *  (the same time this is the number of face in 'tet' adjacent to
 *  'neighbour')
 * ARGUMENTS
 *   tet - pointer to tetrahedron (IN)
 *   neighbour - pointer to tetrahedron neighbour (IN)
 * RETURN VALUE
 *   If 'tet' and 'neighbour' are realy neighboure return value 0-3
 *   otherwise raise error message.
 ******/
int WhichNeighbour (Tetrah *tet, Tetrah *neighbour)
{
  if (TET_NEIGHBOUR (tet, 0) == neighbour) return 0; 
  if (TET_NEIGHBOUR (tet, 1) == neighbour) return 1; 
  if (TET_NEIGHBOUR (tet, 2) == neighbour) return 2; 
  if (TET_NEIGHBOUR (tet, 3) == neighbour) return 3; 

  e_GeneralError ("WhichNeighbour : Therahedra %d %d aren't neighbours", 
                  tet->id, neighbour->id); 
  return -1;
} /* end of WhichNeighbour */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   WhichVertex 
 * DESCRIPTION
 *   Find generic number of 'vertex' in 'tet'
 * ARGUMENTS
 *   tet - pointer to tetrah
 *   vetex - pointer to vertex
 * RETURN VALUE
 *   If 'vertex' is a real vertex of tet return value 0-3
 *   otherwise raise error message.
 ******/
int WhichVertex (Tetrah *tet, Vertex *vertex)
{
  if (TET_VERTEX (tet, 0) == vertex)
  {
    return 0; 
  }
  else if (TET_VERTEX (tet, 1) == vertex) 
  {
    return 1; 
  }
  else if (TET_VERTEX (tet, 2) == vertex) 
  {
    return 2; 
  }
  else if (TET_VERTEX (tet, 3) == vertex)
  {
    return 3; 
  }

  e_GeneralError ("WhichVertex : Vertex %d is not int tetrahedron %d", 
                  vertex->id, tet->id); 
  return -1;
} /* end of WhichVertex */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   IsEdgeOnStack 
 * DESCRIPTION
 *   Checks if edge with given end points is on the stack.
 * ARGUMENTS
 *   edgeStack - stack of pointers to edges
 *   v1 - pointer to first vertex of edge
 *   v2 - pointer to second vertex of edge
 * RETURN VALUE
 *   1 if edge is on the stack oterwise 0
 ******/
int IsEdgeOnStack (Tos_MStack edgeStack, Vertex *v1, Vertex *v2)
{
  Edge *edge;
  /* Prepare to traverse edges stack */
  Tos_MStackTraversalInit (edgeStack); 
  while ((edge = (Edge *) Tos_MStackTraverse (edgeStack)) != NULL)
  {
    if (   (EDGE_VERTEX(edge, 0) == v1 && EDGE_VERTEX(edge, 1) == v2)
        || (EDGE_VERTEX(edge, 0) == v2 && EDGE_VERTEX (edge, 1) == v1))
    {
      return 1;
    }
  } 
  return 0;
} /* end of IsEdgeOnStack */

/* Modifications:
 * $Log: basic_op.c,v $
 * Revision 1.7  2002/07/26 11:34:39  peteri
 * minor editing
 *
 * Revision 1.6  2002/02/24 19:00:57  romek
 * *** empty log message ***
 *
 * Revision 1.5  2002/02/16 18:55:12  romek
 * Doc fix.
 *
 * Revision 1.4  2002/02/16 14:03:14  romek
 * Doc fix.
 *
 */
