/* $Id: check.c,v 1.13 2002/07/26 18:53:30 peteri Exp $ */

#include <assert.h>
#include "eg_lib.h"
#include "find.h"
#include "basic_op.h"
#include "convert.h"
#include "check.h"
#include "boundary.h"
#include "debug.h"

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   CheckCoplanarity
 * DESCRIPTION
 *   Examine all tetrahedra in the mesh and check if their 
 *   vertices aren't coplanar. If any tetrahedra has four
 *   coplanar vertices a message is printed.
 *   This function is used only for debuging.
 * ARGUMENTS
 *   tetMesh - pointer to tetrahedral mesh data structure;
 * RETURN VALUE
 *   0 - if all elements are non-degenerate
 *   otherwise number of degenerate elements.     
 ******/
int CheckCoplanarity (TetrahMesh *tetMesh)
{
  Tetrah *tet;
  double volCheck;  
  double volEps = 1.0e-14; 
  int degenerate = 0; /* counter of degenerate elements */

  tet = LIST_TAIL (tetMesh->tList);
  while (tet  != NULL)
  {
    volCheck = eg_Orient3D (TET_VCOORDS (tet, 0), 
                            TET_VCOORDS (tet, 1),
                            TET_VCOORDS (tet, 2),
                            TET_VCOORDS (tet, 3)); 
    if (E_ABS (volCheck) < volEps)
    {
      degenerate++;
      fprintf (stderr, "Degenerate element (%d %d %d %d)\n", 
                     TET_VERTEX (tet, 0)->id,  
                     TET_VERTEX (tet, 1)->id,  
                     TET_VERTEX (tet, 2)->id,  
                     TET_VERTEX (tet, 3)->id);
      fprintf (stderr, "\t\tVertices:\n");
      fprintf (stderr, "\t\t\t%f %f %f\n", 
                      TET_VCOORDS (tet, 0)[0], 
                      TET_VCOORDS (tet, 0)[1], 
                      TET_VCOORDS (tet, 0)[2]); 
      fprintf (stderr, "\t\t\t%f %f %f\n", 
                      TET_VCOORDS (tet, 1)[0], 
                      TET_VCOORDS (tet, 1)[1], 
                      TET_VCOORDS (tet, 1)[2]); 
      fprintf (stderr, "\t\t\t%f %f %f\n", 
                      TET_VCOORDS (tet, 2)[0], 
                      TET_VCOORDS (tet, 2)[1], 
                      TET_VCOORDS (tet, 2)[2]); 
      fprintf (stderr, "\t\t\t%f %f %f\n", 
                      TET_VCOORDS (tet, 3)[0], 
                      TET_VCOORDS (tet, 3)[1], 
                      TET_VCOORDS (tet, 3)[2]); 
      fprintf (stderr, "\t\tCheck volume: %16.13f\n", volCheck);
    }
    tet = TET_NEXT (tet);
  }
  return degenerate;
} /* end of CheckCoplanarity */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   CheckMeshConformity 
 * DESCRIPTION
 *   For each element check if meshpoints different than 
 *   its own vertices are located strictly outside
 * ARGUMENTS
 *   tetMesh - pointer to mesh structure
 * RETURN VALUE
 *   0  if mesh is conforming
 *   otherwise number of elements which violate mesh conformity 
 ******/
int CheckMeshConformity (TetrahMesh *tetMesh)
{
  Tetrah *tet; 
  Vertex *p;
  int i;
  int dummy1, dummy2;
  int where; 
  int skip = 0; 
  int invalidTet = 0;  /* counter of tetrahedron violation conformity */
  int tetcount=0;

  tet = LIST_TAIL (tetMesh->tList);
  while (tet != NULL)
  {
    p = LIST_TAIL (tetMesh->pList);
    while (p != NULL)
    {
      for (i=0; i<4; i++)
      {
         if (TET_VERTEX (tet, i) == p)
         {
           skip = 1;
           break;
         }
      }
      if (skip == 1)
      {
        skip = 0; 
        p = VERTEX_NEXT(p);
        continue; 
      }
      else 
      {
        where =  WhereInTetrah (tet, p, &dummy1, &dummy2);  
        switch (where)
        {
          case OUTSIDE: {break;}
          case INSIDE: /* FALLTHRU */
          case ONEDGE: /* FALLTHRU */
          case ONFACE: /* FALLTHRU */
          case ONVERTEX: 
          {              
            invalidTet++; 
          }
          default:
          {
            e_GeneralError ("CheckMeshConformity: "
                            "internal error -- unknown point position %d", 
                            where);
          }
        }
      }
      p = VERTEX_NEXT(p);
    }
    tetcount++;
    tet = TET_NEXT(tet); 
  }
  return invalidTet; 
} /* end of CheckMeshConformity */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   CheckTetrahAdjacency --
 * DESCRIPTION
 *   Examine all tetrahedra in the mesh and check if their 
 *   adjacency relationships are correct.
 * ARGUMENTS
 *   tetMesh - pointer to tetrahedral mesh data structure
 * RETURN VALUE
 *   0 - if all elements have correct adjacency 
 *   otherwise number of invalid elements.     
 ******/
int CheckTetrahAdjacency (TetrahMesh *tetMesh)
{
  register int i;
  register int j;
  register int nfound;
  register int cuboid; 
  register int invElemCount=0;
  Tetrah *tet, *neighbour;
  Vertex *point; 
  Face  dummy_face;
  Face *face;
  Tetrah *faceshell[2];
  int fsc;
  int found;
  
  face = &dummy_face;
  AssignNumbersToTetrah (tetMesh); 
  tet = LIST_TAIL (tetMesh->tList);
  while (tet != NULL)
  {
    cuboid = 0;
    /* check if element has cuboid vertices */
    for (i=0; i<4; i++)
    { 
      point = TET_VERTEX(tet, i); 
      if ((point->flag & POINT_CUBOID) == POINT_CUBOID)
      {
        cuboid = 1; 
        break;
      }    
    }
    for (i=0; i<4; i++)
    {
      found = 0;
      neighbour = TET_NEIGHBOUR(tet, i);
      for (j=0; j<4; j++)
      {
        if (j != i && neighbour != NULL && TET_NEIGHBOUR(tet, j) == neighbour)
        {
          fprintf (stderr, "Duplicated neighbour %d (%d and %d) in %d", 
                   neighbour->id, i, j, tet->id); 
          invElemCount++;
        }
        if (neighbour != NULL && TET_NEIGHBOUR(tet, j) == neighbour) found++;
      }
      assert (found == 1 || found == 0);
    }
    for (i=0; i<4; i++)
    {
      GetFaceFromTet (tet, i, face);
      fsc = FindFaceShell (face, faceshell);
      if (fsc > 2)
      {
        fprintf (stderr,
                 "Too many neighbours at face %d in element %d (%d)\n", 
                 i, TET_NUMBER(tet), tet->id); 
        invElemCount++;
      }
      if (TET_NEIGHBOUR (tet, i) == NULL)
      {
        if (cuboid != 1)
        {
          for (j=0; j<3; j++)
          {
            if ((TET_VFLAG(tet, faceDef[i][j]) & POINT_BOUNDARY) !=
                 POINT_BOUNDARY)
            {
              fprintf (stderr, "Face %d of tetrah %d has null neighbour"
                       " but is not boundary face\n", i, tet->id);
              invElemCount++;
              break;
            }
          }
        }
        continue;
      }
      nfound = 0; 
      for (j=0; j<4; j++)
      {
        if (TET_NEIGHBOUR (TET_NEIGHBOUR(tet, i), j) == tet)
        {
          nfound = 1;
          break;
        }
      }
      if (nfound != 1) 
      {
        fprintf (stderr, "Adjacency %d in element %d (%p) is not symetric\n", 
                 i, TET_NUMBER(tet), (void *) tet), 
        invElemCount++;
      }
    }
    tet = TET_NEXT(tet); 
  }
  return invElemCount;  
} /* end of CheckTetrahAdjacency */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   CheckCavity
 * DESCRIPTION
 *  Check if cavity is properly constructed. 
 *  The check consists of:
 *     - check if all element on cavity stack are unique
 *     - FIXME -- add more checks
 * ARGUMENTS
 *   cavity - cavity stack
 * RETURN VALUE
 *  0 if cavity passes all checks otherwise number of invalid cavity elements
 ******/
int CheckCavity (Tos_MStack cavity, Tos_MStack newCavity)
{
  int counter;
  Tetrah *tet;
  double cavvol, newcavvol, vol;
  
  cavvol = 0;
  Tos_MStackTraversalInit (cavity); 
  while ((tet = Tos_MStackTraverse (cavity)) != NULL)
  {
    counter = Tos_MStackSearchN (cavity, tet);
    if (counter != 1)
    {
       e_GeneralWarning ("CheckCavity: cavity element %d appears %d times", 
                         tet->id, counter);
       return 0;
    }

    vol = eg_TetrahVolume(TET_VERTEX(tet, 0)->coords,
                          TET_VERTEX(tet, 1)->coords,
                          TET_VERTEX(tet, 2)->coords,
                          TET_VERTEX(tet, 3)->coords );
    cavvol += vol;
  }

  newcavvol = 0;
  Tos_MStackTraversalInit (newCavity); 
  while ((tet = Tos_MStackTraverse (newCavity)) != NULL)
  {
    vol = eg_TetrahVolume(TET_VERTEX(tet, 0)->coords,
                          TET_VERTEX(tet, 1)->coords,
                          TET_VERTEX(tet, 2)->coords,
                          TET_VERTEX(tet, 3)->coords );
    newcavvol += vol;
  }

  if(fabs((cavvol - newcavvol) / cavvol) < 0.001)
  {
    return 1;
  }
  else
  {
    fprintf(stderr, "Original volume %f -- new volume %f\n",
            cavvol, newcavvol);
    return 0;
  }
  
  /* return (result); */
} /* end of CheckCavity */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   CheckSurfaceCompletness
 * DESCRIPTION
 *   Check if splited faces add up to original ones. 
 * ARGUMENTS
 *   mgds - mesh generator data structure.
 * RETURN VALUE
 *  0 if cavity passes all checks otherwise number of invalid cavity elements
 ******/
int CheckSurfaceCompletness (MeshGenData *mgds)
{
  int     i, surfNum;
  int     count;
  int     badFaces=0;
  double  origArea, splitArea;
  Face   *face, *splitface;
  struct Tos_MStack faceIter = {0, NULL, NULL};

  for(surfNum = 0; surfNum < mgds->NMeshes; surfNum++)
  {
    for(i = 0; i < mgds->fTableLength[surfNum]; i++)
    {
      face = mgds->fTable[surfNum][i];
      count = Tos_MStackCount(face->splitFaces);
      assert(count == 0 || count > 1);

      if(count)
      {
        origArea = eg_TriangArea3D(FACE_VCOORDS(face,0),
                                   FACE_VCOORDS(face,1),
                                   FACE_VCOORDS(face,2));

        splitArea=0;
        for(splitface = InitSplitFace(face, &faceIter);
            splitface != NULL;
            splitface = NextSplitFace(&faceIter))
        {
          splitArea+= eg_TriangArea3D(FACE_VCOORDS(splitface,0),
                                      FACE_VCOORDS(splitface,1),
                                      FACE_VCOORDS(splitface,2));
        }
        
        if(fabs((origArea - splitArea) / origArea) > 0.001)
        {
          fprintf (stderr, "Discrepancy in area, origArea=%f"
                   " splitArea=%f, face=%d\n", origArea, splitArea, i);
          
          badFaces++;
        }
      }
    }
  }
  
  return badFaces;
} /* end of CheckSurfaceCompletness */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   CheckMeshpointsFlags
 * DESCRIPTION
 *   Count points having different flags set.
 * ARGUMENTS
 *   tetMesh - pointer to tetrahedral data structure.
 * RETURN VALUE
 *   NONE
 ******/
void CheckMeshpointsFlags (TetrahMesh *tetMesh)
{
  Vertex *v;
  int boundary=0;
  int ext_boundary=0;
  int int_boundary=0;
  int on_edge=0;
  int on_face=0;
  int connected=0;
  int cuboid=0;
  v = LIST_TAIL (tetMesh->pList);
  while (v != NULL)
  {
    if ((VERTEX_FLAG(v) & POINT_INT_BOUNDARY) == POINT_INT_BOUNDARY)
        int_boundary++;
    if ((VERTEX_FLAG(v) & POINT_EXT_BOUNDARY) == POINT_EXT_BOUNDARY)
        ext_boundary++;
    if ((VERTEX_FLAG(v) & POINT_BOUNDARY    ) == POINT_BOUNDARY)
        boundary++;
    if ((VERTEX_FLAG(v) & POINT_ON_EDGE     ) == POINT_ON_EDGE)
        on_edge++;
    if ((VERTEX_FLAG(v) & POINT_ON_FACE     ) == POINT_ON_FACE)
        on_face++;
    if ((VERTEX_FLAG(v) & POINT_CUBOID      ) == POINT_CUBOID)
        cuboid++;
    if ((VERTEX_FLAG(v) & POINT_CONNECTED   ) == POINT_CONNECTED)    
        connected++;
    v = VERTEX_NEXT(v); 
   }
   fprintf (stdout, "Vertices: total number %d\n", tetMesh->NMeshPoints);
   fprintf (stdout, "flaged with POINT_BOUNDARY: %d\n", boundary);
   fprintf (stdout, "flaged with POINT_EXT_BOUNDARY: %d\n", ext_boundary);
   fprintf (stdout, "flaged with POINT_INT_BOUNDARY: %d\n", int_boundary);
   fprintf (stdout, "flaged with POINT_ON_EDGE: %d\n", on_edge);
   fprintf (stdout, "flaged with POINT_ON_FACE: %d\n", on_face);
   fprintf (stdout, "flaged with POINT_CUBOID: %d\n", cuboid);
   fprintf (stdout, "flaged with POINT_CONNECTED: %d\n", connected);
} /* end of CheckMeshpointsFlags */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   CheckAlphaParam
 * DESCRIPTION
 *   This function checks that the area of a split face is equal to
 *   the area of the sub-faces which make-up the split face.
 * ARGUMENTS
 *   mgds - mesh geeration data structure pointer
 * RETURN VALUE
 *   Number of elements where the original area is not equal to
 *   the area of split faces
 ******/
int CheckAlphaParam(MeshGenData *mgds)
{
  double req_alpha;
  
  req_alpha = FindAlphaParam(mgds, NULL, NULL);

  if(req_alpha > mgds->mgCtrl->alpha)
  {
    fprintf(stderr, "Required alpha: %f -- current alpha: %f\n",
            req_alpha, mgds->mgCtrl->alpha);
    return 0;
  }
  else
      return 1;
}


/* Modifications:               
 * $Log: check.c,v $
 * Revision 1.13  2002/07/26 18:53:30  peteri
 * - minor editing
 * - updated for new edge, face data structure
 *
 * Revision 1.12  2002/07/26 15:48:39  peteri
 * - checking the alpha parameter of the Delaunay process is
 *   integrated
 *
 * Revision 1.11  2002/07/26 13:06:59  peteri
 * minor modifiactions to remove compiler warnings
 *
 * Revision 1.10  2002/07/26 11:36:47  peteri
 * cavity check is extended, it checks for equal volumes as well
 *
 * Revision 1.9  2002/03/08 13:01:16  peteri
 * only re-formatting of code
 *
 * Revision 1.8  2002/02/27 22:44:47  romek
 * Now include debug.h.
 *
 * Revision 1.7  2002/02/26 12:02:05  romek
 * Bug fix (infinite loop) in CheckMeshConformity.
 *
 * Revision 1.6  2002/02/24 19:01:00  romek
 * *** empty log message ***
 *
 * Revision 1.5  2002/02/16 18:58:42  romek
 * Doc fix.
 *
 */
