/* $Id: convert.c,v 1.4 2002/07/28 13:27:37 peteri Exp $ */

#include <assert.h>
#include "list.h"
#include "basic_op.h"
#include "find.h"
#include "convert.h"

/* integer comparison functions for quick sort */
#if 0
static int IntCompare (const void *a, const void *b)
{
  if (*((int *)a) < *((int *)b)) return(-1);
  if (*((int *)a) > *((int *)b)) return(1);
  return (0);
}
#endif 

/* for descending sorting */
static int IntRevCompare (const void *a, const void *b)
{
  if (*((int *)a) > (*(int *)b)) return(-1);
  if (*((int *)a) < (*(int *)b)) return(1);
  return (0);
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   AssignNumbersToTetrah 
 * DESCRIPTION
 *   Assign numbers (starting from 1) to all tetrahedra in the mesh.
 *   Numbers are assigned starting from the tail of tetrahedra list
 *   (the oldest tetrahedron has number 1).
 * ARGUMENTS
 *   tetMesh - pointer to tetrahedral mesh data structure
 * RETURN VALUE
 *   NONE
 ******/
void AssignNumbersToTetrah (TetrahMesh *tetMesh)
{
  register int i=0;
  Tetrah *tet;
  tet = LIST_TAIL (tetMesh->tList);
  while (tet != NULL)
  {
    i++;
    TET_NUMBER (tet) = i;
    tet = TET_NEXT (tet);
  } 

  /* Check number of tetrahedra */
  assert (i == tetMesh->NTetrah);

} /* end of AssignNumbersToTetrah */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   AssignNumbersToPoints 
 * DESCRIPTION
 *   Assign numbers (starting from 1) to all points in the mesh.
 *   Numbers are assigned starting from the tail of points  list
 *   (the oldest point has the number 1). 
 * ARGUMENTS
 *   tetMesh - poiter to tetrahedral mesh data structure
 * RETURN VALUE
 *   NONE
 ******/
void AssignNumbersToPoints (TetrahMesh *tetMesh)
{
  register int i=0;
  Vertex *point;
  point = LIST_TAIL (tetMesh->pList);
  while (point != NULL)
  {
    i++;
    point->number = i;
    point = VERTEX_NEXT (point); 
  } 

  /* Check number of points */
  assert (i == tetMesh->NMeshPoints);

} /* end of AssignNumbersToPoints */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   TetrahMeshToMESH
 * DESCRIPTION
 *   Transfer TetrahMesh data structure to MESH data structure.
 * ARGUMENTS
 *   tetMesh - pointer to  tetrahedral mesh data structure
 *   mesh  - pointer to MESH data structure 
 * RETURN VALUE
 ******/
void TetrahMeshToMESH (TetrahMesh *tetMesh, MESH *mesh)
{
  register int i;
  Vertex *point;
  Tetrah *tet;

  /* prepare MESH structure */
  e_NullifyMesh (mesh); 

  /* set number of Tetrah1 element in the mesh */
  mesh->NElemsTetrah1 = tetMesh->NTetrah;

  /* set total number of element in mesh */
  mesh->TotalNElems = mesh->NElemsTetrah1; 

  /* set number of nodes and mesh points */
  mesh->NNodes = tetMesh->NMeshPoints;
  mesh->NMeshPoints = tetMesh->NMeshPoints;
 
  /* allocate space for mesh members */
  mesh->MeshPointCoords = 
          e_alloc_cont_d_matr (mesh->NMeshPoints, 3);
  mesh->NodesTetrah1 = 
          e_alloc_cont_i_matr (mesh->NElemsTetrah1, E_NVPE_TETRAH1); 
  mesh->Decomp.SubDoms = e_alloc_i_vect (mesh->TotalNElems); 

  /* set subdomain information */
  mesh->Decomp.SubDomsTetrah1 = mesh->Decomp.SubDoms;
  mesh->Decomp.NSubDoms = 0;

  /* copy mesh coordinates */
  VertexListTraversalInit (tetMesh->pList, TAIL_HEAD);
  i=0;
  while ((point = (Vertex *) VertexListTraverse (tetMesh->pList, TAIL_HEAD)) 
                  != NULL)
  {
    /* check if not going out of allocated space */
    assert (i < tetMesh->NMeshPoints);
    memcpy (mesh->MeshPointCoords[i], point->coords,
                      3*sizeof(double));
    i++;
  }
  
  AssignNumbersToPoints (tetMesh);
  
  /* copy adjacency information */
  TetListTraversalInit (tetMesh->tList, TAIL_HEAD);
  i=0;

  while ((tet = TetListTraverse(tetMesh->tList, TAIL_HEAD)) != NULL)
  {
    /* check if not going out of allocated space */
    assert (i < tetMesh->NTetrah);
    /* fill domain indormation */
    if (mesh->Decomp.NSubDoms < TET_SUBDOMAIN(tet))
    {
      mesh->Decomp.NSubDoms = TET_SUBDOMAIN(tet);
    }

    /* set element vertices */
    mesh->NodesTetrah1[i][0] = TET_VERTEX(tet, 0)->number; 
    mesh->NodesTetrah1[i][1] = TET_VERTEX(tet, 1)->number; 
    mesh->NodesTetrah1[i][2] = TET_VERTEX(tet, 2)->number; 
    mesh->NodesTetrah1[i][3] = TET_VERTEX(tet, 3)->number; 
    i++;
  }

  /* set other required members of MESH structure */
  /* set first net index vector */
  mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] = 1;   

  /* allocate memory for global element indices */
  mesh->GEIs = e_alloc_i_vect (mesh->NElemsTetrah1); 
  /* set GEIs vector */
  for (i=0; i<mesh->NElemsTetrah1; i++)
  {
    mesh->GEIs[i] = i+1;
  }
  mesh->GEIsTetrah1 = mesh->GEIs + mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] -1;

  /* allocate memory for global node indices */
  mesh->GNIs = e_alloc_i_vect (mesh->NNodes);
  /* set GNIs vector */
  for (i=0; i<mesh->NNodes; i++)
  {
    mesh->GNIs[i] = i+1;
  }

  /* allocate memory for title string */
  mesh->Title = e_alloc_c_vect (100);
  /* set mesh title */
  strncpy (mesh->Title, "tetrah mesh", 99);
          
  mesh->KeywordSet[E_KW_NUM_TITLE] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NMESHPOINTS] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NNODES] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1] = TRUE;
  mesh->KeywordSet[E_KW_NUM_MESHPOINT_COORDS] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NODES_TETRAH1] = TRUE;

  mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE;
 
} /* end of TetrahMeshToMESH */


#if 0
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 TetrahMeshToMESH_V1 --
        Transfer TetrahMesh data structure to MESH data structure.
        Additionaly add surface mesh to MESH data structure.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void TetrahMeshToMESH_V1 (MeshGenData *mgds, MESH *mesh)
{
  register int i, k;
  Vertex *point;
  Tetrah *tet;
  Face *face;
  TetrahMesh *tetMesh;

  tetMesh = &(mgds->tetMesh);
  
  /* prepare MESH structure */
  e_NullifyMesh (mesh); 

  /* set number of Tetrah1 element in the mesh */
  mesh->NElemsTetrah1 = tetMesh->NTetrah;

  /* set number of Triang1 elements in the mesh */
  mesh->NElemsTriang1 = mgds->fTableLength[1] + 
                        mgds->addedFacesListLength[1];

  /* set total number of element in mesh */
  mesh->TotalNElems = mesh->NElemsTetrah1+mesh->NElemsTriang1; 

  /* set number of nodes and mesh points */
  mesh->NNodes = tetMesh->NMeshPoints;
  mesh->NMeshPoints = tetMesh->NMeshPoints;
 
  /* allocate space for mesh members */
  mesh->MeshPointCoords = 
          e_alloc_cont_d_matr (mesh->NMeshPoints, 3);
  mesh->NodesTetrah1 = 
          e_alloc_cont_i_matr (mesh->NElemsTetrah1, E_NVPE_TETRAH1); 
  mesh->NodesTriang1 = 
          e_alloc_cont_i_matr (mesh->NElemsTriang1, E_NVPE_TRIANG1); 
  mesh->Decomp.SubDoms = e_alloc_i_vect (mesh->TotalNElems); 

  /* set subdomain information */
  /*
  mesh->Decomp.SubDomsTetrah1 = mesh->Decomp.SubDoms;
  mesh->Decomp.NSubDoms = 8;
  */

  /* copy mesh coordinates */
  VertexListTraversalInit (tetMesh->pList, TAIL_HEAD);
  i=0;
  while ((point = (Vertex *) VertexListTraverse (tetMesh->pList, TAIL_HEAD)) 
                  != NULL)
  {
    /* check if not going out of allocated space */
    assert (i < tetMesh->NMeshPoints);
    memcpy (mesh->MeshPointCoords[i], point->coords, 3*sizeof(double));
    i++;
  }
  
  AssignNumbersToPoints (tetMesh);
  
  /* copy node definition */
  TetListTraversalInit (tetMesh->tList, TAIL_HEAD);
  i=0;
  while ((tet = TetListTraverse(tetMesh->tList, TAIL_HEAD)) != NULL)
  {
    /* check if not going out of allocated space */
    assert (i < tetMesh->NTetrah);
    /* fill domain indormation */
    mesh->Decomp.SubDoms[i] = tet->subdomain;

    /* set element vertices */
    mesh->NodesTetrah1[i][0] = TET_VERTEX(tet, 0)->number; 
    mesh->NodesTetrah1[i][1] = TET_VERTEX(tet, 1)->number; 
    mesh->NodesTetrah1[i][2] = TET_VERTEX(tet, 2)->number; 
    mesh->NodesTetrah1[i][3] = TET_VERTEX(tet, 3)->number; 
    i++;
  }


  i=0;
  FaceListTraversalInit (mgds->addedFacesList[1], TAIL_HEAD); 
  while ((face = FaceListTraverse (mgds->addedFacesList[1], TAIL_HEAD))!= NULL)
  {
    mesh->NodesTriang1[i][0] = FACE_VERTEX(face, 0)->number; 
    mesh->NodesTriang1[i][1] = FACE_VERTEX(face, 1)->number; 
    mesh->NodesTriang1[i][2] = FACE_VERTEX(face, 2)->number; 
    i++; 
    FACE_FLAG (FACE_NEIGHBOUR (face, 3)) = -2;  
  }


  for (k=0; k<mgds->fTableLength[1]; k++)
  {
    face = mgds->fTable[1][k];
    if (FACE_FLAG(face) != -2)
    {
      mesh->NodesTriang1[i][0] = FACE_VERTEX(face, 0)->number; 
      mesh->NodesTriang1[i][1] = FACE_VERTEX(face, 1)->number; 
      mesh->NodesTriang1[i][2] = FACE_VERTEX(face, 2)->number; 
      i++; 
    }
  }

  /* set other required members of MESH structure */
  /* set first net index vector */
  /* mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] = 1;   */

  /* allocate memory for global element indices */
  /* mesh->GEIs = e_alloc_i_vect (mesh->NElemsTetrah1); */
  /* set GEIs vector */
  /*
  for (i=0; i<mesh->NElemsTetrah1; i++)
  {
    mesh->GEIs[i] = i+1;
  }
  mesh->GEIsTetrah1 = mesh->GEIs + mesh->FirstNetIndex[E_ELEM_TYPE_TETRAH1] -1;
  */
  /* allocate memory for global node indices */
 /* mesh->GNIs = e_alloc_i_vect (mesh->NNodes); */
  /* set GNIs vector */
  /* 
  for (i=0; i<mesh->NNodes; i++)
  {
    mesh->GNIs[i] = i+1;
  }
  */

  /* allocate memory for title string */
  mesh->Title = e_alloc_c_vect (100);
  /* set mesh title */
  strncpy (mesh->Title, "tetrah mesh", 99);
          
  mesh->KeywordSet[E_KW_NUM_TITLE] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NMESHPOINTS] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NNODES] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_TETRAH1] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1] = TRUE;
  mesh->KeywordSet[E_KW_NUM_MESHPOINT_COORDS] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NODES_TETRAH1] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NODES_TRIANG1] = TRUE;

  /* mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE; */

  return;
} /* end of TetrahMeshToMESH_V1 */
#endif


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   MESHToTetrahMesh  
 * DESCRIPTION
 *   Transfer data from MESH data structure to TetrahMesh one.
 *   There is no one to one correspondence between those data structures,
 *   only meshpoint coordiantes and adjacency relationships are transfered.
 * ARGUMENTS
 *   mesh - pointer to MESH data structre
 *   tetMesh - pointer to tetrahedral mesh data structure
 * RETURN VALUE
 *   NONE
 ******/
void MESHToTetrahMesh (MESH *mesh, TetrahMesh *tetMesh)
{
  int i, vert_num;
  int facnum;
  Face *face;
  Vertex *point;
  Tetrah *tet, *adjtet;

  tetMesh->NMeshPoints = mesh->NMeshPoints;
  tetMesh->NTetrah = mesh->NElemsTetrah1;

  /* check if there are any tetrahedal elements in the mesh */
  if (tetMesh->NTetrah == 0)
  {
    e_GeneralError ("Tos_MESHToTetrahMesh : no Tetrah1 elements in mesh");
  }
  /* create points, copy mesh points coordinates and put points on list*/
  for (i=0; i<tetMesh->NMeshPoints; i++)
  {
    point = VertexCreate ();
    memcpy (point->coords, mesh->MeshPointCoords[i], 3*sizeof(double));
    point->number = i+1;
    PutVertexOnListHead(tetMesh->pList, point);
  }

  for (i=0; i<tetMesh->NTetrah; i++)
  {
    /* create new tetrahedron */
    tet = TetrahCreate ();
    /* put tetrah on the list */
    PutTetOnListHead(tetMesh->tList, tet);
    TET_NUMBER(tet) = i+1; 
    /* sort array of vertices numbers */
    qsort ((void *) mesh->NodesTetrah1[i], E_NVPE_TETRAH1, sizeof (int), IntRevCompare);
    vert_num=0; /*number of tet vertex */
    /* prepare to traverse points list */
    /* CAUTION: order of traversal maters - it should be the matched
     * with sorting of the node numbers in elements
     */  
    point = LIST_HEAD (tetMesh->pList); 
    /* traverse point list and set pointers to vertices in tetrahedron */
    while (point  != NULL && vert_num <4)
    {
      if (point->number == *(mesh->NodesTetrah1[i]+vert_num))
      {
        TET_VERTEX(tet, vert_num) = point;

        /* add tet to point adjacency list */
        Tos_MStackPush (point->adjElem, (void *) tet);
        
        vert_num++;
      }
      point = VERTEX_PREV (point);
    }
    /* if point list has bee  dn traversed and not all four vertices
     *  in tetrahedron has been set - something went wrong 
     */  
    assert (vert_num == 4);
  }

  /* Fill adjacency informations in tetrahedra */
    /* create temporary Face object to make easier looking for adjacency */
  face = FaceCreate();
    /* prepare to traverse list of tetrahedra */
  TetListTraversalInit (tetMesh->tList, HEAD_TAIL);
  while ((tet = TetListTraverse (tetMesh->tList, HEAD_TAIL)) != NULL)
  {
    /* for each face 'i' */
    for (i=0; i<4; i++)
    {
      /* check if neighbouring information has been set for that face */
      if (TET_NEIGHBOUR (tet, i) != NULL)
      {
        continue; /* face adjacency has been already set */
      }
      else
      {
        GetFaceFromTet (tet, i, face);
        /* prepare to traverse adjacency list of one of element vertices 
           optimizing remark:
                it would be beneficial to pick vertex with the smallest
                adjacency list
           at the moment I simply pick the vertex 0     
         */
        point = FACE_VERTEX (face, 0); 
        Tos_MStackTraversalInit (point->adjElem);
        while ((adjtet = Tos_MStackTraverse (point->adjElem)) != NULL)
        {
          if (adjtet != tet)
          {
            if (TetHasFace (adjtet, face, &facnum) == TRUE)
            {
               /* set adjacency information */
               TET_NEIGHBOUR (tet, i) = adjtet;
               TET_NEIGHBOUR (adjtet, facnum)  = tet;
               break;
            }
          }
        } /* end of traversal over adjacent elements */
      }
    } /* end of loop over tet faces */ 
  }
  /* delete temporary Face object */
  FaceDelete (face); 
} /* end of MESHToTetrahMesh */

/* Modifications:
 * $Log: convert.c,v $
 * Revision 1.4  2002/07/28 13:27:37  peteri
 * - list creation interface is modified
 * - it is possible to add to the head or the tail of the list
 * - all related code is updated
 *
 * Revision 1.3  2002/02/16 18:58:44  romek
 * Doc fix.
 *
 */
