/********************************************************************
 ctr_read.c --     
        Function for reading mesh generator control data structure. 
 
 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University Edinburgh    
 ********************************************************************/

/* $Id: ctr_read.c,v 1.6 2002/07/28 16:57:01 peteri Exp $ */

#include <stdio.h>
#include "ctr_guard.h"
#include "ctr_read.h"

/* Some error messages uses gcc __FUNCTION__ macro which 
 * is not defined in Sun cc  
 */
#ifndef __GNUC__
#   define __FUNCTION__ __FILE__
#endif

extern int yyparse();
extern void ctr_InitLexer (FILE *filePtr);           


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   MeshGenCtrlRead 
 * DESCRIPTION
 *   Read mesh generator control data from a file 
 * ARGUMENTS
 * RETURN VALUE
 *   NONE
 ******/
void MeshGenCtrlRead (MeshGenCtrl *mctrl, const char *filename) 
{
  FILE* filePtr;
  char * logNameSuffix = ".log";
  filePtr = fopen (filename, "r");
  if (filePtr == NULL)
  {
    e_GeneralError("%s(%d) : cannot open data file '%s' for reading", 
                   __FILE__, __LINE__, filename);
  } 
  mctrl->ctrlFileName = strdup (filename);
  ctr_InitLexer (filePtr);
  yyparse(mctrl);
  if (mctrl->logLevel != VM_NO_LOG)
  {
    if (mctrl->logFileName == NULL)
    { 
      mctrl->logFileName = (char *) 
             e_alloc_c_vect (strlen(mctrl->outputFileName)
                     + strlen(logNameSuffix)+ 1);
      if (mctrl->logFileName == NULL)
      {
        e_GeneralError("%s(%d) : cannot allocate memory for log file name", 
                   __FILE__, __LINE__);
      }
      strcpy (mctrl->logFileName, mctrl->outputFileName);
      strcat (mctrl->logFileName, logNameSuffix);
    }
    mctrl->logFile = fopen (mctrl->logFileName, "w");
    if (filePtr == NULL)
    {
      e_GeneralError("%s(%d) : cannot open log file '%s' for writing", 
                   __FILE__, __LINE__, mctrl->logFileName);
    } 
  }
} /* end of MeshGenCtrlRead */

int MGCSet_nmeshes (MeshGenCtrl *mctrl, int nmeshes)
{
   if (nmeshes > VM_MAX_NMESHES || nmeshes < 1)
   {
     fprintf (stderr, 
         "Invalid number of meshes %d in file %s(must be: 0 < nmehses < %d\n",
          nmeshes, mctrl->ctrlFileName, VM_MAX_NMESHES);
     return 0;
   }
   mctrl->nmeshes = nmeshes;
   return 1;
} /* end of MGCSet_nmeshes */

int  MGCSet_inputFileNames (MeshGenCtrl *mctrl, int meshNum, char* fileName)
{
   assert (meshNum >=0 );
   if (meshNum >= VM_MAX_NMESHES)
   {
     e_GeneralError("Too many input meshes in file %s(expecting exactly %d\n", 
                    mctrl->ctrlFileName, mctrl->nmeshes);
     return 0;
   }
   mctrl->inputFileNames[meshNum] = fileName;
   return 1;
} /* end of MGCSet_inputFileNames */

int  MGCSet_outputFileName (MeshGenCtrl *mctrl, char* fileName)
{
   mctrl->outputFileName = fileName;
   return 1;
} /* end of MGCSet_outputFileName */

int  MGCSet_alpha (MeshGenCtrl *mctrl, double alpha)
{
   mctrl->alpha = alpha;
   return 1;
} /* end of MGCSet_alpha */

int  MGCSet_beta (MeshGenCtrl *mctrl, double beta)
{
   mctrl->beta = beta;
   return 1;
} /* end of MGCSet_beta */

int MGCSet_refineMethod (MeshGenCtrl *mctrl, vm_RefineMethod refineMethod)
{
  switch(refineMethod)
  {
    case VM_REFINE_EDGE_LEN:
        mctrl->refineMethod = refineMethod;
        return 1;
    default:
        return 0;
  }
} /* end of MGCSet_refineMethod */

int  MGCSet_logFileName (MeshGenCtrl *mctrl, char* fileName)
{
   mctrl->logFileName = fileName;
   return 1;
} /* end of MGCSet_logFileName */


int MGCSet_logLevel (MeshGenCtrl *mctrl, vm_LogLevel logLevel)
{
   switch (logLevel)
   {
     case VM_NO_LOG:
       mctrl->logLevel = logLevel;
       return 1;
     case VM_MINIMAL_LOG:
       mctrl->logLevel = logLevel;
       return 1;
     case VM_NORMAL_LOG:
       mctrl->logLevel = logLevel;
       return 1;
     case VM_FULL_LOG:
       mctrl->logLevel = logLevel;
       return 1;
     default:
       return 0;
   }
} /* end of MGCSet_logLevel */


int  MGCSet_optimization (MeshGenCtrl *mctrl, int optimization)
{
  mctrl->optimizeMethod = optimization;

  return 1;
}



/* Modifications:
 * $Log: ctr_read.c,v $
 * Revision 1.6  2002/07/28 16:57:01  peteri
 * - the required optimization method is specified in the file
 *   OPTIMIZATION INSERTING LAPLACE SWAPPING COLLAPSING
 *
 * Revision 1.5  2002/07/26 17:53:34  peteri
 * - Internal point is not needed for flooding any more
 *
 * Revision 1.4  2002/07/26 13:06:29  peteri
 * CENTR_VERT_LEN refinement method is removed
 *
 * Revision 1.3  2002/01/31 22:37:39  peteri
 * Minor modifications, some renaming.
 *
 * Revision 1.2  2002/01/31 21:27:48  peteri
 * Major reorganization in the code.
 * - It is split into smaller modules.
 * - Some types, functions are renamed
 * - Proper flipping of edges is implemented
 *
 * Revision 1.1.1.1  2002/01/30 10:20:56  peteri
 * Imported using TkCVS
 *
 * Revision 1.2  2001/10/27 18:31:22  romek
 * ,
 *
 * Revision 1.1  2001/10/04 17:59:07  romek
 * Initial revision
 *
 */
