#include <assert.h>
#include "find.h"
#include "eg_lib.h"
#include "basic_op.h"
#include "debug.h"
#include "boundary.h"

#ifdef WIN32
extern double copysign(double x, double y);
#endif

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *  WhereInTetrah 
 * DESCRIPTION
 *  Find position of point in tetrahedron.
 * ARGUMENTS
 *  tetrah -  tetrahedron (IN)
 *  point  -  point (IN)
 *  faceNum - number of face (OUT) 
 *  faceSum - sum of number of faces (OUT) 
 * RETURN VALUE
 *  OUTSIDE - if point is not enclosed within tetrahedron
 *  INSIDE - if point is inside tetrahedron
 *  ONFACE - if point is on tetrahedron face
 *  ONEDGE - if point is on tetrahedron edge
 *  ONVERTEX - if point falls on vertex
 *  Additionally if point is located on face or on edge or on 
 *  vertex 'faceNum' is set to the bigest generic number of face
 *  (0, 1, 2, or 3) on which point lies and 'faceSum' is sum
 *  is sum of generic number of faces on which poit lies.
 ******/
int WhereInTetrah (Tetrah *tet, Vertex *point,
                   int *faceNum, int *faceSum)
{
  register int face;
  register int loccard;
  double position, position1;
  double posSign, pos1Sign;
  enum PointLocation where = OUTSIDE; /* this assigment is only to suppress
                                       * warning message
                                       */  
  double *p[5];

  assert (tet != NULL);
  assert (point != NULL);

  loccard = 0; /* location cardinality -1=outside, 0=inside, 1=onface
                * 2=onedge, 3=onvertex 
                */
  *faceNum = -1;
  *faceSum = 0;
  p[4] = point->coords;        
  p[0] = TET_VCOORDS(tet, 0); 
  p[1] = TET_VCOORDS(tet, 1); 
  p[2] = TET_VCOORDS(tet, 2); 
  p[3] = TET_VCOORDS(tet, 3);         
  for (face=0; face <4; face++)
  {
    position = eg_Orient3D (p[faceDef[face][0]], p[faceDef[face][1]], 
                            p[faceDef[face][2]], p[4]);  
    position1 = eg_Orient3D (p[faceDef[face][0]], p[faceDef[face][1]], 
                             p[faceDef[face][2]], p[face]);
    posSign = copysign (1.0,  position);
    pos1Sign = copysign (1.0,  position1);  

    if ((-POSITION_TOL <= position) && (position <= POSITION_TOL) ) 
    {
      (*faceSum) += face;
      loccard++; /* count how many times point falls on element face */
      *faceNum = face;  
      if (loccard == 3) break;
    }
    else if ((posSign * pos1Sign) < 0.0) 
    {
      loccard = -1; 
      *faceNum = face;  
      (*faceSum) = 0;
      break;  
    }

  }
  switch (loccard)
  {
    case -1: {where = OUTSIDE;  break;}
    case  0: {where = INSIDE;   break;}
    case  1: {where = ONFACE;   break;}
    case  2: {where = ONEDGE;   break;}
    case  3: {where = ONVERTEX; break;}
    default:
      e_GeneralError("WhereInTetrah: Unknown location cardinality %d\n", 
                     loccard);
  }
  return where;
} /* end of WhereInTertah */




/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/****f*
 * NAME
 *   FindSharedEdge 
 * DESCRIPTION
 *   Finds edge which is shared by two face of tetrahedron.
 * ARGUMENTS
 *   tet - pointer to tetrahedron (IN)
 *   face1 - index of face in tet (IN)
 *   face2 - index of face in tet (IN)
 *   e  - edge shared between face1 and face2 (OUT)
 * RETURN VALUE
 *   NONE
 ******/
void FindSharedEdge (Tetrah *tet, int face1, int face2, Edge *e)
{
  int vert1, vert2;
  TetGenericMapping (face1, face2, &vert1, &vert2);
  EDGE_VERTEX (e, 0) = TET_VERTEX (tet, vert1);
  EDGE_VERTEX (e, 1) = TET_VERTEX (tet, vert2);
 return;
} /* end of FindSharedEdge */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*
 * This is like FindBaseWalk but it does an exhaustive search !
 */
int FindBaseSearchAll(Tetrah *st, Vertex *point, Tos_MStack base)
{
  int face, sumface;
  int where;
  int baseFound = FALSE; 
  int result = 1;
  int sanityCheck; 

  assert (st != NULL); 

  while(st != NULL && baseFound != TRUE && result != -1)
  {
    where = WhereInTetrah (st, point, &face, &sumface);

    switch (where)
    {
      case OUTSIDE: 
      {
        st = TET_PREV(st);
        break;
      }
      case INSIDE:
      {
        Tos_MStackPush (base, (void *) st);
        baseFound = TRUE;
        break; 
      }
      case ONFACE:
      {
        Tos_MStackPush (base, (void *) st);
        assert (TET_NEIGHBOUR(st, face) != NULL); 
        Tos_MStackPush (base, (void *) TET_NEIGHBOUR(st, face));        
        baseFound = TRUE;
        break;
      }
      case ONEDGE:
      {
        Edge edge;
        int facebis = sumface - face;

        /* find edge common to two faces */
        FindSharedEdge (st, face, facebis, &edge);
        /* find elements of shell associated with edge e */
        sanityCheck = FindEdgeShell (&edge, base);
        assert (sanityCheck > 2);
        baseFound = TRUE;
        break;
      }
      case ONVERTEX:
      {
        e_GeneralWarning ("FindBaseSearchAll: Duplicate point location at: "
                          "(%f %f %f)\n\t(point %p, tetrah %p)\n", 
                          point->coords[0], point->coords[1],
                          point->coords[2], point, st);
        result = -1;
        break;
      }
      default:
         e_GeneralError ("FindBaseSearchAll: Unknown point location %d", where);
    }
  }
  assert(st != NULL);

  return result; 
}


/* FIXME: Does not work, it may get into a loop,
   the neighbour access should be randomized
 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/****f*
 * NAME
 *   FindBaseWalk  
 * DESCRIPTION
 *   Finds base of given meshpoint using walking search algorithm.
 *   For a given conforming triangulation T and a point P enclosed in
 *   T base is a set of elements in T enclosing P. 
 * ARGUMENTS
 *   startTet - tetrahedron from which the search starts (IN)
 *   point  -  point for which base is searched (IN)
 *   base - pointer to point base stack (OUT)
 * RETURN VALUE
 *   If base has been found return 1.
 *   If point falls on existing vertex returns -1 and prints
 *   warning message.
 ******/
int FindBaseWalk (Tetrah *st, Vertex *point, Tos_MStack base)
{
  int face, sumface;
  int where;
  int baseFound = FALSE; 
  int result = 1;
  int sanityCheck; 

  assert (st != NULL); 

  while (baseFound != TRUE && result != -1)
  {
    where = WhereInTetrah (st, point, &face, &sumface);

    switch (where)
    {
      case OUTSIDE: 
      {
        assert (TET_NEIGHBOUR(st, face) != NULL);
        st = TET_NEIGHBOUR (st, face);
        break;
      }
      case INSIDE:
      {
        Tos_MStackPush (base, (void *) st);
        baseFound = TRUE;
        break; 
      }
      case ONFACE:
      {
        Tos_MStackPush (base, (void *) st);
        assert (TET_NEIGHBOUR(st, face) != NULL); 
        Tos_MStackPush (base, (void *) TET_NEIGHBOUR(st, face));        
        baseFound = TRUE;
        break;
      }
      case ONEDGE:
      {
        Edge edge;
        int facebis = sumface - face;

        /* find edge common to two faces */
        FindSharedEdge (st, face, facebis, &edge);
        /* find elements of shell associated with edge e */
        sanityCheck = FindEdgeShell (&edge, base);
        assert (sanityCheck > 2);
        baseFound = TRUE;
        break;
      }
      case ONVERTEX:
      {
        e_GeneralWarning ("FindBaseWalk: Duplicate point location at: "
                          "(%f %f %f)\n\t(point %p, tetrah %p)\n", 
                          point->coords[0], point->coords[1],
                          point->coords[2], point, st);
        result = -1;
        break;
      }
      default:
         e_GeneralError ("FindBaseWalk: Unknown point location %d", where);
    }
  }
  
  return result; 
} /* end of FindBaseWalk */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindCavity -- 
        Finds cavity associated with given point.
 Cavity associated with a given point is the set of elements whose
 circumballs enclose the point.

 NOTE: IMPORTANT: FindCavity maintains one VERY IMPORTANT stack: 
 outcavStack. It contains elements which were flaged as TET_OUTCAVITY.
 The contents of this stack MUST NOT be changed between call to 
 FindCavity and ReconectCavity.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int FindCavity (Vertex *point, Tos_MStack base, int force, Tos_MStack cavity, 
                Tos_MStack outcavStack)
{
  register int face; 

  Tetrah *tet, *neighbour;
  double *p[5];
  double where;
  double orient;
  int cavok = 1;
  struct Tos_MStack cStack = {0, NULL, NULL};
  struct Tos_MStack vStack = {0, NULL, NULL};
  Tos_MStack checkStack = &cStack;
  Tos_MStack visitedStack = &vStack;
  
  /* Take elements from base and put them into checking stack */
  while (Tos_MStackEmpty (base) != 1)
  {
    tet = (Tetrah *) Tos_MStackPop (base);
    Tos_MStackPush (checkStack, (void *) tet);
  }

  while (Tos_MStackEmpty (checkStack) != 1)
  {
    tet = (Tetrah *) Tos_MStackPop (checkStack);
    if ((TET_FLAG (tet) & TET_VISITED) != TET_VISITED)
    {
      TET_SET_FLAG (tet, visitedStack, TET_VISITED);
      /* get coords of tet vertices */
      p[0] = TET_VCOORDS(tet, 0);
      p[1] = TET_VCOORDS(tet, 1);
      p[2] = TET_VCOORDS(tet, 2);
      p[3] = TET_VCOORDS(tet, 3);
      p[4] = point->coords;
      /* the order of arguments below is essential */ 
      orient = eg_Orient3D (p[1], p[2], p[3], p[0]);
      where = eg_InSphere (p[1], p[2], p[3], p[0], p[4]); 

      if (orient < 0.0 ) where = -where;
        
      if ((force != 1) && (0.0 < where && where < COSPHER_TRESHOLD))
      {
        cavok = -1; 
        break;
      }
      
      if (where > 0.0) /* point inside sphere */
      {
        Tos_MStackPush (cavity, (void *)tet);
        for (face = 0; face <4; face++)
        {
          neighbour = TET_NEIGHBOUR (tet, face);
          if (   neighbour != NULL
              && ((TET_FLAG(neighbour) & TET_VISITED) != TET_VISITED))
          {
            Tos_MStackPush (checkStack, (void *) neighbour);
          }
        }
      }
      else
      {
        TET_SET_FLAG (tet, outcavStack, TET_OUTCAVITY);  
      }
    }
  }

  /* Remove all elements for checking stack */
  Tos_MStackFlush (checkStack);
  TetrahClearFlag (visitedStack, TET_VISITED);
  
  return cavok;
} /* end of FindCavity */




/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindConnectedPoints --
        Build stack of points connected to a given one through 
        common edge.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FindConnectedPoints (Vertex* point, Tos_MStack connectedPoints)
{
  Vertex* cpoint;
  Tetrah *adjtet;
  int i;
  Tos_MStackTraversalInit (point->adjElem);
  while ((adjtet = Tos_MStackTraverse (point->adjElem)) != NULL)
  {
    for (i=0; i<4; i++)
    {
      if (TET_VERTEX(adjtet, i) == point)
      {
        continue;
      }
      if ((TET_VERTEX(adjtet, i)->flag & POINT_CONNECTED) != POINT_CONNECTED)
      {
        TET_VERTEX(adjtet, i)->flag |= POINT_CONNECTED;       
        Tos_MStackPush (connectedPoints, TET_VERTEX(adjtet, i));
      }
    }
  }
  /* clear the flag POINT_CONNECTED */
  Tos_MStackTraversalInit (connectedPoints); 
  while ((cpoint = (Vertex*) Tos_MStackTraverse (connectedPoints)) != NULL)
  {
    cpoint->flag &= ~(POINT_CONNECTED);
  }
} /* end of FindConnectedPoints */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindEdgesIntersectingFacePlane --
        Finds edges of tetrahedron 'tet' which intersect plane
       of given face and puts it on 'intersectingEdges' stack. 
       Array 'nodes' contain generic numbers of tetrahedron vertices
       shared by tetrahedron and face.
       Number of shared vertices is hold in 'cnodes'.

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FindEdgesIntersectingFacePlane (Tetrah *tet, Face *face,
                                     int *nodes, int cnodes,
                                     Tos_MStack intersectingEdges)
{
  register int i;
  double pos[3]; 
  int ne[3]; 
  int ie1, ie2;
  Edge *edge;

  if (cnodes == 2) /* two nodes of tetrah are shared with those which 
                    * are lying on face so only one edge can intersect
                    * face plane 
                    */    
  {
    /* find generic numbers of nodes which aren't in nodes array */
    /* those nodes possibly form intersectin edge */
    TetGenericMapping (nodes[0], nodes[1], &(ne[0]), &(ne[1])); 
  
    /* check if edge of specified node is already on the stack */
    if (IsEdgeOnStack(intersectingEdges,
                      TET_VERTEX (tet, ne[0]),
                      TET_VERTEX (tet, ne[1]) ) == FALSE)
    {
      /* check position of the nodes */
        pos[0] = eg_Orient3D (FACE_VCOORDS (face, 0), FACE_VCOORDS (face, 1), 
                              FACE_VCOORDS (face, 2), TET_VCOORDS (tet, ne[0])); 
        pos[1] = eg_Orient3D (FACE_VCOORDS (face, 0), FACE_VCOORDS (face, 1), 
                              FACE_VCOORDS (face, 2), TET_VCOORDS (tet, ne[1])); 
       
      if ((pos[0] <=0 && pos[1] >= 0) || (pos[0] >=0 && pos[1]<=0))
      {
        /* create intersection edge and push it on stack */
        edge = EdgeCreate ();
        EDGE_VERTEX (edge, 0) = TET_VERTEX (tet, ne[0]); 
        EDGE_VERTEX (edge, 1) = TET_VERTEX (tet, ne[1]);
        Tos_MStackPush (intersectingEdges, (void *) edge); 
      }  
    }
  }
  else if (cnodes == 1) /* one node is shared by face and tetrah
                         * so possibly two of three tetrahedron edges
                         * can intersect face plane
                         */
  {
    /* find nodes which aren't shared with face (aren't in 'nodes' array) */
    ne[0] = faceDef[nodes[0]][0]; 
    ne[1] = faceDef[nodes[0]][1]; 
    ne[2] = faceDef[nodes[0]][2]; 

    /* check position of the nodes */
    pos[0] = eg_Orient3D (FACE_VCOORDS (face, 0), FACE_VCOORDS (face, 1), 
                          FACE_VCOORDS (face, 2), TET_VCOORDS  (tet, ne[0])); 
    pos[1] = eg_Orient3D (FACE_VCOORDS (face, 0), FACE_VCOORDS (face, 1), 
                          FACE_VCOORDS (face, 2), TET_VCOORDS  (tet, ne[1])); 
    pos[2] = eg_Orient3D (FACE_VCOORDS (face, 0), FACE_VCOORDS (face, 1), 
                          FACE_VCOORDS (face, 2), TET_VCOORDS  (tet, ne[2])); 
    for (i=0; i<3; i++)
    {
      ie1=sideDef[i][0];
      ie2=sideDef[i][1];
      
      /* create and push edge on stack only if it hasn't been examined */
      if (IsEdgeOnStack(intersectingEdges,
                        TET_VERTEX (tet, ne[ie1]), 
                        TET_VERTEX (tet, ne[ie2]) ) == FALSE)
      {
        /* check if edge intersect face plane */
        if ((pos[ie1] <=0 && pos[ie2] >= 0) || (pos[ie1] >=0 && pos[ie2]<=0))
        {
          /* create intersection edge and push it on stack */
          edge = EdgeCreate ();
          EDGE_VERTEX (edge, 0) = TET_VERTEX (tet, ne[ie1]); 
          EDGE_VERTEX (edge, 1) = TET_VERTEX (tet, ne[ie2]);
          Tos_MStackPush (intersectingEdges, (void *) edge); 
        } 
      }
    }
  }
  else 
  {
    e_GeneralError("FindEdgesIntersectingFacePlane:\n"
                   "Invalid number of shared nodes (cnodes=%d)\n",
                   cnodes); 
  }
} /* end of FindEdgesIntersectingFacePlane */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindEdgeShell --
        Finds shell associated with given edge.
 Shell associated with edge is the set of elements sharing this edge.       

 Arguments:
        e - edge
        shell - stack of shell elements
 Return value:
        On succes returns number of elements in the shell, 
        otherwise returns 0.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int FindEdgeShell (Edge *edge, Tos_MStack shell)
{
  register int vertex; 
  register int counter=0; /* counts number of elements in shell */
  Vertex *v1, *v2, *vd;
  Tetrah *tet;


  v1 = EDGE_VERTEX (edge, 0); 
  v2 = EDGE_VERTEX (edge, 1);
  /* check which vertex has smaller adjacency stack and set v1 to
   * that vertex
   */
  if (Tos_MStackCount (v1->adjElem) > Tos_MStackCount (v2->adjElem))
  {
    vd = v1; v1 = v2; v2 = vd;
  }
  /* traverse stack of adjacent elements at vertex v1 */
  Tos_MStackTraversalInit (v1->adjElem);
  while ((tet = (Tetrah *) Tos_MStackTraverse (v1->adjElem)) != NULL)
  {
    /* check if vertex v2 is in tet */
    /* This is trick to make the loop a bit faster and is equivalent to:
     *  for (vertex=0; vertex<4; vertex++)
     */
    for (vertex=4; vertex--; )
    {
      if (TET_VERTEX (tet, vertex) == v2) 
      {
        counter++; /* advabnce number of elements in shell */
        Tos_MStackPush (shell, (void *)tet); 
        break; 
      }
    } 
  }
  return counter;
} /* end of FindEdgeShell */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindEdgeEquator --
     Finds all vertices which belongs to edge shell but are different
     than edge endpoints. 
 Arguments:
        e - edge
        tet - one element form edge shell
        shell - stack of equator points 
 Return value:
        On succes returns number of vertices on edge equator, 
        otherwise returns 0.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int FindEdgeEquator (Edge *edge, Tetrah *startTet, Tos_MStack equator)
{
  register int counter=0; /* counts number of elements in shell */
  Vertex *v1, *v2; /* vertices of the edge */
  Vertex *v3, *v4; /* vertices different than edge endpoints */
  int iv1, iv2;
  int iv3, iv4;
  
  Tetrah *tet, *neighbour;

  v1 = EDGE_VERTEX (edge, 0); 
  v2 = EDGE_VERTEX (edge, 1);

  iv1 = WhichVertex (startTet, v1);
  iv2 = WhichVertex (startTet, v2);
  
  TetGenericMapping (iv1, iv2, &iv3, &iv4);

  tet = startTet; 
  v3 = TET_VERTEX(tet, iv3);
  v4 = TET_VERTEX(tet, iv4);
  while(1) 
  {
    Tos_MStackPush (equator, (void *)v3);
    /* find neighbour oposite v3 */
    neighbour = TET_NEIGHBOUR(tet, iv3);
    if (neighbour == startTet)
    {
      break;
    }
    iv1 = WhichVertex (neighbour, v1);  
    iv2 = WhichVertex (neighbour, v2);
    TetGenericMapping (iv1, iv2, &iv3, &iv4);
    if (TET_VERTEX(neighbour, iv3) == v4)
    {
       v3 = v4;
       v4 = TET_VERTEX(neighbour, iv4);
    }
    else
    {
       v3 = TET_VERTEX(neighbour, iv4);
       v4 = TET_VERTEX(neighbour, iv3);
       /* */
       iv3 = iv4;
    }
    tet = neighbour;
  }
  counter = Tos_MStackCount (equator);
  return counter;
} /* end of FindEdgeEquator */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindFaceShell --
        Finds face shell -- at most two tetrahedra having specified 
        face.
 Return value : 0 - if no tetrahedra adjacent to face is found
                1 - if only one tetrahedron is adjacent to face 
                    (eg. boundary face)
                2 - if two tetrahedra are adjacent to face    
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int FindFaceShell (Face *face, Tetrah **shell)
{
  register int vertex;
  int v2found, v3found;
  int faceNum=6;
  int counter=0; /* number of tetrahedra adjacent to face */
  Vertex *v1, *v2, *v3, *vd;
  Tetrah *tet;
  
  v1 = FACE_VERTEX (face, 0);
  v2 = FACE_VERTEX (face, 1);
  v3 = FACE_VERTEX (face, 2);

  /* arrange vertices according to number of adjacent tetrahedra in 
   * such a way that v1 has the smallest number of adjacent elements 
   */
   
  if (Tos_MStackCount (v1->adjElem) > Tos_MStackCount (v2->adjElem))
  {
    vd = v1; v1 = v2; v2 = vd;
  }  
  if (Tos_MStackCount (v1->adjElem) > Tos_MStackCount (v3->adjElem))
  {
    vd = v1; v1 = v3; v3 = vd;
  }

  /* Check number of tetrahedra adjacent to vertex1 */
  if (Tos_MStackCount (v1->adjElem) < 1)
  {
    /* fprintf (stderr, "Not connected vertex\n"); */
    return 0; 
  }
  /* Traverse adjacency stack at vertex v1 
   * and search for tetrahedron adjacent to face
   * the same time find generic face number of that tetrahedron  
   */
  Tos_MStackTraversalInit (v1->adjElem);
  while ((tet = (Tetrah *) Tos_MStackTraverse (v1->adjElem)) != NULL)
  {
    v2found = FALSE; 
    v3found = FALSE;
    faceNum = 6;
    /* check if vertices v2 and v3 are in tet */
    /* Trick to make loop faster, this is equivalent to:
     *  for (vertex=0; vertex<4; vertex++)
     */  
    for (vertex=4; vertex--; )
    {
      if (TET_VERTEX (tet, vertex) == v1)
      {
        faceNum = faceNum - vertex; 
        continue;
      }
      if (TET_VERTEX (tet, vertex) == v2)
      {
        v2found = TRUE;
        faceNum = faceNum - vertex; 
        continue;
      }
      if (TET_VERTEX (tet, vertex) == v3)
      {
        v3found = TRUE;
        faceNum = faceNum - vertex; 
        continue;
      }
    }
    if (v2found == TRUE && v3found == TRUE)
    {
      counter = 1;
      break;
    }
  }

  if (counter == 1)
  {
    shell[0] = tet;
    if (TET_NEIGHBOUR (tet, faceNum) != NULL)
    {
      shell[1] = TET_NEIGHBOUR (tet, faceNum);
      counter = 2;
    }
  }
  return counter;
} /* end of FindFaceShell */

/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

static void DoesEdgeExist(Edge *edge, Tos_MStack missingEdges)
{
  int     edgeExist, vertex;
  Vertex *v1, *v2, *vd;
  Tetrah *tet;
  
  edgeExist = 0;
  v1 = EDGE_VERTEX (edge, 0);
  v2 = EDGE_VERTEX (edge, 1);

  /* check which vertex has smaller adjacency stack */
  if (Tos_MStackCount (v1->adjElem) > Tos_MStackCount (v2->adjElem))
  {
    vd = v1; v1 = v2; v2 = vd;
  }
  
  /* prepare to traverse adjacency stack at vertex v1 */
  Tos_MStackTraversalInit (v1->adjElem);
  while ((tet = (Tetrah *) Tos_MStackTraverse (v1->adjElem)) != NULL)
  {
    for (vertex = 0; vertex < 4; vertex++)
    {
      if (TET_VERTEX (tet, vertex) == v2)
      {
        edgeExist = 1;
        break;
      }
    }  
    if (edgeExist == 1) break;
  }
  
  if (edgeExist == 0) /* edge hasn't been encountered in tetrah mesh */
  {
    Tos_MStackPush (missingEdges, (void *)edge);
  }
  
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindMissingEdges --
        Find edges present in edges list 'eTable' but missing from 
        tetraherdal mesh.

 Notes: This function operates on global variable 'mgds'
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FindMissingEdges (MeshGenData *mgds, Tos_MStack missingEdges)
{
  Edge  *edge, *splitedge;
  int    i, iedge;
  struct Tos_MStack edgeIter = {0, NULL, NULL};

  /* go through all boundary edges */
  for (i = 0; i < mgds->NMeshes; i++)
  {
    /* prepare to traverse boundary edges table */
    for(iedge = 0; iedge < mgds->eTableLength[i]; iedge++)
    {
      edge = mgds->eTable[i][iedge];
      
      for(splitedge = InitSplitEdge(edge, &edgeIter);
          splitedge != NULL;
          splitedge = NextSplitEdge(&edgeIter))
      {
        DoesEdgeExist(splitedge, missingEdges);
      }

    } /* end of loop over edges */
  } /* end of loop over meshes */

  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 -- Number of missing edges : %d\n", 
           Tos_MStackCount (missingEdges));
  }
  
} /* end of FindMissingEdges */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 DoesFaceExist --
        Check if face is present in mesh construction.

 Arguments:
        tetMesh - tetrahedral mesh structure (IN)
        face - face in question (IN)
        tetWithFace - tetrahedron having given face (OUT)
        faceNum - generic number of face in tet (OUT)
 Return value:
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void DoesFaceExist(Face *face, Tos_MStack missingFaces)
{
  register int vertex; 
  register int v2found;
  register int v3found;
  register int fnum;

  Tetrah *tet;
  Vertex *vd, *v1, *v2, *v3;
  
  v1 = FACE_VERTEX (face, 0); 
  v2 = FACE_VERTEX (face, 1); 
  v3 = FACE_VERTEX (face, 2); 

  /* arranging v1, v2, v3 so v1 has the smallest adjacency stack */
  if (Tos_MStackCount (v1->adjElem) > Tos_MStackCount (v2->adjElem))
  {
    vd = v1; v1 = v2; v2 = vd;
  }
  if (Tos_MStackCount (v1->adjElem) > Tos_MStackCount (v3->adjElem))
  {
    vd = v1; v1 = v3; v3 = vd;
  }

  /* traverse adjacency stack at vertex v1 */
  Tos_MStackTraversalInit (v1->adjElem);
  while ((tet = (Tetrah *) Tos_MStackTraverse (v1->adjElem)) != NULL)
  {
    fnum = 6;
    v2found = 0;
    v3found = 0;
    /* check if vertices v2 and v3 are present in adjacent tetrahedra */
    for (vertex = 0; vertex <4; vertex++)
    {
      if (TET_VERTEX (tet, vertex) == v1) 
      {
        fnum -= vertex;
        continue; 
      } 
      if (TET_VERTEX (tet, vertex) == v2)
      {
        fnum -= vertex;
        v2found = 1;
        continue;
      }
      if (TET_VERTEX (tet, vertex) == v3)
      {
        fnum -= vertex; 
        v3found = 1;
        continue;
      }
    }

    if(v2found == 1 && v3found == 1)
    {
      /* face [v1, v2, v3] is the 'fnum' face of tetrahedron 'tet' */
      return;
    }
  }

  Tos_MStackPush (missingFaces, (void *)face);
} /* end of IsFaceInMesh */
        
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindMissingFace --
        Find faces present in face list 'fTable' but missing from 
        tetraherdal mesh.

 Notes: This function operates on global variable 'mgds'
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FindMissingFaces (MeshGenData *mgds, Tos_MStack missingFaces)
{
  Face  *face, *splitface;
  int    imesh, iface;
  struct Tos_MStack faceIter = {0, NULL, NULL};
  
  
  /* go through all boundary faces */
  for (imesh = 0; imesh < mgds->NMeshes; imesh++)
  {
    /* prepare to traverse boundary faces table */
    for(iface = 0; iface < mgds->fTableLength[imesh]; iface++)
    {
      face = mgds->fTable[imesh][iface];
      for(splitface = InitSplitFace(face, &faceIter);
          splitface != NULL;
          splitface = NextSplitFace(&faceIter))
      {
        DoesFaceExist(splitface, missingFaces);
      }
    } /* end of loop over faces */
  } /* end of loop over meshes */

  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 -- Number of missing faces : %d\n", 
           Tos_MStackCount (missingFaces));
  }

} /* end of FindMissingFaces */



static void FindPointsOnEdge(Edge *edge, Tos_MStack pointsOnEdges)
{
  /* first edge point is on stack ? */
  if(!Tos_MStackSearch(pointsOnEdges, (void *)EDGE_VERTEX(edge, 0)))
  {
    Tos_MStackPush (pointsOnEdges, (void *)EDGE_VERTEX(edge, 0));
  }
  /* second edge point is on stack ? */
  if(!Tos_MStackSearch(pointsOnEdges, (void *)EDGE_VERTEX(edge, 1)))
  {
    Tos_MStackPush (pointsOnEdges, (void *)EDGE_VERTEX(edge, 1));
  }
}



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindPointsOnFaceEdges --
        For each edge of a face push points lying on that edge
        on the stack (vertices points aren't duplicated).
 Return value:
        Return number of points on the stack
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int FindPointsOnFaceEdges (Face *face, Tos_MStack pointsOnEdges)
{
  Edge  *edge, *splitedge;
  int    iside;
  struct Tos_MStack edgeIter = {0, NULL, NULL};

  assert(face != NULL);
  assert(FACE_NEIGHBOUR(face, 0) != NULL);
  assert(FACE_NEIGHBOUR(face, 1) != NULL);
  assert(FACE_NEIGHBOUR(face, 2) != NULL);

  for(iside = 0; iside < 3; iside++)
  {
    edge = FACE_NEIGHBOUR(face, iside);
    
    for(splitedge = InitSplitEdge(edge, &edgeIter);
        splitedge != NULL;
        splitedge = NextSplitEdge(&edgeIter))
    {
      FindPointsOnEdge(splitedge, pointsOnEdges);
    }
  }
  
  /* return number of points on the stack */
  return (Tos_MStackCount (pointsOnEdges));
} /* end of FindPointsOnFaceEdges */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindPointsTetrahAdjSet --
        Finds all tetrahedra which
        have at last one vertex among points from stack and push
        those tetrahedra on new stack
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FindPointsTetrahAdjSet (Tos_MStack pointsStack, Tos_MStack tetrahStack)
{
  Tetrah *tet; 
  Vertex *point;
  /* prepare to traverse points stack */
  Tos_MStackTraversalInit (pointsStack);

  while ((point = (Vertex *) Tos_MStackTraverse (pointsStack)) != NULL)
  {
     /* traverse adjacency stack at 'point' */
     Tos_MStackTraversalInit (point->adjElem);
     while ((tet = (Tetrah *) Tos_MStackTraverse (point->adjElem)) != NULL)
     {
       /* check if tetrahedron has been visited */
       /* if no push it on stack and set visited flag */
        if ((TET_FLAG (tet) & TET_PTAS) != TET_PTAS)
        {
          /* set flag TET_PTAS to mark tetrahedron visited */
          TET_FLAG (tet) |= TET_PTAS; 
          /* push tetrahedron on stack */
          Tos_MStackPush (tetrahStack, (void *) tet);
        }
     }
  }
  /* clear PTAS flag */
  Tos_MStackTraversalInit (tetrahStack);
  while ((tet = (Tetrah *) Tos_MStackTraverse (tetrahStack)) != NULL)
  {
    TET_FLAG (tet) &= ~TET_PTAS; 
  }
  return;
} /* end of FindPointsTetrahAdjSet */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindSharedNodes --
        Find nodes which are shared by tetrahedron and set of points.
 Return value:
        Number of shared nodes (0-4);
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int FindSharedNodes (Tetrah *tet, Tos_MStack pointsStack, int *nodes)
{
  register int i;
  Vertex *point;
  int cnodes;
  
  cnodes = 0;
  Tos_MStackTraversalInit (pointsStack);
  while ((point = (Vertex *) Tos_MStackTraverse (pointsStack)) != NULL)
  {
    /* This trick is to make loop a bit faster, this is equivalent to
     *  for (i=0; i<4; i++)
     */
    for (i=4; i--; )
    {
      if (TET_VERTEX (tet, i) == point)
      {
        nodes[cnodes]=i;
        cnodes++;
      }
    }
    if (cnodes==4) /* only 4 nodes can be shared as tetrah has four nodes */
    {
      break;
    }
  }
  return cnodes; 
} /* end of FindSharedNodes */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindSubdomainTetrah --
        Find tetrahedron enclosed in given subdomain.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* FIXME : this is brute force, it should be optimised,
           maybe a variation of FindBaseWalk */
Tetrah *FindSubdomainTetrah (MeshGenData *mgds, int subdomain)
{
  register int tetFound = FALSE; 
  int face, sumface;
  int where;
  Tetrah *startTet;
  Vertex subdomainPoint;

  subdomainPoint.coords[0] = mgds->mgCtrl->subdomainPoint[subdomain-1][0];
  subdomainPoint.coords[1] = mgds->mgCtrl->subdomainPoint[subdomain-1][1];
  subdomainPoint.coords[2] = mgds->mgCtrl->subdomainPoint[subdomain-1][2];

  startTet = LIST_HEAD((mgds->tetMesh).tList);

  while (tetFound != TRUE)
  {
    where = WhereInTetrah (startTet, &subdomainPoint , &face, &sumface);  
    switch (where)
    {
      case OUTSIDE: 
      {
        startTet = TET_NEIGHBOUR (startTet, face);
        break;
      }
      case INSIDE:
      case ONEDGE:
      case ONVERTEX:
      {
        tetFound = TRUE;
        break; 
      }
      case ONFACE:
      {
        e_GeneralWarning ("FindSubdomainTetrah: Point on surface\n");
        tetFound = TRUE;
        break;
      }
      default:
         e_GeneralError ("FindSubdomainTetrah: Unknown point location %d",
                          where);
    }
  } 
  return startTet;
} /* end of FindSubdomainTetrah */

/* Modifications:
 * $Log: find.c,v $
 * Revision 1.17  2002/07/26 18:54:16  peteri
 * - editing of error messages
 *
 * Revision 1.16  2002/07/26 13:20:24  peteri
 * - recursive functions are replaced with iterative functions
 *
 * Revision 1.15  2002/07/26 13:07:00  peteri
 * minor modifiactions to remove compiler warnings
 *
 * Revision 1.14  2002/07/26 11:43:47  peteri
 * - base search uses exhaustive search, (walking is not perfect now)
 * - find boundary edges and faces is modified due to new dat astructure
 *   (functions are still recursive, could be changed to iterative)
 *
 * Revision 1.13  2002/07/20 15:52:34  peteri
 * Interface of FindCavity is simplified
 *
 * Revision 1.12  2002/03/18 15:55:39  peteri
 * FindEdgeShellRank is removed, not necessary at all !
 *
 * Revision 1.11  2002/02/26 21:36:51  romek
 * Bug fixed with clearing OUTCAVITY flag.
 *
 * Revision 1.10  2002/02/24 19:01:07  romek
 * *** empty log message ***
 *
 * Revision 1.9  2002/02/17 14:58:33  romek
 * Function FindBallShell removed.
 * Changed interface to: FindBaseWalk, FindCavity,  FindConnectedPoints,
 * FindEdgeShell,FindMissingEdges.
 *
 */
