/* $Id: kernel.c,v 1.23 2002/07/29 20:46:58 peteri Exp $ */

#include <assert.h>
#include "e_lib.h"
#include "eg_lib.h"
#include "list.h"
#include "vertex.h"
#include "find.h"
#include "basic_op.h"
#include "mesh_op.h"
#include "kernel.h"
#include "check.h"
#include "debug.h"
#include "boundary.h"
#include "measures.h"


#define CLOSE_NODE  1e-6
#define ZERO        1e-8

static int InEdgeSphere(Edge *edge, double *point, double *cc);
static int InFaceSphere(Face *face, double *point, double *cc);


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 MeshParamGetAtPos --
     Calculates the mesh parameter at the given position considering
     all boundary points. It uses the weighted distance.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
double MeshParamGetAtPos(MeshGenData *mgds, double *pos)
{
  double   weight;
  int      imesh, ipoint;
  double   xp, yp, zp, dx, dy, dz, d;
  double   mpr;
  Vertex  *point;
  
  mpr = 0.0;
  weight = 0.0;

  for(imesh = 0; imesh < mgds->NMeshes; imesh++)
  {
    for(ipoint = 0; ipoint < mgds->pTableLength[imesh]; ipoint++)
    {
      point = mgds->pTable[imesh][ipoint];
      xp = point->coords[0];
      yp = point->coords[1];
      zp = point->coords[2];

      dx = xp-pos[0];
      dy = yp-pos[1];
      dz = zp-pos[2];

      if(   E_ABS(dx) < CLOSE_NODE
         && E_ABS(dy) < CLOSE_NODE
         && E_ABS(dz) < CLOSE_NODE )
      {
        return(point->meshParam);
      }
      else
      {
        d = 1.0 / (dx*dx + dy*dy + dz*dz);
        mpr += d * point->meshParam;
        weight += d;
      }
    }

    mpr /= weight;
  }

  /* check for safety */
  if(mpr < ZERO)
  {
    e_GeneralError("zero mesh parameter");
  }
  
  return(mpr);
}



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 ReconnectCavity --
     By connecting the point to the cavity it creates the new cavity.
     No internal data structure changes.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void ReconnectCavity (Tos_MStack cavity,
                      Tos_MStack newCavity,
                      Vertex *point)
{
  Tetrah *tet, *newtet, *neighbour;
  int face;

  /* go through the elements of the cavity */
  Tos_MStackTraversalInit (cavity);
  while ((tet = Tos_MStackTraverse(cavity)) != NULL)
  {
    /* got through the faces of the cavity elements */
    for (face = 0; face < 4; face++)
    {
      /* check adjacency */
      neighbour = TET_NEIGHBOUR (tet, face);
      /* face on cuboid surface or cavity surface */
      if (   neighbour == NULL
          || ((TET_FLAG (neighbour) & TET_OUTCAVITY) == TET_OUTCAVITY) )
      {
        /* create new tetrahedron */
        newtet = TetrahCreate ();
        TET_VERTEX (newtet, 0) = TET_VERTEX (tet, faceDef[face][2]);
        TET_VERTEX (newtet, 1) = TET_VERTEX (tet, faceDef[face][1]);
        TET_VERTEX (newtet, 2) = TET_VERTEX (tet, faceDef[face][0]);
        TET_VERTEX (newtet, 3) = point;

        Tos_MStackPush (newCavity, newtet);
      }
    }
  }
} /* end of ReconnectCavity */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 UpdateCavity --
     Removes the elements of the old cavity and inserts the elements
     of the new cavity
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void UpdateCavity(TetrahMesh *tetMesh,
                  Tos_MStack cavity,
                  Tos_MStack newCavity)
{
  Tetrah *tet;
  
  /* remove the old cavity elements */
  while (Tos_MStackCount (cavity) > 0)
  {
    tet = (Tetrah *)  Tos_MStackPop (cavity);
    RemoveTetFromMesh (tetMesh, tet);
    TetrahDelete (tet);
  }

  /* add the new cavity elements, adjacency is updated */
  while (Tos_MStackCount (newCavity) > 0)
  {
    tet = (Tetrah *)  Tos_MStackPop (newCavity);
    AddTetToMesh (tetMesh, tet);
  }
}


/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 DelaunayKernel --
        Insert point into existing dealunay triangulation creating
        new triangulation.

 Result :
        1  - successfull insertion
        -1 - point pushed back on waiting stack; 
        -2 - duplicate point location, insertion of point skipped
   ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
int DelaunayKernel (TetrahMesh *tetMesh, Tos_MStack waitingPoints,
                    int force)
{
  int cavok;
  int baseok;
  int result= 1;
  Vertex *point;
  static int skipLimit = 10;
  static int skipCounter = 0;
  /* these are static stacks for the procedure */
  struct Tos_MStack ibase = {0, NULL, NULL};
  struct Tos_MStack icavity = {0, NULL, NULL};
  struct Tos_MStack ioutcavStack = {0, NULL, NULL};
  struct Tos_MStack inewCavity = {0, NULL, NULL};
  Tos_MStack base = &ibase;
  Tos_MStack cavity = &icavity;
  Tos_MStack outcavStack = &ioutcavStack;
  Tos_MStack newCavity = &inewCavity;
  
  assert(Tos_MStackCount(waitingPoints) > 0);
  assert(tetMesh != NULL);
  
  point = (Vertex *) Tos_MStackPop(waitingPoints);

  /* find all tetrahedra which contains the point */
  /* baseok = FindBaseWalk (LIST_HEAD(tetMesh->tList), point, base); */
  baseok = FindBaseSearchAll(LIST_HEAD(tetMesh->tList), point, base);
  if (baseok == -1)
  {
    skipCounter++;
    if (skipCounter >= skipLimit)
    {
      e_GeneralError("DelaunayKernel: Limit of duplicated points reached...");
      exit(1);
    }
    return -2; 
  }

  /* Find cavity associated with given point 
   * cavity - set of elements whose circumballs contain the given point
   *
   * IMPORTANT: FindCavity maintains one VERY IMPORTANT stack: 
   * outcavStack. It contains elements which were flaged as TET_OUTCAVITY.
   * The contents of this stack MUST NOT be changed between calls to 
   * FindCavity and ReconnectCavity.
   *
   * FIXME: This maybe replaced if ReconnectCavity uses the stack,
   * instead of the flag ???
   */
  cavok = FindCavity (point, base, force, cavity, outcavStack);

  assert (CheckTetrahAdjacency (tetMesh) < 1);
  /* assert (CheckMeshConformity(tetMesh) == 0); */

  /* if cavity is OK or we are forced to insert */
  if(cavok == 1 || force == 1)
  {
    /* make up new cavity */
    ReconnectCavity (cavity, newCavity, point);

    assert (CheckCavity(cavity, newCavity) == 1);
    
    /* update cavity */
    UpdateCavity(tetMesh, cavity, newCavity);
    /* add point to list of points in the mesh */
    PutVertexOnListHead(tetMesh->pList, point);
    /* advance counter of points */
    (tetMesh->NMeshPoints)++;
  }
  else
  {
    /* push the point back on the waiting stack */
    Tos_MStackPush(waitingPoints, (void *)point);
    result = -1;
  }

  /* flush stacks and clear flags */
  Tos_MStackFlush (base);
  Tos_MStackFlush (cavity);
  TetrahClearFlag (outcavStack, TET_OUTCAVITY);

  return result;
} /* end of DelaunayKernel */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EmptyTriang --
        Add all points from boundaries to cuboid triangulation
        thus creating initial (so called empty) triangulation.

 Return value:
        On succes retrurns TRUE otherwise returns FALSE.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int EmptyTriang (MeshGenData *mgds)
{
  register int i;
  register int nm;
  Tos_MStack waitingPoints;

  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 -- Empty triangulation\n");
  }
  
  waitingPoints = Tos_MStackNew ();

  /* add points from each 'pTable[nm]' of boundary points to triangulation */
  for(nm = 0; nm < mgds->NMeshes; nm++)
  {
    for(i = 0; i < mgds->pTableLength[nm]; i++)
    {
      /* push point on waiting stack and call Delaunay kernel */
      Tos_MStackPush (waitingPoints, (void *) mgds->pTable[nm][i]);
      /* if it cannot insert the point, the point remains on the stack
       * to insert later by "force" */
      DelaunayKernel (&(mgds->tetMesh), waitingPoints, 0); 
    }
  }

  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 -- Points for forced insertion %d\n",
           Tos_MStackCount(waitingPoints));
  }

  /* insert points from waiting stack which haven't been inserted */
  while (Tos_MStackCount (waitingPoints) > 0)
  {
    DelaunayKernel(&(mgds->tetMesh), waitingPoints, 1); 
  }

  Tos_MStackFree(&waitingPoints);

  return TRUE;
} /* end of EmptyTriang */


/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 GoodPointSelection --
     This function generates a number of points, constructs the new
     cavity for each point, determines the minimum quality in the
     new cavity and selects the point which maximizes the minimum
     quality.
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
Vertex *GoodPointSelection(MeshGenData *mgds, TetrahMesh *tetMesh,
                           double *coords, double radius,
                           int encroachment, int optimize,
                           double ms_min, double ms_max,
                           double md_min, double md_max)
{
  Vertex  points[NMAX_POINT];
  Edge   *edges[NMAX_POINT];
  Face   *faces[NMAX_POINT];
  double  maxquality, minquality, quality;
  int     maxpoint = -1;
  int     ipoint, idim;
  double  delta, dd, sign;
  double  newcenter[3];
  struct Tos_MStack ibase = {0, NULL, NULL};
  struct Tos_MStack icavity = {0, NULL, NULL};
  struct Tos_MStack ioutcavStack = {0, NULL, NULL};
  struct Tos_MStack inewCavity = {0, NULL, NULL};
  Tos_MStack base = &ibase;
  Tos_MStack cavity = &icavity;
  Tos_MStack outcavStack = &ioutcavStack;
  Tos_MStack newCavity = &inewCavity;
  int     baseok, cavok;
  Tetrah *tet;
  Vertex *p;
  double  dist[3];
  int     cavmin, ncav, nnewcav;
  double  cavminradius, sliverradius;
  

  /* max radius of picking area */
  delta = radius * PICKING_SIZE;
  
  /* generate some random points */
  for(ipoint = 0; ipoint < NMAX_POINT; ipoint++)
  {
    for(idim = 0; idim < 3; idim++)
    {
      dd = (double)(rand()) / (double)(RAND_MAX);
      sign = (double)(rand()) / (double)(RAND_MAX);
      if(sign > 0.5) dd *= -1;
      
      points[ipoint].coords[idim] = coords[idim] + delta * dd;
    }
  
    edges[ipoint] = NULL;
    faces[ipoint] = NULL;
    if(encroachment)
    {
      if(CheckEdgeEncroachment(mgds, points[ipoint].coords,
                               newcenter, &(edges[ipoint])))
      {
        memcpy(points[ipoint].coords, newcenter, 3*sizeof (double));
      }
      else if(CheckFaceEncroachment(mgds, points[ipoint].coords, newcenter,
                                    &(faces[ipoint]), &(edges[ipoint])))
      {
        assert(edges[ipoint] != NULL || faces[ipoint] != NULL);
        memcpy(points[ipoint].coords, newcenter, 3*sizeof (double));
      }
    }
  }

  /* check quality */
  for(ipoint = 0; ipoint < NMAX_POINT; ipoint++)
  {
    /* if point falls out of cuboid */
    if(   points[ipoint].coords[0] < mgds->low[0]
       || points[ipoint].coords[0] > mgds->high[0]
       || points[ipoint].coords[1] < mgds->low[1]
       || points[ipoint].coords[1] > mgds->high[1]
       || points[ipoint].coords[2] < mgds->low[2]
       || points[ipoint].coords[2] > mgds->high[2])
    {
      continue;
    }

    baseok = FindBaseSearchAll(LIST_HEAD(tetMesh->tList),
                               &(points[ipoint]), base);
    if (baseok == -1)
    {
      continue;
    }
  
    cavok = FindCavity (&(points[ipoint]), base, 0, cavity, outcavStack);
    if(cavok != 1)
    {
      continue;
    }
    
    /* connect up new cavity */
    ReconnectCavity(cavity, newCavity, &(points[ipoint]));

    /* if optimization, get minimum sliver radius */
    if(optimize)
    {
      ncav = 0;
      cavmin = -1;
      cavminradius = 0.0;
      Tos_MStackTraversalInit(cavity);
      while((tet = (Tetrah *)Tos_MStackTraverse(cavity)) != NULL)
      {
        if(tet->flag & TET_SLIVER)
        {
          eg_TetCircumcenter (TET_VERTEX(tet, 0)->coords,
                              TET_VERTEX(tet, 1)->coords,
                              TET_VERTEX(tet, 2)->coords,
                              TET_VERTEX(tet, 3)->coords,
                              dist, NULL, NULL, NULL);
          sliverradius = sqrt(dist[0] * dist[0] +
                              dist[1] * dist[1] +
                              dist[2] * dist[2]  );
          if(cavmin == -1 || sliverradius < cavminradius)
          {
            cavminradius = sliverradius;
          }
        }
      }
    }
    
    /* get worst quality in the new cavity */
    minquality = 1000;
    Tos_MStackTraversalInit(newCavity);
    while((tet = (Tetrah *)Tos_MStackTraverse(newCavity)) != NULL)
    {
      quality = Tetrah_Measure(tet, SIGMA_MEASURE);
      if(quality < minquality)
      {
        minquality = quality;
      }

      /* if optimization, check the radius of circum-sphere */
      if(optimize)
      {
        TetrahSetTypeFlag(tet, ms_min, ms_max, md_min, md_max);
        if(tet->flag & TET_SLIVER)
        {
          eg_TetCircumcenter (TET_VERTEX(tet, 0)->coords,
                              TET_VERTEX(tet, 1)->coords,
                              TET_VERTEX(tet, 2)->coords,
                              TET_VERTEX(tet, 3)->coords,
                              dist, NULL, NULL, NULL);
          sliverradius = sqrt(dist[0] * dist[0] +
                              dist[1] * dist[1] +
                              dist[2] * dist[2]  );
          /* if the radius of the created sliver is smaller than
             some percentage of the minimum radius of the original cavity
             then the point is skipped */
          if(sliverradius < cavminradius * 0.8)
          {
            continue;
          }
        }
      }
      
    }
    assert(fabs(minquality - 1000) > 1e-8);
    
    if(   maxpoint == -1
       || minquality > maxquality)
    {
      maxquality = minquality;
      maxpoint   = ipoint;
    }
    
    /* reset/clear data structure */
    while(Tos_MStackCount(newCavity) > 0)
    {
      tet = (Tetrah *)Tos_MStackPop(newCavity);
      TetrahDelete(tet);
    }
    Tos_MStackFlush (base);
    Tos_MStackFlush (cavity);
    TetrahClearFlag (outcavStack, TET_OUTCAVITY);
  }
  
  /* this means, there was always a cavity or base problem,
     so it never found any point to check for quality OR
     when we optimize the maximum quality must be higher than 0.01
  */
  if(   (maxpoint == -1)
     || (maxquality < 0.00001)
     || (optimize == 1 && maxquality < 0.01))
  {
    return NULL;
  }

  if(vm_verbose_level >= 5)
  {
    printf("+ debug 5 -- %d Insertion quality: %f\n", optimize, maxquality);
  }

  /* create the point */
  p = VertexCreate ();
  memcpy (p->coords, points[maxpoint].coords, 3*sizeof (double));
  /* check point position */
  assert(p->coords[0] > mgds->low[0] && p->coords[0] < mgds->high[0]);
  assert(p->coords[1] > mgds->low[1] && p->coords[1] < mgds->high[1]);
  assert(p->coords[2] > mgds->low[2] && p->coords[2] < mgds->high[2]);
  /* the actual mesh parameter */
  p->meshParam = MeshParamGetAtPos(mgds, p->coords);

  /* if there was encroachment, insert point into the boundary structure */
  if(edges[maxpoint] != NULL)
  {
    SplitBoundaryEdge(mgds, edges[maxpoint], p);
  }
  else if(faces[maxpoint] != NULL)
  {
    SplitBoundaryFace(mgds, faces[maxpoint], p);
  }
  
  return(p);
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 VolumeTriang --
     Refines the initial tetrahedra mesh.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int VolumeTriang(int method, MeshGenData *data, TetrahMesh *tetMesh,
                 double alpha, double beta, int mode)
{
  int         iedge;
  double      s; /* distances */
  double      cpMeshParam;
  double      minMeshParam;
  double      radius;
  double      center[3];
  char        reject = 0;
  char        accept = 0;
  int         internalPoints = 0;
  Vertex     *v[4]; /* tetrahedron vertices */
  Vertex     *p, *prevPoint;
  Tetrah     *tet;
  Tos_MStack  waitingPoints;
  int         numberOfPasses = 0;
  int         ntet;
  
  waitingPoints = Tos_MStackNew();

  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 -- VolumeTriang\n");
  }

  /* initialise random number generation */
  srand(246234);
  /* CONSTCOND */
  while (1)
  {
    numberOfPasses++;
    ntet = 0;
    
    tet = LIST_HEAD(tetMesh->tList);
    while (tet != NULL)
    {
      ntet++;
      /* check if any of tet vertices is a cuboid point - if so, skip this
       * tetrahedron
       */
      if(   ((tet->flag & TET_FINISHED) != 0)
         || ((TET_VERTEX (tet, 0)->flag & POINT_CUBOID)  == POINT_CUBOID)
         || ((TET_VERTEX (tet, 1)->flag & POINT_CUBOID)  == POINT_CUBOID)
         || ((TET_VERTEX (tet, 2)->flag & POINT_CUBOID)  == POINT_CUBOID)
         || ((TET_VERTEX (tet, 3)->flag & POINT_CUBOID)  == POINT_CUBOID) )
      {
        tet = TET_PREV(tet);
        continue;
      }
      
      /* get tet vertices */
      v[0] = TET_VERTEX(tet, 0);
      v[1] = TET_VERTEX(tet, 1);
      v[2] = TET_VERTEX(tet, 2);
      v[3] = TET_VERTEX(tet, 3);
      
      /* Check whether tet should be refined */
      accept=0;
      switch(method)
      {
        case VM_REFINE_EDGE_LEN:
        {
          for(iedge = 0; iedge < 6; iedge++)
          {
            s = eg_PointDist3D(v[edgeDef[iedge][0]]->coords,
                               v[edgeDef[iedge][1]]->coords);
            minMeshParam = E_MIN (v[edgeDef[iedge][0]]->meshParam, 
                                  v[edgeDef[iedge][1]]->meshParam);
            if (s > alpha * minMeshParam)
            {
              accept = 1;
              break;
            }
          }
          break;
        }
        default:
        {
          e_GeneralError("Invalid refinement method !!!");
          exit(1);
        }
      }
      
      
      /* this tetrahedron does not have to be refined, continue */
      if(accept == 0)
      {
        tet->flag |= TET_FINISHED;
        tet = TET_PREV(tet);
        continue;
      }

      /* the inserted point is the centre of the circum-sphere
         of the tetrahedron !!! */
      eg_TetCircumcenter (v[0]->coords, v[1]->coords,
                          v[2]->coords, v[3]->coords,
                          center, NULL, NULL, NULL);
      radius = sqrt(center[0]*center[0] +
                    center[1]*center[1] +
                    center[2]*center[2]   );
      center[0] += v[0]->coords[0];
      center[1] += v[0]->coords[1];
      center[2] += v[0]->coords[2];

      if(mode & USE_GOODPOINT)
      {
        /* select a good point */
        p = GoodPointSelection(data, tetMesh, center, radius,
                               mode & USE_ENCROACHMENT,
                               0, 0.0, 0.0, 0.0, 0.0);
        if(p == NULL)
        {
          /* in this case it is only finished, since no point could be
             found, so lets hope it will be refined by other insertion */
          tet->flag |= TET_FINISHED;
          tet = TET_PREV(tet);
          continue;
        }
      }
      /* select the circumcentre of the sphere */
      else
      {
        /* create point */
        p = VertexCreate ();
        memcpy (p->coords, center, 3*sizeof (double));

        /* the actual mesh parameter */
        p->meshParam = MeshParamGetAtPos(data, p->coords);
        /* check point position */
        assert(p->coords[0] > data->low[0] && p->coords[0] < data->high[0]);
        assert(p->coords[1] > data->low[1] && p->coords[1] < data->high[1]);
        assert(p->coords[2] > data->low[2] && p->coords[2] < data->high[2]);
      }
      
      
      if(mode & USE_BETA)
      {
        assert(beta > 0);
        
        /* second check with beta parameter,
           the inserted points should not be close to each other */
        reject=0;
        
        Tos_MStackTraversalInit (waitingPoints);
        while((prevPoint = (Vertex *)Tos_MStackTraverse(waitingPoints))
              != NULL)
        {
          s = eg_PointDist3D (prevPoint->coords, center);
          if (s < beta * cpMeshParam)
          {
            reject = 1;
            break; 
          }
        }
        
        if (reject == 1)
        {
          /* delete the point */
          VertexDelete(p);
          
          tet = TET_PREV(tet);
          continue;
        }
      }

      /* store the point for insertion */
      Tos_MStackPush (waitingPoints, (void *)p);

      if(mode & USE_BETA)
      {
        /* other tetrahedron meybe refined */
        tet = TET_PREV(tet);
      }
      else
      {
        if(vm_verbose_level >= 6)
        {
          printf("+ debug 6 -- total: %d selected: %d\n",
                 tetMesh->NTetrah, ntet);
        }
        /* no beta parameter, so only one point is on the waiting stack */
        break;
      }
    }
    
    if((Tos_MStackCount (waitingPoints) == 0))
    {
      break; /* refinement finished */
    }
    else
    {
      /* insert points from waiting stack into triangulation */
      while (Tos_MStackCount (waitingPoints) > 0)
      {
        DelaunayKernel (tetMesh, waitingPoints, 1);
        internalPoints++;
      }
    }
  }

  Tos_MStackFree (&waitingPoints);

  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 -- Number of refinement points: %d\n",
           internalPoints);
  }
  
  if(vm_verbose_level >= 3)
  {
    printf("+ debug 3 -- Number of passes: %d\n", numberOfPasses);
  }
  
  return internalPoints;  
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 InEdgeSphere --
     Checks whether the point is in the circum-sphere of the edge
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int InEdgeSphere(Edge *edge, double *point, double *cc)
{
  double rr, dd, shake;
  Vertex *v1;
  Vertex *v2;

  v1 = EDGE_VERTEX(edge, 0);
  v2 = EDGE_VERTEX(edge, 1);

  shake = 2.0 * ((double)(rand()) / (double)(RAND_MAX)) - 1.0;
  
  /* centre of the edge */
  cc[0] = (VERTEX_COORDS(v1)[0] + VERTEX_COORDS(v2)[0]) * 0.5;
  cc[1] = (VERTEX_COORDS(v1)[1] + VERTEX_COORDS(v2)[1]) * 0.5;
  cc[2] = (VERTEX_COORDS(v1)[2] + VERTEX_COORDS(v2)[2]) * 0.5;
  
  /* half length of the edge, radius of diametric sphere */
  rr = 0.5 * eg_PointDist3D(VERTEX_COORDS(v1), VERTEX_COORDS(v2));

  /* distance between the point and the centre of the sphere */
  dd = eg_PointDist3D(point, cc);

  /* point is inside the sphere */
  if(dd < rr)
  {
    
    cc[0] = VERTEX_COORDS(v1)[0] +
        (0.5 + PICKING_SIZE * shake) *
        (VERTEX_COORDS(v2)[0] - VERTEX_COORDS(v1)[0]);
    cc[1] = VERTEX_COORDS(v1)[1] +
        (0.5 + PICKING_SIZE * shake) *
        (VERTEX_COORDS(v2)[1] - VERTEX_COORDS(v1)[1]);
    cc[2] = VERTEX_COORDS(v1)[2] +
        (0.5 + PICKING_SIZE * shake) *
        (VERTEX_COORDS(v2)[2] - VERTEX_COORDS(v1)[2]);
    
    return 1;
  }

  return 0;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 CheckEdgeEncroachment --
     Checks whether the point is in the circum-sphere of any of the
     boundary edges. If yes, creates a new coordinate (middle point
     of the edge) and returns it with the edge as well.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int CheckEdgeEncroachment(MeshGenData *data,
                          double *coords, double *newcoords,
                          Edge **encedge)
{
  int     imesh, iedge;
  Edge   *edge, *splitedge;
  struct Tos_MStack edgeIter = {0, NULL, NULL};
  
  *encedge = NULL;
  for(imesh = 0; imesh < data->NMeshes; imesh++)
  {
    for(iedge = 0; iedge < data->eTableLength[imesh]; iedge++)
    {
      edge = data->eTable[imesh][iedge];
      for(splitedge = InitSplitEdge(edge, &edgeIter);
          splitedge != NULL;
          splitedge = NextSplitEdge(&edgeIter))
      {
        if(InEdgeSphere(splitedge, coords, newcoords))
        {
          /* printf("Edge enc. edge %d\n", iedge); */
          *encedge = splitedge;
          Tos_MStackFlush(&edgeIter);
          return 1;
        }
      }
    }
  }
  Tos_MStackFlush(&edgeIter);
  return 0;
}


static int InFaceSphere(Face *face, double *point, double *cc)
{
  Vertex *v1;
  Vertex *v2;
  Vertex *v3;
  double dv[3];
  double dd, rr;
  
  v1 = FACE_VERTEX(face, 0);
  v2 = FACE_VERTEX(face, 1);
  v3 = FACE_VERTEX(face, 2);
  
  /* centre of circum-sphere */
  eg_TriCircumcenter3D(VERTEX_COORDS(v1),
                       VERTEX_COORDS(v2),
                       VERTEX_COORDS(v3),
                       dv, NULL, NULL);
  
  cc[0] = dv[0] + v1->coords[0];
  cc[1] = dv[1] + v1->coords[1];
  cc[2] = dv[2] + v1->coords[2];
  
  /* radius of circum-sphere */
  rr = sqrt( dv[0]*dv[0] + dv[1]*dv[1] + dv[2]*dv[2] );

  /* distance between the point and the centre of the sphere */
  dd = eg_PointDist3D(point, cc);

  /* point is inside the sphere */
  if(dd < rr)
  {
    return 1;
  }
  
  return 0;
}


/*
 * Returns: 0 - not in sphere
 *          1 - in edge sphere
 *          2 - in face sphere
 */
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 CheckFaceEncroachment --
     Checks whether the point is in the circum-sphere of any of the
     boundary faces. If yes, creates a new coordinate (middle point
     of the face). Then check whether this new point is in the circum-
     sphere of any of the boundary edges which are not the edge of the
     face. If yes, the new point is the middle point of the edge.
     It returns a new coordinate and either an edge of face.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int CheckFaceEncroachment(MeshGenData *data,
                          double *coords, double *newcoords,
                          Face **encface, Edge **encedge)
{
  int     imesh, iface, jmesh, jedge;
  Face   *face, *splitface;
  Edge   *edge, *splitedge;
  double  cc[3];
  struct Tos_MStack faceIter = {0, NULL, NULL};
  struct Tos_MStack edgeIter = {0, NULL, NULL};
  
  for(imesh = 0; imesh < data->NMeshes; imesh++)
  {
    for(iface = 0; iface < data->fTableLength[imesh]; iface++)
    {
      face = data->fTable[imesh][iface];
      
      for(splitface = InitSplitFace(face, &faceIter);
          splitface != NULL;
          splitface = NextSplitFace(&faceIter))
      {
        if(InFaceSphere(splitface, coords, newcoords))
        {
          assert(FACE_NEIGHBOUR(splitface, 0) != NULL);
          assert(FACE_NEIGHBOUR(splitface, 1) != NULL);
          assert(FACE_NEIGHBOUR(splitface, 2) != NULL);
          
          cc[0] = newcoords[0];
          cc[1] = newcoords[1];
          cc[2] = newcoords[2];
        
          /* check whether new point encroaches on any 'other' segment*/
          for(jmesh = 0; jmesh < data->NMeshes; jmesh++)
          {
            for(jedge = 0; jedge < data->eTableLength[jmesh]; jedge++)
            {
              edge = data->eTable[jmesh][jedge];
              
              for(splitedge = InitSplitEdge(edge, &edgeIter);
                  splitedge != NULL;
                  splitedge = NextSplitEdge(&edgeIter))
              {
                /* except triangle edges */
                if(   (splitface->neighbour[0] != splitedge)
                   && (splitface->neighbour[1] != splitedge)
                   && (splitface->neighbour[2] != splitedge) )
                {
                  int sidenum;
                  /* ensure that adjacency information is not corrupt */
                  assert(FaceHasEdge(splitface, splitedge, &sidenum) == 0);
                  
                  if(InEdgeSphere(splitedge, cc, newcoords))
                  {
                    /* printf("Face enc. edge %d\n", jedge); */
                    *encface = NULL;
                    *encedge = splitedge;
                    
                    Tos_MStackFlush(&faceIter);
                    Tos_MStackFlush(&edgeIter);
                    
                    return 1;
                  }
                }
              }
            }
          }
          /* printf("Face enc. face %d\n", iface); */

          /* FIXME: Face encroachment is not enabled
           * The problem with face encroachment is, that the centre of the
           * circum-sphere of the face may fall outside of the face. In
           * that case the point may be or may not be on the boundary surface.
           * So another iteration is required, where all surfaces are checked
           * with the inserted point whether the point is on the surface.
           * However in the case of curved surface the precision problem
           * strikes again. :-)
           */
          
          /* *encface = splitface; */
          *encface = NULL;
          *encedge = NULL;
          
          Tos_MStackFlush(&faceIter);
          Tos_MStackFlush(&edgeIter);

          /* return 2; */
          return 0;
          
        }
      }
    }
  }
  
  *encface = NULL;
  *encedge = NULL;
  
  Tos_MStackFlush(&faceIter);
  Tos_MStackFlush(&edgeIter);

  return 0;
}



/* Modifications:
 * $Log: kernel.c,v $
 * Revision 1.23  2002/07/29 20:46:58  peteri
 * - once an element is examined and no further refinement is necessary
 *   then no further calculation is produced for this element, since it is
 *   flagged and skipped
 *
 * Revision 1.22  2002/07/28 16:13:57  peteri
 * - the point insertion for optimization has changed,
 *   it checks for slivers and there is quality threshold,
 *   though this checks are a little bit ad hoc.
 *
 * Revision 1.21  2002/07/28 13:37:16  peteri
 * - order of refinement of tetrahedra is reversed,
 *   new elements are put on the head of the list,
 *   the refinement is considered from the head
 *
 * Revision 1.20  2002/07/28 13:27:39  peteri
 * - list creation interface is modified
 * - it is possible to add to the head or the tail of the list
 * - all related code is updated
 *
 * Revision 1.19  2002/07/28 12:50:17  peteri
 * - when point falls outside of cuboid, skip it
 * - when there is no point which is "good" skip that tetrahedron
 *
 * Revision 1.18  2002/07/26 13:07:01  peteri
 * minor modifiactions to remove compiler warnings
 *
 * Revision 1.17  2002/07/26 11:47:09  peteri
 * - reconnect cavity is split further
 * - "good-point-selection" is implemented
 * - encroachment is implemented,
 *   though face encroachment requires further checking,
 *   so it is switched off at the moment
 *
 * Revision 1.16  2002/07/20 16:56:07  peteri
 * Correct mesh parameter calculation
 * It uses all boundary points to determine the mesh parameter
 *
 * Revision 1.15  2002/07/20 16:18:22  peteri
 * - Delaunay kernel is much more simplified
 * - internal stacks are static and local to the functions
 * - encroachment is moved to optimization
 *
 * Revision 1.14  2002/07/19 21:02:49  peteri
 * - Interface to volume meshing has changed
 * - Only edge length refinement is valid now
 * - A new point is inserted into the centre of the circum-sphere
 * - Other simplifications, code clean-up
 *
 * Revision 1.13  2002/03/18 15:57:46  peteri
 * More comments, some clean-up
 *
 * Revision 1.12  2002/02/27 22:56:20  romek
 * Fixed bug with Nana configuration.
 *
 * Revision 1.11  2002/02/26 21:37:05  romek
 * Bug fixed with clearing OUTCAVITY flag.
 *
 * Revision 1.10  2002/02/24 19:04:29  romek
 * *** empty log message ***
 *
 */
