/* $Id: meshgen.c,v 1.23 2002/07/28 16:14:00 peteri Exp $ */

#include <assert.h>
#include "e_lib.h"
#include "eg_lib.h"

#include "boundary.h"
#include "bbox.h"
#include "kernel.h"
#include "recover.h"
#include "optimize.h"
#include "subdomain.h"
#include "convert.h"
#include "meshgen.h"
#include "measures.h"

#include "write_log.h"
#include "write_dx.h"
#include "check.h"
#include "debug.h"

/* colors for debug messages for terminals which support colors */
#define TOS_FOREGROUND   "[00m"
#define TOS_RED          "[31m"
#define TOS_GREEN        "[32m"
#define TOS_BRIGHT_RED   "[01;31m"
#define TOS_BRIGHT_GREEN "[01;32m"



MESH *vm_GenerateVolumeMESH(MeshGenData *mgds)
{
  MESH   *outMesh;
  int     counter;
  double  mindangle, maxdangle;

  Tos_LogHeader(mgds);
  Tos_LogNInputMeshes (mgds);
  Tos_LogNInputMeshNames (mgds); 
  Tos_LogNInputNEF (mgds); 
  Tos_LogNTotalInputNEF (mgds); 
  Tos_LogInputMeshParam (mgds);
  Tos_InputEdgeLengthDistribution (mgds); 

  /* create bounding cuboid */
  BoundingBox (mgds->pTableLength[0], mgds->pTable[0],
               mgds->low, mgds->high);
  MakeCuboid (&(mgds->tetMesh), mgds->low, mgds->high);

  Tos_LogBoundingBox (mgds, mgds->low, mgds->high);

  /* create initial/empty tetrahedra mesh */
  EmptyTriang (mgds);

  Tos_LogNTetEmptyTriang (mgds);
  Tos_LogShapeEmptyTriang (mgds);
  Tos_LogShapeCoreEmptyTriang (mgds);
  Tos_LogNTetCoreEmptyTriang (mgds);

  if(vm_verbose_level >=1)
  {
    printf("%s+ debug 1 -- INITIAL TRIANGULATION FINISHED %s\n",
           TOS_BRIGHT_RED, TOS_FOREGROUND);
  }
  
  
#ifdef DEBUG_ADDCHECK
  printf("+ debug A --  Additional checking, it may take a while...");
  CheckCoplanarity (&(mgds->tetMesh));
  CheckTetrahAdjacency (&(mgds->tetMesh));
  CheckMeshConformity (&(mgds->tetMesh)); 
  printf(" OK\n");
#endif

  /* remesh/refine the initial tetrahedra mesh */
  counter = VolumeTriang (mgds->mgCtrl->refineMethod,
                          mgds, &(mgds->tetMesh),
                          mgds->mgCtrl->alpha,
                          mgds->mgCtrl->beta,
                          USE_GOODPOINT | USE_ENCROACHMENT);

  Tos_LogNPointsVolumeTriang (mgds, counter);
  Tos_LogNTetVolumeTriang (mgds);
  Tos_LogNTetCoreVolumeTriang (mgds);
  Tos_LogShapeVolumeTriang (mgds);
  Tos_LogShapeCoreVolumeTriang (mgds);


#ifdef DEBUG_ADDCHECK
  printf("debug A --  Additional checking, it may take a while...");
  CheckCoplanarity (&(mgds->tetMesh));
  CheckTetrahAdjacency (&(mgds->tetMesh));
  CheckMeshConformity (&(mgds->tetMesh)); 
  printf(" OK\n");
#endif 

  if(vm_verbose_level>=1)
  {
    printf("%s+ debug 1 -- VOLUME TRIANGULATION FINISHED%s\n",
           TOS_BRIGHT_RED, TOS_FOREGROUND);

    printf("----------------------------------------\n");
    Mesh_MinAndMaxDihedralAngle(&(mgds->tetMesh), &mindangle, &maxdangle);
    printf("Minimum dihedral angle: %.2f degree\n"
           "Maximum dihedral angle: %.2f degree\n\n",
           mindangle * 180 / M_PI, maxdangle * 180 / M_PI);
  }
    
  /* optimize the mesh, sliver and bad element removal */
  vm_OptimizeMesh (mgds, 1);
  
  /* restore boundary edges and surfaces */
  EnsureBoundaryConformity (mgds); 

  Tos_LogSplitPointsDistribution (mgds);

  assert(CheckSurfaceCompletness(mgds) == 0);
  
  if(vm_verbose_level >= 1)
  {
    printf("%s+ debug 1 -- BOUNDARY RECOVERY FINISHED%s\n",
           TOS_BRIGHT_RED, TOS_FOREGROUND);
  }
  
#ifdef DEBUG_ADDCHECK
  printf(" + debug A --  Additional checking, it may take a while...");
  CheckTetrahAdjacency (&(mgds->tetMesh));
  CheckMeshConformity (&(mgds->tetMesh)); 
  CheckSurfaceCompletness (mgds);  
  printf(" OK\n");
#endif

  ExtractMeshFromCuboid (mgds); 
  
  if(vm_verbose_level >= 1)
  {
    printf("%s+ debug 1 -- DOMAIN RECOVERY FINISHED%s\n",
           TOS_BRIGHT_RED, TOS_FOREGROUND);
  }
  
  RecoverSubdomains(mgds);

  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 --  subdomain recovery finished \n");
  }
  

  /* FIXME: change it for the new data structure */
#if 0
  counter = UnsplitEdges (mgds);  

  Tos_LogUnsplitedEdges (mgds, counter); 

  counter = UnsplitFaces (mgds); 

  Tos_LogUnsplitedFaces (mgds, counter); 

  assert (CheckTetrahAdjacency (&(mgds->tetMesh)) == 0);
#endif

  /* optimize without Delaunay property */
  vm_OptimizeMesh (mgds, 0);

  if(vm_verbose_level>=1)
  {
    printf("%s+ debug 1 -- MESH OPTIMIZATION FINISHED%s\n",
           TOS_BRIGHT_RED, TOS_FOREGROUND);

    printf("----------------------------------------\n");
    Mesh_MinAndMaxDihedralAngle(&(mgds->tetMesh), &mindangle, &maxdangle);
    printf("Minimum dihedral angle: %.2f degree\n"
           "Maximum dihedral angle: %.2f degree\n\n",
           mindangle * 180 / M_PI, maxdangle * 180 / M_PI);
  }

  /* write out the mesh */
  outMesh = vm_GetMdfOutputMesh (mgds); 

  if(vm_verbose_level>=1)
  {
    printf("%s+ debug 1 -- MESH GENERATION FINISHED%s\n",
           TOS_BRIGHT_GREEN, TOS_FOREGROUND);
  }
  
  Tos_LogSplitedFaces (mgds); 
  Tos_LogNTotalTetrah (mgds);
  Tos_LogNTotalMeshPoints (mgds);
  Tos_LogNAddedMeshPoints (mgds);
  Tos_LogShapeFinalTriang (mgds);
  Tos_LogFinalVolumeDistribution (mgds);
  Tos_LogDihedralAngle(mgds);
  
  Tos_OutputDXShape (mgds);
  Tos_OutputOuterSurfaceDX (mgds);
  
  return outMesh;
} /* end of vm_GenerateMESH */ 



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 vm_CreateMeshGenData --
        Create and initialize data structure for mesh generator.

 Return value:
        On succes returns pointer to new mesh generator data
        structure otherwise returns NULL.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
MeshGenData *vm_CreateMeshGenData(MeshGenCtrl *mgCtrl)
{
  int          i;
  MeshGenData *mgds = NULL;

  assert(mgCtrl->nmeshes <= VM_MAX_NMESHES);
  assert(mgCtrl->nmeshes > 0);
  
  /* call initialization */
  vm_Init();

  /* create new MeshGenData structure */
  mgds = (MeshGenData *)e_calloc(1, sizeof (MeshGenData));
  if(mgds == NULL)
  {
     return NULL;
  }

  /* set pointer to control structure */  
  mgds->mgCtrl = mgCtrl;
  mgds->NMeshes = mgCtrl->nmeshes;
  
  /* Initialize structure of tetrahedral mesh */
  vm_InitTetrahMesh (&(mgds->tetMesh));

  /* other initializations */
  /* initialize tables of points, edges and faces of input meshes */
  /* IMPORTANT : check if mesh parameters has been set in input meshes */
  for(i = 0; i < mgds->NMeshes; i++)
  {
    mgds->fTable[i] = NULL;
    mgds->eTable[i] = NULL;
    mgds->pTable[i] = NULL;
    mgds->fTableLength[i] = 0;
    mgds->eTableLength[i] = 0;
    mgds->pTableLength[i] = 0;
    if(mgCtrl->inputMeshes[i]->SubStructSet[E_SS_NUM_MESHPARAM] != TRUE)
    {
      e_GeneralError("Input mesh no %d has no mesh parameters data\n", i+1);
    }
  }
  
  /* transform boundary representation */
  GetBoundaryRep(mgds);  

  assert(CheckAlphaParam(mgds) == 1);
  
  return mgds;
} /* end of vm_CreateMeshGenData */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 vm_InitTetrahMesh --
        Initializes terahedral mehs data structure. 

 Argumens:       
        tetMeshPtr - pointer to structure to be initialized
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void vm_InitTetrahMesh (TetrahMesh *tetMeshPtr)
{
  /* set number of meshes, nodes and tetrah */
  tetMeshPtr->NMeshPoints = 0;
  tetMeshPtr->NTetrah = 0;

  /* initialize lists of tetrah, edges, faces and points */  
  tetMeshPtr->tList = CreateTetList();
  tetMeshPtr->pList = CreateVertexList();

} /* end of vm_InitTetrahMesh */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 vm_Init --
        Tos library initialization function.
        It should be called before any other function form Tos
        library is called. It is safe to call that function many
        time. Any but the first call will be ignored.
 Return value:
        On succes returns TRUE otherwise return FALSE
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int vm_Init(void)
{
  static int initialized = 0;
  if (initialized == 0)
  {        
    initialized = 1;

    TetrahMemPoolInit();
    VertexMemPoolInit();
    EdgeMemPoolInit();
    FaceMemPoolInit();
    MStackMemPoolInit();
  
    /* Initialize library of geometric routines */
    eg_Init(); 
    eg_SetTolerance (EG_ZERO_DIST_TOL,1.0e-3); 
    eg_SetTolerance (EG_ZERO_AREA_TOL,1.0e-3); 
    eg_SetTolerance (EG_PAR_LINE_TOL,1.0e-4); 
  }
  return TRUE;
} /* end of vm_Init */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 vm_GetMdfOutputMesh --
        Creates output a MESH data structure from generated tetrahedral
        mesh.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
MESH *vm_GetMdfOutputMesh (MeshGenData *mgds)
{
  MESH *mesh;
  
  mesh = e_calloc (1, sizeof (MESH)); 
  if (mesh == NULL)
  {
    e_GeneralError ("GetMdfOutputMesh: "
                    "Cannot allocate memory for MESH structure");
  }

  /* fill MESH structure */
  TetrahMeshToMESH (&(mgds->tetMesh), mesh);

  return mesh;
} /* end of vm_GetMdfOutputMesh */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 vm_CreateMeshGenCtrl --
        Allocate memory for new MeshGenCtrl data structure and initialize
        it with default values.
 Return value:
        Pointer to allocated structure.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
MeshGenCtrl *vm_CreateMeshGenCtrl (void)
{
  int i;
  MeshGenCtrl *mgCtrl;
  mgCtrl = (MeshGenCtrl *) e_calloc (1, sizeof (MeshGenCtrl));
  if (mgCtrl == NULL)
  {
    return NULL;
  }  
  mgCtrl->alpha                = 1.0;
  mgCtrl->beta                 = 1.0;
  mgCtrl->includeEmbededMeshes = 0;
  mgCtrl->refineMethod         = VM_REFINE_EDGE_LEN;
  mgCtrl->nmeshes              = 0;
  mgCtrl->outputFileName       = NULL;
  mgCtrl->ctrlFileName         = NULL;
  mgCtrl->logFileName          = NULL;
  mgCtrl->logFile              = NULL;
  mgCtrl->logLevel             = VM_NO_LOG;
  
  for (i=0; i<VM_MAX_NMESHES; i++)
     mgCtrl->inputFileNames[i]= NULL;
  mgCtrl->nsubdomains=0;

  return mgCtrl;
} /* end of vm_CreateMeshGenCtrl */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 vm_DeleteMeshGenCtrl --
        Free memory allocated for Tetrah data structure and sets
        pointer pointing to that memory to NULL.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void vm_DeleteMeshGenCtrl (MeshGenCtrl *mgCtrl)
{
  int i;
  for (i=0; i<mgCtrl->nmeshes; i++)
  {
     e_free (mgCtrl->inputFileNames[i]);
  }
  e_free (mgCtrl->outputFileName);
  e_free (mgCtrl->ctrlFileName);
  e_free (mgCtrl->logFileName);

  e_free (mgCtrl);
} /* end of vm_DeleteMeshGenCtrl */

/* Modifications:
 * $Log: meshgen.c,v $
 * Revision 1.23  2002/07/28 16:14:00  peteri
 * - the point insertion for optimization has changed,
 *   it checks for slivers and there is quality threshold,
 *   though this checks are a little bit ad hoc.
 *
 * Revision 1.22  2002/07/28 13:27:43  peteri
 * - list creation interface is modified
 * - it is possible to add to the head or the tail of the list
 * - all related code is updated
 *
 * Revision 1.21  2002/07/26 18:56:21  peteri
 * - adding the printing of the distribution of dihedral angles
 * - editing of error messages
 *
 * Revision 1.20  2002/07/26 15:48:43  peteri
 * - checking the alpha parameter of the Delaunay process is
 *   integrated
 *
 * Revision 1.19  2002/07/26 14:59:47  peteri
 * - mesh data is separated from boundary data
 * - now it is possible to setup the boundary data separately
 *   and refine that boundary (without having a MESH)
 *
 * Revision 1.18  2002/07/26 13:07:04  peteri
 * minor modifiactions to remove compiler warnings
 *
 * Revision 1.17  2002/07/26 11:56:07  peteri
 * - nana is removed for simplicity
 * - some editing and formating
 * - minor algorithm reorganisation
 *
 * Revision 1.16  2002/07/20 16:24:37  peteri
 * - Bounding box size is stored with the mesh, so centre of circum-sphere
 *   can be checked against it !
 *
 * Revision 1.15  2002/07/19 21:01:24  peteri
 * - Mempool removed for simplicity
 * - Interface to volume meshing has changed
 *
 * Revision 1.14  2002/07/16 15:53:26  peteri
 * vm_OptimizeMesh parameters are simplified
 *
 * Revision 1.13  2002/07/03 21:28:09  peteri
 * Quality is printed before and after optimization
 *
 * Revision 1.12  2002/03/18 16:01:01  peteri
 * Some renaming, simple modifications.
 *
 * Revision 1.11  2002/02/26 12:04:32  romek
 * *** empty log message ***
 *
 * Revision 1.10  2002/02/24 19:08:27  romek
 * *** empty log message ***
 *
 */
