#include <assert.h>
#include "eg_lib.h"
#include "edge.h"
#include "face.h"
#include "tetrah.h"
#include "basic_op.h"
#include "find.h"
#include "mesh_op.h"
#include "measures.h"
#include "recover.h"
#include "optimize.h"
#include "flip.h"
#include "check.h"
#include "debug.h"
#include "kernel.h"

#ifdef WIN32
extern double copysign(double x, double y);
#endif

static int vm_OptimizePointBall (Vertex *org, Vertex *dest, int measureType);
static void vm_LaplacianSmoothing (TetrahMesh *tetMesh, int max_pass, 
                                   int measureType);
static int vm_OptimizeBySwapping(TetrahMesh *tetMesh, int max_pass,
                                 int measureType);
static int vm_InsertForQuality(MeshGenData *mgds);

static int vm_OptimizeByCollapsing (TetrahMesh *tetMesh, 
                                    enum MeasureType measure);


static int CheckEdgeCollapse (Vertex *org, Vertex *dest,
                              int measureType, double minMeasure);
static int CollapseEdge (TetrahMesh *tetMesh, Vertex *org, Vertex *dest,
                         Tos_MStack auxstack, int measureType,
                         double minMeasure);




/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 vm_OptimizeMesh --
        Main optimization routine for 3D triangulations.
 Arguments:
 Return value:
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int vm_OptimizeMesh (MeshGenData *mgds, int delaunay)
{
  int swapped;
  int added;
  TetrahMesh *tetMesh;
  int tetTypes[NTYPES_TET];
  int i;

  tetMesh = &(mgds->tetMesh);

  /* assert (CheckCoplanarity (tetMesh) == 0);
     assert (CheckMeshConformity (tetMesh) == 0);
     assert (CheckTetrahAdjacency (tetMesh) == 0);
     CheckMeshpointsFlags (tetMesh); */
  
  /* print quality before optimization */
  if(vm_verbose_level >= 2)
  {
    printf("Quality before optimization\n");
    CountTetrahType(tetMesh, tetTypes);
    for (i=0; i<NTYPES_TET; i++)
    {
      printf("%s\t%d\n", tetTypeNames[i], tetTypes[i]);
    }
  }
  
  added = 0;
  swapped = 0;
  do
  {
    if(delaunay)
    {
      if(mgds->mgCtrl->optimizeMethod & OPT_INSERTING)
      {
        added = vm_InsertForQuality(mgds);
      }
    }
    else
    {
      if(mgds->mgCtrl->optimizeMethod & OPT_SWAPPING)
      {
        swapped = vm_OptimizeBySwapping (tetMesh, 5, SIGMA_MEASURE);
      }
      if(mgds->mgCtrl->optimizeMethod & OPT_LAPLACE)
      {
        vm_LaplacianSmoothing (tetMesh, 5, SIGMA_MEASURE);
      }
      if(mgds->mgCtrl->optimizeMethod & OPT_COLLAPSING)
      {
        /* vm_OptimizeByCollapsing (tetMesh, SIGMA_MEASURE); */
      }
      
      /* assert (CheckTetrahAdjacency (tetMesh) == 0); */
    }
  }
  while(added != 0 || swapped != 0);
  
  /* print quality after optimization */
  if(vm_verbose_level >= 2)
  {
    printf("Quality after optimization\n");
    CountTetrahType(tetMesh, tetTypes);
    for (i=0; i<NTYPES_TET; i++)
    {
      printf("%s\t%d\n", tetTypeNames[i], tetTypes[i]);
    }
  }
  
  return 0;
} /* end of vm_OptimizeMesh */ 


/* ####################################################################### */
/*                       Laplacian smoothing                               */
/* ####################################################################### */

/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 vm_OptimizePointBall --
        Check if moving point 'org' to point 'dest' will not 
        invalidate mesh construction and if it will improve mesh quality.
        The check is based on calculation of signed volume
        of every tetrahedron in point ball. If the sign of
        volume changes it means that the tetrahedron will be 'twisted'
        and moving 'org' to 'dest' will invalidate mesh.
        At the same time if the tetrahedron becomes degenerate (zero volume)
        and does not have 'dest' vertex it means that we cannot move
        'org' to dest.
        
 Arguments:
        org - origin point
        dest - destination point
 Return value:
        1 - if mesh will not be invalidated
        0 - otherwise 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int vm_OptimizePointBall (Vertex *org, Vertex *dest, int measureType)
{
  Vertex *vertTable[4]; 
  Tetrah *tet; 
  double  vol1, vol2, vol1Sign, vol2Sign;
  int     orgNum, i; 
  double  minMeasure = 2.0;     /* bounded by 1.0 from above */
  double  minNewMeasure = 2.0;  /* bounded by 1.0 from above */
  double  curMeasure;
  Tetrah  dummyTetStruct;
  Tetrah *dummyTet = &dummyTetStruct;
 
  Tos_MStackTraversalInit (org->adjElem); 
  while((tet = Tos_MStackTraverse (org->adjElem)) != NULL)
  {
    for(i = 0; i < 4; i++)
    {
      /* check which vertex number has 'org' */
      if(TET_VERTEX(tet, i) == org)
      {
        orgNum = i;
      }
      vertTable[i] = TET_VERTEX(tet, i);
    }
    
    /* calculate the measure of the original tetrahedra */
    curMeasure = Tetrah_Measure(tet, measureType);
    if(minMeasure > curMeasure)
    {
      minMeasure = curMeasure;
    } 
    /* calculate signed volume of tet */
    vol1 = eg_Orient3D (vertTable[0]->coords,  
                        vertTable[1]->coords, 
                        vertTable[2]->coords, 
                        vertTable[3]->coords); 

    /* substitute 'dest' to 'org' in 'vertTable' */
    vertTable[orgNum] = dest; 
    vol2 = eg_Orient3D (vertTable[0]->coords,  
                        vertTable[1]->coords, 
                        vertTable[2]->coords, 
                        vertTable[3]->coords);
    
    /* set vertices of dummy tetrahedron */
    dummyTet->vertex[0] = vertTable[0];
    dummyTet->vertex[1] = vertTable[1];
    dummyTet->vertex[2] = vertTable[2];
    dummyTet->vertex[3] = vertTable[3];


    /* calculate the measure of the new tetrahedra */
    curMeasure = Tetrah_Measure(dummyTet, measureType);
    if(minNewMeasure > curMeasure)
    {
      minNewMeasure = curMeasure;
    }
    
    
    vol1Sign = copysign (1.0, vol1);
    vol2Sign = copysign (1.0, vol2);
    if(vol1Sign * vol2Sign < 0)
    {
      /* node move will invalidate mesh construction */
      return 0; 
    }
    if(fabs(vol2) < ZERO_VOLUME)
    {
       /* node move will invalidate mesh construction */
      return 0;
    }
  }
  
  /* check quality improvements */
  if(minNewMeasure < minMeasure)
      return 0;
  else
      return 1;
  
} /* end of OptimizePointBall */



void vm_LaplacianSmoothing(TetrahMesh *tetMesh, int max_pass,
                           int measureType)
{
  Tos_MStack connectedPoints;
  Vertex *vertex, *v;
  int N;
  double x, y, z;
  Vertex dummyVert;
  Vertex* newPos;
  int pass, nsmooth;
  double l,li;

  newPos = &dummyVert;

  connectedPoints = Tos_MStackNew();

  for (pass=0; pass < max_pass; pass++)
  {
    if(vm_verbose_level >= 5)
    {
      printf("+ debug 5 -- Laplacian smoothing pass: %d\n", pass+1);
    }
    
    nsmooth = 0;
    
    vertex = LIST_TAIL(tetMesh->pList);

    while (vertex != NULL)
    {
      if ((VERTEX_FLAG(vertex) & POINT_BOUNDARY) != POINT_BOUNDARY)
      {
        FindConnectedPoints (vertex, connectedPoints);
        /* calculate new position */
        N = Tos_MStackCount (connectedPoints);
        l = 0.0; 
        x = y = z = 0.0;
        while (Tos_MStackCount (connectedPoints) > 0)
        {
          v = (Vertex *)Tos_MStackPop(connectedPoints);
          li = eg_PointDist3D(v->coords, vertex->coords);
          x += v->coords[0]*li;
          y += v->coords[1]*li;
          z += v->coords[2]*li;
          l += li;
        }
        newPos->coords[0] = x/l;
        newPos->coords[1] = y/l;
        newPos->coords[2] = z/l;
        if (vm_OptimizePointBall (vertex, newPos, measureType) != 0)
        {
          nsmooth++;
          vertex->coords[0] = x/l;
          vertex->coords[1] = y/l;
          vertex->coords[2] = z/l;
        }
      }
      vertex = VERTEX_NEXT(vertex);  
    }
    if(vm_verbose_level >= 5)
    {
      printf("+ debug 5 -- Number of smoothed points: %d\n", nsmooth);
    }

  }

}

/* ####################################################################### */
/*                                 Swapping                                */
/* ####################################################################### */

/*
 * Swaps edges if the quality is improving. However to not swap always the
 * same tetrahedra, it does not restart from the beginning of the list of
 * tetrahedra but tries to continue at position in the list where the last
 * time there was a swap.
 */
static int vm_OptimizeBySwapping(TetrahMesh *tetMesh, int max_pass,
                                 int measureType)
{
  Tetrah  *tet; 
  Vertex  *v1, *v2;
  int      swapped, i;
  int      swapped_edges = 0;
  
  do
  {
    tet = LIST_TAIL (tetMesh->tList);
    swapped=0;
    while(tet != NULL)
    {
      for(i = 0; i < 6; i++)
      {
        v1 = TET_VERTEX(tet, edgeDef[i][0]);
        v2 = TET_VERTEX(tet, edgeDef[i][1]);
        if (   (VERTEX_FLAG(v1) & POINT_BOUNDARY) != POINT_BOUNDARY
            && (VERTEX_FLAG(v2) & POINT_BOUNDARY) != POINT_BOUNDARY)
        {
          swapped = SwapEdgeForOptimization(tetMesh, v1, v2, measureType);
        }
        if (swapped == 1)
        {
          swapped_edges++;
          /* CheckTetrahAdjacency(tetMesh); */
          break;
        }
      }
      if(swapped == 1) break;

      tet = TET_NEXT(tet);
    }
  }
  while(tet != NULL);
  
  if(vm_verbose_level >= 2)
  {
    printf ("+debug 2 -- number of swapped edges %d\n", swapped_edges);
  }
  
  return swapped_edges;
}


/* ####################################################################### */
/*              Point Insertion for quality improvement                    */
/* ####################################################################### */

static int vm_InsertForQuality(MeshGenData *mgds)
{
  Tetrah     *tet;
  TetrahMesh *tetMesh;
  Vertex     *v[4]; /* tetrahedron vertices */
  Vertex     *p;
  double      center[3];
  double      radius;
  Tos_MStack  waitingPoints;
  int         result = 0;
  double ms_min, ms_max;  /* small and large mesh solid angle */
  double md_min, md_max;  /* small and large mesh dihedral angle */
  
  waitingPoints = Tos_MStackNew ();

  tetMesh = &(mgds->tetMesh);

  GoodAnglesLimits (tetMesh, &ms_min, &ms_max, &md_min, &md_max);
  ClassifyTetrah (tetMesh);

  /* slivers */
  tet = LIST_TAIL (tetMesh->tList);
  while (tet != NULL)
  {
    if ((tet->flag & TET_SLIVER) == TET_SLIVER)
    {
      v[0] = TET_VERTEX(tet, 0);
      v[1] = TET_VERTEX(tet, 1);
      v[2] = TET_VERTEX(tet, 2);
      v[3] = TET_VERTEX(tet, 3);
      
      eg_TetCircumcenter (v[0]->coords, v[1]->coords,
                          v[2]->coords, v[3]->coords,
                          center, NULL, NULL, NULL);
      radius = sqrt(center[0] * center[0] +
                    center[1] * center[1] +
                    center[2] * center[2]  );
      center[0] += v[0]->coords[0];
      center[1] += v[0]->coords[1];
      center[2] += v[0]->coords[2];

      
      p = GoodPointSelection(mgds, tetMesh, center, radius, 1,
                             1, ms_min, ms_max, md_min, md_max);
      
      if(p != NULL)
      {
        Tos_MStackPush (waitingPoints, (void *)p);
        DelaunayKernel (tetMesh, waitingPoints, 1);
        result = 1;

        break;
      }
    }
    
    tet = TET_NEXT(tet);
  }
  
  Tos_MStackFree (&waitingPoints);

  return result;
}


/* ####################################################################### */
/*                         Collapsing edges                                */
/* ####################################################################### */


/* construct table of edges sorted by length. The shortest edge come
   first.
*/
static void sort_edges_in_tet (Tetrah *tet, int *edgeIndices)
{
   double *ltable[4];
   double *vcoords[4];
   double data[4][4];
   int i, j, e1, e2, swap;
   ltable[0] = &(data[0][0]);
   ltable[1] = &(data[1][0]);
   ltable[2] = &(data[2][0]);
   ltable[3] = &(data[3][0]);
 
   vcoords[0] = TET_VCOORDS(tet,0);
   vcoords[1] = TET_VCOORDS(tet,1);
   vcoords[2] = TET_VCOORDS(tet,2);
   vcoords[3] = TET_VCOORDS(tet,3);
   GetTetEdgesLengthTable (vcoords, ltable);

   for (i=0; i<6; i++)
     edgeIndices[i] = i;

  
   for (i=0; i<6; i++)
   {
     for (j=i+1; j<6; j++)
     {
       e1 = edgeIndices[j];
       e2 = edgeIndices[i];
       if (  data[edgeDef[e1][0]][edgeDef[e1][1]] 
           < data[edgeDef[e2][0]][edgeDef[e2][1]]) 
       {
          swap = edgeIndices[i];
          edgeIndices[i] = edgeIndices[j];
          edgeIndices[j] = swap;  
       }
     }
   }
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 CheckEdgeCollapsing --
        Check if moving point 'org' to point 'dest' will not 
        invalidate mesh construction and if it will improve mesh quality.
        The check is based on calculation of signed volume
        of every tetrahedron in point ball. If the sign of
        volume changes it means that the tetrahedron will be 'twisted'
        and moving 'org' to 'dest' will invalidate mesh.
        At the same time if the tetrahedron becomes degenerate (zero volume)
        and does not have 'dest' vertex it means that we cannot move
        'org' to dest.
        
 Arguments:
        org - origin point
        dest - destination point
 Return value:
        1 - if mesh will not be invalidated
        0 - otherwise 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int CheckEdgeCollapse (Vertex *org, Vertex *dest,
                              int measureType, double minMeasure)
{
  Vertex *vertTable[4]; 
  Tetrah *tet; 
  int orgNum, i; 
  int result = 1;
  int skip;
  double currentMeasure;
  Tetrah dummyTetStruct;
  Tetrah *dummyTet = &dummyTetStruct;
  double vol1, vol2, vol1Sign, vol2Sign;

  Tos_MStackTraversalInit (org->adjElem); 
  while ((tet = Tos_MStackTraverse (org->adjElem)) != NULL)
  {
    for (i=0; i<4; i++)
    {
      if (TET_VERTEX (tet, i) == dest)
      {
        skip = 1; 
        break;
      }
      /* check which vertex number has 'org' */
      if (TET_VERTEX (tet, i) == org)
      {
        orgNum = i;
      }
      vertTable[i] = TET_VERTEX (tet, i); 
    }
    if (skip == 1)
    { 
      skip = 0; 
      continue;  /* skip 'tet' and go to next one */
    }


    /* calculate signed volume of tet */
    vol1 = eg_Orient3D (vertTable[0]->coords,  
                        vertTable[1]->coords, 
                        vertTable[2]->coords, 
                        vertTable[3]->coords); 


    /* substitute 'dest' to 'org' in 'vertTable' */
    vertTable[orgNum] = dest; 
    vol2 = eg_Orient3D (vertTable[0]->coords,  
                        vertTable[1]->coords, 
                        vertTable[2]->coords, 
                        vertTable[3]->coords); 
    
    vol1Sign = copysign (1.0, vol1); 
    vol2Sign = copysign (1.0, vol2); 

    /* set vertices of dummy tetrahedron */
    dummyTet->vertex[0] = vertTable[0];
    dummyTet->vertex[1] = vertTable[1];
    dummyTet->vertex[2] = vertTable[2];
    dummyTet->vertex[3] = vertTable[3];
    
    currentMeasure = Tetrah_Measure (dummyTet, measureType);

    if (vol1Sign * vol2Sign < 0)
    {
       /* node move will invalidate mesh construction */
       result = 0; 
    }
    if (fabs(vol2) < ZERO_VOLUME)
    {
      result = 0; /* node move will invalidate mesh construction */
    }
    if (currentMeasure <= minMeasure + 0.05)
    {
      result = 0; /* moving node is not benefitial */
    }
  }
  return result;                         
} /* end of CheckEdgeCollapse */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 CollapseEdge 
        Remove edge  from the mesh construction by moving
        'org' vertex to position of 'dest' vertex and then 
        remove 'org' vertex from the mesh.
        Update adjacency information in affected vertices and
        tetrahedra. The tetrahedra forming edge shell are removed 
        from the mesh construction as well as from the supplied stack.
 CAUTION: memory allocated for 'org' vertex is not free.

 Arguments:
        tetMesh -- tetrhaedral mesh data structure
        org -- endpoint of removed edge (endpoint which will be removed)
        dest -- endpoint of removed edge (endpoint which will be preserved)
        auxstack -- auxiliary stack which should be updated while removing
                 tetrahedra
 Return value:
        1 - if edge removed succesfully
        0 - if edge romoval is not possible
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int CollapseEdge (TetrahMesh *tetMesh, Vertex *org, Vertex *dest,
                         Tos_MStack auxstack, int measureType, double minMeasure)
{
  Edge dummyedge; /* to have automatic edge variable */
  Edge *edge; 
  Tos_MStack edgeShell; 
  Tetrah *tet, *oNeighb, *dNeighb;
  int vo, vd, von, vdn;
  int i=0;
  int result = 0; 
  
  edge = &dummyedge; 
  /* check if is it is possible and beneficial to remove edge by moving 'org'
   * vertex to poisition of 'dest' vertex
   */  
  if (CheckEdgeCollapse (org, dest, measureType, minMeasure) != 0)
  {
    result = 1; 
    /* set endpoints of edge beeing removed */
    EDGE_VERTEX (edge, 0) = org; 
    EDGE_VERTEX (edge, 1) = dest;

    /* allocate memory for edge shell */
    edgeShell = Tos_MStackNew ();
    /* find edge shell */
    FindEdgeShell (edge, edgeShell);

    /* process each tetrahedron from edge shell and update
     * neighbouring information in adjacent elements which 
     * doesn't belong to the edge shell 
     */  
     Tos_MStackTraversalInit (edgeShell);
     while ((tet = (Tetrah *) Tos_MStackTraverse (edgeShell)) != NULL)
     {
       if (tet->id == 0)
       {
         continue;
       }
       /* find numbers of 'org' and 'dest' vertices in 'tet' */
       for (i=0; i<4; i++)
       {
         if (TET_VERTEX (tet, i) == org) vo = i;
         if (TET_VERTEX (tet, i) == dest) vd = i;
       }
       /* find neighbour of 'tet' opposite to vertex vo */
       oNeighb = TET_NEIGHBOUR (tet, vo); 
       /* find neighbour of 'tet' opposite to vertex vd */
       dNeighb = TET_NEIGHBOUR (tet, vd); 

       /* find which neighbour numbers has 'tet' in  'oNeighb' and 'dNeighb' 
        * and update adjacency information in 'oNeighb' and 'dNeighb' 
        */
       if (oNeighb != NULL)
       {
         von = WhichNeighbour (oNeighb, tet); 
         TET_NEIGHBOUR (oNeighb, von) = dNeighb; 
       }
       if (dNeighb != NULL)
       {
         vdn = WhichNeighbour (dNeighb, tet); 
         TET_NEIGHBOUR (dNeighb, vdn) = oNeighb; 
       }

       /* remove 'tet' from adjacency lists of 'org' and 'dest' */
       /*
       Tos_MStackRemove (org->adjElem, (void *) tet); 
       Tos_MStackRemove (dest->adjElem, (void *) tet); 
       */
       for (i=0; i<4; i++)
       {
         Tos_MStackRemove (TET_VERTEX(tet,i)->adjElem, (void *)tet);
       }

       tet->id = 0;
     }   
     /* process all remaining tetrahedra which belong to adjacency
      * list of 'org' vertex 
      */
     while (Tos_MStackCount (org->adjElem) > 0)
     {
       tet = (Tetrah *) Tos_MStackPop (org->adjElem); 
       for (i=0; i<4; i++)
       {
         if (TET_VERTEX (tet, i) == org)
         {
           TET_VERTEX (tet, i) = dest;
           break;
         }
       }
       Tos_MStackPush (dest->adjElem, (void *)tet); 
     } 

     /* vertex should not be connected to any tetrahedron */
     assert (Tos_MStackCount (org->adjElem) == 0);

     /* remove 'org' from the list of mesh points */
     RemoveVertexFromList (tetMesh->pList, org); 
     (tetMesh->NMeshPoints)--;

     /* remove edge shell elements form the mesh */
     while (Tos_MStackCount (edgeShell) > 0)
     {
       int co = 0;
       Tetrah *tt;
       tet = (Tetrah *) Tos_MStackPop (edgeShell); 
       Tos_MStackRemove (auxstack, tet); 
       TetListTraversalInit (tetMesh->tList, TAIL_HEAD);
       while ((tt = (Tetrah *) TetListTraverse (tetMesh->tList, TAIL_HEAD))
              !=NULL)
       {
         int i;
         for (i=0; i<4; i++)
           {
             if (tt->neighbour[i] == tet && tet->id!=0) co++; 
           }
       }
       assert (co<=2);
       RemoveTetFromList (tetMesh->tList, tet); 
       (tetMesh->NTetrah)--;  
     }
  }
  return result; 
} /* end of CollapseEdge */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 vm_OptimizeByCollapsing --
 Arguments:
 Return value:
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int vm_OptimizeByCollapsing (TetrahMesh *tetMesh, enum MeasureType measure)
{
  Tetrah *tet;
  Vertex *v1, *v2;
  Tos_MStack badTetStack;
  int edge_collapsed;
  int edge_collapse_counter=0;
  double minMeasure, maxMeasure, currentMeasure;
  double collapseTreshold = 0.2;
  int edgeIndices[6];
  int i, esaindex1, esaindex2;

  badTetStack = Tos_MStackNew ();
  assert(badTetStack != NULL);
  
  if(vm_verbose_level>=2)
  {
    printf("+ debug 2 -- Optimize by collapsing \n");
  }
  
  /* find minimal measure of entire mesh */
  Mesh_MinAndMaxMeasure (tetMesh, measure, &minMeasure, &maxMeasure); 
  
  /* find minimal measure of entire mesh */
  tet = LIST_TAIL(tetMesh->tList);   
  while (tet != NULL)
  {
    currentMeasure = Tetrah_Measure (tet, measure);
    if (currentMeasure < minMeasure + collapseTreshold)
    {
      Tos_MStackPush (badTetStack, tet);
    }
    tet = TET_NEXT(tet);
  } 
  printf ("NUMBER OF BAD TET=%d\n", badTetStack->count);
  /* go through all bad tetrahedra */
  while (Tos_MStackCount (badTetStack) > 0)
  {
    tet = (Tetrah *) Tos_MStackPop (badTetStack);
    currentMeasure = Tetrah_Measure (tet, measure);   
    sort_edges_in_tet (tet, edgeIndices);
    edge_collapsed = 0;
    /* try to collapse edge starting for the shortest one */
    for (i=0; i<1; i++)    
    {
       esaindex1 = edgeDef[i][0];
       esaindex2 = edgeDef[i][1];
    
       v1 = TET_VERTEX(tet, esaindex1);
       v2 = TET_VERTEX(tet, esaindex2);
       /* check which vertex can be moved and collapse edge */
       if ((VERTEX_FLAG(v1) & POINT_BOUNDARY) != POINT_BOUNDARY)
       {
         edge_collapsed = CollapseEdge(tetMesh, v1, v2, badTetStack, measure, 
                          currentMeasure);
       }
       if (edge_collapsed != 0)
         break;

       /* try to collapse in oposite direction */
       if ((VERTEX_FLAG(v2) & POINT_BOUNDARY) != POINT_BOUNDARY)
       {
         edge_collapsed = CollapseEdge(tetMesh, v2, v1, badTetStack, measure,
                          currentMeasure);
       }
       if (edge_collapsed != 0) 
       break;
    }
     edge_collapse_counter += edge_collapsed; 
  }

  if(vm_verbose_level >= 2)
  {
    printf("+ debug 2 -- %d edges collapsed\n",
           edge_collapse_counter );
  }
  
  return edge_collapse_counter;
} /* end of vm_OptimizeByCollapsing */



/* ####################################################################### */
/*                    Unsplit boundary edges/faces                         */
/* ####################################################################### */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 UnsplitFaces --
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int UnsplitFaces (MeshGenData *mgds)
{
  Vertex     *v, *vert;
  Face       *face;
  Tetrah     *tet;
  int         imesh, iface, i, iv;
  int         unsplitCounter=0;
  Tos_MStack  connectedPoints;

  connectedPoints = Tos_MStackNew();

  /* go through all meshes */
  for(imesh = 0; imesh < mgds->NMeshes; imesh++)
  {
    /* go through all faces in the mesh */
    for(iface = 0; iface < mgds->fTableLength[imesh]; iface++)
    {
      /* if the face has three partial faces */
      if(Tos_MStackCount(mgds->fTable[imesh][iface]->splitFaces) == 3)
      {
        /* get the first partial face */
        face = (Face *)Tos_MStackPop(mgds->fTable[imesh][iface]->splitFaces);
        /* find the node inserted in the middle of face */
        for(iv = 0; iv < 3; iv++)
        {
          v = FACE_VERTEX(face, iv);
          if((VERTEX_FLAG(v) & POINT_ON_FACE) == POINT_ON_FACE)
              break;
        }
        assert(iv < 3);
        
        /* find all nodes of the connected tetrahedra */
        FindConnectedPoints (v, connectedPoints);
        /* remove all connected tetrahedra */
        while (Tos_MStackCount (v->adjElem) > 0)
        {
          tet = (Tetrah *) Tos_MStackPop (v->adjElem);
          RemoveTetFromMesh (&(mgds->tetMesh), tet);
          TetrahDelete (tet);
        }
        /* create a new tetrahedron */
        tet = TetrahCreate ();
        /* set the vertices of the tetrahedron */
        i=0;
        while (Tos_MStackCount (connectedPoints) > 0)
        {
          vert = (Vertex *) Tos_MStackPop (connectedPoints);
          TET_VERTEX(tet,i) = vert;
          i++;
        }
        assert(i==4);
        
        /* add the tetrahedra to the mesh */
        AddTetToMesh (&(mgds->tetMesh), tet);
        /* unsplit the face, free up partial faces */
        FaceDelete(face);
        face = (Face *)Tos_MStackPop(mgds->fTable[imesh][iface]->splitFaces);
        FaceDelete(face);
        face = (Face *)Tos_MStackPop(mgds->fTable[imesh][iface]->splitFaces);
        FaceDelete(face);
        assert(Tos_MStackCount(mgds->fTable[imesh][iface]->splitFaces) == 0);
        /* remove the vertex from the list */
        RemoveVertexFromList(mgds->tetMesh.pList, v);
        (mgds->tetMesh.NMeshPoints)--;
        /* delete the vertex itself */
        VertexDelete(v);
        unsplitCounter++;
      }
    }
  }
  Tos_MStackFree(&connectedPoints);
  
  return unsplitCounter;
} /* end of UnsplitFaces */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 UpdateSplitFaces --
    Removes split faces which disappear due to removal of edge segment.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void UpdateSplitFaces(Edge *edge,
                     Vertex *removedPoint, Vertex *destPoint)
{
  int        i, k, dummy;
  Face      *face; /* oritinal face */
  Face      *delFace; /* face to delete */
  Face      *partFace;
  Edge      vedge; 

  /* create an virtual edge */
  EDGE_VERTEX(&vedge, 0) = removedPoint;
  EDGE_VERTEX(&vedge, 1) = destPoint;

  for (k=0; k<2; k++)
  {

    face = EDGE_NEIGHBOUR(edge, k);
 
    Tos_MStackTraversalInit(face->splitFaces);
    while((partFace = Tos_MStackTraverse(face->splitFaces)) != NULL)
    {
      /* check if partial face has the edge */
      if(FaceHasEdge(partFace, &vedge, &dummy) == TRUE)
      {
        delFace = partFace;
        break;
      }
    }
    /* delete split face which disapears due to edge collapsing */
    Tos_MStackRemove(face->splitFaces, delFace);
      
    if (Tos_MStackCount(face->splitFaces) == 1)
    {
       Tos_MStackFlush (face->splitFaces);
    }
    else
    {
      /* update vertices of other split faces */
      Tos_MStackTraversalInit(face->splitFaces);
      while((partFace = Tos_MStackTraverse(face->splitFaces)) != NULL)
      {
        /* change vertex of the other face */
        for(i = 0; i < 3; i++)
        {
          if(face->vertex[i] == removedPoint)
          {
            face->vertex[i] = destPoint;
            break;
          }
        }
      }
    }
  }
} /* end of UpdateSplitFaces */



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 RemoveEdgeFromMesh --
        Remove edge  from the mesh construction by moving
        'org' vertex to position of 'dest' vertex and then 
        remove 'org' vertex from the mesh.
        Update adjacency information in affected vertices and
        tetrahedra. The tetrahedra forming edge shell are removed 
        from the mesh construction.
 CAUTION: memory allocated for 'org' vertex is not free.

 Arguments:
        tetMesh -- tetrhaedral mesh data structure
        org -- endpoint of removed edge (endpoint which will be removed)
        dest -- endpoint of removed edge (endpoint which will be preserved)
 Return value:
        1 - if edge removed succesfully
        0 - if edge romoval is not possible
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int RemoveEdgeFromMesh (TetrahMesh *tetMesh,
                        Vertex *org, Vertex *dest, int *callnum)
{
  Edge dummyedge; /* to have automatic edge variable */
  Edge *edge; 
  Tos_MStack edgeShell; 
  Tetrah *tet, *oNeighb, *dNeighb;
  int vo, vd, von, vdn;
  int i;
  int result = 0; 
  
  edge = &dummyedge; /* edges are implemented with the same structure
                      *  as tetrahedra
                      */  
  *callnum = *callnum + 1;
  /* check if is it is possible to remove edge by moving 'org'
   * vertex to position of 'dest' vertex
   */  
  if (CheckPointBall (org, dest) != 0)
  {
    result = 1; 
    /* set endpoints of edge beeing removed */
    EDGE_VERTEX (edge, 0) = org; 
    EDGE_VERTEX (edge, 1) = dest;

    /* allocate memory for edge shell */
    edgeShell = Tos_MStackNew ();
    /* find edge shell */
    FindEdgeShell (edge, edgeShell);

    /* process each tetrahedron from edge shell and update
     * neighbouring information in adjacent elements which 
     * doesn't belong to the edge shell 
     */  
     Tos_MStackTraversalInit (edgeShell);
     while ((tet = (Tetrah *) Tos_MStackTraverse (edgeShell)) != NULL)
     {
       if (tet->id == 0)
       {
         continue;
       }
       /* find numbers of 'org' and 'dest' vertices in 'tet' */
       for (i=0; i<4; i++)
       {
         if (TET_VERTEX (tet, i) == org) vo = i;
         if (TET_VERTEX (tet, i) == dest) vd = i;
       }
       /* find neighbour of 'tet' opposite to vertex vo */
       oNeighb = TET_NEIGHBOUR (tet, vo); 
       /* find neighbour of 'tet' opposite to vertex vd */
       dNeighb = TET_NEIGHBOUR (tet, vd); 

       /* find which neighbour numbers has 'tet' in  'oNeighb' and 'dNeighb' 
        * and update adjacency information in 'oNeighb' and 'dNeighb' 
        */
       if (oNeighb != NULL)
       {
         von = WhichNeighbour (oNeighb, tet); 
         TET_NEIGHBOUR (oNeighb, von) = dNeighb; 
       }
       if (dNeighb != NULL)
       {
         vdn = WhichNeighbour (dNeighb, tet); 
         TET_NEIGHBOUR (dNeighb, vdn) = oNeighb; 
       }

       /* remove 'tet' from adjacency lists of 'org' and 'dest' */
       /*
       Tos_MStackRemove (org->adjElem, (void *) tet); 
       Tos_MStackRemove (dest->adjElem, (void *) tet); 
       */
       for (i=0; i<4; i++)
       {
         Tos_MStackRemove (TET_VERTEX(tet,i)->adjElem, (void *)tet);
       }

       tet->id = 0;
     }   
     /* process all remaining tetrahedra which belong to adjacency
      * list of 'org' vertex 
      */
     while (Tos_MStackCount (org->adjElem) > 0)
     {
       tet = (Tetrah *) Tos_MStackPop (org->adjElem); 
       for (i=0; i<4; i++)
       {
         if (TET_VERTEX (tet, i) == org)
         {
           TET_VERTEX (tet, i) = dest;
           break;
         }
       }
       Tos_MStackPush (dest->adjElem, (void *)tet); 
     } 

     /* vertex should not be connected to any tetrahedron */
     assert (Tos_MStackCount (org->adjElem) == 0);

     /* remove 'org' from the list of mesh points */
     RemoveVertexFromList (tetMesh->pList, org); 
     (tetMesh->NMeshPoints)--;

     /* remove edge shell elements form the mesh */
     while (Tos_MStackCount (edgeShell) > 0)
     {
       int co = 0;
       Tetrah *tt;
       tet = (Tetrah *) Tos_MStackPop (edgeShell); 
       TetListTraversalInit (tetMesh->tList, TAIL_HEAD);
       while ((tt = (Tetrah *) TetListTraverse (tetMesh->tList, TAIL_HEAD))
              !=NULL)
       {
         int i;
         for (i=0; i<4; i++)
           {
             if (tt->neighbour[i] == tet && tet->id!=0) co++; 
           }
       }
       assert (co<=2);
       RemoveTetFromList (tetMesh->tList, tet); 
       (tetMesh->NTetrah)--;  
     }
  }
  return result; 
} /* end of RemoveEdgeFromMesh */



static void RemoveSplitPoint(Edge *edge, Vertex *pointS, Vertex *pointE,
                             Tos_MStack stackPop, Tos_MStack stackPush,
                             MeshGenData *mgds)
{
  Vertex *pointA, *pointB;
  int     callnum = 0;
  int     removed;
  
  pointA = pointS; 

  while(Tos_MStackCount(stackPop) > 0)
  {
    pointB = Tos_MStackPop(stackPop);
    
    /* try to remove edge segment A-B */
    removed =  RemoveEdgeFromMesh (&(mgds->tetMesh), pointB, pointA,
                                   &callnum);
    if (removed != 0)
    {
      UpdateSplitFaces(edge, pointB, pointA);
      VertexDelete(pointB);
    }
    else
    {
      /* if not end point push it on stack, since we could not remove it */
      if(pointA != pointS)
          Tos_MStackPush(stackPush, (void *)pointA);
      /* try if removal is possible with next segment */
      pointA = pointB;
    }
  }
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 UnsplitEdges --
        For each edge on which additional points were located during
        recovery proces preform unsplit operation.
        The unsplit operation removes added points.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int UnsplitEdges (MeshGenData *mgds)
{
  int    i, iedge;
  Edge  *edge;
  int    unsplitEdgeCounter = 0;
  Tos_MStack splitP;
  
  splitP = Tos_MStackNew();
  
  /* for each input mesh */
  for (i = 0; i < mgds->NMeshes; i++)
  {
    for(iedge = 0; iedge < mgds->eTableLength[i]; iedge++)
    {
      edge = mgds->eTable[i][iedge];
      
      /* are there split points on the edge ? */
      if(Tos_MStackCount(edge->splitEdges) <= 0)
          continue;
#if 0
      /* empty the temporary stack */
      Tos_MStackFlush (splitP);

      /* try to collapse towards one direction */
      RemoveSplitPoint(edge, EDGE_VERTEX (edge,1), EDGE_VERTEX (edge,0),
                       edge->splitPoints, splitP, mgds);
      /* try to collapse towards the other direction */
      RemoveSplitPoint(edge, EDGE_VERTEX (edge,0), EDGE_VERTEX (edge,1),
                       splitP, edge->splitPoints, mgds);
#endif
      
      /* if all split points are removed, the edge is fully recovered */
      if(Tos_MStackCount(edge->splitEdges) == 0)
          unsplitEdgeCounter++;

    } /* end of loop over edges */
  } /* end of loop over meshes */

  Tos_MStackFree(&splitP);
  
  return unsplitEdgeCounter;
} /* end of UnsplitEdges */








