#include <assert.h>

#include "eg_lib.h"
#include "find.h"
#include "check.h"
#include "basic_op.h"
#include "mesh_op.h"
#include "flip.h"
#include "recover.h"
#include "boundary.h"
#include "debug.h"
#include "write_log.h"  /* FIXME: it should not be here !!!!! */

#ifdef WIN32
extern double copysign(double x, double y);
#endif

static void RecoverBoundaryEdges (MeshGenData *mgds);
static void RecoverEdge (TetrahMesh *tetMesh, Edge *edge);
static int TryMovePointRecovery (TetrahMesh *tetMesh, Tetrah *tetA, 
                                 int *vertices, Vertex *intersectionPoint);

static void RecoverBoundaryFaces (MeshGenData *mgds);
static void DoRecoverFace(Face *face, Tos_MStack pointsOnFace);
static int RecoverFace (TetrahMesh *tetMesh, Face *face);


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 EnsureBoundaryConformity --
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int EnsureBoundaryConformity (MeshGenData *mgds)
{
  if(vm_verbose_level>=2)
  {
    printf("+ debug 2 -- EnsureBoundaryConformity\n");
  }
  
  RecoverBoundaryEdges (mgds); 
  RecoverBoundaryFaces (mgds); 

  return 1;
}

/* **************************************************************** */
/* **************************************************************** */
/* **************************************************************** */
/*                            Edges                                 */
/* **************************************************************** */
/* **************************************************************** */
/* **************************************************************** */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 RecoverBoundaryEdges --
        Recover boundary edges by direct mesh modification
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void RecoverBoundaryEdges (MeshGenData *mgds)
{
  int iedge = 0;
  Tos_MStack missingEdges; /* stack of missing edges */
  Edge *edge;
  
  /* initialize stack of missing edges */
  missingEdges = Tos_MStackNew ();

  FindMissingEdges (mgds, missingEdges); 

  Tos_LogNMissingEdges (mgds, Tos_MStackCount(missingEdges));

  /* recover each edge */
  while (Tos_MStackCount (missingEdges) > 0)
  {
    edge = (Edge *) Tos_MStackPop (missingEdges);

    RecoverEdge (&(mgds->tetMesh), edge);
    
    /* update boundary representation if edge was split */
    if(Tos_MStackCount (edge->splitEdges) > 0)
    {
      /* UpdateSplitEdgeToFace(mgds, edge); */
    }
    
    iedge++;
  }

  if(vm_verbose_level>=3)
  {
    printf("+ debug 3 -- Number of recovered edges: %d\n", iedge);
  }
  
  /* delete stack of missing edges */
  Tos_MStackFree (&missingEdges); 

  return;
} /* end of RecoverBoundaryEdges */


static void RecoverEdge (TetrahMesh *tetMesh, Edge *edge)
{
  int     edgeRecovered;
  int     moveSucceded;
  int     intersectType; /* type of intersection */
  int     vertices[3];   /* vertices of intersected item */
  Tetrah *tet;   /* tetrhaedron containing intersected face or edge */
  Vertex *nodeA, *nodeB;
  Vertex *intersectionPoint; 
  Edge    intersectedEdge;
  Face    intersectedFace;
  int     boundaryFlag = 0;
  Edge   *splitEdge;
  
  nodeA = EDGE_VERTEX (edge, 0);
  nodeB = EDGE_VERTEX (edge, 1);

  /* check the flagging of the edge */
  if ((nodeA->flag & POINT_EXT_BOUNDARY) == POINT_EXT_BOUNDARY)
  {
     if ((nodeB->flag & POINT_EXT_BOUNDARY) == POINT_EXT_BOUNDARY)
     {
       boundaryFlag = POINT_EXT_BOUNDARY;
     } 
     else
     {
       e_GeneralError ("Inconsistent flagging of boundary points\n");
     }
  }
  else if ((nodeA->flag & POINT_INT_BOUNDARY) == POINT_INT_BOUNDARY)
  {
     if ((nodeB->flag & POINT_INT_BOUNDARY) == POINT_INT_BOUNDARY)
     {
       boundaryFlag = POINT_INT_BOUNDARY;
     } 
     else
     {
       e_GeneralError("Inconsistent flagging of boundary points\n");
     }
  }
  else
  {
    e_GeneralError("Edge to be recovered has no properly flaged nodes\n");
  }

  edgeRecovered = FALSE;

  /* CONSTCOND */
  while (TRUE)
  {
    /* find intersection of edge A-B and faces of tetrahedral
     * mesh closest to node A
     */
    edgeRecovered = SegmentTetMeshIntersect (tetMesh, nodeA, nodeB, 
                                             &tet, &intersectType,
                                             &intersectionPoint, vertices);
    if (edgeRecovered == TRUE)
    {
      break;  
    }
    
    /* try to recover the edge by swapping edges */
    edgeRecovered = TrySwapEdgeRecovery (tetMesh, nodeA, nodeB, tet,
                                         intersectType, vertices);
    if (edgeRecovered == TRUE)
    {
      break;
    }

    /* try to collapse the point on an edge */
    /* FIXME: actually we should consider the quality of the created
     * configuration of tetrahedra, for example we may try to simply
     * insert a point on the edge (as below) rather than move a point */
    if (intersectType == EG_ONEDGE)
    {
      moveSucceded = TryMovePointRecovery (tetMesh, tet, vertices,
                                           intersectionPoint);
      if (moveSucceded == TRUE)
      {
        intersectType = EG_ONVERTEX;
      }
    }
    
    /* update the mesh */
    switch (intersectType)
    {
      case (EG_ONEDGE):
      {
        EDGE_VERTEX (&intersectedEdge, 0) = TET_VERTEX (tet, vertices[0]);
        EDGE_VERTEX (&intersectedEdge, 1) = TET_VERTEX (tet, vertices[1]);
        InsertPointOnEdge (tetMesh, intersectionPoint, &intersectedEdge);
        intersectionPoint->flag |= POINT_BOUNDARY;
        intersectionPoint->flag |= POINT_ON_EDGE;
        intersectionPoint->flag |= boundaryFlag;
        break;
      }
      case (EG_INSIDE):
      {
        FACE_VERTEX (&intersectedFace, 0) = TET_VERTEX (tet, vertices[0]);
        FACE_VERTEX (&intersectedFace, 1) = TET_VERTEX (tet, vertices[1]);
        FACE_VERTEX (&intersectedFace, 2) = TET_VERTEX (tet, vertices[2]);
        InsertPointOnFace (tetMesh, intersectionPoint, &intersectedFace);
        intersectionPoint->flag |= POINT_BOUNDARY;
        intersectionPoint->flag |= POINT_ON_EDGE;
        intersectionPoint->flag |= boundaryFlag;
        break;
      }
      case (EG_ONVERTEX):
      {
        /* intersection is on existing vertex so the vertex should not
         * be duplicated
         */
        VertexDelete (intersectionPoint);
        /* set intersectionPoint to intersected vertex */
        intersectionPoint = TET_VERTEX (tet, vertices[0]);
        break;
      }
      default:
      {
        e_GeneralError("RecoverEdge: unknown intersection type %d\n",
                        intersectType);
      }
    }

    splitEdge = EdgeCreate();
    EDGE_VERTEX(splitEdge, 0) = nodeA;
    EDGE_VERTEX(splitEdge, 1) = intersectionPoint;
    
    /* store the split edge with the original edge */
    Tos_MStackPush(edge->splitEdges, (void *)splitEdge);
    
    /* move nodeA to intersection point */
    nodeA = intersectionPoint;
  }

  /* if the edge was recovered by point insertion */
  if(Tos_MStackCount (edge->splitEdges) > 0)
  {
    /* add the last "bit" of edge */
    splitEdge = EdgeCreate();
    EDGE_VERTEX(splitEdge, 0) = nodeA;
    EDGE_VERTEX(splitEdge, 1) = nodeB;
    Tos_MStackPush(edge->splitEdges, (void *)splitEdge);
  }
  
} /* end of RecoverEdge */


static int TryMovePointRecovery (TetrahMesh *tetMesh, Tetrah *tetA, 
                                 int *vertices, Vertex *intersectionPoint)
{
  Vertex *nodeE[2];
  double dE[2];
  int imin;
  
  /* get the vertices of intersected edge */
  nodeE[0] = TET_VERTEX (tetA, vertices[0]);
  nodeE[1] = TET_VERTEX (tetA, vertices[1]);
  
  /* calculate distance between edge vertices
   * and intersection point */
  dE[0] = eg_PointDist3D (nodeE[0]->coords, intersectionPoint->coords);
  dE[1] = eg_PointDist3D (nodeE[1]->coords, intersectionPoint->coords);

  /* get the index of edge vertex closest to the intersection point */
  if (dE[0] < dE[1])
  {
    imin = 0;
  }
  else
  {
    imin = 1;
  }

  if ((nodeE[imin]->flag & POINT_BOUNDARY) != POINT_BOUNDARY)
  { /* if point is not a boundary one */
    if (CheckPointBall (nodeE[imin], intersectionPoint) == 1)
    {
      /* change coordinates of edge end point to coordinates of 
       * intersection point
       */
      memcpy (nodeE[imin]->coords, intersectionPoint->coords, 
              sizeof(double) * 3);
      vertices[0] = vertices[imin];
      return TRUE;
    }
  }
  
  if ((nodeE[!imin]->flag & POINT_BOUNDARY) != POINT_BOUNDARY)
  {
    if (CheckPointBall (nodeE[!imin], intersectionPoint) == 1)
    {
      /* change coordinates of edge end point to coordinates of 
       * intersection point
       */
      memcpy (nodeE[!imin]->coords, intersectionPoint->coords, 
              sizeof(double) * 3);
      vertices[0] = vertices[!imin];
      return TRUE;
    }
  }
  return FALSE;
}

/* **************************************************************** */
/* **************************************************************** */
/* **************************************************************** */
/*                            Faces                                 */
/* **************************************************************** */
/* **************************************************************** */
/* **************************************************************** */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 RecoverBoundaryFaces --
        For each face in surfaces array check if face is present
        in tetrahedral mesh and if not recover it by direct
        mesh modification.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void RecoverBoundaryFaces (MeshGenData *mgds)
{
  Tos_MStack  missingFaces;
  Face       *face;
  int         nsplit, iface = 0;
  int         maxFaceSplit = 0;

  missingFaces = Tos_MStackNew ();

  FindMissingFaces(mgds, missingFaces); 

  Tos_LogNMissingFaces(mgds, Tos_MStackCount(missingFaces));

  
  /* recover each face */
  while (Tos_MStackCount(missingFaces) > 0)
  {
    face = (Face *)Tos_MStackPop(missingFaces);

    nsplit = RecoverFace(&(mgds->tetMesh), face);
    if(maxFaceSplit < nsplit)
        maxFaceSplit = nsplit;
    
    /* update boundary representation if face was split */
    if(Tos_MStackCount (face->splitFaces) > 0)
    {
      /* UpdateSplitFaceToEdge(mgds, face); */
    }
    
    iface++;
  }

  Tos_LogMaxFaceSplit(mgds, maxFaceSplit);

  if(vm_verbose_level>=3)
  {
    printf("+ debug 3 -- Number of recovered faces: %d\n", iface);
  }
  
  /* delete stack of missing faces */
  Tos_MStackFree(&missingFaces);
} /* end of RecoverBoundaryFaces */


/* this function does the actual face recovery by creating the
   subfaces which make up the whole face */
static void DoRecoverFace(Face *face, Tos_MStack pointsOnFace)
{
  Tos_MStack   adjTetrah;
  Tetrah      *tet;
  int          nodes[4];
  int          cnodes, facenum;
  Face        *partFace;
  
  adjTetrah = Tos_MStackNew();

  /* get all attached tetrahedra */
  FindPointsTetrahAdjSet (pointsOnFace, adjTetrah);
  while (Tos_MStackCount (adjTetrah) > 0)
  {
    tet = (Tetrah *) Tos_MStackPop (adjTetrah);
      
    /* find number of nodes shared by 'face' and 'tet' */
    cnodes = FindSharedNodes (tet, pointsOnFace, nodes);
  
    /* process cases of 1, 2, 3, and 4 common nodes */
    switch (cnodes)
    {
      case 1:
      case 2:
      {
        break;
      } 
      case 3:
      {
        int found = 0;
        /* check whether partial face has been already recovered */
        Tos_MStackTraversalInit(face->splitFaces);
        while((partFace = Tos_MStackTraverse(face->splitFaces)) != NULL)
        {
          if(TetHasFace(tet, partFace, &facenum) == TRUE)
          {
            found = 1;
          }
        }
        /* recover partial face only once */
        if(found == 0)
        {
          /* determine face number
           * the node and the opposite face has the same index,
           * therefore substracting the indices of the nodes of
           * the face gives you the face index,
           * Roman was tricky :-) */
          facenum = 6 - nodes[0] - nodes[1] - nodes[2];
          /* create a face */
          partFace = FaceCreate();
          partFace->mesh = face->mesh;
          /* set vertices of partial face */
          GetFaceFromTet (tet, facenum, partFace);
          /* store partial face with the stack */
          Tos_MStackPush(face->splitFaces, (void *)partFace);
        }
        break;
      }
      default:
      {
        e_GeneralError ("RecoverFace: Internal error %s %d (%s%d)\n",
                        "invalid number of common nodes ",
                        cnodes, "line: ", __LINE__);
      }
    } /* end of switch (cnodes) */
  }
  
  Tos_MStackFree(&adjTetrah);
  
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 RecoverFace -- 
        Check if face is present in mesh construction and if not
        recover it by direct mesh modification.

 CAUTION : Face recovery will only work if face is completly
           enclosed within mesh (see note on counting Nt below). 
 Return value:
        Returns 0 if face was present in tetrahedral mesh 
        otherwise return number of triangles forming that face. 

 Note: This function has two return points.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int RecoverFace (TetrahMesh *tetMesh, Face *face)
{
  Tos_MStack pointsOnFace;
  Tos_MStack adjTetrah; /* tetrah adjacent to point on face */
  Tos_MStack intersectingEdges; /* edges intersectin face */
  int        Nh; /* number of meshpoint on face edges */
  int        Nt; /* number of triangles forming face */
  int        Nn; /* number of meshpoints on face (including face edges) */
  int        cnodes; /* number of nodec shared between face andt
                        terahedron adjacent to it */
  int        nodes[4]; /* generic numbers of vertices shared between
                          tetrahedron and face */
  int        where, gennum;
  double     isCoords[3], mu, dummy[3]; 
  Vertex    *intersectionPoint;
  Tetrah    *tet;
  Edge      *edge; 
  int        boundaryFlag=0;


  if(   (FACE_VERTEX(face, 0)->flag & POINT_EXT_BOUNDARY)
     && (FACE_VERTEX(face, 1)->flag & POINT_EXT_BOUNDARY)
     && (FACE_VERTEX(face, 2)->flag & POINT_EXT_BOUNDARY) )
  {
    boundaryFlag = POINT_EXT_BOUNDARY;
  }
  else if(   (FACE_VERTEX(face, 0)->flag & POINT_INT_BOUNDARY)
          && (FACE_VERTEX(face, 1)->flag & POINT_INT_BOUNDARY)
          && (FACE_VERTEX(face, 2)->flag & POINT_INT_BOUNDARY) )
  {
    boundaryFlag = POINT_INT_BOUNDARY;
  }
  else
  {
    e_GeneralError("Face to be recovered has no properly flaged nodes\n");
  }
  
  /* create stacks necessary in recovery */
  pointsOnFace      = Tos_MStackNew ();
  adjTetrah         = Tos_MStackNew ();
  intersectingEdges = Tos_MStackNew ();

  Nt = 0;
  /* find points lying on face edges, all of them, inc. corners */
  Nh = FindPointsOnFaceEdges(face, pointsOnFace);
  Nn = Nh; 
  
  /* modify mesh until face is recovered */
  /* CONSTCOND */
  while (1)
  {
    assert (Tos_MStackCount (adjTetrah) == 0);
    /* find all tetrahedra which has at least one of the
       points on the stack */
    FindPointsTetrahAdjSet (pointsOnFace, adjTetrah);
    Nt=0; 
    /* examine adjacent tetrahedra and find edges which possibly
     * intersect face 
     */
    while (Tos_MStackCount (adjTetrah) > 0)
    {
      tet = (Tetrah *) Tos_MStackPop (adjTetrah);
      
      /* find number of shared nodes between 'tet' and
         list of points on the edges of the face */
      cnodes = FindSharedNodes (tet, pointsOnFace, nodes);
      
      /* proces cases of 1, 2, 3, and 4 common nodes */
      switch (cnodes)
      {
        case 1:
        case 2:
            FindEdgesIntersectingFacePlane (tet, face, nodes, cnodes, 
                                            intersectingEdges);
            break;
        case 3:
            /* increment number of triangle forming face */
            Nt++;
            break;
        case 4:
            e_GeneralError 
                ("RecoverFace: Invalid mesh construction %s (%s%d)\n", 
                 "four points of tet on face", "line: ", __LINE__);  
        default:
            e_GeneralError 
                ("RecoverFace: Internal error %s %d (%s%d)\n", 
                 "invalid number of common nodes ",cnodes, "line: ", __LINE__);
      } /* end of switch (cnodes) */
    }

    /* NOTE ON COUNTING Nt
       As we are assuming that the recovered face is completly enclosed within
       the mesh each subface will be a side of two tetrahedra
       and therefore will be counted twice
     */
    assert (Nt%2 == 0);
    Nt = Nt/2;
    
    /* Check if face has been recovered, Euler formula */
    if (Nt == 2 * (Nn-1) - Nh)
    {
      /* do the actual recovery */
      DoRecoverFace(face, pointsOnFace);
      
      /* delete created stacks */
      /* CAUTION - EDGES SHOULD BE DELETED BEFORE STACK IS DELETED */
      while (Tos_MStackCount (intersectingEdges) >0)
      {
         edge = (Edge *) Tos_MStackPop (intersectingEdges); 
         EdgeDelete (edge);
      }
      Tos_MStackFree (&intersectingEdges);
      
      Tos_MStackFree (&pointsOnFace); 
      assert (Tos_MStackCount (adjTetrah) == 0);
      Tos_MStackFree (&adjTetrah); 
      return Nt;
    }
    
    /* Traverse stack of intersecting edges, find intersecting points
     * and insert them on intersecting edges in order to recover face
     */  
    Tos_MStackTraversalInit (intersectingEdges);
    while (Tos_MStackCount (intersectingEdges) > 0)
    {
      edge = (Edge *) Tos_MStackPop (intersectingEdges);
      /* find edge face intersection point */
      where = eg_V1_TriangLineIntersect3D (FACE_VCOORDS (face, 0),
                                           FACE_VCOORDS (face, 1), 
                                           FACE_VCOORDS (face, 2), 
                                           EDGE_VCOORDS (edge, 0),
                                           EDGE_VCOORDS (edge, 1),
                                           isCoords, &gennum, dummy, &mu);

      if (where == EG_INSIDE)
      {
        if (mu>=0.0 && mu<1.0e-8)
        { /* edge only touches face with its 0 node*/
          /* Check if that node is already on stack */
          if(Tos_MStackSearch(pointsOnFace,
                              (void *)(EDGE_VERTEX(edge, 0))) == 0)
          {
            Tos_MStackPush (pointsOnFace, (void *)(EDGE_VERTEX(edge, 0)));
            Nn++;
          }
        }
        else if (mu<=1.0 && mu>(1.0-1.0e-8))
        { /* edge only touches face with its 1 node*/
          /* Check if that node is already on stack */
          if(Tos_MStackSearch(pointsOnFace,
                              (void *)(EDGE_VERTEX (edge, 1))) == 0)
          {
            Tos_MStackPush(pointsOnFace, (void *)(EDGE_VERTEX(edge, 1)));
            Nn++;
          }
        }
        else
        {
          /* create the intersection point */
          intersectionPoint = VertexCreate();
          intersectionPoint->coords[0] = isCoords[0];
          intersectionPoint->coords[1] = isCoords[1];
          intersectionPoint->coords[2] = isCoords[2];
          /* inserts the intersecting point and
             updates the mesh all around, tetrahedra, etc. */
          InsertPointOnEdge (tetMesh, intersectionPoint, edge);
          /* flag it properly */
          intersectionPoint->flag |= POINT_BOUNDARY;
          intersectionPoint->flag |= POINT_ON_FACE;
          intersectionPoint->flag |= boundaryFlag;
          /* this is a new point on the face */
          Tos_MStackPush (pointsOnFace, (void *) intersectionPoint);  
          Nn++;
        }
      }
      /* These would signal invalid mesh configuration */
      else if (where == EG_ONEDGE || where == EG_ONVERTEX)
      {
        e_GeneralError("RecoverFace: Internal error %s %d (%s%d)\n", 
                       "Edge does not intersect face interior",
                       where, "line: ", __LINE__);
      }
      /* delete edge */
      EdgeDelete(edge);
    }
  }

  /* This statement will never be reached */
  /* NOTREACHED */ 
  return 0; 
} /* end of RecoverFace */


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 CheckPointBall --
        Check if moving point 'org' to point 'dest' will not 
        invalidate mesh construction.
        The check is based on calculation of signed volume
        of every tetrahedron in point ball. If the sign of
        volume changes it means that the tetrahedron will be 'twisted'
        and moving 'org' to 'dest' will invalidate mesh.
        At the same time if the tetrahedron becomes degenerate (zero volume)
        and does not have 'dest' vertex it means that we cannot move
        'org' to dest.
        
 Arguments:
        org - origin point
        dest - destination point
 Return value:
        1 - if mesh will not be invalidated
        0 - otherwise 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int CheckPointBall (Vertex *org, Vertex *dest)
{
  Vertex *vertTable[4]; 
  Tetrah *tet; 
  double vol1, vol2, vol1Sign, vol2Sign;
  int orgNum, i; 
  int skip = 0; 
  int result = 1;

  Tos_MStackTraversalInit (org->adjElem); 
  while ((tet = Tos_MStackTraverse (org->adjElem)) != NULL)
  {
    for (i=0; i<4; i++)
    {
      /* check if 'tet' has 'dest' as a vertex -- 
       * if yes skip that tetrahedron (as it vol2 will be 0 (zero)
       */  
      if (TET_VERTEX (tet, i) == dest)
      {
        skip = 1; 
        break;
      }
      /* check which vertex number has 'org' */
      if (TET_VERTEX (tet, i) == org)
      {
        orgNum = i;
      }
      vertTable[i] = TET_VERTEX (tet, i); 
    }
    if (skip == 1)
    { 
      skip = 0; 
      continue;  /* skip 'tet' and go to next one */
    }

    /* calculate signed volume of tet */
    vol1 = eg_Orient3D (vertTable[0]->coords,  
                        vertTable[1]->coords, 
                        vertTable[2]->coords, 
                        vertTable[3]->coords); 

    /* substitute 'dest' to 'org' in 'vertTable' */
    vertTable[orgNum] = dest; 
    vol2 = eg_Orient3D (vertTable[0]->coords,  
                        vertTable[1]->coords, 
                        vertTable[2]->coords, 
                        vertTable[3]->coords); 
    
    vol1Sign = copysign (1.0, vol1); 
    vol2Sign = copysign (1.0, vol2); 
    if (vol1Sign * vol2Sign < 0)
    {
       /* node move will invalidate mesh construction */
       result = 0; 
    }
    if (fabs(vol2) < ZERO_VOLUME)
    {
      /* node move will invalidate mesh construction */
      result = 0;
    }
  }

  return result;                         
} /* end of CheckPointBall */



/* Modifications:
 * $Log: recover.c,v $
 * Revision 1.11  2002/07/26 14:58:19  peteri
 * - fix number of recovered faces
 *
 * Revision 1.10  2002/07/26 11:53:45  peteri
 * - nana is removed for simplicity
 * - algorithm is adjusted to new boundary data structure
 *
 * Revision 1.9  2002/03/18 16:00:05  peteri
 * Algorithm clean-up, some functions are removed, more comments.
 *
 * Revision 1.8  2002/02/24 19:10:32  romek
 * *** empty log message ***
 *
 */
