/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                           The E-Lib Header File                            */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* An appropriate selection of macros must be defined before the inclusion of
   this header file. The following table summarises the choice of macros for
   each E-Lib library.


   Library                 Description                Compilation Macros
   -----------------       ------------------------   ------------------------ 

   e_lib_sparc.a           Command line Sparc         __E_CLANSIC__
   e_lib_sparc_xwin.a      X-Windows Sparc            __E_XWIN__
   e_lib_sparc_mpi.a       MPI command line Sparc     __E_MPI__, __E_CLANSIC__
   e_lib_sparc_pvm.a       PVM command line Sparc     __E_PVM__, __E_CLANSIC__
   e_lib_m.t8              Transputer full run-time   __E_T800RTL__
   e_lib_w.t8              Transputer stand-alone     __E_T800SAL__

   e_libdos.lib            Command line DOS           __E_CLANSIC__
   e_libw16.lib            MS-Win16 API               __E_MSWIN16__
   e_libw32.lib            MS-Win32 API               __E_MSWIN32__
   e_lib_w32_con.lib       MS-Win32 console           __E_CLANSIC__
   e_lib_w32_con_mpi.lib   MPI MS-Win32 console       __E_MPI__, __E_CLANSIC__


   The macro __E_CLANSIC__ implies that standard input-output is via the
   command line, whether this be through a DOS terminal screen, a UNIX or
   MS-Win32 console window or some such equivalent. The three macros
   __E_XWIN__,  __E_MSWIN16__ and __E_MSWIN32__ denote that the program has a
   graphical user interface.

   Although transputer programs are run from the command line, the macros
   __E_T800RTL__ and __E_T800SAL__ must not be defined along with
   __E_CLANSIC__.

   The macro __E_SEQUENTIAL__ can also be defined in source files to denote
   that the program is sequential. Note that the E-Lib library is not
   dependent upon __E_SEQUENTIAL__.

*//* //////////////////////////////////////////////////////////////////////// */

/* to protect against recursive inclusion */
#ifndef _E_LIB_H
#define _E_LIB_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


#ifdef __E_MSWIN16__
#define __E_MSWIN__
#endif

#ifdef __E_MSWIN32__
#define __E_MSWIN__
#endif

#ifdef __E_T800RTL__
#define __E_T800__
#endif

#ifdef __E_T800SAL__
#define __E_T800__
#endif


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                          Element Type Definitions                          */
/*                                                                            */
/* //////////////////////////////////////////////////////////////////////////  /


      Element Type    Element Definition
      ------------    ------------------
      LINK1           Truss with 2 end nodes
      LINK2           Cable with 2 end nodes
      LINK3           Fixed tension 1-D link with 2 end nodes
      LINK4           Fixed force density 1-D link with 2 end nodes
      LINK5           G-string with 2 end nodes
      TRIANG1         Constant strain triangle with 3 vertex nodes
      TRIANG2         Combined constant strain plane stress and constant
                      moment plate bending simple facet triangular element
                      with multiple layers, 3 vertex nodes and 3 mid-side
                      nodes, giving 12 degrees of freedom
      TRIANG3         Solid triangle
      TRIANG4         Triangular membrane
      TRIANG5         Constant stress triangle
      QUAD1           Plane stress quadrilateral with 4 vertex nodes
      QUAD3           Plane and bending stress quadrilateral with 9 nodes
                      Each node has 5 degree of freedom
      TETRAH1         Tetrahedral with 4 vertex nodes
      BLOCK1          Block element with 8 vertex nodes


*//* //////////////////////////////////////////////////////////////////////// */


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                           MESH Structure Macros                            */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* Array length macros */

#define E_NDIMS                3     /* number of spatial dimensions */
#define E_DOF_MAX              6     /* maximum nodal degree of freedom */
#define E_NELEM_TYPES         14     /* number of element types */
#define E_NKEYWORDS           75     /* number of .mdf keywords */

#define E_NMAT_PROP_KEYWORDS  35     /* number of material property keywords */
#define E_NSUBSTRUCTS          6     /* number of sub-structure file types */
#define E_NC_NLOAD_TYPES       3     /* number of NURBS curve load types */

/* Element type indices */

#define E_ELEM_TYPE_LINK1            0
#define E_ELEM_TYPE_LINK2            1
#define E_ELEM_TYPE_LINK3            2
#define E_ELEM_TYPE_LINK4            3
#define E_ELEM_TYPE_LINK5            4
#define E_ELEM_TYPE_TRIANG1          5
#define E_ELEM_TYPE_TRIANG2          6
#define E_ELEM_TYPE_TRIANG3          7
#define E_ELEM_TYPE_TRIANG4          8
#define E_ELEM_TYPE_TRIANG5          9
#define E_ELEM_TYPE_QUAD1           10
#define E_ELEM_TYPE_QUAD3           11
#define E_ELEM_TYPE_TETRAH1         12
#define E_ELEM_TYPE_BLOCK1          13

/* Vertex node degrees of freedom */

#define E_VERTEX_NODE_DOF_LINK1      3
#define E_VERTEX_NODE_DOF_LINK2      3
#define E_VERTEX_NODE_DOF_LINK3      3
#define E_VERTEX_NODE_DOF_LINK4      3
#define E_VERTEX_NODE_DOF_LINK5      3
#define E_VERTEX_NODE_DOF_TRIANG1    3
#define E_VERTEX_NODE_DOF_TRIANG2    3
#define E_VERTEX_NODE_DOF_TRIANG3    3
#define E_VERTEX_NODE_DOF_TRIANG4    3
#define E_VERTEX_NODE_DOF_TRIANG5    3
#define E_VERTEX_NODE_DOF_QUAD1      3
#define E_VERTEX_NODE_DOF_QUAD3      5
#define E_VERTEX_NODE_DOF_TETRAH1    3
#define E_VERTEX_NODE_DOF_BLOCK1     3

/* Non-vertex node degrees of freedom */

#define E_MIDSIDE_NODE_DOF_TRIANG2   1
#define E_MIDSIDE_NODE_DOF_QUAD3     5

/* Number of vertices per element */

#define E_NVPE_LINK1                 2
#define E_NVPE_LINK2                 2
#define E_NVPE_LINK3                 2
#define E_NVPE_LINK4                 2
#define E_NVPE_LINK5                 2
#define E_NVPE_TRIANG1               3
#define E_NVPE_TRIANG2               3
#define E_NVPE_TRIANG3               3
#define E_NVPE_TRIANG4               3
#define E_NVPE_TRIANG5               3
#define E_NVPE_QUAD1                 4
#define E_NVPE_QUAD3                 4
#define E_NVPE_TETRAH1               4
#define E_NVPE_BLOCK1                8

/* Number of nodes per element */

#define E_NNPE_LINK1                 2
#define E_NNPE_LINK2                 2
#define E_NNPE_LINK3                 2
#define E_NNPE_LINK4                 2
#define E_NNPE_LINK5                 2
#define E_NNPE_TRIANG1               3
#define E_NNPE_TRIANG2               6
#define E_NNPE_TRIANG3               3
#define E_NNPE_TRIANG4               3
#define E_NNPE_TRIANG5               3
#define E_NNPE_QUAD1                 4
#define E_NNPE_QUAD3                 9
#define E_NNPE_TETRAH1               4
#define E_NNPE_BLOCK1                8

/* Number of interfaces per element */

#define E_NIPE_TRIANG1               3
#define E_NIPE_TRIANG2               3
#define E_NIPE_TRIANG3               3
#define E_NIPE_TRIANG4               3
#define E_NIPE_TRIANG5               3
#define E_NIPE_QUAD1                 4
#define E_NIPE_QUAD3                 4
#define E_NIPE_TETRAH1               4
#define E_NIPE_BLOCK1                6

/* Number of vertices per interface */

#define E_NVPI_TRIANG1               2
#define E_NVPI_TRIANG2               2
#define E_NVPI_TRIANG3               2
#define E_NVPI_TRIANG4               2
#define E_NVPI_TRIANG5               2
#define E_NVPI_QUAD1                 2
#define E_NVPI_QUAD3                 2
#define E_NVPI_TETRAH1               3
#define E_NVPI_BLOCK1                4

/* Number of nodes per interface */

#define E_NNPI_TRIANG1               2
#define E_NNPI_TRIANG2               3
#define E_NNPI_TRIANG3               2
#define E_NNPI_TRIANG4               2
#define E_NNPI_TRIANG5               2
#define E_NNPI_QUAD1                 2
#define E_NNPI_QUAD3                 3
#define E_NNPI_TETRAH1               3
#define E_NNPI_BLOCK1                4

/* Number of edges per element */

#define E_NEPE_TRIANG1               3
#define E_NEPE_TRIANG2               3
#define E_NEPE_TRIANG3               3
#define E_NEPE_TRIANG4               3
#define E_NEPE_TRIANG5               3
#define E_NEPE_QUAD1                 4
#define E_NEPE_QUAD3                 4
#define E_NEPE_TETRAH1               6
#define E_NEPE_BLOCK1               12

/* Number of edges per interface */

#define E_NEPI_TRIANG1               1
#define E_NEPI_TRIANG2               1
#define E_NEPI_TRIANG3               1
#define E_NEPI_TRIANG4               1
#define E_NEPI_TRIANG5               1
#define E_NEPI_QUAD1                 1
#define E_NEPI_QUAD3                 1
#define E_NEPI_TETRAH1               3
#define E_NEPI_BLOCK1                4

/* "SubStructSet" Boolean indices */

#define E_SS_NUM_GEOM                0
#define E_SS_NUM_MAT_PROPS           1
#define E_SS_NUM_DECOMP              2
#define E_SS_NUM_MESHPARAM           3
#define E_SS_NUM_STRESS              4
#define E_SS_NUM_FEERROR             5

/* Direction indices */

#define E_XDIR   0              /* x-direction */
#define E_YDIR   1              /* y-direction */
#define E_ZDIR   2              /* z-direction */

/* Stress components */

#define E_NUM_STRESS_COMPONENTS    6

#define E_SIGMA_X    0
#define E_SIGMA_Y    1
#define E_SIGMA_Z    2
#define E_TAU_XY     3
#define E_TAU_YZ     4
#define E_TAU_ZX     5

/* Segment refinement */

#define E_SEGREF_NONE 	0
#define E_SEGREF_ODD 	1
#define E_SEGREF_EVEN	2


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                              NURBS Structure                               */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

typedef struct
{
   char *Name;                  /* name */

   int Degree;                  /* degree */
   int EndPoint1;               /* first end-point */
   int EndPoint2;               /* second end-point */
   double **CntlPoints;         /* control point coordinates and weights */

   int *BCTypes;                /* boundary condition types */
   double *BCDispls;            /* initial nodal displacements */
   double *BCSpringConsts;      /* spring constants per unit length */
   int FixityType;              /* global or local coord. sys. */

   double *PointLoad;           /* point load */
   double  PointLoadParam;      /* point load curve parameter */
   double *LinLoadGCBeg;        /* initial linear load in global coord. sys. */
   double *LinLoadGCEnd;        /* final linear load in global coord. sys. */
   double *LinLoadCCBeg;        /* initial linear load in curve coord. sys. */
   double *LinLoadCCEnd;        /* final linear load in curve coord. sys. */
   int LoadTypeSet[E_NC_NLOAD_TYPES];   /* which load types are set */
}
   NURBSCURVE;


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                               GEOM Structure                               */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

typedef struct
{
   int NEndPoints;              /* number of end-points */
   int NNurbsCurves;            /* number of NURBS curves */

   double **EndPointCoords;     /* end-point coordinates */

   NURBSCURVE *NurbsCurves;     /* NURBS curves */
}
   GEOM;


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                               MAT Structure                                */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

typedef struct
{
   char *Name;                  /* material name */

   int MatType;                 /* index for a user defined material model */

   double Density;              /* density */
   double YMod;                 /* isotropic Young's modulus */
   double YMod_x;               /* x-direction Young's modulus */
   double YMod_y;               /* y-direction Young's modulus */
   double YMod_z;               /* z-direction Young's modulus */
   double PoissonsRatio;        /* Poisson's ratio */
   double Thickness;            /* thickness */

   int iParam1;                 /* /////////////////////////////////////// */
   int iParam2;                 /*                                         */
   int iParam3;                 /* These variables enable users to define  */
   int iParam4;                 /* their own material properties.          */
   int iParam5;                 /*                                         */
   int iParam6;                 /* The interpretation of the properties    */
                                /* can be made dependent upon the material */
   double dParam1;              /* model indexed by the integer variable   */
   double dParam2;              /* "MatType". The definition of each       */
   double dParam3;              /* property should be clearly documented.  */
   double dParam4;              /*                                         */
   double dParam5;              /* It is up to the user to check the       */
   double dParam6;              /* validity of user defined property       */
   double dParam7;              /* values once the material data have been */
   double dParam8;              /* read.                                   */
   double dParam9;              /*                                         */
   double dParam10;             /*                                         */
   double dParam11;             /*                                         */
   double dParam12;             /*                                         */
   double dParam13;             /*                                         */
   double dParam14;             /*                                         */
   double dParam15;             /*                                         */
   double dParam16;             /*                                         */
   double dParam17;             /*                                         */
   double dParam18;             /*                                         */
   double dParam19;             /*                                         */
   double dParam20;             /* /////////////////////////////////////// */

   int KeywordSet[E_NMAT_PROP_KEYWORDS];
                                /* material keyword set Booleans */
}
   MAT;


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                             COMPMAT1 Structure                             */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

typedef struct
{
   char *Name;                          /* composite material name */
   int NLayers;                         /* number of layers */
   MAT **LayerMats;                     /* layer materials */
   double *Thicknesses;                 /* layer thicknesses */
}
   COMPMAT1;


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                              DECOMP Structure                              */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

typedef struct
{
   int NSubDoms;                        /* number of subdomains  */
   int *SubDoms;                        /* subdomain indices */
   int *SubDomsLink1;                   /* LINK1 subdomain indices */
   int *SubDomsLink2;                   /* LINK2 subdomain indices */
   int *SubDomsLink3;                   /* LINK3 subdomain indices */
   int *SubDomsLink4;                   /* LINK4 subdomain indices */
   int *SubDomsLink5;                   /* LINK5 subdomain indices */
   int *SubDomsTriang1;                 /* TRIANG1 subdomain indices */
   int *SubDomsTriang2;                 /* TRIANG2 subdomain indices */
   int *SubDomsTriang3;                 /* TRIANG3 subdomain indices */
   int *SubDomsTriang4;                 /* TRIANG4 subdomain indices */
   int *SubDomsTriang5;                 /* TRIANG5 subdomain indices */
   int *SubDomsQuad1;                   /* QUAD1 subdomain indices */
   int *SubDomsQuad3;                   /* QUAD3 subdomain indices */
   int *SubDomsTetrah1;                 /* TETRAH1 subdomain indices */
   int *SubDomsBlock1;                  /* BLOCK1 subdomain indices */
}
   DECOMP;


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                            MESHPARAM Structure                             */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

typedef struct
{
   double *NodalMeshParams;             /* nodal mesh parameters */
   double *ElemMeshParams;              /* element mesh parameters */
   double *ElemMeshParamsLink1;         /* LINK1 element mesh parameters */
   double *ElemMeshParamsLink2;         /* LINK2 element mesh parameters */
   double *ElemMeshParamsLink3;         /* LINK3 element mesh parameters */
   double *ElemMeshParamsLink4;         /* LINK4 element mesh parameters */
   double *ElemMeshParamsLink5;         /* LINK5 element mesh parameters */
   double *ElemMeshParamsTriang1;       /* TRIANG1 element mesh parameters */
   double *ElemMeshParamsTriang2;       /* TRIANG2 element mesh parameters */
   double *ElemMeshParamsTriang3;       /* TRIANG3 element mesh parameters */
   double *ElemMeshParamsTriang4;       /* TRIANG4 element mesh parameters */
   double *ElemMeshParamsTriang5;       /* TRIANG5 element mesh parameters */
   double *ElemMeshParamsQuad1;         /* QUAD1 element mesh parameters */
   double *ElemMeshParamsQuad3;         /* QUAD3 element mesh parameters */
   double *ElemMeshParamsTetrah1;       /* TETRAH1 element mesh parameters */
   double *ElemMeshParamsBlock1;        /* BLOCK1 element mesh parameters */

   int NodalMeshParamsSet;              /* nodal mesh parameters set? */
   int ElemMeshParamsSet;               /* element mesh parameters set? */
}
   MESHPARAM;


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                             FEERROR Structure                              */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

typedef struct
{
   double *FEErrors;                 /* finite element errors */
   double *FEErrorsLink1;            /* LINK1 finite element errors */
   double *FEErrorsLink2;            /* LINK2 finite element errors */
   double *FEErrorsLink3;            /* LINK3 finite element errors */
   double *FEErrorsLink4;            /* LINK4 finite element errors */
   double *FEErrorsLink5;            /* LINK5 finite element errors */
   double *FEErrorsTriang1;          /* TRIANG1 finite element errors */
   double *FEErrorsTriang2;          /* TRIANG2 finite element errors */
   double *FEErrorsTriang3;          /* TRIANG3 finite element errors */
   double *FEErrorsTriang4;          /* TRIANG4 finite element errors */
   double *FEErrorsTriang5;          /* TRIANG5 finite element errors */
   double *FEErrorsQuad1;            /* QUAD1 finite element errors */
   double *FEErrorsQuad3;            /* QUAD3 finite element errors */
   double *FEErrorsTetrah1;          /* TETRAH1 finite element errors */
   double *FEErrorsBlock1;           /* BLOCK1 finite element errors */
}
   FEERROR;


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                             STRESSES Structure                             */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

typedef struct
{
   double  ***StressesLink1;            /* LINK1 finite element stresses */
   double  ***StressesLink2;            /* LINK2 finite element stresses */
   double  ***StressesLink3;            /* LINK3 finite element stresses */
   double  ***StressesLink4;            /* LINK4 finite element stresses */
   double  ***StressesLink5;            /* LINK5 finite element stresses */

   double  ***StressesTriang1;          /* TRIANG1 finite element stresses */
   double ****StressesTriang2;          /* TRIANG2 finite element stresses */
   double  ***StressesTriang3;          /* TRIANG3 finite element stresses */
   double  ***StressesTriang4;          /* TRIANG4 finite element stresses */
   double  ***StressesTriang5;          /* TRIANG5 finite element stresses */

   double  ***StressesQuad1;            /* QUAD1 finite element stresses */
   double ****StressesQuad3;            /* QUAD3 finite element stresses */
   double  ***StressesTetrah1;          /* TETRAH1 finite element stresses */
   double  ***StressesBlock1;           /* BLOCK1 finite element stresses */
}
   STRESSES;



/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                               MESH Structure                               */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* NOTE: See above for a brief definition of each element type.               */


typedef struct
{
   char *Title;                      /* title */

   int NMeshPoints;                  /* number of mesh-points */
   int NNodes;                       /* number of finite element nodes */
   int NElemsLink1;                  /* number of LINK1 elements */
   int NElemsLink2;                  /* number of LINK2 elements */
   int NElemsLink3;                  /* number of LINK3 elements */
   int NElemsLink4;                  /* number of LINK4 elements */
   int NElemsLink5;                  /* number of LINK5 elements */
   int NElemsTriang1;                /* number of TRIANG1 elements */
   int NElemsTriang2;                /* number of TRIANG2 elements */
   int NElemsTriang3;                /* number of TRIANG3 elements */
   int NElemsTriang4;                /* number of TRIANG4 elements */
   int NElemsTriang5;                /* number of TRIANG5 elements */
   int NElemsQuad1;                  /* number of QUAD1 elements */
   int NElemsQuad3;                  /* number of QUAD3 elements */
   int NElemsTetrah1;                /* number of TETRAH1 elements */
   int NElemsBlock1;                 /* number of BLOCK1 elements */
   int TotalNElems;                  /* total number of elements */
   int NBCNodes;                     /* number of boundary condition nodes */
   int NLoadedNodes;                 /* number of loaded nodes */
   int NMats;                        /* number of materials */
   int NMdfMats;                     /* number of .mdf declared materials */
   int NCompMatsType1;               /* number of type 1 composite materials */
   int NTimeSteps;                   /* number of time-steps */
   int NInternalTimeSteps;           /* number of internal time-steps */
   int NExternalTimeSteps;           /* number of external time-steps */
   int FirstNetIndex[E_NELEM_TYPES]; /* first net index of each element type */
   double TimeStep;                  /* time-step */
   double DampingFactor;             /* viscous damping factor */
   double Beta;                      /* Newmark's beta integration constant */
   double Gamma;                     /* Newmark's gamma integration constant */

   int NStressPointsLink1;           /* number of stress points in LINK1 */
   int NStressPointsLink2;           /* number of stress points in LINK2 */
   int NStressPointsLink3;           /* number of stress points in LINK3 */
   int NStressPointsLink4;           /* number of stress points in LINK4 */
   int NStressPointsLink5;           /* number of stress points in LINK5 */
   int NStressPointsTriang1;         /* number of stress points in TRIANG1 */
   int NStressPointsTriang2;         /* number of stress points in TRIANG2 */
   int NStressPointsTriang3;         /* number of stress points in TRIANG3 */
   int NStressPointsTriang4;         /* number of stress points in TRIANG4 */
   int NStressPointsTriang5;         /* number of stress points in TRIANG5 */
   int NStressPointsQuad1;           /* number of stress points in QUAD1 */
   int NStressPointsQuad3;           /* number of stress points in QUAD3 */
   int NStressPointsTetrah1;         /* number of stress points in TETRAH1 */
   int NStressPointsBlock1;          /* number of stress points in BLOCK1 */

   double **MeshPointCoords;         /* mesh-point coordinates */

   int **NodesLink1;                 /* LINK1 node indices */
   int **NodesLink2;                 /* LINK2 node indices */
   int **NodesLink3;                 /* LINK3 node indices */
   int **NodesLink4;                 /* LINK4 node indices */
   int **NodesLink5;                 /* LINK5 node indices */
   int **NodesTriang1;               /* TRIANG1 node indices */
   int **NodesTriang2;               /* TRIANG2 node indices */
   int **NodesTriang3;               /* TRIANG3 node indices */
   int **NodesTriang4;               /* TRIANG4 node indices */
   int **NodesTriang5;               /* TRIANG5 node indices */
   int **NodesQuad1;                 /* QUAD1 node indices */
   int **NodesQuad3;                 /* QUAD3 node indices */
   int **NodesTetrah1;               /* TETRAH1 node indices */
   int **NodesBlock1;                /* BLOCK1 node indices */

   int TotalNodalDOF;                /* total nodal degrees of freedom */
   int *NodalDOF;                    /* nodal degrees of freedom */

   int  *BCNodes;                    /* boundary condition nodes */
   int **BCTypes;                    /* boundary condition types */
   double **BCDispls;                /* initial nodal displacements */
   double **BCSpringConsts;          /* spring constants */

   int *LoadedNodes;                 /* loaded nodes */
   double **Loads;                   /* loads */

   void **Mats;                      /* materials */
   MAT  **MatsLink1;                 /* LINK1 materials */
   MAT  **MatsLink2;                 /* LINK2 materials */
   MAT  **MatsLink3;                 /* LINK3 materials */
   MAT  **MatsLink4;                 /* LINK4 materials */
   MAT  **MatsLink5;                 /* LINK5 materials */
   MAT  **MatsTriang1;               /* TRIANG1 materials */
   COMPMAT1 **CompMatsTriang2;       /* TRIANG2 composite materials */
   MAT  **MatsTriang3;               /* TRIANG3 materials */
   MAT  **MatsTriang4;               /* TRIANG4 materials */
   MAT  **MatsTriang5;               /* TRIANG5 materials */
   MAT  **MatsQuad1;                 /* QUAD1 materials */
   COMPMAT1 **CompMatsQuad3;         /* QUAD3 composite materials */
   MAT  **MatsTetrah1;               /* TETRAH1 materials */
   MAT  **MatsBlock1;                /* BLOCK1 materials */

   int *NodeTypes;                   /* remeshing node types */
   int *NNurbsCurves;                /* number of NURBS curves per node */
   double **NurbsCurveParams;        /* NURBS curve parameters */
   NURBSCURVE ***NurbsCurves;        /* node NURBS curves */

   int *GNIs;                        /* global node indices */
   int *GEIs;                        /* global element indices */

   int *GEIsLink1;                   /* LINK1 global element indices */
   int *GEIsLink2;                   /* LINK2 global element indices */
   int *GEIsLink3;                   /* LINK3 global element indices */
   int *GEIsLink4;                   /* LINK4 global element indices */
   int *GEIsLink5;                   /* LINK5 global element indices */
   int *GEIsTriang1;                 /* TRIANG1 global element indices */
   int *GEIsTriang2;                 /* TRIANG2 global element indices */
   int *GEIsTriang3;                 /* TRIANG3 global element indices */
   int *GEIsTriang4;                 /* TRIANG4 global element indices */
   int *GEIsTriang5;                 /* TRIANG5 global element indices */
   int *GEIsQuad1;                   /* QUAD1 global element indices */
   int *GEIsQuad3;                   /* QUAD3 global element indices */
   int *GEIsTetrah1;                 /* TETRAH1 global element indices */
   int *GEIsBlock1;                  /* BLOCK1 global element indices */

   GEOM Geom;                        /* geometric model */
   MAT *Props;                       /* material properties */
   COMPMAT1 *CompPropsType1;         /* type 1 composite material properties */
   DECOMP Decomp;                    /* decomposition data */
   MESHPARAM MeshParam;              /* mesh parameters */
   FEERROR FEError;                  /* finite element errors */
   STRESSES Stresses;                /* finite element stresses */

   int KeywordSet[E_NKEYWORDS];      /* keyword Booleans */
   int SubStructSet[E_NSUBSTRUCTS];  /* sub-structure Booleans */
}
   MESH;


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                                  Keywords                                  */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#define E_KEYWORD_MAX  48                   /* maximum keyword length */

/* Mesh definition file keywords */

#define E_KW_TITLE                          "TITLE"
#define E_KW_NMESHPOINTS                    "NMESHPOINTS"
#define E_KW_NNODES                         "NNODES"
#define E_KW_NELEMS_LINK1                   "NELEMENTS_LINK1"
#define E_KW_NELEMS_LINK2                   "NELEMENTS_LINK2"
#define E_KW_NELEMS_LINK3                   "NELEMENTS_LINK3"
#define E_KW_NELEMS_LINK4                   "NELEMENTS_LINK4"
#define E_KW_NELEMS_LINK5                   "NELEMENTS_LINK5"
#define E_KW_NELEMS_TRIANG1                 "NELEMENTS_TRIANG1"
#define E_KW_NELEMS_TRIANG2                 "NELEMENTS_TRIANG2"
#define E_KW_NELEMS_TRIANG3                 "NELEMENTS_TRIANG3"
#define E_KW_NELEMS_TRIANG4                 "NELEMENTS_TRIANG4"
#define E_KW_NELEMS_TRIANG5                 "NELEMENTS_TRIANG5"
#define E_KW_NELEMS_QUAD1                   "NELEMENTS_QUAD1"
#define E_KW_NELEMS_QUAD3                   "NELEMENTS_QUAD3"
#define E_KW_NELEMS_TETRAH1                 "NELEMENTS_TETRAH1"
#define E_KW_NELEMS_BLOCK1                  "NELEMENTS_BLOCK1"
#define E_KW_NBC_NODES                      "NBOUNDARY_CONDITION_NODES"
#define E_KW_NLOADED_NODES                  "NLOADED_NODES"
#define E_KW_NMATS                          "NMATERIALS"
#define E_KW_NCOMP_MATS_TYPE1               "NCOMP_MATERIALS_TYPE1"
#define E_KW_NNURBS_CURVES                  "NNURBS_CURVES"
#define E_KW_TIMESTEP                       "TIMESTEP"
#define E_KW_NTIMESTEPS                     "NTIMESTEPS"
#define E_KW_NINTERNAL_TIMESTEPS            "NINTERNAL_TIMESTEPS"
#define E_KW_NEXTERNAL_TIMESTEPS            "NEXTERNAL_TIMESTEPS"
#define E_KW_DAMPING_FACTOR                 "DAMPING_FACTOR"
#define E_KW_BETA                           "BETA"
#define E_KW_GAMMA                          "GAMMA"
#define E_KW_NSTRESS_POINTS_LINK1           "NSTRESS_POINTS_LINK1"
#define E_KW_NSTRESS_POINTS_LINK2           "NSTRESS_POINTS_LINK2"
#define E_KW_NSTRESS_POINTS_LINK3           "NSTRESS_POINTS_LINK3"
#define E_KW_NSTRESS_POINTS_LINK4           "NSTRESS_POINTS_LINK4"
#define E_KW_NSTRESS_POINTS_LINK5           "NSTRESS_POINTS_LINK5"
#define E_KW_NSTRESS_POINTS_TRIANG1         "NSTRESS_POINTS_TRIANG1"
#define E_KW_NSTRESS_POINTS_TRIANG2         "NSTRESS_POINTS_TRIANG2"
#define E_KW_NSTRESS_POINTS_TRIANG3         "NSTRESS_POINTS_TRIANG3"
#define E_KW_NSTRESS_POINTS_TRIANG4         "NSTRESS_POINTS_TRIANG4"
#define E_KW_NSTRESS_POINTS_TRIANG5         "NSTRESS_POINTS_TRIANG5"
#define E_KW_NSTRESS_POINTS_QUAD1           "NSTRESS_POINTS_QUAD1"
#define E_KW_NSTRESS_POINTS_QUAD3           "NSTRESS_POINTS_QUAD3"
#define E_KW_NSTRESS_POINTS_TETRAH1         "NSTRESS_POINTS_TETRAH1"
#define E_KW_NSTRESS_POINTS_BLOCK1          "NSTRESS_POINTS_BLOCK1"
#define E_KW_MESHPOINT_COORDS               "MESHPOINT_COORDINATES"
#define E_KW_NODES_LINK1                    "NODES_LINK1"
#define E_KW_NODES_LINK2                    "NODES_LINK2"
#define E_KW_NODES_LINK3                    "NODES_LINK3"
#define E_KW_NODES_LINK4                    "NODES_LINK4"
#define E_KW_NODES_LINK5                    "NODES_LINK5"
#define E_KW_NODES_TRIANG1                  "NODES_TRIANG1"
#define E_KW_NODES_TRIANG2                  "NODES_TRIANG2"
#define E_KW_NODES_TRIANG3                  "NODES_TRIANG3"
#define E_KW_NODES_TRIANG4                  "NODES_TRIANG4"
#define E_KW_NODES_TRIANG5                  "NODES_TRIANG5"
#define E_KW_NODES_QUAD1                    "NODES_QUAD1"
#define E_KW_NODES_QUAD3                    "NODES_QUAD3"
#define E_KW_NODES_TETRAH1                  "NODES_TETRAH1"
#define E_KW_NODES_BLOCK1                   "NODES_BLOCK1"
#define E_KW_BCS                            "BOUNDARY_CONDITIONS"
#define E_KW_LOADS                          "LOADS"
#define E_KW_MATS_LINK1                     "MATERIALS_LINK1"
#define E_KW_MATS_LINK2                     "MATERIALS_LINK2"
#define E_KW_MATS_LINK3                     "MATERIALS_LINK3"
#define E_KW_MATS_LINK4                     "MATERIALS_LINK4"
#define E_KW_MATS_LINK5                     "MATERIALS_LINK5"
#define E_KW_MATS_TRIANG1                   "MATERIALS_TRIANG1"
#define E_KW_COMP_MATS_TRIANG2              "COMP_MATERIALS_TRIANG2"
#define E_KW_MATS_TRIANG3                   "MATERIALS_TRIANG3"
#define E_KW_MATS_TRIANG4                   "MATERIALS_TRIANG4"
#define E_KW_MATS_TRIANG5                   "MATERIALS_TRIANG5"
#define E_KW_MATS_QUAD1                     "MATERIALS_QUAD1"
#define E_KW_COMP_MATS_QUAD3                "COMP_MATERIALS_QUAD3"
#define E_KW_MATS_TETRAH1                   "MATERIALS_TETRAH1"
#define E_KW_MATS_BLOCK1                    "MATERIALS_BLOCK1"
#define E_KW_REMESH_DATA                    "REMESH_DATA"

/* Remeshing data keywords */

#define E_RMKW_NNURBS_CURVES                "NNURBS_CURVES"
#define E_RMKW_NURBS_CURVE                  "NURBS_CURVE"
#define E_RMKW_NC_PARAM                     "PARAMETER"

/* Remeshing data keyword arguments */

#define E_RM_EXTERIOR                       "\"EXTERIOR\""
#define E_RM_INTERIOR                       "\"INTERIOR\""
#define E_RM_ISD_BOUNDARY                   "\"ISD_BOUNDARY\""

/* Boundary condition keyword arguments */

#define E_BC_FREE                           "\"FREE\""
#define E_BC_FIXED                          "\"FIXED\""
#define E_BC_SPRING                         "\"SPRING\""


/* ////////////////////////////////////////////////////////////////////////// */
/* Geometric model file keywords */

#define E_GKW_NENDPOINTS                    "NENDPOINTS"
#define E_GKW_NNURBS_CURVES                 "NNURBS_CURVES"
#define E_GKW_ENDPOINT_COORDS               "ENDPOINT_COORDINATES"

#define E_GKW_NC                            "NURBS_CURVE"
#define E_GKW_NC_DEGREE                     "DEGREE"
#define E_GKW_NC_CNTL_POINTS                "CONTROL_POINTS"
#define E_GKW_NC_WEIGHTS                    "WEIGHTS"
#define E_GKW_NC_FIXITY_GC                  "FIXITY_IN_GLOBAL_COORDS"
#define E_GKW_NC_FIXITY_CC                  "FIXITY_IN_CURVE_COORDS"
#define E_GKW_NC_POINT_LOAD                 "POINT_LOAD"
#define E_GKW_NC_POINT_LOAD_PARAM           "POINT_LOAD_PARAM"
#define E_GKW_NC_LIN_LOAD_GC_BEG         "LIN_DISTRIB_LOAD_IN_GLOBAL_COORDS_BEG"
#define E_GKW_NC_LIN_LOAD_GC_END         "LIN_DISTRIB_LOAD_IN_GLOBAL_COORDS_END"
#define E_GKW_NC_LIN_LOAD_CC_BEG         "LIN_DISTRIB_LOAD_IN_CURVE_COORDS_BEG"
#define E_GKW_NC_LIN_LOAD_CC_END         "LIN_DISTRIB_LOAD_IN_CURVE_COORDS_END"
#define E_GKW_NC_END                        "END"


/* ////////////////////////////////////////////////////////////////////////// */
/* Materials file keywords */

#define E_MKW_MAT                           "MATERIAL"
#define E_MKW_COMP_MAT_TYPE1                "COMP_MATERIAL_TYPE1"

#define E_MKW_NLAYERS                       "NLAYERS"
#define E_MKW_LAYER_MATS                    "LAYER_MATERIALS"
#define E_MKW_LAYER_THICKNESSES             "LAYER_THICKNESSES"


/* ////////////////////////////////////////////////////////////////////////// */
/* Material property keywords */

#define E_MKW_MAT_TYPE                      "MAT_TYPE"

#define E_MKW_DENSITY                       "DENSITY"
#define E_MKW_YMOD                          "YMOD"
#define E_MKW_YMOD_X                        "YMOD_X"
#define E_MKW_YMOD_Y                        "YMOD_Y"
#define E_MKW_YMOD_Z                        "YMOD_Z"
#define E_MKW_POISSONS_RATIO                "POISSONS_RATIO"
#define E_MKW_THICKNESS                     "THICKNESS"

#define E_MKW_IPARAM1                       "IPARAM1"
#define E_MKW_IPARAM2                       "IPARAM2"
#define E_MKW_IPARAM3                       "IPARAM3"
#define E_MKW_IPARAM4                       "IPARAM4"
#define E_MKW_IPARAM5                       "IPARAM5"
#define E_MKW_IPARAM6                       "IPARAM6"

#define E_MKW_DPARAM1                       "DPARAM1"
#define E_MKW_DPARAM2                       "DPARAM2"
#define E_MKW_DPARAM3                       "DPARAM3"
#define E_MKW_DPARAM4                       "DPARAM4"
#define E_MKW_DPARAM5                       "DPARAM5"
#define E_MKW_DPARAM6                       "DPARAM6"
#define E_MKW_DPARAM7                       "DPARAM7"
#define E_MKW_DPARAM8                       "DPARAM8"
#define E_MKW_DPARAM9                       "DPARAM9"
#define E_MKW_DPARAM10                      "DPARAM10"
#define E_MKW_DPARAM11                      "DPARAM11"
#define E_MKW_DPARAM12                      "DPARAM12"
#define E_MKW_DPARAM13                      "DPARAM13"
#define E_MKW_DPARAM14                      "DPARAM14"
#define E_MKW_DPARAM15                      "DPARAM15"
#define E_MKW_DPARAM16                      "DPARAM16"
#define E_MKW_DPARAM17                      "DPARAM17"
#define E_MKW_DPARAM18                      "DPARAM18"
#define E_MKW_DPARAM19                      "DPARAM19"
#define E_MKW_DPARAM20                      "DPARAM20"

#define E_MKW_END                           "END"


/* ////////////////////////////////////////////////////////////////////////// */
/* Mesh decomposition file keywords */

#define E_KW_NSUBDOMS                       "NSUBDOMAINS"
#define E_KW_SUBDOMS_LINK1                  "SUBDOMAINS_LINK1"
#define E_KW_SUBDOMS_LINK2                  "SUBDOMAINS_LINK2"
#define E_KW_SUBDOMS_LINK3                  "SUBDOMAINS_LINK3"
#define E_KW_SUBDOMS_LINK4                  "SUBDOMAINS_LINK4"
#define E_KW_SUBDOMS_LINK5                  "SUBDOMAINS_LINK5"
#define E_KW_SUBDOMS_TRIANG1                "SUBDOMAINS_TRIANG1"
#define E_KW_SUBDOMS_TRIANG2                "SUBDOMAINS_TRIANG2"
#define E_KW_SUBDOMS_TRIANG3                "SUBDOMAINS_TRIANG3"
#define E_KW_SUBDOMS_TRIANG4                "SUBDOMAINS_TRIANG4"
#define E_KW_SUBDOMS_TRIANG5                "SUBDOMAINS_TRIANG5"
#define E_KW_SUBDOMS_QUAD1                  "SUBDOMAINS_QUAD1"
#define E_KW_SUBDOMS_QUAD3                  "SUBDOMAINS_QUAD3"
#define E_KW_SUBDOMS_TETRAH1                "SUBDOMAINS_TETRAH1"
#define E_KW_SUBDOMS_BLOCK1                 "SUBDOMAINS_BLOCK1"


/* ////////////////////////////////////////////////////////////////////////// */
/* Mesh parameters file keywords */

#define E_KW_NODAL_MESHPARAMS               "NODAL_MESHPARAMS"
#define E_KW_MESHPARAMS_LINK1               "MESHPARAMS_LINK1"
#define E_KW_MESHPARAMS_LINK2               "MESHPARAMS_LINK2"
#define E_KW_MESHPARAMS_LINK3               "MESHPARAMS_LINK3"
#define E_KW_MESHPARAMS_LINK4               "MESHPARAMS_LINK4"
#define E_KW_MESHPARAMS_LINK5               "MESHPARAMS_LINK5"
#define E_KW_MESHPARAMS_TRIANG1             "MESHPARAMS_TRIANG1"
#define E_KW_MESHPARAMS_TRIANG2             "MESHPARAMS_TRIANG2"
#define E_KW_MESHPARAMS_TRIANG3             "MESHPARAMS_TRIANG3"
#define E_KW_MESHPARAMS_TRIANG4             "MESHPARAMS_TRIANG4"
#define E_KW_MESHPARAMS_TRIANG5             "MESHPARAMS_TRIANG5"
#define E_KW_MESHPARAMS_QUAD1               "MESHPARAMS_QUAD1"
#define E_KW_MESHPARAMS_QUAD3               "MESHPARAMS_QUAD3"
#define E_KW_MESHPARAMS_TETRAH1             "MESHPARAMS_TETRAH1"
#define E_KW_MESHPARAMS_BLOCK1              "MESHPARAMS_BLOCK1"


/* ////////////////////////////////////////////////////////////////////////// */
/* Stress file keywords */

#define E_KW_STRESSES_LINK1                 "STRESSES_LINK1"
#define E_KW_STRESSES_LINK2                 "STRESSES_LINK2"
#define E_KW_STRESSES_LINK3                 "STRESSES_LINK3"
#define E_KW_STRESSES_LINK4                 "STRESSES_LINK4"
#define E_KW_STRESSES_LINK5                 "STRESSES_LINK5"
#define E_KW_STRESSES_TRIANG1               "STRESSES_TRIANG1"
#define E_KW_STRESSES_TRIANG2               "STRESSES_TRIANG2"
#define E_KW_STRESSES_TRIANG3               "STRESSES_TRIANG3"
#define E_KW_STRESSES_TRIANG4               "STRESSES_TRIANG4"
#define E_KW_STRESSES_TRIANG5               "STRESSES_TRIANG5"
#define E_KW_STRESSES_QUAD1                 "STRESSES_QUAD1"
#define E_KW_STRESSES_QUAD3                 "STRESSES_QUAD3"
#define E_KW_STRESSES_TETRAH1               "STRESSES_TETRAH1"
#define E_KW_STRESSES_BLOCK1                "STRESSES_BLOCK1"


/* ////////////////////////////////////////////////////////////////////////// */
/* Finite element error file keywords */
#define E_KW_FEERRORS_LINK1                 "FEERRORS_LINK1"
#define E_KW_FEERRORS_LINK2                 "FEERRORS_LINK2"
#define E_KW_FEERRORS_LINK3                 "FEERRORS_LINK3"
#define E_KW_FEERRORS_LINK4                 "FEERRORS_LINK4"
#define E_KW_FEERRORS_LINK5                 "FEERRORS_LINK5"
#define E_KW_FEERRORS_TRIANG1               "FEERRORS_TRIANG1"
#define E_KW_FEERRORS_TRIANG2               "FEERRORS_TRIANG2"
#define E_KW_FEERRORS_TRIANG3               "FEERRORS_TRIANG3"
#define E_KW_FEERRORS_TRIANG4               "FEERRORS_TRIANG4"
#define E_KW_FEERRORS_TRIANG5               "FEERRORS_TRIANG5"
#define E_KW_FEERRORS_QUAD1                 "FEERRORS_QUAD1"
#define E_KW_FEERRORS_QUAD3                 "FEERRORS_QUAD3"
#define E_KW_FEERRORS_TETRAH1               "FEERRORS_TETRAH1"
#define E_KW_FEERRORS_BLOCK1                "FEERRORS_BLOCK1"



/* ////////////////////////////////////////////////////////////////////////// */
/* Mesh definition file "KeywordSet" Boolean indices */

#define E_KW_NUM_TITLE                               0
#define E_KW_NUM_NMESHPOINTS                         1
#define E_KW_NUM_NNODES                              2
#define E_KW_NUM_NELEMS_LINK1                        3
#define E_KW_NUM_NELEMS_LINK2                        4
#define E_KW_NUM_NELEMS_LINK3                        5
#define E_KW_NUM_NELEMS_LINK4                        6
#define E_KW_NUM_NELEMS_LINK5                        7
#define E_KW_NUM_NELEMS_TRIANG1                      8
#define E_KW_NUM_NELEMS_TRIANG2                      9
#define E_KW_NUM_NELEMS_TRIANG3                     10
#define E_KW_NUM_NELEMS_TRIANG4                     11
#define E_KW_NUM_NELEMS_TRIANG5                     12
#define E_KW_NUM_NELEMS_QUAD1                       13
#define E_KW_NUM_NELEMS_QUAD3                       14
#define E_KW_NUM_NELEMS_TETRAH1                     15
#define E_KW_NUM_NELEMS_BLOCK1                      16
#define E_KW_NUM_NBC_NODES                          17
#define E_KW_NUM_NLOADED_NODES                      18
#define E_KW_NUM_NMATS                              19
#define E_KW_NUM_NCOMP_MATS_TYPE1                   20
#define E_KW_NUM_NNURBS_CURVES                      21
#define E_KW_NUM_TIMESTEP                           22
#define E_KW_NUM_NTIMESTEPS                         23
#define E_KW_NUM_NINTERNAL_TIMESTEPS                24
#define E_KW_NUM_NEXTERNAL_TIMESTEPS                25
#define E_KW_NUM_DAMPING_FACTOR                     26
#define E_KW_NUM_BETA                               27
#define E_KW_NUM_GAMMA                              28
#define E_KW_NUM_NSTRESS_POINTS_LINK1               29
#define E_KW_NUM_NSTRESS_POINTS_LINK2               30
#define E_KW_NUM_NSTRESS_POINTS_LINK3               31
#define E_KW_NUM_NSTRESS_POINTS_LINK4               32
#define E_KW_NUM_NSTRESS_POINTS_LINK5               33
#define E_KW_NUM_NSTRESS_POINTS_TRIANG1             34
#define E_KW_NUM_NSTRESS_POINTS_TRIANG2             35
#define E_KW_NUM_NSTRESS_POINTS_TRIANG3             36
#define E_KW_NUM_NSTRESS_POINTS_TRIANG4             37
#define E_KW_NUM_NSTRESS_POINTS_TRIANG5             38
#define E_KW_NUM_NSTRESS_POINTS_QUAD1               39
#define E_KW_NUM_NSTRESS_POINTS_QUAD3               40
#define E_KW_NUM_NSTRESS_POINTS_TETRAH1             41
#define E_KW_NUM_NSTRESS_POINTS_BLOCK1              42
#define E_KW_NUM_MESHPOINT_COORDS                   43
#define E_KW_NUM_NODES_LINK1                        44
#define E_KW_NUM_NODES_LINK2                        45
#define E_KW_NUM_NODES_LINK3                        46
#define E_KW_NUM_NODES_LINK4                        47
#define E_KW_NUM_NODES_LINK5                        48
#define E_KW_NUM_NODES_TRIANG1                      49
#define E_KW_NUM_NODES_TRIANG2                      50
#define E_KW_NUM_NODES_TRIANG3                      51
#define E_KW_NUM_NODES_TRIANG4                      52
#define E_KW_NUM_NODES_TRIANG5                      53
#define E_KW_NUM_NODES_QUAD1                        54
#define E_KW_NUM_NODES_QUAD3                        55
#define E_KW_NUM_NODES_TETRAH1                      56
#define E_KW_NUM_NODES_BLOCK1                       57
#define E_KW_NUM_BCS                                58
#define E_KW_NUM_LOADS                              59
#define E_KW_NUM_MATS_LINK1                         60
#define E_KW_NUM_MATS_LINK2                         61
#define E_KW_NUM_MATS_LINK3                         62
#define E_KW_NUM_MATS_LINK4                         63
#define E_KW_NUM_MATS_LINK5                         64
#define E_KW_NUM_MATS_TRIANG1                       65
#define E_KW_NUM_COMP_MATS_TRIANG2                  66
#define E_KW_NUM_MATS_TRIANG3                       67
#define E_KW_NUM_MATS_TRIANG4                       68
#define E_KW_NUM_MATS_TRIANG5                       69
#define E_KW_NUM_MATS_QUAD1                         70
#define E_KW_NUM_COMP_MATS_QUAD3                    71
#define E_KW_NUM_MATS_TETRAH1                       72
#define E_KW_NUM_MATS_BLOCK1                        73
#define E_KW_NUM_REMESH_DATA                        74

/* Nodal remeshing type indices */

#define E_RM_NUM_EXTERIOR                            0
#define E_RM_NUM_INTERIOR                            1
#define E_RM_NUM_ISD_BOUNDARY                        2

/* Boundary condition type indices */

#define E_BC_NUM_FREE                                0
#define E_BC_NUM_FIXED                               1
#define E_BC_NUM_SPRING                              2


/* ////////////////////////////////////////////////////////////////////////// */
/* Geometric model file type indices */

#define E_NC_NUM_NO_FIXITY                           0
#define E_NC_NUM_FIXITY_GC                           1
#define E_NC_NUM_FIXITY_CC                           2

#define E_NC_NUM_POINT_LOAD                          0
#define E_NC_NUM_LIN_LOAD_GC                         1
#define E_NC_NUM_LIN_LOAD_CC                         2


/* ////////////////////////////////////////////////////////////////////////// */
/* Materials file "KeywordSet" Boolean indices */

#define E_MKW_NUM_MAT_TYPE                           0

#define E_MKW_NUM_DENSITY                            1
#define E_MKW_NUM_YMOD                               2
#define E_MKW_NUM_YMOD_X                             3
#define E_MKW_NUM_YMOD_Y                             4
#define E_MKW_NUM_YMOD_Z                             5
#define E_MKW_NUM_POISSONS_RATIO                     6
#define E_MKW_NUM_THICKNESS                          7

#define E_MKW_NUM_IPARAM1                            8
#define E_MKW_NUM_IPARAM2                            9
#define E_MKW_NUM_IPARAM3                           10
#define E_MKW_NUM_IPARAM4                           11
#define E_MKW_NUM_IPARAM5                           12
#define E_MKW_NUM_IPARAM6                           13

#define E_MKW_NUM_DPARAM1                           14
#define E_MKW_NUM_DPARAM2                           15
#define E_MKW_NUM_DPARAM3                           16
#define E_MKW_NUM_DPARAM4                           17
#define E_MKW_NUM_DPARAM5                           18
#define E_MKW_NUM_DPARAM6                           19
#define E_MKW_NUM_DPARAM7                           20
#define E_MKW_NUM_DPARAM8                           21
#define E_MKW_NUM_DPARAM9                           22
#define E_MKW_NUM_DPARAM10                          23
#define E_MKW_NUM_DPARAM11                          24
#define E_MKW_NUM_DPARAM12                          25
#define E_MKW_NUM_DPARAM13                          26
#define E_MKW_NUM_DPARAM14                          27
#define E_MKW_NUM_DPARAM15                          28
#define E_MKW_NUM_DPARAM16                          29
#define E_MKW_NUM_DPARAM17                          30
#define E_MKW_NUM_DPARAM18                          31
#define E_MKW_NUM_DPARAM19                          32
#define E_MKW_NUM_DPARAM20                          33

#define E_MKW_NUM_END                               34


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                              File Extensions                               */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#define E_MDF  "mdf"
#define E_GMF  "gmf"
#define E_MAT  "mat"
#define E_DOM  "dom"
#define E_MPR  "mpr"
#define E_STE  "ste"
#define E_FEE  "fee"


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                         Input-Output Functions                             */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#ifndef __E_T800SAL__

int e_GetLine(char *line, int MaxLineLength, int *LineNumber, FILE *fp);
int e_GetNextLine(char *line, int *LineNumber, char *FullPathName, FILE *fp);

/*
char *OpenInputFile(char *FileName, char *extension, FILE **fp);
char *OpenOutputFile(char *FileName, char *extension, FILE **fp);
int GetKeyword(char *line, char *word, int *LineNumber, 
               char *FullPathName, FILE *fp);
*/

char *GetStrArg(char *pLine, char *format, void *argument);

void e_beep(void);

#endif

void e_PlatformMessage(char *message);


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                         Warning Message Functions                          */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

void e_GeneralWarning(char *Warning, ...);
void e_FileWarning(char *FullPathName, char *Warning);
void e_LineWarning(int LineNumber, char *Warning);
void e_FileLineWarning(int LineNumber, char *FullPathName, char *Warning);


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                          Error Message Functions                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* NOTE: The following functions output an error message and, with the
         exception of __E_MSWIN__, terminate execution.                       */

void e_GeneralError(char *ErrorMessage, ...);
void e_FileError(char *FullPathName, char *ErrorMessage);
void e_LineError(int LineNumber, char *ErrorMessage);
void e_FileLineError(int LineNumber, char *FullPathName, char *ErrorMessage);
void e_SystemError(void);


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                               Exit Function                                */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

void e_exit(int ExitStatus);


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                      Memory Allocation Functions                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

void   *e_calloc(unsigned int length, unsigned int NBytes);

int    *e_alloc_i_vect(unsigned int length);
float  *e_alloc_f_vect(unsigned int length);
double *e_alloc_d_vect(unsigned int length);
char   *e_alloc_c_vect(unsigned int length);
void **e_alloc_vp_vect(unsigned int length);

int    **e_alloc_frag_i_matr(unsigned int NRows, unsigned int NCols);
float  **e_alloc_frag_f_matr(unsigned int NRows, unsigned int NCols);
double **e_alloc_frag_d_matr(unsigned int NRows, unsigned int NCols);
char   **e_alloc_frag_c_matr(unsigned int NRows, unsigned int NCols);
void ***e_alloc_frag_vp_matr(unsigned int NRows, unsigned int NCols);

int    **e_alloc_cont_i_matr(unsigned int NRows, unsigned int NCols);
float  **e_alloc_cont_f_matr(unsigned int NRows, unsigned int NCols);
double **e_alloc_cont_d_matr(unsigned int NRows, unsigned int NCols);
char   **e_alloc_cont_c_matr(unsigned int NRows, unsigned int NCols);
void ***e_alloc_cont_vp_matr(unsigned int NRows, unsigned int NCols);

int    ***e_alloc_frag_i_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols );
float  ***e_alloc_frag_f_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols );
double ***e_alloc_frag_d_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols );
char   ***e_alloc_frag_c_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols );
void ****e_alloc_frag_vp_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols );

int    ***e_alloc_cont_i_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols );
float  ***e_alloc_cont_f_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols );
double ***e_alloc_cont_d_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols );
char   ***e_alloc_cont_c_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols );
void ****e_alloc_cont_vp_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                      unsigned int NCols );

int    ****e_alloc_frag_i_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  );
float  ****e_alloc_frag_f_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  );
double ****e_alloc_frag_d_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  );
char   ****e_alloc_frag_c_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  );
void *****e_alloc_frag_vp_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  );

int    ****e_alloc_cont_i_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  );
float  ****e_alloc_cont_f_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  );
double ****e_alloc_cont_d_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  );
char   ****e_alloc_cont_c_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  );
void *****e_alloc_cont_vp_4D_matr(unsigned int NDepth, unsigned int NWidth,
                                  unsigned int NRows,  unsigned int NCols  );

int    **e_init_frag_i_matr(unsigned int NRows);
float  **e_init_frag_f_matr(unsigned int NRows);
double **e_init_frag_d_matr(unsigned int NRows);
char   **e_init_frag_c_matr(unsigned int NRows);
void ***e_init_frag_vp_matr(unsigned int NRows);

int    ***e_init_frag_i_3D_matr(unsigned int NDepth);
float  ***e_init_frag_f_3D_matr(unsigned int NDepth);
double ***e_init_frag_d_3D_matr(unsigned int NDepth);
char   ***e_init_frag_c_3D_matr(unsigned int NDepth);
void ****e_init_frag_vp_3D_matr(unsigned int NDepth);

int    ****e_init_frag_i_4D_matr(unsigned int NDepth);
float  ****e_init_frag_f_4D_matr(unsigned int NDepth);
double ****e_init_frag_d_4D_matr(unsigned int NDepth);
char   ****e_init_frag_c_4D_matr(unsigned int NDepth);
void *****e_init_frag_vp_4D_matr(unsigned int NDepth);

int    *e_incr_i_vect(unsigned int length, int *vector,
                      unsigned int LengthIncrem        );
float  *e_incr_f_vect(unsigned int length, float *vector,
                      unsigned int LengthIncrem          );
double *e_incr_d_vect(unsigned int length, double *vector,
                      unsigned int LengthIncrem           );
char   *e_incr_c_vect(unsigned int length, char *vector,
                      unsigned int LengthIncrem         );
void **e_incr_vp_vect(unsigned int length, void **vector,
                      unsigned int LengthIncrem          );

int    **e_incr_frag_i_matr(unsigned int NRows, int **matrix,
                            unsigned int NRowsIncrem         );
float  **e_incr_frag_f_matr(unsigned int NRows, float **matrix,
                            unsigned int NRowsIncrem           );
double **e_incr_frag_d_matr(unsigned int NRows, double **matrix,
                            unsigned int NRowsIncrem            );
char   **e_incr_frag_c_matr(unsigned int NRows, char **matrix,
                            unsigned int NRowsIncrem          );
void ***e_incr_frag_vp_matr(unsigned int NRows, void ***matrix,
                            unsigned int NRowsIncrem           );

int    **e_incr_cont_i_matr(unsigned int NRows, unsigned int NCols,
                            int **matrix,    unsigned int NRowsIncrem);
float  **e_incr_cont_f_matr(unsigned int NRows, unsigned int NCols,
                            float **matrix,  unsigned int NRowsIncrem);
double **e_incr_cont_d_matr(unsigned int NRows, unsigned int NCols,
                            double **matrix, unsigned int NRowsIncrem);
char   **e_incr_cont_c_matr(unsigned int NRows, unsigned int NCols,
                            char **matrix,   unsigned int NRowsIncrem);
void ***e_incr_cont_vp_matr(unsigned int NRows, unsigned int NCols,
                            void ***matrix,  unsigned int NRowsIncrem);


int    *e_ext_i_vect(unsigned int length, int *vector,
                     unsigned int LengthIncrem        );
float  *e_ext_f_vect(unsigned int length, float *vector,
                     unsigned int LengthIncrem          );
double *e_ext_d_vect(unsigned int length, double *vector,
                     unsigned int LengthIncrem           );
char   *e_ext_c_vect(unsigned int length, char *vector,
                     unsigned int LengthIncrem         );
void  **e_ext_vp_vect(unsigned int length, void **vector,
                      unsigned int LengthIncrem          );

int    **e_ext_frag_i_matr(unsigned int NRows, int **matrix,
                           unsigned int NRowsIncrem         );
float  **e_ext_frag_f_matr(unsigned int NRows, float **matrix,
                           unsigned int NRowsIncrem           );
double **e_ext_frag_d_matr(unsigned int NRows, double **matrix,
                           unsigned int NRowsIncrem            );
char   **e_ext_frag_c_matr(unsigned int NRows, char **matrix,
                           unsigned int NRowsIncrem          );
void ***e_ext_frag_vp_matr(unsigned int NRows, void ***matrix,
                           unsigned int NRowsIncrem           );

int    **e_ext_cont_i_matr(unsigned int NRows, unsigned int NCols,
                           int **matrix,    unsigned int NRowsIncrem);
float  **e_ext_cont_f_matr(unsigned int NRows, unsigned int NCols,
                           float **matrix,  unsigned int NRowsIncrem);
double **e_ext_cont_d_matr(unsigned int NRows, unsigned int NCols,
                           double **matrix, unsigned int NRowsIncrem);
char   **e_ext_cont_c_matr(unsigned int NRows, unsigned int NCols,
                           char **matrix,   unsigned int NRowsIncrem);
void ***e_ext_cont_vp_matr(unsigned int NRows, unsigned int NCols,
                           void ***matrix,  unsigned int NRowsIncrem);


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                       Memory Deallocation Functions                        */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

void e_free(void *vector);

void  e_free_frag_i_matr(unsigned int NRows,    int **matrix);
void  e_free_frag_f_matr(unsigned int NRows,  float **matrix);
void  e_free_frag_d_matr(unsigned int NRows, double **matrix);
void  e_free_frag_c_matr(unsigned int NRows,   char **matrix);
void e_free_frag_vp_matr(unsigned int NRows,  void ***matrix);

void  e_free_cont_i_matr(unsigned int NRows,    int **matrix);
void  e_free_cont_f_matr(unsigned int NRows,  float **matrix);
void  e_free_cont_d_matr(unsigned int NRows, double **matrix);
void  e_free_cont_c_matr(unsigned int NRows,   char **matrix);
void e_free_cont_vp_matr(unsigned int NRows,  void ***matrix);

void  e_free_frag_i_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                     int ***matrix3D);
void  e_free_frag_f_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                   float ***matrix3D);
void  e_free_frag_d_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                  double ***matrix3D);
void  e_free_frag_c_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                    char ***matrix3D);
void e_free_frag_vp_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                   void ****matrix3D);

void  e_free_cont_i_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                     int ***matrix3D);
void  e_free_cont_f_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                   float ***matrix3D);
void  e_free_cont_d_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                  double ***matrix3D);
void  e_free_cont_c_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                    char ***matrix3D);
void e_free_cont_vp_3D_matr(unsigned int NDepth, unsigned int NRows,
                                                   void ****matrix3D);

void  e_free_frag_i_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  int ****matrix4D    );
void  e_free_frag_f_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  float ****matrix4D  );
void  e_free_frag_d_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  double ****matrix4D );
void  e_free_frag_c_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  char ****matrix4D   );
void e_free_frag_vp_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  void *****matrix4D  );

void  e_free_cont_i_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  int ****matrix4D    );
void  e_free_cont_f_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  float ****matrix4D  );
void  e_free_cont_d_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  double ****matrix4D );
void  e_free_cont_c_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  char ****matrix4D   );
void e_free_cont_vp_4D_matr(unsigned int NDepth, unsigned int NWidth,
                            unsigned int NRows,  void *****matrix4D  );


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                        Mesh Initialisation Function                        */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

void e_NullifyMesh(MESH *mesh);


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                            Mesh Type Functions                             */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

int e_IsOfOneElementType(MESH *mesh);
int e_LinkElementsOnly(MESH *mesh);
int e_PlanarElementsOnly(MESH *mesh);


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                          Mesh Type Specific Data                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* NOTE: It is assumed that the mesh is composed of only one element type. */

int   e_GetElementType(MESH *mesh);
int   e_GetNInterfacesPerElement(MESH *mesh);
int   e_GetNVerticesPerElement(MESH *mesh);
int   e_GetNNodesPerElement(MESH *mesh);
int   e_GetNVerticesPerInterface(MESH *mesh);
int **e_GetNodes(MESH *mesh);
int **e_GetVertices(MESH *mesh);


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                            Mesh Data Functions                             */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */


#ifndef __E_T800SAL__

int   e_GetMeshData(MESH**mesh, char *InputFile);
int   e_PutMeshData(MESH *mesh, char *OutputFile);
int  e_GetGeomModel(MESH *mesh, char *InputFile);
int  e_PutGeomModel(MESH *mesh, char *OutputFile);
int  e_GetMaterials(MESH *mesh, char *InputFile);
int  e_PutMaterials(MESH *mesh, char *OutputFile);
int e_GetDecompData(MESH *mesh, char *InputFile);
int e_PutDecompData(MESH *mesh, char *OutputFile);
int e_GetMeshParams(MESH *mesh, char *InputFile);
int e_PutMeshParams(MESH *mesh, char *OutputFile);
int   e_GetStresses(MESH *mesh, char *InputFile);
int   e_PutStresses(MESH *mesh, char *OutputFile);
int   e_GetFEErrors(MESH *mesh, char *InputFile);
int   e_PutFEErrors(MESH *mesh, char *OutputFile);

#endif

void        e_FreeMeshData(MESH *mesh);
void       e_FreeGeomModel(MESH *mesh);
void        e_FreeMatProps(MESH *mesh);
void      e_FreeDecompData(MESH *mesh);
void      e_FreeMeshParams(MESH *mesh);
void e_FreeNodalMeshParams(MESH *mesh);
void  e_FreeElemMeshParams(MESH *mesh);
void        e_FreeStresses(MESH *mesh);
void        e_FreeFEErrors(MESH *mesh);

int e_CreateSubMesh(MESH *mesh, MESH **SubMesh, char *SubMeshTitle,
                                                 int SubdomainIndex);

#ifdef __E_T800__

#include <chan.h>

void e_chan_out_Buffer(int msglength, void *buffer, CHAN *chan);
int   e_chan_in_Buffer(               void *buffer, CHAN *chan);
void   e_chan_out_MeshData(MESH *mesh, CHAN *chan);
void    e_chan_in_MeshData(MESH**mesh, CHAN *chan);
void e_chan_out_DecompData(MESH *mesh, CHAN *chan);
void  e_chan_in_DecompData(MESH *mesh, CHAN *chan);
void e_chan_out_MeshParams(MESH *mesh, CHAN *chan);
void  e_chan_in_MeshParams(MESH *mesh, CHAN *chan);
void       e_chan_out_Mats(MESH *mesh, CHAN *chan);
void        e_chan_in_Mats(MESH *mesh, CHAN *chan);

#endif


#ifdef __E_MPI__

void e_MPI_Send_MeshData(   MESH  *mesh, int dest  );
void e_MPI_Recv_MeshData(   MESH **mesh, int source);
void e_MPI_Send_Mats(       MESH  *mesh, int dest  );
void e_MPI_Recv_Mats(       MESH  *mesh, int source);
void e_MPI_Send_MeshParams( MESH  *mesh, int dest  );
void e_MPI_Recv_MeshParams( MESH  *mesh, int source);
void e_MPI_Send_DecompData( MESH  *mesh, int dest  );
void e_MPI_Recv_DecompData( MESH  *mesh, int source);
void e_MPI_Send_FEErrorData(MESH *mesh, int dest);
void e_MPI_Recv_FEErrorData(MESH *mesh, int source);
void e_MPI_Send_Stresses(   MESH  *mesh, int dest);
void e_MPI_Recv_Stresses(   MESH  *mesh, int source);

#endif


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                          Miscellaneous Functions                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

int e_SetNodalMeshParams(MESH *mesh);
int e_SetupStresses(MESH *mesh);
int e_SetupDecompData(MESH *mesh);

int e_SetAdjacency(MESH *mesh, int **Degree, int ***AdjElems);
void e_FreeAdjacency(MESH *mesh, int *Degree, int **AdjElems);

int e_SetAdjElems(MESH *mesh, int ***AdjElems);
void e_FreeAdjElems(MESH *mesh, int **AdjElems);

int e_GetNCutInterfaces(MESH *mesh);
int e_GetNInterSubDomNodes(MESH *mesh, int *NInterSubDomNodes,
                                int *WeightedNInterSubDomNodes);

int e_GetInterSubDomNodes(MESH *mesh, int *NInterSubDomNodes,
                          int **InterSubDomNodes, int **NNodeSubDoms,
                                                   int ***NodeSubDoms);
void e_FreeInterSubDomNodes(int NInterSubDomNodes, int *InterSubDomNodes,
                                     int *NNodeSubDoms, int **NodeSubDoms);

int e_SubDomBoundaries2PolyPolygon(MESH *mesh, int **PolygonVertices, 
                                   int **NPolygonVertices, int *NPolygons);
void e_FreeSubDomBoundaries2PolyPolygon(int *PolygonVertices,
                                        int *NPolygonVertices);

int e_SegmentRefine(double   *start_crd, double    start_mpr, 
                    double   *end_crd,   double    end_mpr, 
		    double ***seg_crd,   double  **seg_mpr, 
		    int generate_nodes,  int mode);
int e_CurveRefine(NURBSCURVE *curve,
    	    	  double    start_t, 
		  double    start_mpr, 
		  double    end_t, 
		  double    end_mpr, 
		  double ***seg_crd, 
		  double  **seg_mpr,
		  double  **seg_t, 
		  int       generate_nodes,
		  int       mode);

int e_GetCurve(MESH *mesh, int node1, int node2, NURBSCURVE ** curve,
               double *t1, double *t2);


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                            Miscellaneous Macros                            */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#ifndef NULL
#define NULL ((void *)0)
#endif

#ifndef TRUE
#define TRUE   1
#endif

#ifndef FALSE
#define FALSE  0
#endif

#define E_LINE_MAX  400               /* maximum input line length */


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                           Mathematical Functions                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

int e_LUFact(int Order, double **A, int *pivot);
int e_LUSolve(int Order, double **LU, double *b, int *pivot);

unsigned long e_Factorial(unsigned long Number);

double e_FactorialDP(unsigned long Number);
double e_VectDotProd(unsigned int Order, double *x, double *y);

void e_VectCrossVect(double *a, double *b, double *c);
void     e_VectScale(unsigned int Order, double Factor, double *x, double *y);

void   e_MatrTransp(unsigned int Order, double **A, double **B);
void e_MatrVectMult(unsigned int Order, double **A, double *x,  double *y);
void e_VectMatrMult(unsigned int Order, double *x,  double **A, double *y);
void     e_MatrMult(unsigned int Order, double **A, double **B, double **C);

void   e_MatrTranspNS(unsigned int NRows, unsigned int NCols,
                                       double **A, double **B);
void e_MatrVectMultNS(unsigned int NRows, unsigned int NCols,
                             double **A, double *x, double *y);
void e_VectMatrMultNS(unsigned int NRows, unsigned int NCols,
                             double *x, double **A, double *y);
void     e_MatrMultNS(unsigned int NRowsA, unsigned int NColsA_NRowsB,
                      unsigned int NColsB, double **A, double **B, double **C);

#define E_ABS(X)         (((X) > 0) ? (X) : (-(X)))
#define E_MIN(A,B)       (((A) < (B)) ? (A) : (B))
#define E_MAX(A,B)       (((A) > (B)) ? (A) : (B))
#define E_MIN3(A,B,C)    E_MIN(E_MIN(A,B),C)
#define E_MAX3(A,B,C)    E_MAX(E_MAX(A,B),C)
#define E_MIN4(A,B,C,D)  E_MIN(E_MIN3(A,B,C),D)
#define E_MAX4(A,B,C,D)  E_MAX(E_MAX3(A,B,C),D)
#define E_SQUARE(X)      ((X)*(X))


#endif
