#include "e_lib.h"
#include "adv_apex.h"
#include "eg_lib.h"     /* geometric library */
#include "adv_config.h" /* parameters for the advancing front algorithm */
#include "common.h"

#define VALID_TRIANG    1
#define INVALID_TRIANG  2

static void PerpChord(double *coord1, double *coord2, 
                      double delta, double chordnodes[][2]);
static int GetNearNodes(int node1, int node2, double *node_coord, 
                        double **coord, int nseg, int **seg,
                        double delta, int *nstrict, int *nn, int *nearnodes);
static double TriangQuality(double *a, double *b, double *c);
static int CheckInside(double *a, double *b, double *c, double *p);
static int CheckNewTriang(int iseg, int selnode, double *selcoord, 
                          double **coord, int nsgt, int **seg);
static int NearSegments(int iseg, double *selcoord, 
                        double **coord, int nseg, int **seg);


/* //////////////////////////////////////////////////////////////////////////
Function:     SelectNode
Descriptions: selects a node to form a triangle
Arguments:
              iseg: selected shortest segment
              delta: required size of the triangle sides
              nnode: number of nodes
              coords: coordinates of nodes
              nseg: number of segments
              seg: list of segments on the active front
Return:       n -> if successfull, the selected node
             -1 -> if no node could be selected
///////////////////////////////////////////////////////////////////////////// */
int SelectNode(int iseg, double delta, int *nnode, double ***coord, 
               int nseg, int **seg)
{
  int i, node1, node2;
  int nnear;
  int nstrict;
  int minnear;
  int nearnode[MAX_NO_NEAR_NODES];
  double chordnode[NO_CHORD_NODES][2];
  double lmin, l1, l2, quality;
  
  minnear = -1;
  /* beginning and end nodes of the selected segment */
  node1 = seg[iseg][0];
  node2 = seg[iseg][1];

  /* draw a perpendicular chord and generated five points along chord. */
  PerpChord((*coord)[node1-1], (*coord)[node2-1], delta, chordnode);

  /* collect the closest nodes around the apex */
  nnear = 0;
  GetNearNodes(node1, node2, chordnode[0],
               *coord, nseg, seg, delta, &nstrict, &nnear, nearnode);

  /* check all near nodes */
  for(i = 0; i < nnear; i++)
  {
    /* check whether the formed triangle is valid */
    if(CheckNewTriang(iseg, nearnode[i], (*coord)[nearnode[i]-1], 
                      *coord, nseg, seg) == VALID_TRIANG)
    {
      /* check the quality of the triangle */
      quality = TriangQuality((*coord)[node1-1], 
                              (*coord)[node2-1], 
                              (*coord)[nearnode[i]-1] );
      /* check the length of the sides of the triangle */
      l1 = eg_PointDist2D((*coord)[nearnode[i]-1], (*coord)[node1-1]);
      l2 = eg_PointDist2D((*coord)[nearnode[i]-1], (*coord)[node2-1]);
      /* if size is acceptable accept the node immediately */
      if(l1 < CRITERIA*delta && l2 < CRITERIA*delta && quality > 0.3)
      {
        return(nearnode[i]);
      }
      else
      {
        /* store the closest node */
        if(lmin > l1+l2 || minnear == -1)
        {
          lmin = l1+l2;
          minnear = i;
        }
      }
    }
  }

  /* check all chord nodes, from apex towards the segment */
  for(i = 0; i < NO_CHORD_NODES; i++)
  {
    if(CheckNewTriang(iseg, -1, chordnode[i], (*coord), nseg, seg) == VALID_TRIANG)
    {
      break;
    }
  }
  
  /* check it against near segments;
     if there are nearby segments try a chord node closer to the segment */
  for( ; i < NO_CHORD_NODES; i++)
  {
    /* check near segment */
    if(NearSegments(iseg, chordnode[i], *coord, nseg, seg) == FALSE)
    {
      /* no segment is near, node is accepted 
         create new node */
      *coord = e_incr_cont_d_matr(*nnode, E_NDIMS, *coord, INCREMENT);
      (*coord)[*nnode][0] = chordnode[i][0];
      (*coord)[*nnode][1] = chordnode[i][1];
      (*coord)[*nnode][2] = 0.0;
      (*nnode)++;
      /* return its node number */
      return(*nnode);
    }
  }
  
  /* last chance
     there are near nodes, but the size was not perfect */
  if(minnear == -1)
    return(-1);
  else
    return(nearnode[minnear]);

}

/* //////////////////////////////////////////////////////////////////////////
Function:     PerpChord
Descriptions: generate a perpendicular chord from the point specified by
              the mid point of line segment x1,y1 ->  x2 y2. The chord
              length is del1. the number of points generated along chord 
              is NO_CHORD_NODES.
Arguments:
              coord1 the first end point of line segment.
              coord2 the second end point of line segment.
              delta: the required length of the triangle sides
              chordnodes: the array holds generated new point coordinates.
Return:       None.
///////////////////////////////////////////////////////////////////////////// */
static void PerpChord(double *coord1, double *coord2, 
                      double delta, double chordnodes[][2])
{
  int i;
  double x3, y3, L, s[NO_CHORD_NODES];
  double ang, incr, dinc, limit;

  /* length of the segment */
  L = eg_PointDist2D(coord1, coord2);

  /* calculate the "angle", limit the value for acos */
  limit = L / (2 * delta);
  if(limit > 1.0)
    limit = 1.0;
  if(limit < 0.0)
    limit = 0.0;
  ang = acos(limit);

  /* peak of chord */
  s[0] = sin(ang) * delta / L;

  /* increments */
  incr = 1.0 / NO_CHORD_NODES;
  dinc = incr;
  /* generate other nodes on the chord */
  for(i = 1; i < NO_CHORD_NODES; i++)
  {
    s[i] = s[0] * (1.0 - dinc);
    dinc += incr;
  }
  
  /* middle of the segment */
  x3 = (coord1[0]+coord2[0]) / 2.0;
  y3 = (coord1[1]+coord2[1]) / 2.0;

  /* generate NO_CHORD_NODES points along the perpendicular chord */
  for(i = 0; i < NO_CHORD_NODES; i++)
  {
    chordnodes[i][0] = x3-s[i]*(coord2[1]-coord1[1]);
    chordnodes[i][1] = y3+s[i]*(coord2[0]-coord1[0]);
  }

}


/* //////////////////////////////////////////////////////////////////////////
Function:     GetNearNodes
Descriptions: finds MAX_NO_NEAR_NODES existing nodes which
              are the closest to the point specified by x, y except 
              node1 and node2.
              Finally determine how many nodes satisfy the stricter
              condition, which is that they are inside a circle of 
              radius 'delta'. These nodes are on the front of the list
Arguments:		
              node1:  the node ID for the left end point of line segment.
              node2:  the node ID for the right end point of line segment.
              node_coord: the original point coordinate.
              coord:      all node coordinates.
              nseg:       the number of segments.
              seg:        the array holds all front line segments.
              delta:      the radius which all selected nodes should be within.
              nstrict:    no of nodes satisfying the stricter condition
              nearnodes:  the array holds the all near node IDs
Return:       no of near nodes (0 - MAX_NO_NEAR_NODES)
///////////////////////////////////////////////////////////////////////////// */
static int GetNearNodes(int node1, int node2, double *node_coord, 
                        double **coord, int nseg, int **seg,
                        double delta, int *nstrict, int *nn, int *nearnodes)
{
  int    i, j, k;
  int    maxi;     /* index of the furthest node */
  /* distance between current node and near node */
  double dist[MAX_NO_NEAR_NODES]; 
/*  int    nn; */      /* no of stored nodes */
  int    nd;       /* node to examine */
  double l1;       /* distance between given and examined node */
  int    ntemp;

/*  nn = 0; */
  /* check all segment on the front */
  for(i = 0; i < nseg; i++)
  {
    /* there are two nodes per segments */
    for(j = 0; j < E_NVPI_TRIANG1; j++)
    {
      /* we check this node */
      nd = seg[i][j];
      
      /* skip the nodes of the selected segment */
      if(nd == node1 || nd == node2)
        continue;
      
      /* if we already stored it */
      for(k = 0; k < *nn; k++)
      {
        if(nd == nearnodes[k])
          break;
      }
      /* skip further check */
      if(k != *nn)
        continue;

      /* distance between given and examined node */
      l1 = eg_PointDist2D(node_coord, coord[nd-1]);
      
      /* if node is on the wrong side of the segment skip again;
         node should be on the left hand side, 
         orientation counter-clockwise, > 0.0 */
      if(eg_Orient2D(coord[node1-1], coord[node2-1], coord[nd-1]) < 0.0)
        continue;

      /* there is still space to store */
      if(*nn < MAX_NO_NEAR_NODES)
      {
        if(*nn == 0)
        {
          maxi = 0;
        }
        else if(l1 > dist[maxi])
        {
          maxi = *nn;
        }
        nearnodes[*nn] = nd;
        dist[*nn] = l1;
        (*nn)++;
      }
      /* no more space, replace if it is smaller than max distance */
      else if(l1 < dist[maxi])
      {
        nearnodes[maxi] = nd;
        dist[maxi]  = l1;

        /* find new maximum distance */
        maxi = 0;
        for(k = 1; k < MAX_NO_NEAR_NODES; k++)
        {
          if(dist[k] > dist[maxi])
          {
            maxi = k;
          }
        }
      }
    }
  }

  /* sort according to distance, closest first */
  for(i = 0; i < *nn; i++)
  {
    int mini = i;
    for(k = i+1; k < *nn; k++)
    {
      if(dist[k] < dist[mini])
      {
        mini = k;
      }
    }
    if(mini != i)
    {
      ntemp = nearnodes[i];
      nearnodes[i] = nearnodes[mini];
      nearnodes[mini] = ntemp;
      dist[mini] = dist[i];
    }
  }

  /* how many nodes satisfy the stricter condition */
  *nstrict = 0;
  for(i = 0; i < *nn; i++)
  {
    /* in circle ? */
    if(dist[i] > delta)
    {
      break;
    }
    (*nstrict)++;
  }

  return(*nn);

}


/* //////////////////////////////////////////////////////////////////////////
Function:     TriangQuality
Descriptions: determines the quality of the triangle, uses scalar products
              of side vectors
Arguments:		
              a, b, c: coordinates of the corner nodes
Return:       value between 0.0 and 1.0
              1.0 -> best quality
              0.0 -> worst quality
///////////////////////////////////////////////////////////////////////////// */
static double TriangQuality(double *a, double *b, double *c)
{
  double scalar;
  
  scalar = 0.0;
  scalar += ScalarProd(a[0], a[1], b[0], b[1], a[0], a[1], c[0], c[1]);
  scalar += ScalarProd(b[0], b[1], a[0], a[1], b[0], b[1], c[0], c[1]);
  scalar += ScalarProd(c[0], c[1], b[0], b[1], c[0], c[1], a[0], a[1]);
  
  return(1.0 - (scalar / 3.0));
}

/* //////////////////////////////////////////////////////////////////////////
Function:     CheckInside
Descriptions: determines whether the point is inside or on the side of the
              triangle element
Arguments:		
              a, b, c: coordinates of the corner nodes of the triangle element
              p: coordinate of the point 
Return:       TRUE -> point is inside
              FALSE -> point is outside
///////////////////////////////////////////////////////////////////////////// */
static int CheckInside(double *a, double *b, double *c, double *p)
{
  double r1, r2, r3;

  r1 = eg_Orient2D(a, b, p);
  r2 = eg_Orient2D(b, c, p);
  r3 = eg_Orient2D(c, a, p);

  if(   ((E_ABS(r1) < TOLERANCE) && r2 > 0.0 && r3 > 0.0)
     || ((E_ABS(r2) < TOLERANCE) && r1 > 0.0 && r3 > 0.0)
     || ((E_ABS(r3) < TOLERANCE) && r1 > 0.0 && r2 > 0.0)
     || (r1 > 0.0 && r2 > 0.0 && r3 > 0.0) )
  {
    return(TRUE);
  }
  else
  {
    return(FALSE);
  }
}


/* //////////////////////////////////////////////////////////////////////////
Function:     CheckNewTriang
Descriptions: checks that the new triangle intersects any of the segments of
              the active front or any of the segments is inside the triangle
Arguments:		
              iseg: selected segment index
              selnode: selected node 
                       (when chord node it is -1, node does not exist yet)
              selcoord: coordinates of the selected node
              coord: coordinates of existing points
              nsgt: number of segments
              seg: list of segments
Return:       INVALID_TRIANG -> if any problem
              VALID_TRIANG -> if triangle is without intersection or inclusion
///////////////////////////////////////////////////////////////////////////// */
static int CheckNewTriang(int iseg, int selnode, double *selcoord, 
                          double **coord, int nsgt, int **seg)
{
  int i, n1, n2;
  int res;
  int segbeg, segend;
  double mu1, mu2;

  segbeg = seg[iseg][0];
  segend = seg[iseg][1];

  for(i = 0; i < nsgt; i++)
  {
    /* skip selected segment */
    if(i == iseg)
    {
      continue;
    }
    
    /* nodes of the segments */
    n1 = seg[i][0];
    n2 = seg[i][1];
    
    /* intersection with one side, but not when segment shares the same node */
    if(   n1 != segbeg && n2 != segbeg
       && n1 != selnode && n2 != selnode )
    {
      res = eg_LineLineIntersect2D(coord[n1-1], coord[n2-1], 
                                   selcoord, coord[segbeg-1], 
                                   &mu1, &mu2);
      if(res == EG_INVALID)
      {
        e_GeneralError("CheckNewTriang: geometry error 1 !");
      }
      
      if(   (res == EG_INTERSECT)
         && (mu1 >= 0.0 && mu1 <= 1.0)
         && (mu2 >= 0.0 && mu2 <= 1.0) )
      {
        return(INVALID_TRIANG);
      }
    }
    

    if(   n1 != segend && n2 != segend
       && n1 != selnode && n2 != selnode )
    {
      res = eg_LineLineIntersect2D(coord[n1-1], coord[n2-1], 
                                   selcoord, coord[segend-1], 
                                   &mu1, &mu2);

      if(res == EG_INVALID)
      {
        e_GeneralError("CheckNewTriang: geometry error 2 !");
      }

      if(   (res == EG_INTERSECT)
         && (mu1 >= 0.0 && mu1 <= 1.0)
         && (mu2 >= 0.0 && mu2 <= 1.0) )
      {
        return(INVALID_TRIANG);
      }
    }
    
    /* check that any of the nodes inside the triangle or on
       the edge of the triangle */
    if(   (   n1 != segend && n1 != segbeg && n1 != selnode
           && CheckInside(coord[segbeg-1], coord[segend-1], 
                          selcoord, coord[n1-1]) == TRUE)
       || (   n2 != segend && n2 != segbeg && n2 != selnode
           && CheckInside(coord[segbeg-1], coord[segend-1], 
                          selcoord, coord[n2-1]) == TRUE) )
    {
      return(INVALID_TRIANG);
    }
    
  }

  return(VALID_TRIANG);
}


/* //////////////////////////////////////////////////////////////////////////
Function:     NearSegments
Descriptions: determines whether a segment is too close to a chord node,
              the threshold distance is the distance from the chord node 
              to the reference segment; 
              if the perpendicular distance between the chord node and 
              another segment is smaller than half of the threshold 
              distance, then the chord node is not acceptable
              or one of the end nodes of the segment is within the threshold
              distance then the chord node is not acceptable
Arguments:		
              iseg: selected segment index
              selcoord: coordinates of the selected node
              coord: coordinates of points
              nseg: number of segments
              seg: list of segments
Return:       TRUE -> if a segment is too close
              FALSE -> if there is no segment near
///////////////////////////////////////////////////////////////////////////// */
static int NearSegments(int iseg, double *selcoord, 
                        double **coord, int nseg, int **seg)
{
  int    i, n1, n2;
  double delta, mu, dist;
  double l1, l2;

  n1 = seg[iseg][0];
  n2 = seg[iseg][1];
  /* threshold distance */
  delta = eg_PointLineDist2D(coord[n1-1], coord[n2-1], 
                             selcoord, &mu);

  if(mu < 0.0 || mu > 1.0 || delta < 0.0)
    e_GeneralError("NearSegments: internal error 1 !");

  for(i = 0; i < nseg; i++)
  {
    /* selected segment */
    if(i == iseg)
    {
      continue;
    }

    n1 = seg[i][0];
    n2 = seg[i][1];
    /* distance betwen the selected chord node and another segment */
    dist = eg_PointLineDist2D(coord[n1-1], coord[n2-1], 
                              selcoord, &mu);
    
    if(dist < 0.0)
      e_GeneralError("NearSegments: internal error 2 !");

    if(   dist < (delta / 2.0)
       && mu >= 0.0 && mu <= 1.0)
    {
      return(TRUE);
    }
    
    /* check the distance between the selected node and the end
       nodes of the segment */
    l1 = eg_PointDist2D(selcoord, coord[n1-1]);
    l2 = eg_PointDist2D(selcoord, coord[n2-1]);
    if(l1 < delta || l2 < delta)
    {
      return(TRUE);
    }
  }

  return(FALSE);
}











