/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                Triangular Element Generation Functions                     */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */
#include "adv_elem.h"
#include "eg_lib.h"       /* geometric operators */
#include "adv_config.h"   /* parameters for the advancing front algorithm */
#include "adv_apex.h"     /* node selection */
#include "adv_mpar.h"     /* mesh parameter determination */

#include "mgn_graphics.h" /* graphics implementation */
extern int ShowGraphics;
extern int PrintProgress;

/* local functions */
static int ShortestSegment(double **coords, int nsegt, int **seg);
static void RemoveSegment(int iseg, int *nsgt, int **segments);
static void AdjustSegments(int iseg, int selnod, int *nsgt, int ***seg);
static int SegmentExist(int n11, int n12, int nsgt, int **seg);

/* //////////////////////////////////////////////////////////////////////////
Function:     AdvTriang
Descriptions: This is the main function for generating triangular mesh
              using advancing front method. The basic principle is generating
              new triangle based on the shortest line segment of the current
              active front. Such process will only be terminated if there is
              no more line segment on the active front or the advancing front
              method failed.
Arguments:
              nnodegen:  the number of meshpoints
              coords:    the nodal coordinates
              nelemgen:  the number of elements defined in the node list
              elemnodes: the node list defining the elements
              nedges:    the number of edges passed in the list
              edges:     the list of edges
              ctrl:      mesh parameter control flag
              bgd:       all data for the background mesh and sources
              mpr:       mesh parameters of the boundary nodes if CTRL_MPR
              h:         uniform mesh parameter if CTRL_UNIFORM
Return:       None.

///////////////////////////////////////////////////////////////////////////// */

void AdvTriang(int *nnodegen, double ***coords,
               int *nelemgen, int ***elemnodes,
               int nedges, EDGE *edges,
               int ctrl, BG_DATA *bgd, double *mpr, double h)
{
  int 	  sum, i, j;
  int 	  nod1,nod2;
  int 	  iseg;
  int 	  selnode;
  double  x1,y1,x2,y2;
  double  middle_pos[2];
  double  d1, orient;
  int     nsegt;
  int   **segments;

  /* make a continuous and modifyable data structure from the edges */
  sum = 0;
  for(i = 0; i < nedges; i++)
  {
    sum += edges[i].nen-1;
  }
  segments = e_alloc_cont_i_matr(INCREMENT * ((int)(sum / INCREMENT)+1), 2);
  iseg = 0;
  for(i = 0; i < nedges; i++)
  {
    for(j = 0; j < edges[i].nen-1; j++)
    {
      segments[iseg][0] = edges[i].en[j];
      segments[iseg][1] = edges[i].en[j+1];
      iseg++;
    }
  }
  iseg = 0;
  nsegt = sum;

  /* triangs will be generated until nsegt > 0 */
  while(nsegt > 0)
  {
    /* if there are only three front segments on the boundary
       then only one triangle will be generated. */
    if(nsegt == 3)
    {
      *elemnodes =
        e_incr_cont_i_matr(*nelemgen, E_NNPE_TRIANG1, *elemnodes, INCREMENT);
      /* check the element for counter clockwise definition */
      orient = eg_Orient2D((*coords)[segments[0][0]-1],
                           (*coords)[segments[1][0]-1],
                           (*coords)[segments[2][0]-1] );
      if(orient > 0.0)
      {
        (*elemnodes)[*nelemgen][0] = segments[0][0];
        (*elemnodes)[*nelemgen][1] = segments[1][0];
        (*elemnodes)[*nelemgen][2] = segments[2][0];
      }
      else if(orient < 0.0)
      {
        (*elemnodes)[*nelemgen][0] = segments[2][0];
        (*elemnodes)[*nelemgen][1] = segments[1][0];
        (*elemnodes)[*nelemgen][2] = segments[0][0];
      }
      else if(orient == 0.0)
      {
        e_GeneralError("AdvTriang: internal error 1 !");
      }
      (*nelemgen)++;

      /* generation is finished */
      break;
    }
    else if(nsegt < 3)
    {
      e_GeneralError("AdvTriang: internal error 2 !");
    }

    /* determine the shortest segment */
    iseg = ShortestSegment(*coords, nsegt, segments);

    /* this is the shortest/selected segment */
    nod1 = segments[iseg][0]-1;
    nod2 = segments[iseg][1]-1;

    /* get the coordinates of the shortest segment. */
    x1 = (*coords)[nod1][0];
    y1 = (*coords)[nod1][1];
    x2 = (*coords)[nod2][0];
    y2 = (*coords)[nod2][1];

    /* middle point of the segment */
    middle_pos[0] = (x1+x2)/2.0;
    middle_pos[1] = (y1+y2)/2.0;

    /* get the meshparameter at the middle node */
    d1 = AdvGetMeshParam(middle_pos, ctrl, bgd, nedges, edges, *coords, mpr, h);

    /* select a node for the segment to form a triangle */
    selnode = SelectNode(iseg, d1, nnodegen, coords, nsegt, segments);

    /* still no node selected even after relaxation, print error message. */
    if(selnode == -1)
    {
      e_GeneralError("AdvTriang: No apex point could be selected.");
    }

    *elemnodes =
      e_incr_cont_i_matr(*nelemgen, E_NNPE_TRIANG1, *elemnodes, INCREMENT);
    (*elemnodes)[*nelemgen][0] = nod1+1;
    (*elemnodes)[*nelemgen][1] = nod2+1;
    (*elemnodes)[*nelemgen][2] = selnode;
    /* increase the new generated element number */
    (*nelemgen)++;

    if(PrintProgress && (*nelemgen) != 0 && ((*nelemgen) % 100) == 0)
    {
      printf("\nNumber of elements: %d", (*nelemgen) );
    }

    /* draw the triangle on the screen. */
    if(ShowGraphics)
    {
      mgnDrawAll(*nnodegen, *coords, *nelemgen, *elemnodes, 0,
                 NULL, -1, NULL, NULL,"element creation");
    }

    /* remove from and add segments to the front */
    AdjustSegments(iseg, selnode, &nsegt, &segments);

  }

  e_free(segments[0]);
  e_free(segments);
}


/* //////////////////////////////////////////////////////////////////////////
Function:     ShortestSegment
Descriptions:	determines the index of the shortest segment
Arguments:
              coords: array for nodal coordinates.
              nsegt: number of segments.
              seg: array of segments
Return:       None.
///////////////////////////////////////////////////////////////////////////// */
static int ShortestSegment(double **coords, int nsegt, int **seg)
{
  int    i;
  int    nod1, nod2, sgmin;
  double L, Lmin;

  /* take the first as the shortest segment */
  sgmin = 0;
  nod1 = seg[0][0]-1;
  nod2 = seg[0][1]-1;

  Lmin = eg_PointDist2D(coords[nod1], coords[nod2]);

  /* look for a shorter segment */
  for(i = 1; i < nsegt; i++)
  {
    nod1 = seg[i][0]-1;
    nod2 = seg[i][1]-1;

    L = eg_PointDist2D(coords[nod1], coords[nod2]);

    if(Lmin > L)
    {
      Lmin  = L;
      sgmin = i;
    }
  }

  return(sgmin);
}

/* //////////////////////////////////////////////////////////////////////////
Function:     RemoveSegment
Descriptions: removes the given segment from the active front
Arguments:
              iseg: the segment ID to remove
              nsgt: no of segments on the front
              segments: the list of segments
Return:       None.
///////////////////////////////////////////////////////////////////////////// */
static void RemoveSegment(int iseg, int *nsgt, int **segments)
{
  if((*nsgt) - iseg - 1)
  {
    memmove((void *)(segments[iseg]),
            (void *)(segments[iseg+1]),
            sizeof(int) * E_NVPI_TRIANG1 * ((*nsgt) - iseg - 1) );
  }
  (*nsgt)--;
}

/* //////////////////////////////////////////////////////////////////////////
Function:    SegmentExist
Description: checks if the generated side (n11, n12) already exist
Arguments:
             n11, n12 : nodes of the new segment
             nsgt: numbef of segments
             seg: list of segments
Return:  0-nsgt -> the index of the segment, if the segment exists
         -1     -> if it does not exist
///////////////////////////////////////////////////////////////////////////// */
static int SegmentExist(int n11, int n12, int nsgt, int **seg)
{
  int i;

  for(i = 0; i < nsgt; i++)
  {
    if(   (n11==seg[i][0] && n12==seg[i][1])
       || (n12==seg[i][0] && n11==seg[i][1]) )
    {
      return(i);
    }
  }
  return(-1);
}


/* //////////////////////////////////////////////////////////////////////////
Function:     AdjustSegments
Descriptions:	add the new generated segments into the active front,
              and if a segment exists, delete it
Arguments:
              iseg: the new segment ID.
              selnod: the ID for the new apex node.
              nsgt: the number of segments along active front.
              segments: the array for active front.
Return:       None
///////////////////////////////////////////////////////////////////////////// */
static void AdjustSegments(int iseg, int selnode, int *nsgt, int ***segments)
{
  int test1, test2;
  int n11, n22;

  n11 = (*segments)[iseg][0];
  n22 = (*segments)[iseg][1];

  /* delete the reference segment */
  RemoveSegment(iseg, nsgt, *segments);

  /* check the first segment on the active advancing front */
  test1 = SegmentExist(n11, selnode, *nsgt, *segments);

  /* add the left segment if it does not exist */
  if(test1 == -1)
  {
    (*segments) = e_incr_cont_i_matr(*nsgt, E_NVPI_TRIANG1, *segments, INCREMENT);
    (*segments)[*nsgt][0] = n11;
    (*segments)[*nsgt][1] = selnode;
    (*nsgt)++;
  }
  /* remove the segment if it does exist */
  else
  {
    RemoveSegment(test1, nsgt, *segments);
  }

  /* check the second segment on the active advancing front */
  test2 = SegmentExist(selnode, n22, *nsgt, *segments);

  /* add the right segment if it does not exist */
  if(test2 == -1)
  {
    (*segments) = e_incr_cont_i_matr(*nsgt, E_NVPI_TRIANG1, *segments, INCREMENT);
    (*segments)[*nsgt][0] = selnode;
    (*segments)[*nsgt][1] = n22;
    (*nsgt)++;
  }
  /* remove the segment if it does exist */
  else
  {
    RemoveSegment(test2, nsgt, *segments);
  }
}







