/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                       Advancing front main function                        */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_lib.h"
#include "mgni.h"
#include "getvertices.h"
#include "adv_elem.h"
#include "adv_edge.h"
#include "bgdata.h"

#include "mgn_graphics.h"
extern int ShowGraphics;

static MESH *OutMesh(int nnodes, double **coords, int nelem, int **elemnodes);

void adv_main(MGNI *mgni)
{
  int      nnodes;
  double **coords;
  double  *mpr;
  int      nelem;
  int    **elemnodes;
  EDGE   **edges;
  int      i, nvertices, *vertices;
  int      nedge;
  EDGE    *meshedges;
  BG_DATA *bgd;

  /* */
  if(mgni->ctrl == CTRL_BACKGROUND || mgni->ctrl == CTRL_SOURCES)
  {
    get_bg_data(&bgd, mgni->bgmesh, mgni->pcs);
  }
  else
  {
    bgd = NULL;
  }

  /* initialise variables */
  nnodes = 0;
  coords = NULL;
  mpr    = NULL;
  nelem  = 0;
  elemnodes = NULL;
  edges  = NULL;

  /* element by element */
  if(mgni->task==TASK_EBE)
  {
    /* create element edge data structures */
    CreateElemEdges(mgni->mesh, bgd,
                    mgni->ctrl, mgni->geomset,
                    (mgni->opt == CONV_OPT_FISSION_TO_QUAD),
                    &nnodes, &coords, &mpr, mgni->h, &edges);

    /* display initial nodes */
    if(ShowGraphics)
    {
       mgnDrawAll(nnodes, coords, 0, NULL, 0, NULL, -1, NULL, NULL, NULL);
    }

    /* go through all elements, use the bounding edges as the initial front */
    for(i = 0; i < mgni->mesh->TotalNElems; i++)
    {
      e_GetElementVertices(i, mgni->mesh, &nvertices, &vertices);
      AdvTriang(&nnodes, &coords, &nelem, &elemnodes,
                nvertices, edges[i], mgni->ctrl, bgd, mpr, mgni->h);
    }
  }
  /* mesh refmesh */
  else
  {
    /* create the mesh boundary as edge data structure */
    CreateMeshEdges(mgni->mesh, bgd,
                    mgni->ctrl, mgni->geomset,
                    (mgni->opt == CONV_OPT_FISSION_TO_QUAD),
                    &nnodes, &coords, &mpr, mgni->h, &nedge, &meshedges);

    /* display initial nodes */
    if(ShowGraphics)
    {
      mgnDrawAll(nnodes, coords, 0, NULL, 0, NULL, -1, NULL, NULL, NULL);
    }

    /* remesh the whole boundary */
    AdvTriang(&nnodes, &coords, &nelem, &elemnodes,
              nedge, meshedges, mgni->ctrl, bgd, mpr, mgni->h);
  }

  /* generate an E-lib output mesh */
  mgni->outputmesh = OutMesh(nnodes, coords, nelem, elemnodes);

  if(bgd != NULL)
  {
    free_bg_data(bgd);
  }

}


/* generate a valid E-lib mesh output */
static MESH *OutMesh(int nnodes, double **coords, int nelem, int **elemnodes)
{
  MESH *mesh;
  int   i;

  mesh = (MESH *)e_calloc(1,sizeof(MESH));

  mesh->Title = e_alloc_c_vect(1);

  mesh->NMeshPoints = nnodes;
  mesh->NNodes = nnodes;
  mesh->NElemsTriang1 = nelem;
  mesh->TotalNElems = nelem;

  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] = 1;

  mesh->MeshPointCoords = coords;

  mesh->NodesTriang1 = elemnodes;

  mesh->KeywordSet[E_KW_NUM_TITLE] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NMESHPOINTS] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NNODES] = TRUE;
  mesh->KeywordSet[E_KW_NUM_MESHPOINT_COORDS] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NODES_TRIANG1] = TRUE;

  mesh->TotalNodalDOF = mesh->NMeshPoints * E_VERTEX_NODE_DOF_TRIANG1;
  mesh->NodalDOF = e_alloc_i_vect(mesh->TotalNodalDOF);
  mesh->GNIs = e_alloc_i_vect(mesh->NMeshPoints);
  for(i = 0; i < mesh->NMeshPoints; i++)
  {
    mesh->NodalDOF[i] = E_VERTEX_NODE_DOF_TRIANG1;
    mesh->GNIs[i] = i+1;
  }
  mesh->GEIs = e_alloc_i_vect(mesh->TotalNElems);
  for(i = 0; i < mesh->TotalNElems; i++)
  {
    mesh->GEIs[i] = i+1;
  }
  mesh->GEIsTriang1 = mesh->GEIs;

  return mesh;
}



