#include "boundary.h"
#include "getvertices.h"

#define ALLOC_SIZE 10

static int FindEdge(double px, double py, double cx, double cy,
                    int nmeshes, MESH **mesh, int *nbnd, BND **bnd,
                    int *m, int *b, int *l, int *e, double epsilon);

/*
 * Generate the boundary of a mesh.
 *
 * The function assumes, that all elements are defined in anti-clockwise manner.
 *
 * The algorithm gets one boundary edge and starts walking in anti-clockwise
 * direction on the boundary. The mesh can have disconnected pieces, and
 * it handles tha case when the boundary is self-intersecting. This means
 * that two separate pieces of the mesh touches each other only at one node.
 *
 * The output will be stored in 'BND' structures, which contain one external
 * boundary and some holes if there are any. Holes are properly determined.
 *
 */

void Gen2DBoundary(MESH *mesh, int *NoNodeElem, int **NodeElem,
                   int *NBoundaries, BND **Boundaries)

{
  int    NBList;      /* no of boundaries */
  int   *BListLength; /* length of the boundary */
  int  **BList;       /* boundary lists */
  int    i, j, k, ii, i_node; /* loop variables */
  int  **mark;                /* marking already used edges */
  int    found, elem;
  int    ns, nx;
  int    req_pt, end_pt, prev_pt;
  int    NVertices;           /* the number of nodes the selected element has */
  int   *Vertices;            /* the nodes the selected element has */

  /* for the final stage, when everything is converted to BND struct */
  double x, y, x1, y1, area;
  int    state;
  void  *new_ptr;

  /* clear variables */
  NBList = 0;
  BListLength = NULL;
  BList = NULL;
  *NBoundaries = 0;
  *Boundaries  = NULL;

  /* checking that only good meshes are processed */
  if((mesh->NElemsTriang1 +
      mesh->NElemsTriang3 +
      mesh->NElemsTriang4 +
      mesh->NElemsTriang5 +
      mesh->NElemsQuad1    ) != mesh->TotalNElems)
    e_GeneralError("Mesh contains unsupported element(s) !");

  mark = e_init_frag_i_matr(mesh->TotalNElems);

  /* going through all elements */
  for(i = 0; i < mesh->TotalNElems; i++)
  {
    e_GetElementVertices(i, mesh, &NVertices, &Vertices);

    if(mark[i] == NULL)
      mark[i] = e_alloc_i_vect(NVertices);
    else
    {
      /* checking whether we considered all edges already ? */
      for(j = 0; j < NVertices; j++)
        if(mark[i][j] == 0)
          break;

      if(j == NVertices) /* The answer is yes. Well the answer is always 42. */
        continue;
    }

    /* search for a boundary element */
    ns = 0;
    for(i_node = 0; i_node < NVertices; i_node++)
    {
      nx = NVertices - i_node - 1;
      if(mark[i][nx] == 0)
      {
        /* check all possible combination of elements */
        found = 0;
        for(j = 0; j < NoNodeElem[Vertices[ns]-1]; j++)
        {
          for(k = 0; k < NoNodeElem[Vertices[nx]-1]; k++)
          {
            /* the element contains the two nodes and it is not itself */
            if(   (NodeElem[Vertices[ns]-1][j] == NodeElem[Vertices[nx]-1][k])
               && (NodeElem[Vertices[ns]-1][j] != i+1) )
            {
              found = 1; /* found adjacent element */
              mark[i][nx] = 1; /* mark the side */
              break;
            }
          }
          if(found) break; /* found adjacent element */
        }
        if(!found) break; /* found boundary side */
      }
      ns = nx;
    }

    /* if there is no adjacent element it is a free side -> boundary */
    if(i_node != NVertices)
    {
      /* mark the boundary side */
      mark[i][nx] = 1;
      BListLength    = e_incr_i_vect(NBList, BListLength, 1);
      BList          = e_incr_frag_i_matr(NBList, BList, 1);
      BList[NBList] = e_alloc_i_vect(ALLOC_SIZE);

      /* if going around in anti-clockwise direction, the boundary will
         be in anti-clockwise direction as well */
      BList[NBList][0] = Vertices[nx];
      BList[NBList][1] = Vertices[ns];
      BListLength[NBList] += 2;

      req_pt  = Vertices[ns];
      end_pt  = Vertices[nx];
      prev_pt = end_pt;
      while(req_pt != end_pt)
      {
        /* check all elements around the node */
        for(ii = 0; ii < NoNodeElem[req_pt-1]; ii++)
        {
          elem = NodeElem[req_pt-1][ii]-1;
          e_GetElementVertices(elem, mesh, &NVertices, &Vertices);

          if(mark[elem] == NULL)
            mark[elem] = e_alloc_i_vect(NVertices);

          ns = 0;
          /* go to the side where the starting node is req_pt */
          for(i_node = 0; i_node < NVertices; i_node++)
          {
            nx = NVertices - i_node - 1;
            if(   Vertices[nx] == req_pt  /* continuation of the list */
               && Vertices[ns] != prev_pt /* not going back, probably redundant check */
               && mark[elem][nx] == 0)    /* edge has not been used */
            {
              /* check whether it is a boundary */
              found = 0;
              for(j = 0; j < NoNodeElem[Vertices[ns]-1]; j++)
              {
                for(k = 0; k < NoNodeElem[Vertices[nx]-1]; k++)
                {
                  /* the element contains the two nodes and it is not itself */
                  if(   (NodeElem[Vertices[ns]-1][j] == NodeElem[Vertices[nx]-1][k])
                     && (NodeElem[Vertices[ns]-1][j] != elem+1) )
                  {
                    found = 1; /* so there is an adjacent element */
                    mark[elem][nx] = 1;
                    break;
                  }
                }
                if(found) break; /* we already know, that there is an element */
              }
              if(!found) break; /* starting node is good and the side is boundary, quit*/
            } /* if required node */
            else
              found = 1;
            ns = nx;
          } /* go to proper side */
          if(!found) break;
        }

        /* ok, so we have the side and the new node */
        if(ii != NoNodeElem[req_pt-1])
        {
          mark[elem][nx] = 1;
          prev_pt = req_pt;
          req_pt = Vertices[ns];
          /* Check whether we have it on the list, then it is self-intersection. */
          found = 0;
          for(ii = 0; ii < BListLength[NBList]; ii++)
          {
            if(BList[NBList][ii] == req_pt)
            {
              if(ii == 0)
              {
                /* e_GeneralWarning("Boundary closed !"); */
                found = 1;
                break;
              }
              e_GeneralWarning("Self-intersecting boundary !");
              /* the intersecting part will be cut off
                 the cut off list comes first, then the "old" list, because the old
                 list should be continued */
              BList = e_incr_frag_i_matr(NBList+1, BList, 1);
              BListLength = e_incr_i_vect(NBList+1, BListLength, 1);
              /* move the list ahead */
              BList[NBList+1]       = BList[NBList];
              /* adjust the size of the old list */
              BListLength[NBList+1] = ii;

              /* adjust the size of the cut-off part */
              BListLength[NBList] -= ii;
              /* allocate a new list for the cut-off part */
              BList[NBList] = e_alloc_i_vect(BListLength[NBList]);
              /* copy over the cut-off part */
              memcpy((void *)BList[NBList],
                     (void *)&(BList[NBList+1][ii]),
                     sizeof(int) * BListLength[NBList]);
              /* increase the number of lists */
              NBList++;
              break;
            }
          }
          /* not found the end, add the node */
          if(!found)
          {
            BList[NBList] = e_incr_i_vect(BListLength[NBList],
                                          BList[NBList],
                                          ALLOC_SIZE);
            BList[NBList][BListLength[NBList]] = req_pt;
            BListLength[NBList]++;
          }
        }
        else
          e_GeneralError("Could not continue the boundary line !");
      }

      NBList++;
    }
  } /* all element */

  e_free_frag_i_matr(mesh->TotalNElems, mark);

  /* /////////////////////////////////////////////////////////////////// */

  /* now we have all lists, but do not know which are actual holes */
  /* create outer boundaries */
  for(i = 0; i < NBList; i++)
  {
    if(BListLength[i] < 3)
      e_GeneralError("Boundary is one line !");

    /* determine whether the order is anti-clockwise or not */
    area = 0.0;
    for(j = 1; j < BListLength[i]-1; j++)
    {
      x = mesh->MeshPointCoords[BList[i][j]-1][0] -
          mesh->MeshPointCoords[BList[i][0]-1][0];
      y = mesh->MeshPointCoords[BList[i][j]-1][1] -
          mesh->MeshPointCoords[BList[i][0]-1][1];

      x1 = mesh->MeshPointCoords[BList[i][j+1]-1][0] -
           mesh->MeshPointCoords[BList[i][j]  -1][0];
      y1 = mesh->MeshPointCoords[BList[i][j+1]-1][1] -
           mesh->MeshPointCoords[BList[i][j]  -1][1];

      /* cross product of the two vector */
      area += x * y1 - y * x1;
    }
    /* this is anti-clockwise order, this is an outer boundary */
    if(area > 0.0)
    {
      /* store it */
      new_ptr = e_calloc((*NBoundaries)+1, sizeof(BND));
      memcpy((void *)new_ptr, (void *)(*Boundaries),
             sizeof(BND) * (*NBoundaries));
      e_free(*Boundaries);
      (*Boundaries) = (BND *)new_ptr;

      (*Boundaries)[*NBoundaries].oe.nen = BListLength[i];
      (*Boundaries)[*NBoundaries].oe.en  = BList[i];
      (*Boundaries)[*NBoundaries].nie    = 0; /* to be sure that no hole yet */
      (*NBoundaries)++;

      /* remove from lists */
      for(j = i; j < NBList-1; j++)
      {
        BList[j] = BList[j+1];
        BListLength[j] = BListLength[j+1];
      }
      NBList--;
      i--;
    }
  }

  /* one method to determine where the hole belongs to is to check
     that a hole node is inside of which external boundary */
  for(i = 0; i < NBList; i++)
  {
    for(j = 0; j < *NBoundaries; j++)
    {

      for(k = 0; k < BListLength[i]; k++)
      {
        state = Check2DNodeInside(mesh->MeshPointCoords,
                                  (*Boundaries)[j].oe.nen,
                                  (*Boundaries)[j].oe.en,
                                  BList[i][k],
                                  mesh->MeshPointCoords[BList[i][k]-1][0],
                                  mesh->MeshPointCoords[BList[i][k]-1][1]  );
        if(state != -1)
          break;
      }

      if(k == BListLength[i])
        e_GeneralError("Internal and external boundary is identical !");

      /* we found its external boundary */
      if(state == 1)
        break;
    }

    if(j == *NBoundaries)
      e_GeneralError("Internal boundary cannot be placed into an external one !");

    /* so we have the external, then place the current into it as a hole */
    new_ptr = e_calloc(((*Boundaries)[j].nie)+1, sizeof(EDGE));
    memcpy((void *)new_ptr, (void *)((*Boundaries)[j].ie),
           sizeof(EDGE) * (*Boundaries)[j].nie);
    e_free((*Boundaries)[j].ie);
    (*Boundaries)[j].ie = (EDGE *)new_ptr;

    (*Boundaries)[j].ie[(*Boundaries)[j].nie].nen = BListLength[i];
    (*Boundaries)[j].ie[(*Boundaries)[j].nie].en  = BList[i];
    (*Boundaries)[j].nie++;
  }

}


void Free2DBoundary(int NBList, BND *BList)
{
  int i, j;

  for(i = 0; i < NBList; i++)
  {
    e_free(BList[i].oe.en);
    for(j = 0; j < BList[i].nie; j++)
    {
      e_free(BList[i].ie[j].en);
    }
  }
  e_free(BList);
}

/* #################################################################### */

static int FindEdge(double px, double py, double cx, double cy,
                    int nmeshes, MESH **mesh, int *nbnd, BND **bnd,
                    int *m, int *b, int *l, int *e, double epsilon)
{
  int imesh, ibnd, i, j;
  int pre, npre, ncur;
  double x1, x2, y1, y2;

  for(imesh = 0; imesh < nmeshes; imesh++)
  {
    for(ibnd = 0; ibnd < nbnd[imesh]; ibnd++)
    {
      /* check external boundary */
      pre  = bnd[imesh][ibnd].oe.nen-1;
      npre = bnd[imesh][ibnd].oe.en[pre]-1;
      for(i = 0; i < bnd[imesh][ibnd].oe.nen; i++)
      {
        ncur = bnd[imesh][ibnd].oe.en[i]-1;
        x1 = mesh[imesh]->MeshPointCoords[npre][0];
        y1 = mesh[imesh]->MeshPointCoords[npre][1];
        x2 = mesh[imesh]->MeshPointCoords[ncur][0];
        y2 = mesh[imesh]->MeshPointCoords[ncur][1];

        if(   (E_ABS(x1-px) < epsilon && E_ABS(y1-py) < epsilon)
           && (E_ABS(x2-cx) < epsilon && E_ABS(y2-cy) < epsilon) )
        {
          *m = imesh;
          *b = ibnd;
          *l = 0;
          *e = pre;
          return 1;
        }

        pre  = i;
        npre = ncur;
      }
      /* check internal boundaries */
      for(j = 0; j < bnd[imesh][ibnd].nie; j++)
      {
        pre  = bnd[imesh][ibnd].ie[j].nen-1;
        npre = bnd[imesh][ibnd].ie[j].en[pre]-1;
        for(i = 0; i < bnd[imesh][ibnd].ie[j].nen; i++)
        {
          ncur = bnd[imesh][ibnd].ie[j].en[i]-1;
          x1 = mesh[imesh]->MeshPointCoords[npre][0];
          y1 = mesh[imesh]->MeshPointCoords[npre][1];
          x2 = mesh[imesh]->MeshPointCoords[ncur][0];
          y2 = mesh[imesh]->MeshPointCoords[ncur][1];

          if(   (E_ABS(x1-px) < epsilon && E_ABS(y1-py) < epsilon)
             && (E_ABS(x2-cx) < epsilon && E_ABS(y2-cy) < epsilon) )
          {
            *m = imesh;
            *b = ibnd;
            *l = j+1;
            *e = pre;
            return 1;
          }

          pre  = i;
          npre = ncur;
        }
      }
    }
  }
  return 0;
}


void MatchAll2DBoundaries(int noldmesh, MESH **oldmesh, int nnewmesh, MESH **newmesh,
                          int *noldbnd, BND **oldbnd, int *nnewbnd, BND **newbnd,
                          MATCH_BND ***oldmatch, double epsilon, int perfect)
{
  int imesh, ibnd, i, j;
  int pre, npre, ncur;
  double x1, x2, y1, y2;
  int    m, b, l, e;

  (*oldmatch) = (MATCH_BND **)e_calloc(noldmesh, sizeof(MATCH_BND *) );

  for(imesh = 0; imesh < noldmesh; imesh++)
  {

    (*oldmatch)[imesh] = (MATCH_BND *)e_calloc(noldbnd[imesh], sizeof(MATCH_BND) );

    for(ibnd = 0; ibnd < noldbnd[imesh]; ibnd++)
    {
      (*oldmatch)[imesh][ibnd].oe.n     = oldbnd[imesh][ibnd].oe.nen;
      (*oldmatch)[imesh][ibnd].oe.found = e_alloc_i_vect(oldbnd[imesh][ibnd].oe.nen);
      (*oldmatch)[imesh][ibnd].oe.mesh  = e_alloc_i_vect(oldbnd[imesh][ibnd].oe.nen);
      (*oldmatch)[imesh][ibnd].oe.bnd   = e_alloc_i_vect(oldbnd[imesh][ibnd].oe.nen);
      (*oldmatch)[imesh][ibnd].oe.list  = e_alloc_i_vect(oldbnd[imesh][ibnd].oe.nen);
      (*oldmatch)[imesh][ibnd].oe.edge  = e_alloc_i_vect(oldbnd[imesh][ibnd].oe.nen);

      /* check external boundary */
      pre  = oldbnd[imesh][ibnd].oe.nen-1;
      npre = oldbnd[imesh][ibnd].oe.en[pre]-1;
      for(i = 0; i < oldbnd[imesh][ibnd].oe.nen; i++)
      {
        ncur = oldbnd[imesh][ibnd].oe.en[i]-1;

        x1 = oldmesh[imesh]->MeshPointCoords[npre][0];
        y1 = oldmesh[imesh]->MeshPointCoords[npre][1];
        x2 = oldmesh[imesh]->MeshPointCoords[ncur][0];
        y2 = oldmesh[imesh]->MeshPointCoords[ncur][1];

        if(FindEdge(x1, y1, x2, y2, nnewmesh, newmesh, nnewbnd, newbnd,
                    &m, &b, &l, &e, epsilon) )
        {
          (*oldmatch)[imesh][ibnd].oe.found[pre] = TRUE;
          (*oldmatch)[imesh][ibnd].oe.mesh[pre] = m;
          (*oldmatch)[imesh][ibnd].oe.bnd [pre] = b;
          (*oldmatch)[imesh][ibnd].oe.list[pre] = l;
          (*oldmatch)[imesh][ibnd].oe.edge[pre] = e;
        }
        else
        {
          if(perfect)
            e_GeneralError("MatchAllBoundaries: External edge is not found !");
        }

        pre  = i;
        npre = ncur;
      }

      /* check internal boundaries */
      (*oldmatch)[imesh][ibnd].nie = oldbnd[imesh][ibnd].nie;
      if(oldbnd[imesh][ibnd].nie)
      {
        (*oldmatch)[imesh][ibnd].ie =
          (MATCH_EDGE *)e_calloc(oldbnd[imesh][ibnd].nie, sizeof(MATCH_EDGE) );
      }
      else
      {
        (*oldmatch)[imesh][ibnd].ie = NULL;
      }

      for(j = 0;  j < oldbnd[imesh][ibnd].nie; j++)
      {
        (*oldmatch)[imesh][ibnd].ie[j].n     = oldbnd[imesh][ibnd].ie[j].nen;
        (*oldmatch)[imesh][ibnd].ie[j].found = e_alloc_i_vect(oldbnd[imesh][ibnd].ie[j].nen);
        (*oldmatch)[imesh][ibnd].ie[j].mesh  = e_alloc_i_vect(oldbnd[imesh][ibnd].ie[j].nen);
        (*oldmatch)[imesh][ibnd].ie[j].bnd   = e_alloc_i_vect(oldbnd[imesh][ibnd].ie[j].nen);
        (*oldmatch)[imesh][ibnd].ie[j].list  = e_alloc_i_vect(oldbnd[imesh][ibnd].ie[j].nen);
        (*oldmatch)[imesh][ibnd].ie[j].edge  = e_alloc_i_vect(oldbnd[imesh][ibnd].ie[j].nen);

        pre  = oldbnd[imesh][ibnd].ie[j].nen-1;
        npre = oldbnd[imesh][ibnd].ie[j].en[pre]-1;
        for(i = 0; i < oldbnd[imesh][ibnd].ie[j].nen; i++)
        {
          ncur = oldbnd[imesh][ibnd].ie[j].en[i]-1;

          x1 = oldmesh[imesh]->MeshPointCoords[npre][0];
          y1 = oldmesh[imesh]->MeshPointCoords[npre][1];
          x2 = oldmesh[imesh]->MeshPointCoords[ncur][0];
          y2 = oldmesh[imesh]->MeshPointCoords[ncur][1];

          if(FindEdge(x1, y1, x2, y2, nnewmesh, newmesh, nnewbnd, newbnd,
                      &m, &b, &l, &e, epsilon) )
          {
            (*oldmatch)[imesh][ibnd].ie[j].found[pre] = TRUE;
            (*oldmatch)[imesh][ibnd].ie[j].mesh[pre] = m;
            (*oldmatch)[imesh][ibnd].ie[j].bnd [pre] = b;
            (*oldmatch)[imesh][ibnd].ie[j].list[pre] = l;
            (*oldmatch)[imesh][ibnd].ie[j].edge[pre] = e;
          }
          else
          {
            if(perfect)
              e_GeneralError("MatchAllBoundaries: Internal edge is not found !");
          }

          pre  = i;
          npre = ncur;
        }
      }
    }
  }
}


void Free2DMatchBnd(int NBList, MATCH_BND *BList)
{
  int i, j;

  for(i = 0; i < NBList; i++)
  {
    e_free(BList[i].oe.found);
    e_free(BList[i].oe.mesh);
    e_free(BList[i].oe.bnd);
    e_free(BList[i].oe.list);
    e_free(BList[i].oe.edge);
    for(j = 0; j < BList[i].nie; j++)
    {
      e_free(BList[i].ie[j].found);
      e_free(BList[i].ie[j].mesh);
      e_free(BList[i].ie[j].bnd);
      e_free(BList[i].ie[j].list);
      e_free(BList[i].ie[j].edge);
    }
  }
  e_free(BList);
}




