/********************************************************************
 mtrf.c --
        Mapping (and reverse mapping) to coordinate system
        associated with triangle

 Jelle Muylle <jelle@sect.mce.hw.ac.uk>
 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University, Edinburgh
********************************************************************/

#include <math.h>               /* for sqrt */
#include "e_lib.h"
#include "mtrf.h"


/* Tolerance for collinearity test */
#define TOLERANCE 1.e-10

/* Vector operation macros */

/* add vectors: result = a + b */
#define VEC3ADD(result,a,b) \
{  \
   *((result)+0) = (*((a)+0)) + (*((b)+0)); \
   *((result)+1) = (*((a)+1)) + (*((b)+1)); \
   *((result)+2) = (*((a)+2)) + (*((b)+2)); \
}

/* substract vectors: result = a -b */
#define VEC3SUB(result,a,b) \
{  \
   *((result)+0) = (*((a)+0)) - (*((b)+0)); \
   *((result)+1) = (*((a)+1)) - (*((b)+1)); \
   *((result)+2) = (*((a)+2)) - (*((b)+2)); \
}

/* Euclidean norm of a vector */
#define VEC3LENGTH(a) (sqrt ((*((a)+0)) * (*((a)+0)) \
                            +(*((a)+1)) * (*((a)+1)) \
                            +(*((a)+2)) * (*((a)+2)) ))

/* copy vectors: result = p */
#define VEC3COPY(result,p) \
{ \
  *((result)+0) = *((p)+0); \
  *((result)+1) = *((p)+1); \
  *((result)+2) = *((p)+2);\
}

/* normalize vector */
#define VEC3NORMALIZE(a) \
{ \
    double length; \
    length = VEC3LENGTH (a); \
    *((a)+0) /= length; \
    *((a)+1) /= length; \
    *((a)+2) /= length; \
}

/* cross product: result = a x b */
#define VEC3CROSS(result, a, b) \
{ \
  double cx, cy; \
  cx = (*((a)+1)) * (*((b)+2)) - (*((a)+2)) * (*((b)+1)); \
  cy = (*((a)+2)) * (*((b)+0)) - (*((a)+0)) * (*((b)+2)); \
  *((result)+2) = (*((a)+0)) * (*((b)+1)) - (*((a)+1)) * (*((b)+0)); \
  *((result)+1) = cy; \
  *((result)+0) = cx; \
}

/* dot product */
#define VEC3DOT(a,b)  ((*((a)+0)) * (*((b)+0)) \
                           + (*((a)+1)) * (*((b)+1)) \
                           + (*((a)+2)) * (*((b)+2)))


/* linear combination of vectors: result = alpha*a + beta*b */
#define VEC3COMBINE(result, alpha, a, beta, b) \
{ \
  *((result)+0) = (alpha) * (*((a)+0)) + (beta) * (*((b)+0));  \
  *((result)+1) = (alpha) * (*((a)+1)) + (beta) * (*((b)+1));  \
  *((result)+2) = (alpha) * (*((a)+2)) + (beta) * (*((b)+2));  \
}


/* retrieve the mapping transformation data based
 * on the triangle whose corners coordinates are
 * stored in coords on index n[0], n[1], and n[3]
 *
 * returns TRUE if mtrf is retrieved successfully
 * and FALSE if any problems occur such as collinearity
 */
int get_mtrf(double **coords, int *n, MTRF * mtrf)
{
  double v[3];
  VEC3SUB(mtrf->u, coords[n[1]], coords[n[0]]);
  VEC3SUB(v, coords[n[2]], coords[n[0]]);

  /* set  the point vector of system origin */
  VEC3COPY(mtrf->r, coords[n[0]]);

  /* calculate 'side' (u) and 'aside' (s) base vectors */

  VEC3CROSS(mtrf->s, mtrf->u, v);
  VEC3COPY(mtrf->w, mtrf->s);
  /* check if triangle point are not collinear */
  if ((VEC3LENGTH(mtrf->s) < TOLERANCE) || (VEC3LENGTH(mtrf->u) < TOLERANCE))
  {
    return FALSE;
  }
  VEC3CROSS(mtrf->s, mtrf->s, mtrf->u);

  /* normalize base vectors */
  VEC3NORMALIZE(mtrf->u);
  VEC3NORMALIZE(mtrf->s);

  return TRUE;
}


/* map the point src into point dest using the
 * mapping transformation based on data from mtrf.
 * This mapping can be seen as orthogonal projection
 * on the coordinate plane associated with a triangle.
 */
int mtrf_map(double *src, MTRF * mtrf, double *dest)
{
  double rl[3];
  VEC3SUB(rl, src, mtrf->r);
  *(dest + 0) = VEC3DOT(mtrf->u, rl);
  *(dest + 1) = VEC3DOT(mtrf->s, rl);
  *(dest + 2) = 0.0;
  return TRUE;
}

/* map the point src into point dest using the
 * reverse mapping transformation based on data  from mtrf
 */
int mtrf_revmap(double *src, MTRF * mtrf, double *dest)
{
  double xn[3];
  VEC3COMBINE(xn, src[0], mtrf->u, src[1], mtrf->s);
  VEC3ADD(dest, xn, mtrf->r);
  return TRUE;
}


