/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                             PostProcessMesh()                              */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/* NOTE: For planar element meshes, the smoothing process is only applicable  */
/*       to planar meshes, although they need not lie in the x-y-plane. The   */
/*       diagonal exchange function (which is only applied to triangular      */
/*       element meshes) assumes the mesh is planar and lies in the           */
/*       x-y-plane.                                                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "postpro.h"
#include "mgni.h"
#include "eg_lib.h"
#include "adjside.h"
#include "nodelem.h"
#include "enhance.h"

#define ADJ_NODE_ELEMS_INCREM  8

/* ////////////////////////////////////////////////////////////////////////// */
static void InitPostProcessMesh(MESH *mesh, int ***AdjElems,
                                int **NAdjMeshPoints, int ***AdjMeshPoints,
                                int **NodeTypes);
static void FreePostProcessMesh(MESH *mesh, int **AdjElems,
                                int *NAdjMeshPoints, int **AdjMeshPoints,
                                int *NodeTypes);
static void SetAdjMeshPoints(MESH *mesh, int **NAdjMeshPoints,
                             int ***AdjMeshPoints);
static void UpdateMeshPoints(int k1, int k2, int *NAdjMeshPoints,
                             int **AdjMeshPoints);
static int IsANewAdjMeshPoint(int NAdjMeshPoints, int *AdjMeshPoints,
                              int MeshPoint);
static void SetNodeTypes(MESH *mesh, int **AdjElems, int **NodeTypes);
static void SmoothMesh(MESH *mesh, int *NAdjMeshPoints, int **AdjMeshPoints,
                       int *NodeTypes);
static void ExchangeDiagonals(MESH *mesh, int **AdjElems,
                              int *NAdjMeshPoints, int **AdjMeshPoints);
static int ShouldExchangeEdge(MESH *mesh, int **Nodes, int **AdjElems, 
                              int i, int j);
static double GetMaxCosine(MESH *mesh, int k1, int k2, int k3);
static void ExchangeEdge(int ElemType, int **Nodes, int **AdjElems,
                         int *NAdjMeshPoints, int **AdjMeshPoints,
                         int i, int j);
static int GetAdjInterfIndex(int NIpE, int ElemIndex, int *AdjElems);


static int NodeElemsFormATriplet(MESH *mesh, 
                                 int NAdjNodeElems, int *AdjNodeElems);
static int compare(const void *elem1, const void *elem2);
static void CreateNewElem(MESH *mesh, int *NElemsTriang1, int **NodesTriang1,
                          int ElimNode, int *NewNodeNums, int **AdjNodeElems );


#define MEM_INCREM  8        /* memory increment for adjacent mesh-points */


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                             PostProcessMesh()                              */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* PostProcessMesh: To post-process a mesh. If the elements are triangular,
   then the mesh-points are first smoothed followed by "NIterations" of
   diagonal exchange plus smoothing. No post-processing is performed if
   "NIterations" is zero. If the elements are not triangular, then only
   "NIterations" of smoothing are performed.                                  */


int PostProcessMesh(MESH *mesh, int NIterations, int sm_type)
{
   int i;
   int *NodeTypes;
   int **AdjElems;
   int *NAdjMeshPoints;
   int **AdjMeshPoints;


   if(NIterations <= 0)
      return 0;
   
   if(mesh->NElemsTriang1 + mesh->NElemsQuad1 != mesh->TotalNElems)
   {
     printf("\nPostproc support only TRIANG1 and QUAD1 !\n");
     return 0;
   }
   
   if(mesh->NElemsTriang1 == mesh->TotalNElems)
   {
      for(i = 0; i < NIterations-1; i++)
      {
        RemoveTriangTriplets(mesh); 
        InitPostProcessMesh(mesh,&AdjElems,&NAdjMeshPoints,
                            &AdjMeshPoints,&NodeTypes);
        SmoothMesh(mesh,NAdjMeshPoints,AdjMeshPoints,NodeTypes);
        if(sm_type == SM_LAPLACIAN)
        {
          ExchangeDiagonals(mesh,AdjElems,NAdjMeshPoints,AdjMeshPoints);
        }
        FreePostProcessMesh(mesh,AdjElems,NAdjMeshPoints,AdjMeshPoints,NodeTypes);
      }

      RemoveTriangTriplets(mesh);
   }
   else if(mesh->NElemsQuad1 == mesh->TotalNElems)
   {
      InitPostProcessMesh(mesh, &AdjElems,
                          &NAdjMeshPoints, &AdjMeshPoints, &NodeTypes);

      for(i = 0; i < NIterations; i++)
      {
        SmoothMesh(mesh,NAdjMeshPoints,AdjMeshPoints,NodeTypes);
      }
      FreePostProcessMesh(mesh, AdjElems, NAdjMeshPoints, AdjMeshPoints, NodeTypes);
      QuadEnhancement(mesh);

   }
   else
   {
     InitPostProcessMesh(mesh, &AdjElems,
                         &NAdjMeshPoints, &AdjMeshPoints, &NodeTypes);
     for(i = 0; i < NIterations-1; i++)
     {
       SmoothMesh(mesh,NAdjMeshPoints,AdjMeshPoints,NodeTypes);
     }
     FreePostProcessMesh(mesh, AdjElems, NAdjMeshPoints, AdjMeshPoints, NodeTypes);
   }
   
   InitPostProcessMesh(mesh,&AdjElems,&NAdjMeshPoints,
                       &AdjMeshPoints,&NodeTypes);
   SmoothMesh(mesh,NAdjMeshPoints,AdjMeshPoints,NodeTypes);
   FreePostProcessMesh(mesh,AdjElems,NAdjMeshPoints,AdjMeshPoints,NodeTypes);

   return 1;
}



/* ////////////////////////////////////////////////////////////////////////// */
/* InitPostProcessMesh: To initialise post-processing data. */
static void InitPostProcessMesh(MESH *mesh, int ***AdjElems,
                                int **NAdjMeshPoints, int ***AdjMeshPoints,
                                int **NodeTypes)
{
  int *nne;
  int **ne;
  int **side;
  
  nne = e_alloc_i_vect(mesh->NMeshPoints);
  ne  = e_init_frag_i_matr(mesh->NMeshPoints);
  e_InitMeshPointElemConnection(mesh, nne, ne);
  e_SetAdjacencyAndSides(mesh, nne, ne, AdjElems, &side);
  e_free_frag_i_matr(mesh->NMeshPoints, ne);
  e_free_frag_i_matr(mesh->TotalNElems, side);
  
  /* e_SetAdjElems(mesh, AdjElems); */
  SetAdjMeshPoints(mesh, NAdjMeshPoints, AdjMeshPoints);
  SetNodeTypes(mesh, *AdjElems, NodeTypes);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* FreePostProcessMesh: To free memory allocated for the post-processing
   data.                                                                      */
static void FreePostProcessMesh(MESH *mesh, int **AdjElems,
                                int *NAdjMeshPoints, int **AdjMeshPoints,
                                int *NodeTypes)
{
  e_free(NodeTypes);
  e_free_frag_i_matr(mesh->TotalNElems,AdjElems);
  e_free(NAdjMeshPoints);
  e_free_frag_i_matr(mesh->NMeshPoints,AdjMeshPoints);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetAdjMeshPoints: To determine which mesh-points are adjacent to which
   (i.e. share an edge).                                                      */
static void SetAdjMeshPoints(MESH *mesh, int **NAdjMeshPoints,
                             int ***AdjMeshPoints)
{
   int i,j,k1,k2;
   int ElemType;
   int NVpE;
   int **Vertices;


   ElemType = e_GetElementType(mesh);

   *NAdjMeshPoints = e_alloc_i_vect(mesh->NMeshPoints);
   *AdjMeshPoints = e_init_frag_i_matr(mesh->NMeshPoints);

   NVpE = e_GetNVerticesPerElement(mesh);
   Vertices = e_GetVertices(mesh);

   if(   ElemType == E_ELEM_TYPE_TRIANG1
      || ElemType == E_ELEM_TYPE_QUAD1  )
   {
      for(i = 0; i < mesh->TotalNElems; i++)
      {
         for(j = 0; j < NVpE; j++)
         {
            k1 = Vertices[i][j] - 1;
            k2 = Vertices[i][(j + 1) % NVpE] - 1;

            UpdateMeshPoints(k1,k2,*NAdjMeshPoints,*AdjMeshPoints);
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* UpdateMeshPoints: To update the adjacent mesh-point data. */
static void UpdateMeshPoints(int k1, int k2, int *NAdjMeshPoints,
                             int **AdjMeshPoints)
{
   if(IsANewAdjMeshPoint(NAdjMeshPoints[k1],AdjMeshPoints[k1],k2 + 1))
   {
      AdjMeshPoints[k1]
         = e_incr_i_vect(NAdjMeshPoints[k1],AdjMeshPoints[k1],MEM_INCREM);
      AdjMeshPoints[k1][NAdjMeshPoints[k1]++] = k2 + 1;
   }

   if(IsANewAdjMeshPoint(NAdjMeshPoints[k2],AdjMeshPoints[k2],k1 + 1))
   {
      AdjMeshPoints[k2]
         = e_incr_i_vect(NAdjMeshPoints[k2],AdjMeshPoints[k2],MEM_INCREM);
      AdjMeshPoints[k2][NAdjMeshPoints[k2]++] = k1 + 1;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* IsANewAdjMeshPoint: To return whether a mesh-point has a new adjacent
   mesh-point.                                                                */
static int IsANewAdjMeshPoint(int NAdjMeshPoints, int *AdjMeshPoints,
                              int MeshPoint)
{
   int j;


   for (j = 0; j < NAdjMeshPoints; j++)
   {
      if (AdjMeshPoints[j] == MeshPoint)
         return 0;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetNodeTypes: To determine whether the nodes (mesh-points) in a mesh are
   interior (E_RM_NUM_INTERIOR) or exterior (E_RM_NUM_EXTERIOR).              */
static void SetNodeTypes(MESH *mesh, int **AdjElems, int **NodeTypes)
{
   int i,j,k,l;
   int ElemType;
   int NIpE;
   int NVpE;
   int NVpI;
   int **Vertices;


   *NodeTypes = e_alloc_i_vect(mesh->NMeshPoints);

   ElemType = e_GetElementType(mesh);
   NIpE = e_GetNInterfacesPerElement(mesh);
   NVpE = e_GetNVerticesPerElement(mesh);
   NVpI = e_GetNVerticesPerInterface(mesh);
   Vertices = e_GetVertices(mesh);

   for(k = 0; k < mesh->NMeshPoints; k++)
      (*NodeTypes)[k] = E_RM_NUM_INTERIOR;

   for(i = 0; i < mesh->TotalNElems; i++)
   {
      for(k = 0; k < NIpE; k++)
      {
         if(AdjElems[i][k] == 0)
         {
            for(j = 0; j < NVpI; j++)
            {
               l = Vertices[i][(k + j) % NVpE] - 1;
               (*NodeTypes)[l] = E_RM_NUM_EXTERIOR;
            }
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SmoothMesh: To smooth the mesh-point coordinates in a mesh by moving the
   position of all interior mesh-points so that they lie at the mean position
   of their adjacent mesh-points.                                             */
static void SmoothMesh(MESH *mesh, int *NAdjMeshPoints, int **AdjMeshPoints,
                       int *NodeTypes)
{
   int j,k,l;

   double xSum,ySum,zSum;


   for(k = 0; k < mesh->NMeshPoints; k++)
   {
      if(NodeTypes[k] == E_RM_NUM_INTERIOR)
      {
         xSum = ySum = zSum = 0.0;

         for(j = 0; j < NAdjMeshPoints[k]; j++)
         {
            l = AdjMeshPoints[k][j] - 1;
            xSum += mesh->MeshPointCoords[l][0];
            ySum += mesh->MeshPointCoords[l][1];
            zSum += mesh->MeshPointCoords[l][2];
         }

         if(NAdjMeshPoints[k] > 0)
         {
            mesh->MeshPointCoords[k][0] = xSum/NAdjMeshPoints[k];
            mesh->MeshPointCoords[k][1] = ySum/NAdjMeshPoints[k];
            mesh->MeshPointCoords[k][2] = zSum/NAdjMeshPoints[k];
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* ExchangeDiagonals: To perform diagonal exchange on a TRIANG mesh.          */
static void ExchangeDiagonals(MESH *mesh, int **AdjElems,
                              int *NAdjMeshPoints, int **AdjMeshPoints)
{
   int i,j;
   int ElemType;
   int **Nodes;


   ElemType = e_GetElementType(mesh);
   Nodes = e_GetNodes(mesh);

   for(i = 0; i < mesh->TotalNElems; i++)
   {
      for(j = 0; j < E_NIPE_TRIANG1; j++)
      {
         if(ShouldExchangeEdge(mesh,Nodes,AdjElems,i,j))
         {
            ExchangeEdge(ElemType,Nodes,AdjElems,
                         NAdjMeshPoints,AdjMeshPoints,i,j);
         }
      }
   }

}


/* ////////////////////////////////////////////////////////////////////////// */
/* ShouldExchangeEdge: To return whether diagonal exchange should occur
   between two adjacent elements with indices "i + 1" and "i_adj + 1". The
   elements have mesh-point indices "k1 + 1","k2 + 1","k3 + 1" and "k2 + 1",
   "k4 + 1","k3 + 1" and a shared interface with indices "j + 1" and
   "j_adj + 1".                                                               */
static int ShouldExchangeEdge(MESH *mesh, int **Nodes, int **AdjElems, 
                              int i, int j)
{
   int k1,k2,k3,k4;
   int i_adj,j_adj;
   int out;
   double mu1, mu2;


   if (AdjElems[i][j] == 0)
      return FALSE;

   k1 = Nodes[i][(j + 2) % E_NVPE_TRIANG1] - 1;
   k2 = Nodes[i][j] - 1;
   k3 = Nodes[i][(j + 1) % E_NVPE_TRIANG1] - 1;

   i_adj = AdjElems[i][j] - 1;
   j_adj = 0;

   while (   Nodes[i_adj][j_adj] == k2 + 1
          || Nodes[i_adj][j_adj] == k3 + 1)
   {
      j_adj++;
   }

   k4 = Nodes[i_adj][j_adj] - 1;

   out = eg_LineLineIntersect2D(mesh->MeshPointCoords[k1],
                                mesh->MeshPointCoords[k4],
                                mesh->MeshPointCoords[k2],
                                mesh->MeshPointCoords[k3],
                                &mu1, &mu2);
   
   /* if the diagonals intersects and 
      the maximum angle of the new formation is larger than the previous one */
   if(   (   (out == EG_ONVERTEX)
          || (   (out == EG_INTERSECT)
              && (mu1 >= 0.0 && mu1 <= 1.0)
              && (mu2 >= 0.0 && mu2 <= 1.0) ) )
       &&   E_MAX(GetMaxCosine(mesh,k1,k2,k3), GetMaxCosine(mesh,k2,k4,k3))
          > E_MAX(GetMaxCosine(mesh,k1,k4,k3), GetMaxCosine(mesh,k1,k2,k4))    )
   {
      return TRUE;
   }
   else
   {
      return FALSE;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetMaxCosine: To return the maximum cosine of the internal angles of a
   triangular element.                                                        */
static double GetMaxCosine(MESH *mesh, int k1, int k2, int k3)
{
   int j;

   double a,b,c;
   double aSquared,bSquared,cSquared;
   double CosA,CosB,CosC;
   double Delta1[E_NDIMS],Delta2[E_NDIMS],Delta3[E_NDIMS];


   for(j = 0; j < E_NDIMS; j++)
   {
      Delta1[j] = mesh->MeshPointCoords[k1][j] - mesh->MeshPointCoords[k2][j];
      Delta2[j] = mesh->MeshPointCoords[k2][j] - mesh->MeshPointCoords[k3][j];
      Delta3[j] = mesh->MeshPointCoords[k3][j] - mesh->MeshPointCoords[k1][j];
   }

   aSquared = e_VectDotProd(E_NDIMS,Delta1,Delta1);
   bSquared = e_VectDotProd(E_NDIMS,Delta2,Delta2);
   cSquared = e_VectDotProd(E_NDIMS,Delta3,Delta3);

   a = sqrt(aSquared);
   b = sqrt(bSquared);
   c = sqrt(cSquared);

   CosA = (bSquared + cSquared - aSquared)/(2.0*b*c);
   CosB = (cSquared + aSquared - bSquared)/(2.0*c*a);
   CosC = (aSquared + bSquared - cSquared)/(2.0*a*b);

   return E_MAX3(CosA,CosB,CosC);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* ExchangeEdge: To exchange the diagonal edge. If the mesh is of type
   TRIANG1, then consider the two adjacent elements ABD and BCD with
   interfaces AB, BD... etc where A,B,C,D are the vertex nodes. 
   The first element has
   index "i + 1", the second has index "i_adj + 1" and their shared interface
   has indices "j + 1" and "j_adj + 1" respectively.

   NOTE: The element adjacency and mesh-point adjacency matrices are updated
         as appropriate.                                                      */

static void ExchangeEdge(int ElemType, int **Nodes, int **AdjElems,
                         int *NAdjMeshPoints, int **AdjMeshPoints,
                         int i, int j)
{
   int l;
   int i_adj,j_adj;
   int j_minus;
   int j_minus_adj;
   int i_AB_adj,i_CD_adj;
   int j_AB_adj,j_CD_adj;
   int kA,kB,kC,kD;
   int NodeA,NodeB,NodeC,NodeD;

   i_adj = AdjElems[i][j] - 1;
   j_adj = GetAdjInterfIndex(E_NIPE_TRIANG1,i + 1,AdjElems[i_adj]) - 1;
   j_minus = (j + 2) % E_NVPE_TRIANG1;
   j_minus_adj = (j_adj + 2) % E_NVPE_TRIANG1;

   NodeA = Nodes[i][j_minus];
   NodeB = Nodes[i][j];
   NodeC = Nodes[i_adj][j_minus_adj];
   NodeD = Nodes[i_adj][j_adj];

   Nodes[i][j] = NodeC;
   Nodes[i_adj][j_adj] = NodeA;

   i_AB_adj = AdjElems[i][j_minus] - 1;
   i_CD_adj = AdjElems[i_adj][j_minus_adj] - 1;

   if (i_AB_adj >= 0)
   {
      j_AB_adj = GetAdjInterfIndex(E_NIPE_TRIANG1,i + 1,
                                   AdjElems[i_AB_adj]   ) - 1;
      AdjElems[i_AB_adj][j_AB_adj] = i_adj + 1;
   }

   if (i_CD_adj >= 0)
   {
      j_CD_adj = GetAdjInterfIndex(E_NIPE_TRIANG1,i_adj + 1,
                                   AdjElems[i_CD_adj]       ) - 1;
      AdjElems[i_CD_adj][j_CD_adj] = i + 1;
   }

   AdjElems[i][j_minus] = i_adj + 1;
   AdjElems[i][j] = i_CD_adj + 1;
   AdjElems[i_adj][j_minus_adj] = i + 1;
   AdjElems[i_adj][j_adj] = i_AB_adj + 1;

   kA = NodeA - 1;
   kB = NodeB - 1;
   kC = NodeC - 1;
   kD = NodeD - 1;

   AdjMeshPoints[kA] = e_incr_i_vect(NAdjMeshPoints[kA],AdjMeshPoints[kA],1);
   AdjMeshPoints[kC] = e_incr_i_vect(NAdjMeshPoints[kC],AdjMeshPoints[kC],1);

   AdjMeshPoints[kA][NAdjMeshPoints[kA]++] = NodeC;
   AdjMeshPoints[kC][NAdjMeshPoints[kC]++] = NodeA;

   l = 0;

   while (AdjMeshPoints[kB][l] != NodeD)
      l++;

   NAdjMeshPoints[kB]--;

   for (  ; l < NAdjMeshPoints[kB]; l++)
      AdjMeshPoints[kB][l] = AdjMeshPoints[kB][l + 1];

   l = 0;

   while (AdjMeshPoints[kD][l] != NodeB)
      l++;

   NAdjMeshPoints[kD]--;

   for (  ; l < NAdjMeshPoints[kD]; l++)
      AdjMeshPoints[kD][l] = AdjMeshPoints[kD][l + 1];

}

/* ////////////////////////////////////////////////////////////////////////// */
/* GetAdjInterfIndex: To return the index of an interface as defined by an
   adjacent element.                                                          */
/* NOTE: This function is valid for 3-D elements. */

static int GetAdjInterfIndex(int NIpE, int ElemIndex, int *AdjElems)
{
   int j;


   for (j = 0; j < NIpE; j++)
   {
      if (AdjElems[j] == ElemIndex)
         return j + 1;
   }

   return 0;   /* dummy return */
}


/* //////////////////////////////////////////////////////////////////////////
Function:	    RemoveTriangTriplets
Descriptions:	This is a mesh enhancement called point refinement. The basic
                idea is removing three triangles if they share one single node.
Arguments:
              mesh: the mesh to enhance
Return:       number of eliminated nodes or 0
///////////////////////////////////////////////////////////////////////////// */
int RemoveTriangTriplets(MESH *mesh)
{
  int i,j,k,l,m;
  int *NAdjNodeElems;
  int **AdjNodeElems;
  int NElimNodes;
  int *ElimNodes;
  int NElimElems;
  int *ElimElems;
  int *NewNodeNums;
  int NMeshPoints;
  int NElemsTriang1;
  int **NodesTriang1;

  if(mesh->NElemsTriang1 != mesh->TotalNElems)
  {
    return 0;
  }

  NAdjNodeElems = e_alloc_i_vect(mesh->NMeshPoints);
  AdjNodeElems = e_init_frag_i_matr(mesh->NMeshPoints);

  /* first construct the node-element adjacent list. */
  for(i = 0; i < mesh->NElemsTriang1; i++)
  {
    for(j = 0; j < E_NVPE_TRIANG1; j++)
    {
      k = mesh->NodesTriang1[i][j] - 1;
      AdjNodeElems[k]
        = e_incr_i_vect(NAdjNodeElems[k],AdjNodeElems[k],
                        ADJ_NODE_ELEMS_INCREM);
      AdjNodeElems[k][NAdjNodeElems[k]++] = i + 1;
    }
  }

  NElimNodes = 0;
  NElimElems = 0;
  ElimNodes = NULL;
  ElimElems = NULL;

  /* for each element node, find the nodes which have three triangles attached. */
  for(k = 0; k < mesh->NMeshPoints; k++)
  {
    if(NodeElemsFormATriplet(mesh, NAdjNodeElems[k], AdjNodeElems[k]))
    {
      ElimNodes = e_incr_i_vect(NElimNodes, ElimNodes, 1);
      ElimElems = e_incr_i_vect(NElimElems, ElimElems, E_NVPE_TRIANG1);

      ElimNodes[NElimNodes++] = k + 1;

      for(j = 0; j < E_NVPE_TRIANG1; j++)
        ElimElems[NElimElems++] = AdjNodeElems[k][j];
    }
  }

  /* restruct the mesh after deleting one node and three triangles. */
  NMeshPoints = mesh->NMeshPoints - NElimNodes;
  NElemsTriang1 = mesh->NElemsTriang1 - 2*NElimElems/3;

  ElimNodes = e_incr_i_vect(NElimNodes,ElimNodes,1);
  ElimElems = e_incr_i_vect(NElimElems,ElimElems,1);
  NodesTriang1 = e_alloc_cont_i_matr(NElemsTriang1, E_NNPE_TRIANG1);
  NewNodeNums = e_alloc_i_vect(mesh->NMeshPoints);

  qsort(ElimElems, NElimElems, sizeof(int), compare);

  l = m = 0;

  /* remove the nodes */
  for(k = 0; k < mesh->NMeshPoints; k++)
  {
    if(k + 1 == ElimNodes[l])
    {
      l++;
    }
    else
    {
      NewNodeNums[k] = m + 1;
      /*** --jmch-- changed the memcopy into a for loop ***/
      /* memcpy(mesh->MeshPointCoords[m++],mesh->MeshPointCoords[k],
                                                 E_NDIMS*sizeof(double)); */
      for(i = 0; i < E_NDIMS; i++)
        mesh->MeshPointCoords[m][i]=mesh->MeshPointCoords[k][i];
      m++;
    }
  }

  /* remove the triangles */
  NElemsTriang1 = 0;
  l = 0;
  for(i = 0; i < mesh->NElemsTriang1; i++)
  {
    if(i + 1 == ElimElems[l])
    {
      l++;
    }
    else
    {
      for(j = 0; j < E_NVPE_TRIANG1; j++)
      {
        k = mesh->NodesTriang1[i][j] - 1;
        NodesTriang1[NElemsTriang1][j] = NewNodeNums[k];
      }

      NElemsTriang1++;
    }
  }

  /* replace deleted three triangles with a new triangle. */
  for(k = 0; k < NElimNodes; k++)
  {
    CreateNewElem(mesh, &NElemsTriang1, NodesTriang1, ElimNodes[k],
                  NewNodeNums, AdjNodeElems);
  }

  e_free_cont_i_matr(mesh->NElemsTriang1, mesh->NodesTriang1);

  /* final adjustments to the mesh */
  mesh->NMeshPoints = NMeshPoints;
  mesh->NNodes = NMeshPoints;
  mesh->NElemsTriang1 = NElemsTriang1;
  mesh->TotalNElems = NElemsTriang1;
  mesh->NodesTriang1 = NodesTriang1;
  
  mesh->TotalNodalDOF = NMeshPoints * E_VERTEX_NODE_DOF_TRIANG1;
  e_free(mesh->NodalDOF);
  e_free(mesh->GNIs);
  mesh->NodalDOF = e_alloc_i_vect(mesh->TotalNodalDOF);
  mesh->GNIs = e_alloc_i_vect(mesh->NMeshPoints);
  for(i = 0; i < mesh->NMeshPoints; i++)
  {
    mesh->NodalDOF[i] = E_VERTEX_NODE_DOF_TRIANG1;
    mesh->GNIs[i] = i+1;
  }
  e_free(mesh->GEIs);
  mesh->GEIs = e_alloc_i_vect(mesh->TotalNElems);
  for(i = 0; i < mesh->TotalNElems; i++)
  {
    mesh->GEIs[i] = i+1;
  }
  mesh->GEIsTriang1 = mesh->GEIs;

  e_free(NAdjNodeElems);
  e_free_frag_i_matr(mesh->NMeshPoints, AdjNodeElems);
  e_free(ElimNodes);
  e_free(ElimElems);
  e_free(NewNodeNums);

  return(NElimNodes);
}


/* ////////////////////////////////////////////////////////////////////////// 
Function:     NodeElemsFormATriplet
Descriptions: determine if the node should be deleted. It contains two stages:
              1). check the number of triangles attached to this node is three,
                  if not the case return false. else go to the second stage:
              2). check if the attached triangles form a closed domain. if it 
                  is the case then return true, else there must be at least one 
                  boundary triangle, return false.
Arguments:    mesh: the mesh
              NAdjNodeElems: the number of triangles attached to this node.
              AdjNodeElems: the node-element adjacency list.
Return:       return true if the node should be deleted else return false.
//////////////////////////////////////////////////////////////////////////// */
static int NodeElemsFormATriplet(MESH *mesh, 
                                 int NAdjNodeElems, int *AdjNodeElems)
{
  int i,j,l,r;
  int NNodes;
  int Nodes[3*E_NVPE_TRIANG1];


  /* if the number of triangles attached to this node not equal to three, 
     return false. */
  if(NAdjNodeElems != E_NVPE_TRIANG1)
    return FALSE;

  NNodes = 0;

  /* a simple test to detect if the attached three triangles form a 
       closed domain. if so, then return true, else there must be at 
       least one boundary element, return false */
  for(l = 0; l < NAdjNodeElems; l++)
  {
    i = AdjNodeElems[l] - 1;

    for(j = 0; j < E_NVPE_TRIANG1; j++)
    {
      for(r = 0; r < NNodes; r++)
      {
        if(mesh->NodesTriang1[i][j] == Nodes[r])
          break;
      }

      if(r == NNodes)
        Nodes[NNodes++] = mesh->NodesTriang1[i][j];
    }
  }

  if (NNodes == E_NVPE_TRIANG1 + 1)
    return TRUE;
  else
    return FALSE;
}


/* ////////////////////////////////////////////////////////////////////////// 
Function:	    compare
Descriptions: compare two integers.
Arguments:
              elem1: the first integer.
              elem2: the second integer.
return:       -1 if first integer less than the second integer, 0 if they are
				      the same else 1 if the first integer bigger than the second one.
///////////////////////////////////////////////////////////////////////////// */
static int compare(const void *elem1, const void *elem2)
{
   if (*(int *)elem1 < *(int *)elem2)
      return -1;
   else if (*(int *)elem1 == *(int *)elem2)
      return 0;
   else
      return 1;
}


/* ////////////////////////////////////////////////////////////////////////// 
Function:	    CreateNewElem
Descriptions: If there is a node deleted from the mesh. With it there 
              would be three triangles need to be deleted. After that
              the region which is occupied by the three triangles whould 
              be occupy by a new triangle. This function will replace
              three triangles with the new triangle.
Arguments:		mesh: the mesh.
              NElemsTriang1: the generated element ID.
              NodesTriang1: the array holds nodal IDs for each element 
                            in the mesh.
              ElimNode: the node ID for elimination. 
              NewNodeNums: the new traigle corner node IDs.
              AdjNodeElems: the node-element adjacency list for 
                            eliminating node.
Return:       None.
/////////////////////////////////////////////////////////////////////////// */
static void CreateNewElem(MESH *mesh, int *NElemsTriang1, int **NodesTriang1,
                          int ElimNode, int *NewNodeNums, int **AdjNodeElems )
{
  int i,j,k;
  int i1,i2;


  k = ElimNode - 1; /* the eliminating node index in the array. */
  i = (*NElemsTriang1)++; /* increase the number of  elements by one. */
  i1 = AdjNodeElems[k][0] - 1; /* the first element attached to the node. */
  i2 = AdjNodeElems[k][1] - 1; /* the second element attached to the node. */

  for(j = 0; j < E_NVPE_TRIANG1; j++)
  {
    if(mesh->NodesTriang1[i1][j] == ElimNode)
    {
      k = mesh->NodesTriang1[i1][(j + 1) % E_NVPE_TRIANG1] - 1;
      NodesTriang1[i][0] = NewNodeNums[k];
      k = mesh->NodesTriang1[i1][(j + 2) % E_NVPE_TRIANG1] - 1;
      NodesTriang1[i][1] = NewNodeNums[k];

      for(j = 0; j < E_NVPE_TRIANG1; j++)
      {
        if(   mesh->NodesTriang1[i2][j] == mesh->NodesTriang1[i1][0]
           || mesh->NodesTriang1[i2][j] == mesh->NodesTriang1[i1][1]
           || mesh->NodesTriang1[i2][j] == mesh->NodesTriang1[i1][2])
        {
          continue;
        }

        k = mesh->NodesTriang1[i2][j] - 1;
        NodesTriang1[i][2] = NewNodeNums[k];

        return;
      }
    }
  }
}









