#include "mgni.h"
#include "boundary.h"
#include "nodelem.h"
#include "adjside.h"
#include "getvertices.h"
#include "mgn_graphics.h"
#include "split.h"
#include "eg_lib.h"

extern int ShowGraphics;


void split2triang(MESH *mesh)
{
  /* split a quad1 mesh back to triangles using the shortest diagonal */
  int q,t;
  double l1,l2;
  
  /* checking */
  if (mesh->NElemsQuad1!=mesh->TotalNElems)
    e_GeneralError("split2triang: requires QUAD1 mesh for input");
    
  mesh->NElemsTriang1=2*mesh->NElemsQuad1;
  mesh->NodesTriang1=e_alloc_cont_i_matr(mesh->NElemsTriang1,3);
  
  t = 0;
  for (q=0;q<mesh->NElemsQuad1;q++)
  {
    l1=eg_PointSqrDist3D(
      mesh->MeshPointCoords[mesh->NodesQuad1[q][0]-1],
      mesh->MeshPointCoords[mesh->NodesQuad1[q][2]-1] );
    l2=eg_PointSqrDist3D(
      mesh->MeshPointCoords[mesh->NodesQuad1[q][1]-1],
      mesh->MeshPointCoords[mesh->NodesQuad1[q][3]-1] );
    if (l1<l2)
    {
      mesh->NodesTriang1[t][0]=mesh->NodesQuad1[q][0];
      mesh->NodesTriang1[t][1]=mesh->NodesQuad1[q][1];
      mesh->NodesTriang1[t][2]=mesh->NodesQuad1[q][2];
      t++;
      mesh->NodesTriang1[t][0]=mesh->NodesQuad1[q][0];
      mesh->NodesTriang1[t][1]=mesh->NodesQuad1[q][2];
      mesh->NodesTriang1[t][2]=mesh->NodesQuad1[q][3];
      t++;
    }
    else
    {
      mesh->NodesTriang1[t][0]=mesh->NodesQuad1[q][0];
      mesh->NodesTriang1[t][1]=mesh->NodesQuad1[q][1];
      mesh->NodesTriang1[t][2]=mesh->NodesQuad1[q][3];
      t++;
      mesh->NodesTriang1[t][0]=mesh->NodesQuad1[q][1];
      mesh->NodesTriang1[t][1]=mesh->NodesQuad1[q][2];
      mesh->NodesTriang1[t][2]=mesh->NodesQuad1[q][3];
      t++;
    }  
  }
  
  e_free_cont_i_matr(mesh->NElemsQuad1, mesh->NodesQuad1);
  e_free(mesh->GEIs);
  
  mesh->NodesQuad1 = NULL;
  mesh->NElemsQuad1 = 0;
  mesh->TotalNElems = mesh->NElemsTriang1;
  mesh->GEIs = e_alloc_i_vect(mesh->TotalNElems);
  mesh->GEIsTriang1 = mesh->GEIs;
  
  for (t=0;t<mesh->NElemsTriang1;t++)
    mesh->GEIs[t]=t+1;
  
  mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1] = FALSE;
  mesh->KeywordSet[E_KW_NUM_NODES_TRIANG1] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NODES_QUAD1] = FALSE;

  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] = 1;
  mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1]   = 0;
}



void split2quad(MESH *mesh)
{
  int nb;				/* number of boundaries (should be 1) */
  int **ne;			/* req. for node - element connectivity */
  int *nne;			/* req. for node - element connectivity */
  BND *bnd;			/* boundary of the mesh */
  int **adj;		/* adjacent elements */
  int **adjs;		/* sides along which elems are adjacent */
  int *v;				/* vertices of an element */
  int nv;				/* number of vertices for that element */

  int nss;			/* number of shared sides */
  int nbs;			/* number of boundary sides */
  int nev;			/* number of elements * no of sides per elem */

  double **crds;	/* new coords */
  int nc;					/* new number of points */
  int **quads;		/* new quads */
  int nq;					/* new number of quads */
  int cn,sn[4];		/* center and side nodes of split element */

  int b;				/* boundary counter */
  int e;				/* element counter */
  int s;				/* side counter */
  int i;				/* dimension counter */
  int q;				/* quad counter */
  int p;				/* point counter */
  
  
  /* get adjacency */
  nne=e_alloc_i_vect(mesh->NMeshPoints);
  ne=e_init_frag_i_matr(mesh->NMeshPoints);
  e_InitMeshPointElemConnection(mesh, nne, ne);
  e_SetAdjacencyAndSides(mesh,nne,ne,&adj,&adjs);

  /* get boundary */
  Gen2DBoundary(mesh, nne, ne, &nb, &bnd);
  if (nb!=1)
      e_GeneralError("split2quads: this mesh generator only accepts single domain meshes");
	
  /* counting */
  nbs=bnd->oe.nen;
  for (b=0;b<bnd->nie;b++)
      nbs+=bnd->ie[b].nen;
  nev=(mesh->NElemsTriang1+mesh->NElemsTriang3+
       mesh->NElemsTriang4+mesh->NElemsTriang5)*3
      +mesh->NElemsQuad1*4;
  nss=(nev+nbs)/2-nbs;

  /* allocating */
  nc=mesh->NMeshPoints;             /* all the existing points */
  nc+=nss;                          /* one extra point for each shared side */
  nc+=mesh->TotalNElems;            /* one extra point for each element */
  nc+=nbs;                          /* one extra point for each boundary side */
  crds=e_alloc_cont_d_matr(nc,3);
  nq=nev;                           /* the number of quads to be made */
  quads=e_alloc_cont_i_matr(nq,4);

  /* the existing nodes are copied in first */
  memcpy((void *)(crds[0]),
         (void *)(mesh->MeshPointCoords[0]),
         3 * mesh->NMeshPoints * sizeof(double));

  /* initialize counters */
  q = 0;
  p = mesh->NMeshPoints;

  /* start the loop */
  for (e = 0;e < mesh->TotalNElems; e++)
  {
    /* get the vertices for element e */
    e_GetElementVertices(e, mesh, &nv, &v);
    /* make centroid */
    for(i = 0; i < 3; i++)
    {
      crds[p][i] = 0.0;
    }
    
    for(s = 0; s < nv; s++)
    {
      for (i = 0; i < 3; i++)
      {
        crds[p][i] += mesh->MeshPointCoords[v[s]-1][i];
      }
    }
    
    for(i = 0; i < 3; i++)
    {
      crds[p][i] /= (double)nv;
    }
    
    cn=p;
    p++;
    /* make side nodes */
    for(s = 0; s < nv; s++)
    {
      /* see if sidenode has already been made */
      if ((adj[e][s]-1 < e) && (adj[e][s] != 0))
      {
        sn[s] = adjs[adj[e][s]-1][adjs[e][s]-1];
      }
      else
      {
        for(i = 0; i < 3; i++)
        {
          crds[p][i] = (mesh->MeshPointCoords[v[s]-1][i]+
                        mesh->MeshPointCoords[v[(s+1)%nv]-1][i])/2.0;
        }
        
        sn[s] = p;
        /* make node number available for neighbour */
        adjs[e][s]=p;
        p++;
      }
    }
    
    /* make quads */
    for(s = 0; s < nv; s++)
    {
      quads[q][0]=cn+1;
      quads[q][1]=sn[(s+nv-1)%nv]+1;
      quads[q][2]=v[s];
      quads[q][3]=sn[s]+1;
      q++;
    }


    if (ShowGraphics)
        mgnDrawAll(p, crds, 0,NULL, q, quads, -1, NULL, NULL,
                   "halfing to quads");

  }

  /* free-up */
  e_free_frag_i_matr(mesh->NMeshPoints, ne);
  e_free_frag_i_matr(mesh->TotalNElems, adj);
  e_free_frag_i_matr(mesh->TotalNElems, adjs);
  e_free(nne);

  /* and now build the quad mesh */
  /* points */
  e_free_cont_d_matr(mesh->NMeshPoints, mesh->MeshPointCoords);
  mesh->NMeshPoints = mesh->NNodes = nc;
  mesh->MeshPointCoords = crds;

  e_free(mesh->NodalDOF);
  mesh->TotalNodalDOF = 0;
  mesh->GNIs     = e_alloc_i_vect(mesh->NMeshPoints);
  mesh->NodalDOF = e_alloc_i_vect(mesh->NMeshPoints);
  for(i = 0; i < mesh->NMeshPoints; i++)
  {
    mesh->GNIs[i] = i+1;
    mesh->NodalDOF[i] = E_VERTEX_NODE_DOF_QUAD1;
  }

  /* elements */
  if (mesh->NElemsQuad1)
  {
    e_free_cont_i_matr(mesh->NElemsQuad1,mesh->NodesQuad1);
  }
  mesh->TotalNElems = mesh->NElemsQuad1=nq;
  mesh->NodesQuad1  = quads;
  
  if (mesh->NElemsTriang1)
      e_free_cont_i_matr(mesh->NElemsTriang1, mesh->NodesTriang1);
  if (mesh->NElemsTriang3)
      e_free_cont_i_matr(mesh->NElemsTriang3, mesh->NodesTriang3);
  if (mesh->NElemsTriang4)
      e_free_cont_i_matr(mesh->NElemsTriang4, mesh->NodesTriang4);
  if (mesh->NElemsTriang5)
      e_free_cont_i_matr(mesh->NElemsTriang5, mesh->NodesTriang5);

  mesh->NElemsTriang1 = 0;
  mesh->NElemsTriang3 = 0;
  mesh->NElemsTriang4 = 0;
  mesh->NElemsTriang5 = 0;
  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] = 0;
  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] = 0;
  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] = 0;
  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] = 0;
  mesh->FirstNetIndex[E_ELEM_TYPE_QUAD1]   = 1;
  mesh->GEIs = e_alloc_i_vect(mesh->NElemsQuad1);
  mesh->GEIsQuad1 = mesh->GEIs;
  for (q=0;q<mesh->NElemsQuad1;q++)
  {
    mesh->GEIs[q] = q + 1;
  }

  mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1] = FALSE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3] = FALSE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4] = FALSE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5] = FALSE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1]   = TRUE;
  mesh->KeywordSet[E_KW_NUM_NODES_TRIANG1]  = FALSE;
  mesh->KeywordSet[E_KW_NUM_NODES_TRIANG3]  = FALSE;
  mesh->KeywordSet[E_KW_NUM_NODES_TRIANG4]  = FALSE;
  mesh->KeywordSet[E_KW_NUM_NODES_TRIANG5]  = FALSE;
  mesh->KeywordSet[E_KW_NUM_NODES_QUAD1]    = TRUE;

  /* free-up */
  Free2DBoundary(nb,bnd);
}



