#include "e_lib.h"
#include "eg_lib.h"
#include "mgn.h"
#include "dtr.h"
#include "bgdata.h"


/* static auxiliary functions */
static int str8refineseg(double **coords, int fn, int sn, RFS * refset,
                         int ctrltype, BG_DATA *bgd, double *mpr, double h, int mode);
static int nurbsrefseg(double **coords, int fn, int sn, RFS * refset,
                       NURBSCURVE * curve, double tfn, double tsn,
                       int ctrltype, BG_DATA *bgd, double *mpr, double h, int mode);
static void resolvempr(double **coords, int n, int ctrltype,
                       BG_DATA *bgd, double *mpr, double h, double *nmpr);

int make_boundary_nodes(double ***coords, int *np, int *tp, int *stnp,
                        int ctrltype, int geomset, MESH *mesh,
                        BG_DATA *bgd, double **mpr, double h, BND *bnd,
                        int ***tpnl, int force_even)
{
  /* this function refines the boundaries:
   * - it generates the new nodes along the edges
   * - it pastes the nodes in the coords array starting at stnp
   * - it enlarges the tpnl
   * - it updates the boundary info bnd
   *
   * the refining is controlled by
   * - the geometry defined in gm
   * - background mesh if (ctrltype==CTRL_BACKGROUND)
   * - bgmesh and sources if (ctrltype==CTRL_SOURCES)
   * - nodal params (...)
   *
   * it returns the number of new nodes generated
   */

  int nnp;                      /* number of new points generated */
  int fn;                       /* first node of the segment */
  int sn;                       /* second node of the segment */
  RFS *refset;                  /* the set of refined nodes to be inserted */
  int b, e;                     /* bnd counters */
  EDGE dummy;                   /* dummy original edge */
  int sb;                       /* index of node in updated bnd */
  int oldtp;                    /* keep tp before it is updated */
  int reversed;                 /* is the refset reversed */
  int i;                        /* counter */
  int bmax;											/* index of largest outer boundary segment */
  double l,lmax;								/* lengths of largest outer boundary segment */
  double lmin;									/* lengths of the shortest boundary segment */
  int total;										/* total of nodes on the boundary */

  (*stnp) = (*np);
  total=0;
  /* initialize the set of refinement nodes (refset) */
  refset = NULL;
  refset = (RFS *) e_calloc(1, sizeof(RFS));
  refset->np = 0;
  refset->tap = INCR_SIZE;
  refset->cr = NULL;
  refset->cr = e_alloc_cont_d_matr(refset->tap, 4);

  /* if the resulting refined boundary has to be even numbered
     some extra bookkeeping is required */
  if (force_even)
  {
    /* find the longest segment (of outer boundary) */
    lmax=0.0;
    bmax=-1;
    for (b = 0; b < bnd->oe.nen; b++)
    {
      l=eg_PointDist3D((*coords)[bnd->oe.en[b]],
                       (*coords)[bnd->oe.en[(b + 1) % bnd->oe.nen]]);
      if (l>lmax)
      {
        lmax=l;
        bmax=b;
      }
    }
    /* rotate outer boundary round so largest segment is last */
    dummy.en=e_alloc_i_vect(bnd->oe.nen);
    for (b=0;b<bnd->oe.nen-bmax-1;b++)
      dummy.en[b]=bnd->oe.en[bmax+b+1];
    for (b=0;b<bmax+1;b++)
      dummy.en[bnd->oe.nen-bmax-1+b]=bnd->oe.en[b];
    for (b=0;b<bnd->oe.nen;b++)
      bnd->oe.en[b]=dummy.en[b];
    e_free(dummy.en);
  }

  /* first refine hole boundaries */
  for (e = 0; e < bnd->nie; e++)
  {
    total+=bnd->ie[e].nen;
    /* make a dummy copy of an inner edge in the boundary */
    dummy.en = e_alloc_i_vect(bnd->ie[e].nen);
    for (i = 0; i < bnd->ie[e].nen; i++)
      dummy.en[i] = bnd->ie[e].en[i];
    dummy.nen = bnd->ie[e].nen;
    sb = 0;

    /* and refine each hole segment */
    for (b = 0; b < dummy.nen; b++)
    {
      fn = dummy.en[b];
      sn = dummy.en[(b + 1) % dummy.nen];
      refset->np = 0;
      reversed =
        refinesegment((*coords), fn, sn, refset, ctrltype, geomset, mesh,
                      bgd, (*mpr), h, E_SEGREF_NONE);
      updatebnd(&(bnd->ie[e]), sb, refset->np, (*np));
      sb += refset->np + 1;
      total += refset->np;
      oldtp = (*tp);
      addrefsettocoords(coords, np, tp, refset, tpnl, reversed);
      if (ctrltype == CTRL_MPR)
      {
        updatempr(mpr, (*np), &oldtp, refset, reversed);
        if (oldtp != (*tp))
          e_GeneralError("unexpected error in make_boundary_nodes");
      }
      (*np) += refset->np;
    }
    e_free(dummy.en);
  }

  total+=bnd->oe.nen;
  /* make a dummy copy of the outer edge in the boundary */
  dummy.nen = bnd->oe.nen;
  dummy.en = e_alloc_i_vect(bnd->oe.nen);
  for (i = 0; i < bnd->oe.nen; i++)
    dummy.en[i] = bnd->oe.en[i];
  sb = 0;

  /* and refine each outer segment */
  for (b = 0; b < dummy.nen; b++)
  {
    fn = dummy.en[b];
    sn = dummy.en[(b + 1) % dummy.nen];
    refset->np = 0;
    if ((force_even)&&(b==dummy.nen-1))
    {
      if (total%2)
        reversed = refinesegment((*coords), fn, sn, refset, ctrltype, geomset,
                                 mesh, bgd, *mpr, h, E_SEGREF_ODD);
      else
        reversed = refinesegment((*coords), fn, sn, refset, ctrltype, geomset,
                                 mesh, bgd, *mpr, h, E_SEGREF_EVEN);
    }
    else
    {
      reversed = refinesegment((*coords), fn, sn, refset, ctrltype, geomset,
                               mesh, bgd, *mpr, h, E_SEGREF_NONE);
    }
    updatebnd(&(bnd->oe), sb, refset->np, (*np));
    sb += refset->np + 1;
    total += refset->np;
    oldtp = (*tp);
    addrefsettocoords(coords, np, tp, refset, tpnl, reversed);
    if (ctrltype == CTRL_MPR)
    {
      updatempr(mpr, (*np), &oldtp, refset, reversed);
      if (oldtp != (*tp))
        e_GeneralError("unexpected error in make_boundary_nodes");
    }
    (*np) += refset->np;
  }

	/* statistics of the boundary refinement */
	lmin=1e10;
	lmax=-1.0;
	for (b=0;b<bnd->oe.nen;b++)
	{
		l=eg_PointDist3D((*coords)[bnd->oe.en[b]],
		                 (*coords)[bnd->oe.en[(b + 1) % bnd->oe.nen]]);
		if (l>lmax)
		  lmax=l;
		if (l<lmin)
		  lmin=l;
	}
	for (e = 0; e < bnd->nie; e++)
		for (b=0;b<bnd->ie[e].nen;b++)
		{
			l=eg_PointDist3D((*coords)[bnd->ie[e].en[b]],
			                 (*coords)[bnd->ie[e].en[(b + 1) % bnd->ie[e].nen]]);
			if (l>lmax)
			  lmax=l;
			if (l<lmin)
			  lmin=l;
		}
	if (PrintProgress)
		printf("boundary refinement all bnd segments in [%1.3f,%1.3f]\n\n",
			lmin,lmax);

  /* free up that memory */
  e_free(dummy.en);
  e_free_cont_d_matr(refset->tap, refset->cr);
  e_free(refset);

  nnp = (*np) - (*stnp);
  return nnp;
}

int refinesegment(double **coords, int fn, int sn, RFS * refset, int ctrltype,
                  int geomset, MESH * mesh, BG_DATA *bgd, double *mpr,
                  double h, int mode)
{
  int str8;                     /* is the segment straight? */
  NURBSCURVE *curve;            /* the curve the segment belongs to */
  double tfn;                   /* the curve param for first node fn */
  double tsn;                   /* the curve param for second node sn */

  str8 = FALSE;
  if (!geomset)
    str8 = TRUE;
  else if ((mesh->NNurbsCurves[fn] == 0) || (mesh->NNurbsCurves[sn] == 0))
    str8 = TRUE;
  else if (!e_GetCurve(mesh, fn + 1, sn + 1, &curve, &tfn, &tsn))
    str8 = TRUE;

  if (str8)
    return str8refineseg(coords, fn, sn, refset, ctrltype, bgd, mpr, h, mode);
  else
    return nurbsrefseg(coords, fn, sn, refset, curve, tfn, tsn, ctrltype,
                       bgd, mpr, h, mode);
}

static int str8refineseg(double **coords, int fn, int sn, RFS * refset,
                         int ctrltype, BG_DATA *bgd, double *mpr,
                         double h, int mode)
{
  double fnmpr;                 /* mesh parameter for the first node fn */
  double snmpr;                 /* mesh parameter for the second node sn */
  double **seg_crd;
  double *seg_mpr;
  int nnp, n, s;

  seg_crd = NULL;
  seg_mpr = NULL;

  resolvempr(coords, fn, ctrltype, bgd, mpr, h, &fnmpr);
  resolvempr(coords, sn, ctrltype, bgd, mpr, h, &snmpr);

  nnp = e_SegmentRefine(coords[sn], snmpr, coords[fn], fnmpr,
                         &seg_crd, &seg_mpr, TRUE, mode);

  if (DebugOutput)
    printf("segment refine (%d,%d) %d new node\n", fn, sn, nnp);

  while ((refset->tap - refset->np) < nnp)
  {
    refset->cr = e_incr_cont_d_matr(refset->tap, 4, refset->cr, refset->tap);
    refset->tap += refset->tap;
  }
  for (n = 0; n < nnp; n++)
  {
    for (s = 0; s < 3; s++)
      refset->cr[refset->np][s] = seg_crd[n][s];
    /* the interpolated mpr is stored in cr[3] */
    refset->cr[refset->np][3] = seg_mpr[n];
    refset->np++;
  }
  e_free(seg_mpr);
  e_free_cont_d_matr(nnp, seg_crd);

  return TRUE;
}

static int nurbsrefseg(double **coords, int fn, int sn, RFS * refset,
                       NURBSCURVE * curve, double tfn, double tsn,
                       int ctrltype, BG_DATA *bgd, double *mpr,
                       double h, int mode)
{
  double fnmpr;                 /* mesh parameter for the first node fn */
  double snmpr;                 /* mesh parameter for the second node sn */
  int bgstarttr;                /* optimize the searching in case of a background search */

  double **seg_crd;
  double *seg_mpr;
  double *seg_t;
  int nnp, n, s;

  seg_crd = NULL;
  seg_mpr = NULL;

  bgstarttr = 0;
  resolvempr(coords, fn, ctrltype, bgd, mpr, h, &fnmpr);
  resolvempr(coords, sn, ctrltype, bgd, mpr, h, &snmpr);

  nnp =
    e_CurveRefine(curve, tfn, fnmpr, tsn, snmpr, &seg_crd, &seg_mpr, &seg_t,
                  TRUE, mode);

  while ((refset->tap - refset->np) < nnp)
  {
    refset->cr = e_incr_cont_d_matr(refset->tap, 4, refset->cr, refset->tap);
    refset->tap += refset->tap;
  }
  for (n = 1; n < nnp - 1; n++)
  {
    for (s = 0; s < 3; s++)
      refset->cr[refset->np][s] = seg_crd[n][s];
    /* the interpolated mpr is stored in cr[3] */
    refset->cr[refset->np][3] = seg_mpr[n];
    refset->np++;
  }

  /* and update begin and end_coords of the curve as well */
  for (s = 0; s < 3; s++)
  {
    coords[fn][s] = seg_crd[0][s];
    coords[sn][s] = seg_crd[nnp - 1][s];
  }

  e_free(seg_mpr);
  e_free(seg_t);
  e_free_cont_d_matr(nnp, seg_crd);

  return FALSE;
}

static void resolvempr(double **coords, int n, int ctrltype,
                       BG_DATA *bgd, double *mpr, double h, double *nmpr)
{
  /* wrapper for get_mesh_parameter */

  MTRF mtrf;                    /* unused mapping transformation data */
  mtrf.used = FALSE;

  if (ctrltype == CTRL_MPR)
    (*nmpr) = mpr[n];

  else if ((ctrltype == CTRL_BACKGROUND) || (ctrltype == CTRL_SOURCES))
    (*nmpr) = get_bg_mpr(bgd, coords[n], ctrltype, &mtrf);

  else if (ctrltype == CTRL_UNIFORM)
  	(*nmpr) = h;
}

void updatebnd(EDGE * edge, int node, int nnp, int startnp)
{
  int n;
  int l;

  l = edge->nen;
  for (n = 0; n < nnp; n++)
  {
    edge->en = e_incr_i_vect(l, edge->en, EDGE_INCR);
    l++;
  }
  for (n = edge->nen - 1; n > node; n--)
    edge->en[n + nnp] = edge->en[n];
  for (n = 0; n < nnp; n++)
    edge->en[node + 1 + n] = startnp + n;
  edge->nen += nnp;
}

void addrefsettocoords(double ***coords, int *np, int *tp, RFS *refset,
                       int ***tpnl, int reversed)
{
  int n, s;

  while ((*tp) - (*np) < refset->np)
  {
    (*coords) = e_incr_cont_d_matr((*tp), 3, (*coords), (*tp));
    (*tpnl) = (int **) e_incr_vp_vect((*tp), (void **) (*tpnl), (*tp));
    (*tp) += (*tp);
  }
  for (n = 0; n < refset->np; n++)
  {
    if (reversed)
      for (s = 0; s < 3; s++)
        (*coords)[(*np) + n][s] = refset->cr[refset->np - 1 - n][s];
    else
      for (s = 0; s < 3; s++)
        (*coords)[(*np) + n][s] = refset->cr[n][s];
    /* (*coords)[(*np)+n][2]=0.0; */
    (*tpnl)[(*np) + n] = e_alloc_i_vect(2);
    (*tpnl)[(*np) + n][0] = 0;
    (*tpnl)[(*np) + n][1] = 0;
  }
}

void updatempr(double **mpr, int np, int *tp, RFS *refset, int reversed)
{
  int n;

  /* increase the length of the mpr vector if necessary */
  while ((*tp) - np < refset->np)
  {
    if (((*mpr) = e_incr_d_vect((*tp), (*mpr), (*tp))) == NULL)
      e_GeneralError("memory problem in updatempr");
    (*tp) += (*tp);
  }
  /* copy the interpolated mpr from cr[3] */
  for (n = 0; n < refset->np; n++)
    if (reversed)
      (*mpr)[np + n] = refset->cr[refset->np - 1 - n][3];
    else
      (*mpr)[np + n] = refset->cr[n][3];
}

/* end of file */
