#include "eg_lib.h"
#include "mgn.h"
#include "mgn_graphics.h"
#include "dtr.h"
#include "mgn_aux.h"

#define SMALL_MU 1e-6

int exh_tr_search(double *i, double **coords,
                  int **tr, int ntr, int *ft, int *onside, MTRF *mtrf)
{
  /* does the same as trianglesearch but by exhaustively checking all triangles
   * NOTE: starttr is replaced by ntr
   */

  int ccw[3];
  int valid[3];
  double mu;
  int t,s;

  for (t = 0; t < ntr; t++)
  {
    ccw[0]=IsCCW(coords[tr[t][0]], coords[tr[t][1]], i, mtrf, &valid[0]);
    ccw[1]=IsCCW(coords[tr[t][1]], coords[tr[t][2]], i, mtrf, &valid[1]);
    ccw[2]=IsCCW(coords[tr[t][2]], coords[tr[t][0]], i, mtrf, &valid[2]);

    /* triangle found and point is completely inside */
    if (ccw[0] && ccw[1] && ccw[2])
    {
      (*onside) = 0;
      (*ft) = t;
      return 1;
    }
    /* most common case: point not inside or on edges */
    else if (valid[0] && valid[1] && valid[2])
      continue;

    /* point lies therefore on one of the 3 edges or vertices */
    for (s=0;s<3;s++)
    {
      if (!valid[s])
      {
        ColinearTriangle(coords[tr[t][s]], coords[tr[t][(s+1)%3]], i,
          mtrf, &mu);
        if (E_ABS(mu)<SMALL_MU)
        {
          (*onside) = s;
          (*ft) = t;
          return -1;
        }
        else if (E_ABS(mu-1.0)<SMALL_MU)
        {
          (*onside) = (s+1)%3;
          (*ft) = t;
          return -1;
        }
        else if ((mu>0.0)&&(mu<1.0))
        {
          (*onside) = 3+s;
          (*ft) = t;
          return 1;
        }
      }
    }
  }

  /* point can't be found inside the mesh */
  return 0;
}

int trianglesearch(double *i, double **coords,
                   int **tr, int starttr, int *ft,
                   int *onside, MTRF *mtrf)
{
  /* this function returns the index of the triangle in which
   * the node with coords i lies. The algorithm performs a walk
   * starting at triangle starttr.
   *
   * cfr S.H. Lo 3.4.1
   *
   * to make it 3d compliant a check has to be done first that coords[i]
   * is in the plane of the triangles of tr
   *
   * return values:
   *  +1 triangle found, index stored in *ft
   *     if point on a side, onside=side index, else onside=0
   *   0 no triangle found, point outside
   *  -1 point corresponds with node of a found triangle
   *     triangle index stored in ft, node index stored in onside (0,1,2)
   */

  double *center;               /* centroids of candidate triangles */
  int t;                        /* triangle to be considered */
  int s;                        /* counter */
  int ccw[3];										/* IsCCW return values */
  int valid[3]; 								/* IsCCW return values */
  double mu;										/* ColinearTriangle return value */
  int ret_val;									/* return value of this function */

  center = e_alloc_d_vect(3);
  t = starttr;

  /* start the walking loop */
  while (TRUE)
  {
    ccw[0]=IsCCW(coords[tr[t][0]], coords[tr[t][1]], i, mtrf, &valid[0]);
    ccw[1]=IsCCW(coords[tr[t][1]], coords[tr[t][2]], i, mtrf, &valid[1]);
    ccw[2]=IsCCW(coords[tr[t][2]], coords[tr[t][0]], i, mtrf, &valid[2]);

    /* triangle found and point is completely inside */
    if (ccw[0] && ccw[1] && ccw[2])
    {
      (*onside) = 0;
      ret_val=1;
      break;
    }

    /* check for onside or onvertex */
    for (s=0;s<3;s++)
    {
      if (!valid[s])
      {
        ColinearTriangle(coords[tr[t][s]], coords[tr[t][(s+1)%3]], i,
          mtrf, &mu);
        if (E_ABS(mu)<SMALL_MU)
        {
          (*onside) = s;
          ret_val = -1;
          break;
        }
        else if (E_ABS(mu-1.0)<SMALL_MU)
        {
          (*onside) = (s+1)%3;
          ret_val = -1;
          break;
        }
        else if ((mu>0.0)&&(mu<1.0))
        {
          (*onside) = 3+s;
          ret_val = 1;
          break;
        }
      }
    }
    if (s<3)
     break;

    /* move to next triangle */
    centroid(&center, coords[tr[t][0]], coords[tr[t][1]], coords[tr[t][2]]);

    /* standard case */
    for (s=0;s<3;s++)
    {
      ccw[0]=IsCCW(coords[tr[t][s]],  i, center, mtrf, &valid[0]);
      ccw[1]=IsCCW(coords[tr[t][(s+1)%3]], center, i, mtrf, &valid[1]);
      if (ccw[0] && ccw[1])
      {
        if (tr[t][3+s] != -1)
        {
          t = tr[t][3+s];
          break;
        }
        else
        {
          t = -1;
          ret_val = 0;
          break;
        }
      }
    }
    if (s<3)
    {
      if (t==-1)
        break;
      else
        continue;
    }

    /* extreme case where center,corner,i are colinear in that order */
    for (s=0;s<3;s++)
    {
      ccw[0]=IsCCW(coords[tr[t][s]],  i, center, mtrf, &valid[0]);
      if (!valid[0])
      {
        if (tr[t][3+s] != -1)
        {
          t = tr[t][3+s];
          break;
        }
        else if (tr[t][3+(s+2)%3] != -1)
        {
          t = tr[t][3+(s+2)%3];
          break;
        }
        else
        {
          t = -1;
          ret_val = 0;
          break;
        }
      }
    }
    if (s<3)
    {
      if (t==-1)
        break;
      else
        continue;
    }

    e_GeneralError("trianglesearch: problem in triangle search\n\n");
  }

  /* output */
  e_free(center);
  (*ft) = t;
  return ret_val;
}

void insert_node(int i, int t, int *pNtr, int **tr, int **efwl,
                 int *pnwl, int *ptwl, int onside, int **hull, int *nhull,
                 int **tpnl)
{
  /* This function inserts a node i in triangle t,
   * it updates the triangle matrix tr
   * and it prepares an edge flip waiting list.
   *
   * It requires the onside info of node i in triangle t,
   * as prepared by the search function.
   *
   * It distinguishes between nodes who fall on an edge of t
   * or completely inside t.
   *
   * It updates the hull if i falls on it.
   */

  int k, s;                     /* counter */
  int *rp1, *rp2;               /* triangles about to be replaced */
  int Ntr = (*pNtr);
  int nwl = (*pnwl);
  int twl = (*ptwl);

  if (!(rp1 = e_alloc_i_vect(11)))
    e_GeneralError("insert_node: memory (1)\n");
  if (!(rp2 = e_alloc_i_vect(11)))
    e_GeneralError("insert_node: memory (2)\n");

  if (onside == 0)              /* node i lies completely within triangle t */
    /* create tree new triangles */
  {
    for (k = 0; k < 9; k++)
      rp1[k] = tr[t][k];
    tr[Ntr][0] = i;
    tr[Ntr][1] = rp1[1];
    tr[Ntr][2] = rp1[2];
    tr[Ntr][3] = t;
    tr[Ntr][4] = rp1[4];
    tr[Ntr][5] = Ntr + 1;
    tr[Ntr][6] = 5;
    tr[Ntr][7] = rp1[7];
    tr[Ntr][8] = 3;
    tr[Ntr + 1][0] = i;
    tr[Ntr + 1][1] = rp1[2];
    tr[Ntr + 1][2] = rp1[0];
    tr[Ntr + 1][3] = Ntr;
    tr[Ntr + 1][4] = rp1[5];
    tr[Ntr + 1][5] = t;
    tr[Ntr + 1][6] = 5;
    tr[Ntr + 1][7] = rp1[8];
    tr[Ntr + 1][8] = 3;
    tr[t][4] = rp1[3];
    tr[t][5] = Ntr;
    tr[t][3] = Ntr + 1;
    tr[t][2] = rp1[1];
    tr[t][1] = rp1[0];
    tr[t][0] = i;
    tr[t][7] = rp1[6];
    tr[t][6] = 5;
    tr[t][8] = 3;
    if (tr[t][4] != -1)
    {
      tr[tr[t][4]][tr[t][7]] = t;
      tr[tr[t][4]][3 + tr[t][7]] = 4;
    }
    if (tr[Ntr + 1][4] != -1)
    {
      tr[tr[Ntr + 1][4]][tr[Ntr + 1][7]] = Ntr + 1;
      tr[tr[Ntr + 1][4]][3 + tr[Ntr + 1][7]] = 4;
    }
    if (tr[Ntr][4] != -1)
    {
      tr[tr[Ntr][4]][tr[Ntr][7]] = Ntr;
      tr[tr[Ntr][4]][3 + tr[Ntr][7]] = 4;
    }
    tr[t][9] = 1;
    tr[Ntr][9] = 1;
    tr[Ntr + 1][9] = 1;
    tr[Ntr][10] = tr[t][10];
    tr[Ntr + 1][10] = tr[t][10];
    rmvtrftpnlist(t, rp1[2], tpnl);
    addtr2tpnlist(tr, t, tpnl);
    addtr2tpnlist(tr, Ntr, tpnl);
    addtr2tpnlist(tr, Ntr + 1, tpnl);
    Ntr += 2;
    if (twl - nwl < 3)          /* prepare the edge flip waiting list */
    {
      (*efwl) = e_incr_i_vect(twl, (*efwl), twl);
      twl += twl;
    }
    if (tr[t][4] != -1)
    {
      (*efwl)[nwl] = t;
      nwl++;
    }
    if (tr[Ntr - 1][4] != -1)
    {
      (*efwl)[nwl] = Ntr - 1;
      nwl++;
    }
    if (tr[Ntr - 2][4] != -1)
    {
      (*efwl)[nwl] = Ntr - 2;
      nwl++;
    }
  }
  else                          /* node i lies on one of the sides of triangle t */
    /* create two new triangles to start */
  {
    for (k = 0; k < 9; k++)
      rp1[k] = tr[t][k];
    tr[Ntr][0] = i;
    tr[Ntr][1] = rp1[(onside + 1) % 3];
    tr[Ntr][2] = rp1[(onside + 2) % 3];
    tr[t][1] = rp1[(onside + 2) % 3];
    tr[t][2] = rp1[(onside) % 3];
    tr[t][0] = i;
    tr[Ntr][4] = rp1[3 + ((onside + 1) % 3)];
    tr[Ntr][5] = t;
    tr[t][3] = Ntr;
    tr[t][4] = rp1[3 + ((onside + 2) % 3)];
    tr[Ntr][7] = rp1[6 + ((onside + 1) % 3)];
    tr[Ntr][8] = 3;
    tr[t][6] = 5;
    tr[t][7] = rp1[6 + ((onside + 2) % 3)];
    if (tr[t][4] != -1)
    {
      tr[tr[t][4]][tr[t][7]] = t;
      tr[tr[t][4]][3 + tr[t][7]] = 4;
    }
    if (tr[Ntr][4] != -1)
    {
      tr[tr[Ntr][4]][tr[Ntr][7]] = Ntr;
      tr[tr[Ntr][4]][3 + tr[Ntr][7]] = 4;
    }
    if (twl - nwl < 4)          /* prepare the edge flip waiting list */
    {
      (*efwl) = e_incr_i_vect(twl, (*efwl), twl);
      twl += twl;
    }
    if (tr[t][4] != -1)
    {
      (*efwl)[nwl] = t;
      nwl++;
    }
    if (tr[Ntr][4] != -1)
    {
      (*efwl)[nwl] = Ntr;
      nwl++;
    }
    tr[Ntr][9] = 1;
    tr[t][9] = 1;
    tr[Ntr][10] = tr[t][10];
    rmvtrftpnlist(t, rp1[(onside + 1) % 3], tpnl);
    addtr2tpnlist(tr, t, tpnl);
    addtr2tpnlist(tr, Ntr, tpnl);
    if (rp1[onside] != -1)      /* new point falls not on hull */
      /* this means you have to make two more triangles */
    {
      for (k = 0; k < 9; k++)
        rp2[k] = tr[rp1[onside]][k];
      tr[rp1[onside]][0] = i;
      tr[rp1[onside]][1] = rp1[(onside) % 3];
      tr[rp1[onside]][2] = rp2[(rp1[onside + 3] + 2) % 3];
      tr[Ntr + 1][0] = i;
      tr[Ntr + 1][1] = rp2[(rp1[onside + 3] + 2) % 3];
      tr[Ntr + 1][2] = rp1[(onside + 1) % 3];
      tr[Ntr][3] = Ntr + 1;
      tr[t][5] = rp1[onside];
      tr[rp1[onside]][3] = t;
      tr[rp1[onside]][4] = rp2[3 + (rp1[onside + 3] + 1) % 3];
      tr[rp1[onside]][5] = Ntr + 1;
      tr[Ntr + 1][3] = rp1[onside];
      tr[Ntr + 1][4] = rp2[3 + (rp1[onside + 3] + 2) % 3];
      tr[Ntr + 1][5] = Ntr;
      tr[Ntr][6] = 5;
      tr[t][8] = 3;
      tr[rp1[onside]][6] = 5;
      tr[rp1[onside]][7] = rp2[6 + (rp1[onside + 3] + 1) % 3];
      tr[rp1[onside]][8] = 3;
      tr[Ntr + 1][6] = 5;
      tr[Ntr + 1][7] = rp2[6 + (rp1[onside + 3] + 2) % 3];
      tr[Ntr + 1][8] = 3;
      if (tr[rp1[onside]][4] != -1)
      {
        tr[tr[rp1[onside]][4]][tr[rp1[onside]][7]] = rp1[onside];
        tr[tr[rp1[onside]][4]][3 + tr[rp1[onside]][7]] = 4;
      }
      if (tr[Ntr + 1][4] != -1)
      {
        tr[tr[Ntr + 1][4]][tr[Ntr + 1][7]] = Ntr + 1;
        tr[tr[Ntr + 1][4]][3 + tr[Ntr + 1][7]] = 4;
      }
      tr[Ntr + 1][9] = 1;
      tr[rp1[onside]][9] = 1;
      tr[Ntr + 1][10] = tr[rp1[onside]][10];
      rmvtrftpnlist(rp1[onside], rp1[(onside + 1) % 3], tpnl);
      addtr2tpnlist(tr, rp1[onside], tpnl);
      addtr2tpnlist(tr, Ntr + 1, tpnl);
      Ntr += 2;
      if (tr[rp1[onside]][4] != -1) /* more edge flip candidates */
      {
        (*efwl)[nwl] = rp1[onside];
        nwl++;
      }
      if (tr[Ntr - 1][4] != -1)
      {
        (*efwl)[nwl] = Ntr - 1;
        nwl++;
      }
    }
    else                        /* new point falls on hull */
      /* no more triangles have to be created */
    {
      tr[Ntr][3] = -1;
      tr[t][5] = -1;
      Ntr++;
      /* but you have to update of the hull */
      for (k = 0; k < *nhull; k++)
        if ((*hull)[k] == tr[t][2])
          break;
      (*hull) = e_incr_i_vect(*nhull, (*hull), INCR_SIZE);
      for (s = (*nhull)-1; s > k; s--)
        (*hull)[s + 1] = (*hull)[s];
      (*hull)[k + 1] = i;
      (*nhull)++;	/***** move this *****/
    }
  }
  e_free(rp1);
  e_free(rp2);
  (*pNtr) = Ntr;
  (*pnwl) = nwl;
  (*ptwl) = twl;
}

void update_hull(int **hull, int *nhull, int ossi, int osei, int i)
{
  /* This function extends the hull with external point i,
   * which is visible from the hull in range ossi to osei.
   */

  int s;                        /* counter */
  int id;                       /* index difference between ossi & osei */

  id = (osei - ossi + (*nhull)) % (*nhull);
  if (id == 1)
  {
    (*hull) = e_incr_i_vect(*nhull, (*hull), INCR_SIZE);
    for (s = (*nhull)-1; s > ossi; s--)
      (*hull)[s + 1] = (*hull)[s];
    (*hull)[ossi + 1] = i;
    (*nhull)++;
  }
  else if (id == 2)
  {
    /* no memory increase required  */
    (*hull)[(ossi + 1)%(*nhull)] = i;
  }
  else if (id > 2)
  {
    if (osei > ossi)
    {
      for (s = osei; s < *nhull; s++)
        (*hull)[ossi + 2 + s - osei] = (*hull)[s];
      (*hull)[ossi + 1] = i;
    }
    else
    {
      for (s = 0; s < *nhull - id + 1; s++)
        (*hull)[s] = (*hull)[osei + s];
      (*hull)[s] = i;
    }
    (*nhull) -= id - 2;
  }
}

void connect_outside_node(int i, double **coords, int *pNtr, int **tr,
                          int **efwl, int *pnwl, int *ptwl, int **hull,
                          int *nhull, int **tpnl, MTRF *mtrf)
{
  /* This function connects an outside node to the existing triangulation.
   * It updates the triangle matrix tr,
   * and it prepares an edge flip waiting list.
   * It also updates the hull.
   */

  int j, k, s;                  /* counters */
  int ossi, osei;               /* outside start and end hull vector indices */
  int falsestart;               /* indicates that ossi motght be wrong */
  int id;                       /* index difference between ossi & osei */
  int Ntr = (*pNtr);
  int nwl = (*pnwl);            /* shortcuts */
  int twl = (*ptwl);
  int ccw,ccw2,valid,valid2;	  /* IsCCw results */
  double  mu;                   /* ColinearTriangle result */
  int outside;                  /* is the point outside the convex hull ? */

  ossi = -1;
  osei = -1;
  falsestart = FALSE;
  for (j = 0; j < *nhull; j++)   /* search for the visible part of the hull */
  {
    outside = !IsCCW(coords[(*hull)[j]], coords[(*hull)[(j + 1)%(*nhull)]], coords[i], mtrf, &valid);
    if (!valid)
    {
      if (!ColinearTriangle(coords[(*hull)[j]], coords[(*hull)[(j + 1)%(*nhull)]], coords[i],
                            mtrf, &mu))
        e_GeneralError("connect_outside_node: improper call to ColinearTriangle");
      if ((mu<0.0)||(mu>1.0))
        outside=FALSE;
      else
        e_GeneralError("connect_outside_node: called for node on edge");
    }
    if (outside)
    {
      if ((falsestart) && (osei != -1))
      {
        ossi = j;
        falsestart = FALSE;
      }
      else if (ossi == -1)
      {
        ossi = j;
        if (j == 0)
          falsestart = TRUE;
      }
    }
    else if ((ossi != -1) && (osei == -1))
      osei = j;
  }
  if (ossi == -1)
  {
    printf("node=%d nhull=%d ntr=%d\n", i, *nhull, Ntr);
    e_GeneralError("connect_outside_node: outside node seems to be inside\n");
  }
  if (osei == -1)
    osei = 0;
  id = (osei - ossi + *nhull) % (*nhull);

  for (j = 0; j < id; j++)      /* connecting the new triangles */
  {
    /* the definition of the new triangle */
    tr[Ntr][0] = i;
    tr[Ntr][1] = (*hull)[(ossi + j + 1) % (*nhull)];
    tr[Ntr][2] = (*hull)[(ossi + j) % (*nhull)];

    /* info typical for all triangles except the last one */
    if (j < id - 1)
    {
      tr[Ntr][3] = Ntr + 1;
      tr[Ntr][6] = 5;
    }
    else
      tr[Ntr][3] = -1;

    for (s = 0; s < tpnl[(*hull)[(ossi + j) % (*nhull)]][0]; s++)
    {
      for (k = 0; k < 3; k++)
        if ((tr[tpnl[(*hull)[(ossi + j) % (*nhull)]][2 + s]][k] ==
             (*hull)[(ossi + j) % (*nhull)])
            && (tr[tpnl[(*hull)[(ossi + j) % (*nhull)]][2 + s]][(k + 1) % 3] ==
                (*hull)[(ossi + j + 1) % (*nhull)]))
          break;
      if (k < 3)
        break;
    }
    if (s == tpnl[(*hull)[(ossi + j) % (*nhull)]][0])
      e_GeneralError("connect_outside_node: triangle on border not found");
    else
    {
      tr[Ntr][4] = tpnl[(*hull)[(ossi + j) % (*nhull)]][2 + s];
      tr[Ntr][7] = 3 + k;
    }

    tr[Ntr][9] = 1;
    tr[tr[Ntr][4]][tr[Ntr][7]] = Ntr;
    tr[tr[Ntr][4]][3 + tr[Ntr][7]] = 4;

    /* info typical for all triangles except the first one */
    if (j > 0)
    {
      tr[Ntr][5] = Ntr - 1;
      tr[Ntr][8] = 3;
    }
    else                        /* and updating the border table */
      tr[Ntr][5] = -1;

    /* checking the new triangles for adding to the edge flip list */
    ccw=IsCCW(coords[tr[tr[Ntr][4]][(tr[Ntr][7] + 2) % 3]],
              coords[(*hull)[(ossi + j) % (*nhull)]],
              coords[i], mtrf, &valid);
    ccw2=IsCCW(coords[tr[tr[Ntr][4]][(tr[Ntr][7] + 2) % 3]],
               coords[i],
               coords[(*hull)[(ossi + j + 1) % (*nhull)]], mtrf, &valid2);
    if (ccw && ccw2 && valid && valid2)
    {
      if (twl - nwl < 1)        /* prepare the edge flip waiting list */
      {
        (*efwl) = e_incr_i_vect(twl, (*efwl), twl);
        twl += twl;
      }
      (*efwl)[nwl] = Ntr;
      nwl++;
    }

    /* and updating the tpnl */
    addtr2tpnlist(tr, Ntr, tpnl);
    Ntr++;
  }
  update_hull(hull, nhull, ossi, osei, i);
  (*pNtr) = Ntr;
  (*pnwl) = nwl;
  (*ptwl) = twl;
}

void edge_flipping(int i, double **coords, int **efwl, int *pnwl, int *ptwl,
                   int **tr, int **tpnl, MTRF *mtrf)
{
  /* This function carries out the edge flipping and clears
   * the edge flip waiting list.
   *
   * Upon flipping the border table may be updated.
   */

  int k;                        /* counter */
  int *rp1, *rp2;               /* triangles about to be replaced */
  int t;                        /* a triangle index */
  int nwl = (*pnwl);
  int twl = (*ptwl);

  rp1 = e_alloc_i_vect(9);
  rp2 = e_alloc_i_vect(9);

  while (nwl != 0)
  {
    if (tr[tr[(*efwl)[nwl - 1]][4]][10] == -1)
    {
      /* out of domain triangles can't be flipped in order to keep boundaries */
      nwl--;
    }
    else if (j_incircle(coords[tr[tr[(*efwl)[nwl - 1]][4]][0]],
                        coords[tr[tr[(*efwl)[nwl - 1]][4]][1]],
                        coords[tr[tr[(*efwl)[nwl - 1]][4]][2]],
                        coords[i], mtrf))
    {
      t = (*efwl)[nwl - 1];
      for (k = 0; k < 9; k++)
        rp1[k] = tr[t][k];
      for (k = 0; k < 9; k++)
        rp2[k] = tr[rp1[4]][k];
      /* tr[t][0]=rp1[0];  */
      tr[t][1] = rp2[(rp1[7] + 2) % 3];
      /* tr[t][2]=rp1[2];  */
      tr[t][3] = rp1[4];
      tr[t][4] = rp2[3 + (rp1[7] + 2) % 3];
      /* tr[t][5]=rp1[5];  */
      tr[t][6] = 5;
      if (tr[t][4] != -1)
        tr[t][7] = rp2[6 + (rp1[7] + 2) % 3];
      /* if (tr[t][5]!=-1)
         tr[t][8]=rp1[8]; */
      tr[t][9] = 1;
      tr[rp1[4]][0] = rp1[0];
      tr[rp1[4]][1] = rp1[1];
      tr[rp1[4]][2] = rp2[(rp1[7] + 2) % 3];
      tr[rp1[4]][3] = rp1[3];
      tr[rp1[4]][4] = rp2[3 + (rp1[7] + 1) % 3];
      tr[rp1[4]][5] = t;
      tr[rp1[4]][9] = 1;
      if (tr[rp1[4]][3] != -1)
        tr[rp1[4]][6] = rp1[6];
      if (tr[rp1[4]][4] != -1)
        tr[rp1[4]][7] = rp2[6 + (rp1[7] + 1) % 3];
      if (tr[rp1[4]][5] != -1)
        tr[rp1[4]][8] = 3;
      /* if (rp1[5]!=-1)
         {
         tr[rp1[5]][rp1[8]]=t;
         tr[rp1[5]][rp1[8]+3]=5;
         } */
      if (rp1[3] != -1)
      {
        tr[rp1[3]][rp1[6]] = rp1[4];
        tr[rp1[3]][rp1[6] + 3] = 3;
      }
      if (tr[t][4] != -1)
      {
        tr[tr[t][4]][tr[t][7]] = t;
        tr[tr[t][4]][tr[t][7] + 3] = 4;
      }
      if (tr[rp1[4]][4] != -1)
      {
        tr[tr[rp1[4]][4]][tr[rp1[4]][7]] = rp1[4];
        tr[tr[rp1[4]][4]][tr[rp1[4]][7] + 3] = 4;
      }
      rmvtrftpnlist(t, rp1[1], tpnl);
      rmvtrftpnlist(rp1[4], rp1[2], tpnl);
      addtr2tpnlist(tr, t, tpnl);
      addtr2tpnlist(tr, rp1[4], tpnl);
      /* add new candidates for flipping */
      if (tr[t][4] != -1)
      {
        /* (*efwl)[nwl-1]=t; */
      }
      else
      {
        nwl--;
      }
      if (tr[rp1[4]][4] != -1)
      {
        if (twl - nwl < 1)
        {
          (*efwl) = e_incr_i_vect(twl, (*efwl), twl);
          twl += twl;
        }
        (*efwl)[nwl] = rp1[4];
        nwl++;
      }
    }
    else
    {
      nwl--;
    }
  }
  e_free(rp1);
  e_free(rp2);
  (*pnwl) = nwl;
  (*ptwl) = twl;
}

void add_new_nodes(double **coords, int np, int nnp, int stnp, int ***triangles,
                   int *ntriangles, int *tts, int **convhull, int *nhullelems,
                   int **tpnl, MTRF *mtrf)
{
  int n;                        /* counter */
  int t;                        /* triangle that contains new point */
  int sres;                     /* result of search operation */
  int onside;                   /* onside indicator */
  int *efwl;                    /* edge flip wait list */
  int nwl;                      /* length of the wait list */
  int twl;                      /* total allocated length of efwl */

  twl = INCR_SIZE;              /* initialize the edge flip wait list */
  efwl = e_alloc_i_vect(twl);   /* edge flip wait list side 4 to be checked */
  nwl = 0;

  while ((*tts) - (*ntriangles) < E_MAX(4, 3 * nnp))
  {
    (*triangles) = e_incr_cont_i_matr((*tts), 11, (*triangles), (*tts));
    (*tts) += (*tts);
  }

  for (n = 0; n < nnp; n++)
  {
    if (DebugOutput)
      printf("\ninserting node %d", n);
    sres =
      trianglesearch(coords[stnp + n], coords, (*triangles), 0, &t, &onside,
                     mtrf);
    if (sres == 1)
      insert_node(stnp + n, t, ntriangles, (*triangles),
                  &efwl, &nwl, &twl, onside, convhull, nhullelems, tpnl);
    else if (sres == 0)
    {
      /* e_GeneralError("add_new_nodes is called for an outside node"); */
      connect_outside_node(stnp + n, coords, ntriangles, (*triangles), &efwl,
                           &nwl, &twl, convhull, nhullelems, tpnl, mtrf);
    }
    edge_flipping(stnp + n, coords, &efwl, &nwl, &twl, (*triangles), tpnl,
                  mtrf);
    if (ShowGraphics)
      mgnDrawAll(np,coords,*ntriangles,*triangles,0,NULL,0,
        NULL,NULL,"adding new node");
  }

  /* free up that memory */
  e_free(efwl);
}


int delaunay_check(int **tr, int ntr, double **coords, int np)
{
  /* perform an exhaustive check to see if a mesh is delaunay */

  int i, t;                     /* counters */
  int del = TRUE;               /* is the mesh delaunay y/n */
  MTRF mtrf;

  mtrf.used = FALSE;

  for (i = 0; i < np; i++)
  {
    for (t = 0; t < ntr; t++)
    {
      if ((i != tr[t][0]) && (i != tr[t][1]) && (i != tr[t][2]))
        if (j_incircle
            (coords[tr[t][0]], coords[tr[t][1]], coords[tr[t][2]], coords[i],
             &mtrf))
        {
          del = FALSE;
          printf("node %d in circumcircle of triangle %d\n",i+1,t+1);
        }
      if (!del)
        break;
    }
    if (!del)
      break;
  }
  return del;
}

/* end of file */
