#include "mgn.h"
#include "mgn_graphics.h"
#include "dtr.h"

void dtr_main(MGNI *mgni)
{
  /* Delaunay triangulation main function */

  /* variable declarations */
  int **triangles;              /* triangle matrix */
  int ntriangles;               /* number of triangles */
  int ttriangles;               /* allocated spaces triangles */
  double *mpr;                  /* original mesh parameters */
  double **coords;              /* coordinates matrix */
  int np;                       /* number of points */
  int tp;                       /* allocated spaces in coords */
  int *convhull;                /* the convex hull */
  int nhullelems;               /* number of points in the hull */
  int **tpnl;                   /* triangle per node listing */
  EDGE *edgs;                   /* element boundaries for ebe refinement */
  BND *bnd;                     /* boundary information */
  int i;                        /* counters */
  MTRF mtrf;                    /* mapping transformation data */
  BG_DATA *bgd;									/* background data */
  double ho;										/* original h */

  /* initial values */
  coords = NULL;
  triangles = NULL;
  convhull = NULL;
  tpnl = NULL;
  bnd = NULL;

  /* parse the coarse mesh */
  parse_mesh(&(mgni->task), mgni->meth, mgni->mesh, &coords, &np, &triangles,
             &ntriangles, &ttriangles, &tpnl, &nhullelems, &convhull, &bnd);

  /* initialize according to the control type */
  parse_meshsize_ctrl(mgni, &bgd, &mpr);

  /* as the triangle and coords matrix are enlarged incrementally from now */
  tp = np;

  /* EBE or remesh/refine */
  if (mgni->task == TASK_EBE)
  {
    /* setting up the individual element boundaries */
    setup_edges(triangles, ntriangles, &edgs);
    /* element boundary refinement */
    refine_edges(&coords, &np, &tp,
                 mgni->ctrl, mgni->geomset, mgni->mesh,
                 bgd, &mpr, mgni->h, &tpnl, triangles, ntriangles, edgs,
                 (mgni->opt==CONV_OPT_FISSION_TO_QUAD));

    if (ShowGraphics)
      mgnDrawAll(np, coords, 0, NULL, 0, NULL, 0, NULL, NULL,
                 "boundary nodes");

    /* refine the element interior */
    ebe_refine(mgni, &coords, &np, &tp,
               mgni->geomset, mgni->mesh, bgd, &mpr,
               &tpnl, &triangles, &ntriangles, &ttriangles, edgs,
               TRUE);
  }
  else
  {
    /* problem with Chew1 */
    if (mgni->dln == DLN_CHEW1)
    {
    	ho=mgni->h;
    	mgni->h=(sqrt(3)/2.0)*ho;
		}

    /* boundary refinement and recovery by halfwaynode method */
    boundary_refinement_and_recovery(&coords, &np, &tp, mgni->ctrl,
                                     mgni->geomset, mgni->mesh,
                                     bgd, &mpr, mgni->h, bnd, &tpnl,
                                     &triangles, &ntriangles, &ttriangles,
                                     &convhull, &nhullelems,
                                     (mgni->opt==CONV_OPT_FISSION_TO_QUAD));

    /* problem with Chew1 */
    if (mgni->dln == DLN_CHEW1)
    	mgni->h=ho;

    /* setting the areacodes */
    setareacode(bnd, tpnl, triangles, ntriangles);

    /* refine the interior of the domain */
    mtrf.used = FALSE;          /* for compatibility */
    f_refine_del_mesh(&coords, &np, &tp, &triangles, &ntriangles,
                      &ttriangles, &convhull, &nhullelems,
                      mgni->ctrl, mgni->alfa, mgni->beta,
                      bgd, &mpr, mgni->h, &tpnl, NULL, mgni->dln, &mtrf);
  }

  if (ShowGraphics)
    mgnDrawAll(np, coords, ntriangles, triangles, 0, NULL, 0, NULL, NULL,
               "refined mesh");

  /* smoothing by nodal averaging before output */
  if (mgni->smtype == SM_NODAL_AVG)
    smooth_by_nodal_averaging(coords, np, triangles, ntriangles, tpnl,
                              mgni->smiter);

  /* write the output */
  mgni->outputmesh = (MESH *) e_calloc(1, sizeof(MESH));
  e_NullifyMesh(mgni->outputmesh);
  create_outputmesh(coords, np, triangles, ntriangles, mgni->outputmesh);

  if (ShowGraphics)
    mgnDrawAll((mgni->outputmesh)->NMeshPoints,
               (mgni->outputmesh)->MeshPointCoords,
               (mgni->outputmesh)->NElemsTriang1,
               (mgni->outputmesh)->NodesTriang1, 0, NULL, -1, NULL, NULL,
               "output mesh");

  /* for elem controlled meshing, copy nodal meshparams back in */
  if (mgni->ctrl == CTRL_MPR)
  {
    (mgni->outputmesh)->SubStructSet[E_SS_NUM_MESHPARAM] = TRUE;
    (mgni->outputmesh)->MeshParam.NodalMeshParamsSet = TRUE;
    (mgni->outputmesh)->MeshParam.NodalMeshParams = e_alloc_d_vect(np);
    for (i = 0; i < (mgni->outputmesh)->NNodes; i++)
      (mgni->outputmesh)->MeshParam.NodalMeshParams[i] = mpr[i];
  }

  /* free up all consumed memory */
  if (mgni->ctrl == CTRL_MPR)
  {
    e_free(mpr);
  }
  else if ((mgni->ctrl == CTRL_BACKGROUND) || (mgni->ctrl == CTRL_SOURCES))
  {
    free_bg_data(bgd);
  }

  e_free(convhull);
  e_free_cont_i_matr(ttriangles, triangles);
  e_free_cont_d_matr(tp, coords);
  freetpnl(tpnl, tp);
}

void boundary_refinement_and_recovery(double ***coords, int *np, int *tp,
                                      int ctrltype, int geomset, MESH *mesh,
                                      BG_DATA *bgd, double **mpr, double h,
                                      BND *bnd, int ***tpnl,
                                      int ***triangles, int *ntriangles,
                                      int *ttriangles, int **convhull,
                                      int *nhullelems, int force_even)
{
  int stnp;                     /* start with inserting at this node */
  int nnp;                      /* n of new pnts made by boundary refinement */
  MTRF mtrf;                    /* unused mapping transformation data */
  mtrf.used = FALSE;

  /* make a better approximation of the boundary by refining the boundaries */
  nnp = make_boundary_nodes(coords, np, tp, &stnp, ctrltype,
                            geomset, mesh, bgd, mpr, h, bnd, tpnl,
                            force_even);

  if (PrintProgress)
    printf("\n%d new boundary nodes to be inserted\n\n", nnp);

  if (ShowGraphics)
  {
    mgnDrawAll(nnp, &((*coords)[stnp]), 0, NULL, 0, NULL, 0,
               NULL, NULL, "new boundary nodes");
  }

  /* add the new boundary nodes to the triangulation */
  add_new_nodes(*coords, *np, nnp, stnp, triangles, ntriangles, ttriangles,
                convhull, nhullelems, *tpnl, &mtrf);
  if (ShowGraphics)
  {
    mgnDrawAll(*np, *coords, *ntriangles, *triangles, 0, NULL, 0,
               NULL, NULL, "initial triangulation");
  }

  /* the stitching procedure (boundary recovery) */
  nnp = recover_boundaries(coords, np, tp, &stnp, *triangles, *ntriangles,
                           tpnl, bnd, ctrltype, mpr, &mtrf);

  if (ShowGraphics)
    if (nnp > 0)
    {
      mgnDrawAll(nnp, &((*coords)[stnp]), 0, NULL, 0, NULL, 0,
                 NULL, NULL, "nodes caused by stitching");
      mgnDrawAll(*np, *coords, *ntriangles, *triangles, 0, NULL, 0,
                 NULL, NULL, "triangulation after stitching");
    }

  while (nnp > 0)
  {
    if (PrintProgress)
      printf("%d new nodes caused by stitching\n\n", nnp);
    add_new_nodes(*coords, *np, nnp, stnp, triangles, ntriangles, ttriangles,
                  convhull, nhullelems, *tpnl, &mtrf);
    nnp = recover_boundaries(coords, np, tp, &stnp, *triangles, *ntriangles,
                             tpnl, bnd, ctrltype, mpr, &mtrf);

    if (ShowGraphics)
      if (nnp > 0)
      {
        mgnDrawAll(nnp, &((*coords)[stnp]), 0, NULL, 0, NULL, 0,
                   NULL, NULL, "nodes caused by stitching");
        mgnDrawAll(*np, *coords, *ntriangles, *triangles, 0, NULL, 0,
                   NULL, NULL, "triangulation after stitching");
      }
  }
}
