#include "mgn.h"

/* globals */
static int 	g_LineNumber;
static char *g_FullPathName;
static FILE *g_fp;

/* the keywords for the mesh generation batch file */
#define MGJ_KW_TASK                 "TASK"
#define MGJ_KW_METHOD               "METHOD"
#define MGJ_KW_ALFA                 "ALFA"
#define MGJ_KW_BETA                 "BETA"
#define MGJ_KW_CTRLTYPE             "CONTROL_TYPE"
#define MGJ_KW_DLNOPT               "DLN_OPTION"
#define MGJ_KW_OPTION               "MESH_OPTION"
#define MGJ_KW_F_GEOM               "GEOM_MODEL_FILE"
#define MGJ_KW_F_COARSE             "COARSE_MESH_FILE"
#define MGJ_KW_F_RES                "RESULT_MESH_FILE"
#define MGJ_KW_F_BKGRMESH           "BKGRND_MESH_FILE"
#define MGJ_KW_F_BKGRPARAM          "BKGRND_MESH_PARAMS_FILE"
#define MGJ_KW_F_CRSPARAM           "COARSE_MESH_PARAMS_FILE"
#define MGJ_KW_F_SRC                "LOCAL_MESH_CTRL_FILE"
#define MGJ_KW_UNIFMPR              "UNIFORM_MESH_PARAM"
#define MGJ_KW_SMTYPE               "SMOOTHING_TYPE"
#define MGJ_KW_SMITER               "SMOOTHING_ITER"
#define MGJ_KW_OUTFORMAT            "OUTPUT_FORMAT"

/* recognized descriptors */
#define MGJ_D_MESHGEN               "\"MESHGEN\""
#define MGJ_D_REMESH                "\"REMESH\""
#define MGJ_D_REFINE                "\"REFINE\""
#define MGJ_D_EBE_REFINE            "\"EBE_REFINE\""
#define MGJ_D_POSTPROC              "\"POSTPROC\""
#define MGJ_D_ADV_FRONT             "\"ADV_FRONT\""
#define MGJ_D_DELAUNAY              "\"DELAUNAY\""
#define MGJ_D_PAVING                "\"PAVING\""
#define MGJ_D_REG_GRID              "\"REG_GRID\""
#define MGJ_D_CENTROID              "\"CENTROID\""
#define MGJ_D_CENTR_RAD             "\"CENTR_RAD\""
#define MGJ_D_OPTIMAL               "\"PROJECTION\""
#define MGJ_D_CHEW1                 "\"CHEW1\""
#define MGJ_D_BACKGROUND            "\"BACKGROUND\""
#define MGJ_D_MPR                   "\"MPR\""
#define MGJ_D_UNIFORM               "\"UNIFORM\""
#define MGJ_D_SOURCES               "\"SOURCES\""
#define MGJ_D_FISSION_TO_QUAD       "\"FISSION_TO_QUAD\""
#define MGJ_D_RANK_TO_QUAD          "\"RANK_TO_QUAD\""
#define MGJ_D_HALF_TO_QUAD          "\"HALF_TO_QUAD\""
#define MGJ_D_SPLIT_TO_TRIANG       "\"SPLITTING_TRIANG\""
#define MGJ_D_NONE                  "\"NONE\""
#define MGJ_D_LAPLACIAN             "\"LAPLACIAN\""
#define MGJ_D_LAPLACIAN_NDE         "\"LAPLACIAN_NDE\""
#define MGJ_D_NODAL_AVG             "\"NODAL_AVG\""
#define MGJ_D_E_LIB                 "\"E_LIB\""
#define MGJ_D_SMSH                  "\"SMSH\""

static int mg_GetKeyword(char *line, char *word);
static int mg_WordMatchesKeyword(char *Keyword, char *word);
static int mg_IsAKeyword(char *word);
static int mg_GetFileName(char *filename, char *line);

int parse_mgn_into_mgni(MGNI ** mgni, char *filename)
{
  int fullpathnamelength;
  char line[E_LINE_MAX];
  char word[E_LINE_MAX];
  char value[E_LINE_MAX];
  int i;

  fullpathnamelength = strlen(filename) + 5;
  g_FullPathName = e_alloc_c_vect(fullpathnamelength);
  strcpy(g_FullPathName, filename);
  strcat(g_FullPathName, ".mgn");
  if ((g_fp = fopen(g_FullPathName, "r")) == NULL)
  {
    e_FileError(g_FullPathName, "cannot open file");
    e_free(g_FullPathName);
    return 0;
  }
  /* initialization of the mgni structure */
  (*mgni) = (MGNI *) e_calloc(1, sizeof(MGNI));
  (*mgni)->task = -1;
  (*mgni)->meth = -1;
  (*mgni)->dln = -1;
  (*mgni)->ctrl = -1;
  for (i = 0; i < MGJ_NFI; i++)
    strcpy((*mgni)->fname[i], "");
  (*mgni)->smtype = -1;
  (*mgni)->smiter = -1;
  (*mgni)->outf = -1;
  (*mgni)->geomset = FALSE;

  /* read the task description */
  mg_GetKeyword(line, word);
  mg_WordMatchesKeyword(MGJ_KW_TASK, word);
  if (sscanf(line, "%*s %s", value) != 1)
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "can't read task description");
  else if (!strcmp(value, MGJ_D_MESHGEN))
    (*mgni)->task = TASK_MESHGEN;
  else if (!strcmp(value, MGJ_D_REMESH))
    (*mgni)->task = TASK_REMESH;
  else if (!strcmp(value, MGJ_D_REFINE))
    (*mgni)->task = TASK_REFINE;
  else if (!strcmp(value, MGJ_D_EBE_REFINE))
    (*mgni)->task = TASK_EBE;
  else if (!strcmp(value, MGJ_D_POSTPROC))
    (*mgni)->task = TASK_POSTPROC;
  else
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "unrecognized task description");

  /* read the method */
  mg_GetKeyword(line, word);
  mg_WordMatchesKeyword(MGJ_KW_METHOD, word);
  if (sscanf(line, "%*s %s", value) != 1)
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "can't read method description");
  else if (!strcmp(value, MGJ_D_ADV_FRONT))
    (*mgni)->meth = METH_ADV_FRONT;
  else if (!strcmp(value, MGJ_D_DELAUNAY))
    (*mgni)->meth = METH_DELAUNAY;
  else if (!strcmp(value, MGJ_D_PAVING))
    (*mgni)->meth = METH_PAVING;
  else if (!strcmp(value, MGJ_D_REG_GRID))
    (*mgni)->meth = METH_REG_GRID;
  else if (!strcmp(value, MGJ_D_NONE))
    (*mgni)->meth = METH_NONE;
  else
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "unrecognized method description");

  if ((*mgni)->meth == METH_DELAUNAY)
  {
    /* read alfa and beta */
    mg_GetKeyword(line, word);
    mg_WordMatchesKeyword(MGJ_KW_ALFA, word);
    if (sscanf(line, "%*s %lf", &((*mgni)->alfa)) != 1)
      e_FileLineError(g_LineNumber, g_FullPathName, "can't read alfa");
    else if ((*mgni)->alfa < 0.1)
      e_FileLineError(g_LineNumber, g_FullPathName, "alfa should be > 0.1");
    mg_GetKeyword(line, word);
    mg_WordMatchesKeyword(MGJ_KW_BETA, word);
    if (sscanf(line, "%*s %lf", &((*mgni)->beta)) != 1)
      e_FileLineError(g_LineNumber, g_FullPathName, "can't read beta");
    else if ((*mgni)->beta < 0.1)
      e_FileLineError(g_LineNumber, g_FullPathName, "beta should be > 0.1");
    mg_GetKeyword(line, word);
    mg_WordMatchesKeyword(MGJ_KW_DLNOPT, word);
    if (sscanf(line, "%*s %s", value) != 1)
      e_FileLineError(g_LineNumber, g_FullPathName,
                      "can't read delaunay options");
    else if (!strcmp(value, MGJ_D_CENTROID))
      (*mgni)->dln = DLN_CENTR;
    else if (!strcmp(value, MGJ_D_OPTIMAL))
      (*mgni)->dln = DLN_OPT;
    else if (!strcmp(value, MGJ_D_CENTR_RAD))
      (*mgni)->dln = DLN_CENTR_RAD;
    else if (!strcmp(value, MGJ_D_CHEW1))
      (*mgni)->dln = DLN_CHEW1;
    else
      e_FileLineError(g_LineNumber, g_FullPathName,
                      "unrecognized delaunay option");
  }

  if ((*mgni)->meth == METH_REG_GRID)
  {
    /* read alfa (gprm) and beta (rprm) */
    mg_GetKeyword(line, word);
    mg_WordMatchesKeyword(MGJ_KW_ALFA, word);
    if (sscanf(line, "%*s %lf", &((*mgni)->alfa)) != 1)
      e_FileLineError(g_LineNumber, g_FullPathName, "can't read alfa");
    else if ((*mgni)->alfa < 0.1)
      e_FileLineError(g_LineNumber, g_FullPathName, "alfa should be > 0.1");
    mg_GetKeyword(line, word);
    mg_WordMatchesKeyword(MGJ_KW_BETA, word);
    if (sscanf(line, "%*s %lf", &((*mgni)->beta)) != 1)
      e_FileLineError(g_LineNumber, g_FullPathName, "can't read beta");
    else if ((*mgni)->beta < 0.1)
      e_FileLineError(g_LineNumber, g_FullPathName, "beta should be > 0.1");
  }
  
  /* read the control type description */
  mg_GetKeyword(line, word);
  mg_WordMatchesKeyword(MGJ_KW_CTRLTYPE, word);
  if (sscanf(line, "%*s %s", value) != 1)
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "can't read control type description");
  else if (!strcmp(value, MGJ_D_BACKGROUND))
    (*mgni)->ctrl = CTRL_BACKGROUND;
  else if (!strcmp(value, MGJ_D_MPR))
    (*mgni)->ctrl = CTRL_MPR;
  else if (!strcmp(value, MGJ_D_SOURCES))
    (*mgni)->ctrl = CTRL_SOURCES;
  else if (!strcmp(value, MGJ_D_UNIFORM))
    (*mgni)->ctrl = CTRL_UNIFORM;
  else
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "unrecognized control type description");

  /* read the mesh options */
  mg_GetKeyword(line, word);
  mg_WordMatchesKeyword(MGJ_KW_OPTION, word);
  if (sscanf(line, "%*s %s", value) != 1)
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "can't read mesh option description");
  else if (!strcmp(value, MGJ_D_FISSION_TO_QUAD))
    (*mgni)->opt = CONV_OPT_FISSION_TO_QUAD;
  else if (!strcmp(value, MGJ_D_RANK_TO_QUAD))
    (*mgni)->opt = CONV_OPT_RANK_TO_QUAD;
  else if (!strcmp(value, MGJ_D_HALF_TO_QUAD))
    (*mgni)->opt = CONV_OPT_HALF_TO_QUAD;
  else if (!strcmp(value, MGJ_D_SPLIT_TO_TRIANG))
    (*mgni)->opt = CONV_OPT_SPLIT_TO_TRIANG;
  else if (!strcmp(value, MGJ_D_NONE))
    (*mgni)->opt = CONV_OPT_NONE;
  else
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "unrecognized mesh option description");

  /* read filenames */
  mg_GetKeyword(line, word);
  if (!strcmp(MGJ_KW_F_GEOM, word))
  {
    mg_GetFileName((*mgni)->fname[MGJ_FI_GEOM], line);
    (*mgni)->geomset = TRUE;
    mg_GetKeyword(line, word);
  }
  mg_WordMatchesKeyword(MGJ_KW_F_COARSE, word);
  mg_GetFileName((*mgni)->fname[MGJ_FI_COARSE], line);

  mg_GetKeyword(line, word);
  mg_WordMatchesKeyword(MGJ_KW_F_RES, word);
  mg_GetFileName((*mgni)->fname[MGJ_FI_RES], line);

  if (((*mgni)->ctrl == CTRL_BACKGROUND) || ((*mgni)->ctrl == CTRL_SOURCES))
  {
    mg_GetKeyword(line, word);
    mg_WordMatchesKeyword(MGJ_KW_F_BKGRMESH, word);
    mg_GetFileName((*mgni)->fname[MGJ_FI_BKGRMESH], line);
    mg_GetKeyword(line, word);
    mg_WordMatchesKeyword(MGJ_KW_F_BKGRPARAM, word);
    mg_GetFileName((*mgni)->fname[MGJ_FI_BKGRPARAM], line);
  }
  if ((*mgni)->ctrl == CTRL_MPR)
  {
    mg_GetKeyword(line, word);
    mg_WordMatchesKeyword(MGJ_KW_F_CRSPARAM, word);
    mg_GetFileName((*mgni)->fname[MGJ_FI_CRSPARAM], line);
  }
  if ((*mgni)->ctrl == CTRL_SOURCES)
  {
    mg_GetKeyword(line, word);
    mg_WordMatchesKeyword(MGJ_KW_F_SRC, word);
    mg_GetFileName((*mgni)->fname[MGJ_FI_SRC], line);
  }
  (*mgni)->loadfiles = TRUE;

  /* if required read uniform mesh parameter */
	if ((*mgni)->ctrl == CTRL_UNIFORM)
  {
    /* read alfa and beta */
    mg_GetKeyword(line, word);
    mg_WordMatchesKeyword(MGJ_KW_UNIFMPR, word);
    if (sscanf(line, "%*s %lf", &((*mgni)->h)) != 1)
      e_FileLineError(g_LineNumber, g_FullPathName, "can't read uniform mesh parameter");
    else if ((*mgni)->h <= 0.0)
      e_FileLineError(g_LineNumber, g_FullPathName, "uniform mesh parameter should be >= 0.0");
 	}

  /* read smoothing type */
  mg_GetKeyword(line, word);
  mg_WordMatchesKeyword(MGJ_KW_SMTYPE, word);
  if (sscanf(line, "%*s %s", value) != 1)
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "can't read method description");
  else if (!strcmp(value, MGJ_D_LAPLACIAN))
    (*mgni)->smtype = SM_LAPLACIAN;
  else if (!strcmp(value, MGJ_D_LAPLACIAN_NDE))
    (*mgni)->smtype = SM_LAPLACIAN_NDE;
  else if (!strcmp(value, MGJ_D_NODAL_AVG))
    (*mgni)->smtype = SM_NODAL_AVG;
  else if (!strcmp(value, MGJ_D_PAVING))
    (*mgni)->smtype = SM_PAVING;
  else if (!strcmp(value, MGJ_D_NONE))
    (*mgni)->smtype = SM_NONE;
  else
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "unrecognized smoothing type description");

  /* read number of smoothing iterations */
  if ((*mgni)->smtype != SM_NONE)
  {
    mg_GetKeyword(line, word);
    mg_WordMatchesKeyword(MGJ_KW_SMITER, word);
    if (sscanf(line, "%*s %d", &((*mgni)->smiter)) != 1)
      e_FileLineError(g_LineNumber, g_FullPathName,
                      "can't read number of smoothing iterations");
    else if ((*mgni)->smiter < 1)
      e_FileLineError(g_LineNumber, g_FullPathName,
                      "number of smoothing iterations should be > 0");
  }

  /* read the output format */
  mg_GetKeyword(line, word);
  mg_WordMatchesKeyword(MGJ_KW_OUTFORMAT, word);
  if (sscanf(line, "%*s %s", value) != 1)
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "can't read output format description");
  else if (!strcmp(value, MGJ_D_E_LIB))
    (*mgni)->outf = OUTF_E_LIB;
  else if (!strcmp(value, MGJ_D_SMSH))
    (*mgni)->outf = OUTF_SMSH;
  else
    e_FileLineError(g_LineNumber, g_FullPathName,
                    "unrecognized output format description");
  fclose(g_fp);
  return 1;
}

/* ////////////////////////////////////////////////////////////////////////// */
/* standard parsing functions copied from biao */

static int mg_GetKeyword(char *line, char *word)
{
  if (!e_GetNextLine(line, &g_LineNumber, g_FullPathName, g_fp))
    return 0;
  else if (sscanf(line, "%s", word) == 1)
    return 1;
  else
    return 0;
}

static int mg_WordMatchesKeyword(char *Keyword, char *word)
{
  if (strcmp(Keyword, word) != 0)
  {
    if (mg_IsAKeyword(word))
      e_FileLineError(g_LineNumber, g_FullPathName,
                      "keyword out of sequence");
    else
      e_FileLineError(g_LineNumber, g_FullPathName, "undefined keyword");

    return 0;
  }
  else
    return 1;
}

static int mg_IsAKeyword(char *word)
{
  if (!strcmp(MGJ_KW_TASK, word)
      || !strcmp(MGJ_KW_METHOD, word)
      || !strcmp(MGJ_KW_ALFA, word)
      || !strcmp(MGJ_KW_BETA, word)
      || !strcmp(MGJ_KW_CTRLTYPE, word)
      || !strcmp(MGJ_KW_OPTION, word)
      || !strcmp(MGJ_KW_DLNOPT, word)
      || !strcmp(MGJ_KW_F_GEOM, word)
      || !strcmp(MGJ_KW_F_COARSE, word)
      || !strcmp(MGJ_KW_F_RES, word)
      || !strcmp(MGJ_KW_F_BKGRMESH, word)
      || !strcmp(MGJ_KW_F_BKGRPARAM, word)
      || !strcmp(MGJ_KW_F_CRSPARAM, word)
      || !strcmp(MGJ_KW_F_SRC, word)
      || !strcmp(MGJ_KW_UNIFMPR, word)
      || !strcmp(MGJ_KW_SMTYPE, word)
      || !strcmp(MGJ_KW_SMITER, word)
      || !strcmp(MGJ_KW_OUTFORMAT, word))
  {
    return TRUE;
  }
  else
  {
    return FALSE;
  }
}

static int mg_GetFileName(char *filename, char *line)
{
  char *pFirstQuote;
  char *pLastQuote;
  char string[E_LINE_MAX + 1];


  if (sscanf(line, "%*s %s", string) != 1
      || (pFirstQuote = strchr(string, '"')) == NULL
      || (pLastQuote = strrchr(string, '"')) == pFirstQuote)
  {
    e_FileLineError(g_LineNumber, g_FullPathName, "cannot read filename");
    return 0;
  }

  strncpy(filename, pFirstQuote + 1, (int) (pLastQuote - pFirstQuote) - 1);

  return 1;
}

/* ////////////////////////////////////////////////////////////////////////// */

/* just testing
main(int argc, char *argv[])
{
    char    filename[E_LINE_MAX];
    MGNI   *mgni;

    strcpy(filename, argv[1]);
    if (parse_mgn_into_mgni(&mgni, filename))
    {
	printf("parsing successfull\n");
	printf("task=%d\n", mgni->task);
    }
    else
	printf("errors discovered while parsing\n");
}  */
