/************************************************************

	filter translates an e_lib mdf file into smsh format.
	it works for meshes with TRIANG1 triangles and
	regions of surface elements in LINK1-5.

	use smsh_add_surfelem to add LINK1-5 to a TRIANG1/QUAD1 mesh.

 ************************************************************/

#include "e_lib.h"
#include "mgni.h"
#include "nodelem.h"
#include "boundary.h"

int smsh_add_surfelem(MESH *m)
{
	BND *bnd;					/* the structure that contains the surface elems */
	int nb;						/* number of boundaries detected */
	int *nne,**ne;				/* adjacency information */
	int i;						/* counters */

	/* create boundary */
	nne=e_alloc_i_vect(m->NMeshPoints);
	ne=e_init_frag_i_matr(m->NMeshPoints);
	e_InitMeshPointElemConnection(m,nne,ne);
	Gen2DBoundary(m,nne,ne,&nb,&bnd);
	if (nb!=1)
		e_GeneralError("smsh_add_surfelem: this mesh should only have 1 external boundary");
	if (bnd->nie>4)
		e_GeneralError("smsh_add_surfelem: this mesh should have max 4 holes");

	/* add the outer boundary as LINK1 */
	m->NElemsLink1=bnd->oe.nen;
	m->NodesLink1=e_alloc_cont_i_matr(m->NElemsLink1,2);
	m->TotalNElems+=m->NElemsLink1;
	m->KeywordSet[E_KW_NUM_NELEMS_LINK1]=TRUE;
	m->KeywordSet[E_KW_NUM_NODES_LINK1]=TRUE;
	for (i=0;i<bnd->oe.nen;i++)
	{
		m->NodesLink1[i][0]=bnd->oe.en[i];
		m->NodesLink1[i][1]=bnd->oe.en[(i+1)%bnd->oe.nen];
	}
		
  /*** should also update GEIs, etc ***/

	/* add the inner boundaries as LINK2-5 */
#define ADD_HOLE(n,k) \
	m->NElemsLink##k##=bnd->ie[n].nen; \
	m->NodesLink##k##=e_alloc_cont_i_matr(m->NElemsLink##k##,2); \
	m->TotalNElems+=m->NElemsLink##k##; \
	m->KeywordSet[E_KW_NUM_NELEMS_LINK##k##]=TRUE; \
	m->KeywordSet[E_KW_NUM_NODES_LINK##k##]=TRUE; \
	for (i=0;i<bnd->ie[n].nen;i++) \
	{ \
		m->NodesLink##k##[i][0]=bnd->ie[n].en[i]; \
		m->NodesLink##k##[i][1]=bnd->ie[n].en[(i+1)%bnd->ie[n].nen]; \
	}

	if (bnd->nie>0) { ADD_HOLE(0,2) }
	if (bnd->nie>1) { ADD_HOLE(1,3) }
	if (bnd->nie>2) { ADD_HOLE(2,4) }
	if (bnd->nie>3) { ADD_HOLE(3,5) }

#undef ADD_HOLE

	/* free-up */
	Free2DBoundary(nb,bnd);
	
  return 1;
}

int smsh_output(MESH *m, char *fname)
{
	FILE *fp;					/* file pointer for output */
  int nse;					/* number of surface elements */
	int i,j;					/* counters */
  char *fullfname;	/* output filename with extension */

	/* some checking */
	if (m->TotalNElems!=
    	m->NElemsTriang1+
      m->NElemsQuad1+
			m->NElemsLink1+
			m->NElemsLink2+
			m->NElemsLink3+
			m->NElemsLink4+
			m->NElemsLink5)
		e_GeneralError("smsh_output: mesh contains unsupported elements");

	/* prepare output */
	fullfname = e_alloc_c_vect(strlen(fname)+10);
  sprintf(fullfname,"%s.smsh",fname);
	if ((fp=fopen(fullfname,"w"))==NULL)
		e_GeneralError("smsh_output: can't open output file %s, check writability of directory",
      fullfname);

	/* output of nodes */
	fprintf(fp,"NODES\n");
	fprintf(fp,"%8d\n",m->NMeshPoints);
	for (i=0;i<m->NMeshPoints;i++)
	{
		fprintf(fp,"%8d %17.10E %17.10E %17.10E\n",i+1,
			m->MeshPointCoords[i][0],
			m->MeshPointCoords[i][1],
			m->MeshPointCoords[i][2]);
	}

	/* output of interior elements */
	if (m->NElemsTriang1)
  {
    fprintf(fp,"FEM\n");
	  fprintf(fp,"%8d\n",m->NElemsTriang1);
	  for (i=0;i<m->NElemsTriang1;i++)
	  {
		  fprintf(fp,"%8d %7d %7d %7d %7d %7d\n",i+1,1,11,
			  m->NodesTriang1[i][0]+1,
			  m->NodesTriang1[i][1]+1,
			  m->NodesTriang1[i][2]+1);
	  }
  }
  if (m->NElemsQuad1)
  {
    fprintf(fp,"FEM\n");
	  fprintf(fp,"%8d\n",m->NElemsQuad1);
	  for (i=0;i<m->NElemsQuad1;i++)
	  {
		  fprintf(fp,"%8d %7d %7d %7d %7d %7d %7d\n",i+1,1,10,
			  m->NodesQuad1[i][0]+1,
			  m->NodesQuad1[i][1]+1,
			  m->NodesQuad1[i][2]+1,
        m->NodesQuad1[i][3]+1);
	  }
  }

	/* output of surface elements */
	nse=m->NElemsLink1+m->NElemsLink2+m->NElemsLink3+
		m->NElemsLink4+m->NElemsLink5;
	if (nse==0)
		e_GeneralWarning("smsh_output: mesh contains no surface elements");
	else
	{
		fprintf(fp,"FEM\n");
		j=0;
		fprintf(fp,"%8d\n",nse);
		for (i=0;i<m->NElemsLink1;i++,j++)
			fprintf(fp,"%8d %7d %7d %7d %7d\n",j+1,2,20,m->NodesLink1[i][0],m->NodesLink1[i][1]);
		for (i=0;i<m->NElemsLink2;i++,j++)
			fprintf(fp,"%8d %7d %7d %7d %7d\n",j+1,3,20,m->NodesLink2[i][0],m->NodesLink2[i][1]);
		for (i=0;i<m->NElemsLink3;i++,j++)
			fprintf(fp,"%8d %7d %7d %7d %7d\n",j+1,4,20,m->NodesLink3[i][0],m->NodesLink3[i][1]);
		for (i=0;i<m->NElemsLink4;i++,j++)
			fprintf(fp,"%8d %7d %7d %7d %7d\n",j+1,5,20,m->NodesLink4[i][0],m->NodesLink4[i][1]);
		for (i=0;i<m->NElemsLink5;i++,j++)
			fprintf(fp,"%8d %7d %7d %7d %7d\n",j+1,6,20,m->NodesLink5[i][0],m->NodesLink5[i][1]);
		if (j!=nse)
			e_GeneralError("smsh_output: sanity check on no of surf elems failed");
	}

	fclose(fp);

	/* output of decomp data */
	if (m->SubStructSet[E_SS_NUM_DECOMP])
	{
		sprintf(fullfname,"%s_%d.nsplit",fname,m->Decomp.NSubDoms);
		if ((fp=fopen(fullfname,"w"))==NULL)
			e_GeneralError("smsh_output: can't open output file %s, check writability of directory",
        fullfname);

		j=0;
		fprintf(fp,"%8d\n",m->TotalNElems);
		for (i=0;i<m->NElemsTriang1;i++,j++)
			fprintf(fp,"%8d %7d\n",j+1,m->Decomp.SubDomsTriang1[i]);
		for (i=0;i<m->NElemsQuad1;i++,j++)
			fprintf(fp,"%8d %7d\n",j+1,m->Decomp.SubDomsQuad1[i]);
		for (i=0;i<m->NElemsLink1;i++,j++)
			fprintf(fp,"%8d %7d\n",j+1,m->Decomp.SubDomsLink1[i]);
		for (i=0;i<m->NElemsLink2;i++,j++)
			fprintf(fp,"%8d %7d\n",j+1,m->Decomp.SubDomsLink2[i]);
		for (i=0;i<m->NElemsLink3;i++,j++)
			fprintf(fp,"%8d %7d\n",j+1,m->Decomp.SubDomsLink3[i]);
		for (i=0;i<m->NElemsLink4;i++,j++)
			fprintf(fp,"%8d %7d\n",j+1,m->Decomp.SubDomsLink4[i]);
		for (i=0;i<m->NElemsLink5;i++,j++)
			fprintf(fp,"%8d %7d\n",j+1,m->Decomp.SubDomsLink5[i]);

		fclose(fp);
	}

  e_free(fullfname);

	return 1;
}
