/**************************************************************

    Include file for the mesh generation function 'mgni'
    Questions? ask Jelle

    Don't forget to include e_lib.h before you include
    this file!

**************************************************************/

#ifndef MGNI_H
#define MGNI_H

#include "e_lib.h"

/* local mesh control structures */
typedef struct
{
  int NCntlPoints;
  double *InnerRadius;
  double *OuterRadius;
  double **Coords;
  double *Param;
}
PARAMCNTL;

typedef struct
{
  int NCntlSets;
  PARAMCNTL *ParamCntls;
}
PARAMCNTLSETS;

/* the boundary structures */
#define EDGE_INCR	10
typedef struct
{
  int nen;                      /* number of edge nodes */
  int *en;                      /* the edge node indices */
}
EDGE;

typedef struct
{
  EDGE oe;                      /* the outer edge */
  int nie;                      /* number of inner edges (holes) */
  EDGE *ie;                     /* the inner edges */
}
BND;

/* MGN Information Structure */

#define	MGJ_NFI		    7

typedef struct
{
  /* task description */
  int task;                     /* task */
  int meth;                     /* method */
  double alfa;                  /* the alfa used in Delaunay */
  double beta;                  /* the beta */
  int ctrl;                     /* control type */
  int opt;                      /* mesh option */
  int dln;                      /* delaunay options */
  int smtype;                   /* smoothing type */
  int smiter;                   /* no of smoothing iterations */
  int outf;                     /* output format */

  /* input from file */
  int loadfiles;                /* load files or take ones specified */
  char fname[MGJ_NFI][E_LINE_MAX]; /* filenames */

  /* input specified */
  MESH *mesh;                   /* input mesh structure */
  int geomset;                  /* geometry information set or not */
  MESH *outputmesh;             /* output mesh structure */
  MESH *bgmesh;                 /* background mesh structure */
  PARAMCNTLSETS *pcs;           /* local mesh size countrol sets */
  double h;											/* uniform mesh size */

  /* other variables */
  int gra;                      /* graphics output */
  int doa;                      /* printing of debug output */
  int pps;                      /* print progress statement lines */
}
MGNI;

/* possible values for task */
#define TASK_NONE         0     /* do nothing */
#define TASK_MESHGEN      1     /* not used */
#define TASK_REMESH       2     /* keep boundaries, remesh interior */
#define TASK_REFINE       3     /* keep boundaries & interior nodes */
#define TASK_EBE          4     /* element by element refinement */
#define TASK_BND          5     /* generate mesh from boundary */
#define TASK_POSTPROC     6     /* only mesh post-processing */

/* possible values for meth (method) */
#define METH_ADV_FRONT       0     /* advancing front */
#define METH_DELAUNAY        1     /* Delaunay triangulation */
#define METH_PAVING          2     /* paving method */
#define METH_REG_GRID        3     /* regular grid method (aka UG) */
#define METH_NONE            4     /* only with post-processing */

/* possible values for ctrl (control method) */
#define CTRL_BACKGROUND      0     /* background controlled meshing */
#define CTRL_MPR             1     /* mesh parameter controlled meshing */
#define CTRL_SOURCES	     2     /* external sources controlled meshing */
#define CTRL_UNIFORM         3     /* uniform meshing */

/* possible values for dln (Delaunay option) */
#define DLN_CENTR            0     /* centroid insertion */
#define DLN_OPT              1     /* projective point placement */
#define DLN_CENTR_RAD        2     /* centroid based on radians not edges */
#define DLN_CHEW1            3     /* chew1 delaunay refinement */

/* possible values for opt (quad conversion option) */
#define CONV_OPT_NONE             0     /* no conversion */
#define CONV_OPT_FISSION_TO_QUAD  1     /* conversion by fusion with fission element */
#define CONV_OPT_RANK_TO_QUAD     2     /* Rank method, fusion and halfing */
#define CONV_OPT_HALF_TO_QUAD     3     /* halfs all elements in the mesh */
#define CONV_OPT_SPLIT_TO_TRIANG  4     /* splitting quads into triangles */

/* possible values for smtype (smoothing type) */
#define SM_NONE               0     /* no smoothing */
#define SM_LAPLACIAN          1     /* Biao's Laplacian smoothing */
#define SM_LAPLACIAN_NDE      2     /* same but without diagonal exchange */
#define SM_NODAL_AVG          3     /* pure nodal averaging (dtr only) */
#define SM_PAVING             4     /* the smoothing required for paving */

/* possible values for outf (output format) */
#define OUTF_E_LIB            0     /* e_library */
#define OUTF_SMSH             1     /* format for Frederic Magoules */

/* filenames vector indices */
#define MGJ_FI_GEOM           0     /* geometric model */
#define MGJ_FI_COARSE         1     /* coarse mesh */
#define MGJ_FI_RES            2     /* resulting mesh */
#define MGJ_FI_BKGRMESH       3     /* background mesh */
#define MGJ_FI_BKGRPARAM      4     /* background mesh parameters */
#define MGJ_FI_CRSPARAM       5     /* coarse mesh parameters */
#define MGJ_FI_SRC            6     /* control sources */

/* allocates space for the MGNI structure and fill's in all
   members with default values; the member meshes are set to NULL */
void mgni_init(MGNI **mgni);

/* frees up all the member structures */
void mgni_free(MGNI *mgni);

/* load all files as specified by the filenames */
void mgni_loadfiles(MGNI *mgni);

/* check the validity of the info specified in mgni:
   this is only a superficial check. it is not checked whether
    - all structures are filled in
    - filenames are valid
   the developer using this function should know what he/she is doing! */
void mgni_check(MGNI *mgni);

/* the mesh generation function */
void mgni_main(MGNI *mgni);

/* generate meshes from a bnd definition only */
int meshgen_from_bnd(double **crd, int nc, double *mpr, BND * bnd,
                     MGNI *mgni, int keep_internal);



#endif
