/* include files */
#include "e_lib.h"
#include "eg_lib.h"


#include "mgn.h"
#include "mgn_graphics.h"
#include "dtr.h"
#include "split.h"
#include "fusion.h"
#include "postpro.h"

int meshgen_from_bnd(double **crd, int nc, double *mpr, BND *bnd,
                     MGNI *mgni, int keep_internal)
{
  /* generate a proper mesh from a boundary.
     the mgni has to be filled in. */

  double **coords;
  int np;
  int tp;
  int **tr;                     /* triangle matrix */
  int ntr;                      /* number of triangles */
  int ttr;                      /* allocated spaces triangles */
  double *crsnmpr;              /* original mesh nodal parameters */
  int *hull;                    /* the convex hull */
  int nhull;                    /* number of points in the hull */
  int **tpnl;                   /* triangle per node listing */
  int t, i;                     /* counters */
  MTRF mtrf;                    /* mapping transformation data */

  mtrf.used = FALSE;

  /* check the purpose of the function */
  if (mgni->task != TASK_BND)
  {
    printf("the function meshgen_from_bnd can only be called for TASK_BND");
    return FALSE;
  }
  if (mgni->ctrl != CTRL_MPR)
  {
    printf
      ("the function meshgen_from_bnd can only be called for CTRL_MPR");
    return FALSE;
  }
  if (mgni->meth != METH_DELAUNAY)
  {
    printf
      ("the function meshgen_from_bnd can only be called for METH_DELAUNAY");
    return FALSE;
  }

  /* initialise eg_lib */
  eg_Init();

  /* initial values */
  tr = NULL;
  hull = NULL;
  tpnl = NULL;
  DebugOutput = mgni->doa;
  ShowGraphics = mgni->gra;
  PrintProgress = mgni->pps;

  if (ShowGraphics)
  {
    mgnInitGraphics(G2_WINDOW_X, G2_WINDOW_Y,
                    nc, crd);
  }

  /* parse the coarse mesh */
  ntr = 0;
  np = nc;
  inittpnl(&tpnl, np);
  coords = e_alloc_cont_d_matr(np, 3);
  crsnmpr = e_alloc_d_vect(np);
  for (t = 0; t < np; t++)
  {
    for (i = 0; i < 3; i++)
      coords[t][i] = crd[t][i];
    crsnmpr[t] = mpr[t];
  }
  if (keep_internal)
    f_connect_mesh(coords, np, &tr, &ntr, &hull, &nhull, tpnl, &mtrf);
  else
    f_connect_bnd_nodes(coords, np, bnd, &tr, &ntr, &hull,
                        &nhull, tpnl, &mtrf);
  ttr = ntr;
  tp = np;

  /* boundary refinement and recovery by halfwaynode method */
  boundary_refinement_and_recovery(&coords, &np, &tp, CTRL_MPR,
                                   mgni->geomset, mgni->mesh, NULL, &crsnmpr,
                                   mgni->h, bnd, &tpnl, &tr, &ntr, &ttr, &hull,
                                   &nhull, (mgni->opt==CONV_OPT_FISSION_TO_QUAD));

  /* setting the areacodes */
  setareacode(bnd, tpnl, tr, ntr);

  /* refine the interior of the domain */
  mtrf.used = FALSE;            /* for compatibility */
  f_refine_del_mesh(&coords, &np, &tp, &tr, &ntr,
                    &ttr, &hull, &nhull,
                    mgni->ctrl, mgni->alfa, mgni->beta, NULL,
                    &crsnmpr, mgni->h, &tpnl, NULL, mgni->dln, &mtrf);

  /* smoothing by nodal averaging before output */
  if (mgni->smtype == SM_NODAL_AVG)
    smooth_by_nodal_averaging(coords, np, tr, ntr, tpnl, mgni->smiter);

  /* write the output */
  mgni->outputmesh = (MESH *) e_calloc(1, sizeof(MESH));
  e_NullifyMesh(mgni->outputmesh);
  create_outputmesh(coords, np, tr, ntr, mgni->outputmesh);

  (mgni->outputmesh)->SubStructSet[E_SS_NUM_MESHPARAM] = TRUE;
  (mgni->outputmesh)->MeshParam.NodalMeshParamsSet = TRUE;
  (mgni->outputmesh)->MeshParam.NodalMeshParams = e_alloc_d_vect(np);
  for (i = 0; i < (mgni->outputmesh)->NNodes; i++)
    (mgni->outputmesh)->MeshParam.NodalMeshParams[i] = crsnmpr[i];

  /* conversion to quads if required */
  if (mgni->opt == CONV_OPT_HALF_TO_QUAD)
    split2quad(mgni->outputmesh);
  else if (mgni->opt == CONV_OPT_FISSION_TO_QUAD)
  {
    if ((mgni->outputmesh)->NElemsTriang1 % 2 != 0)
    {
      fuse2quad(mgni->outputmesh, 0);
      split2quad(mgni->outputmesh);
    }
    else
    {
      fuse2quad(mgni->outputmesh, 1);
    }
  }

  /* perform the necessary smoothing */
  if (mgni->smtype == SM_NONE)
    PostProcessMesh(mgni->outputmesh, mgni->smiter, mgni->smtype);

  e_free(crsnmpr);
  e_free(hull);
  e_free_cont_i_matr(ttr, tr);
  e_free_cont_d_matr(tp, coords);
  freetpnl(tpnl, tp);

  return TRUE;
}
