#include "pav_main.h"
#include "bgdata.h"


static int AngleRangeValidity(void);
static int MeshExtension(MESH *mesh, PAVEDMESH *pm);
static void ClearSmoothBit(PAVEDMESH *mesh);

/* ######################################################################### */

/* 
 * Check the validity of angle ranges
 */
static int AngleRangeValidity(void)
{

  if(   (S1 < 0.0)
     || (S2 < 0.0)
     || (S3 < 0.0)
     || (S4 < 0.0)
     || (S5 < 0.0)
     || (S6 < 0.0)
     || (ANGLE_90  + S1 > ANGLE_180 - S2)
     || (ANGLE_180 + S3 > ANGLE_270 - S4)
     || (ANGLE_270 + S5 > ANGLE_360 - S6) )
  {
    return(FALSE);
  }
  else
  {
    return(TRUE);
  }

}


/*
 * Determine the mesh extensions
 */
static int MeshExtension(MESH *mesh, PAVEDMESH *pm)
{
  double maxx, maxy, minx, miny, x, y, d;
  int    i;

  maxx = mesh->MeshPointCoords[0][0];
  maxy = mesh->MeshPointCoords[0][1];
  minx = mesh->MeshPointCoords[0][0];
  miny = mesh->MeshPointCoords[0][1];
  for(i = 1; i < mesh->NMeshPoints; i++)
  {
    x = mesh->MeshPointCoords[i][0];
    y = mesh->MeshPointCoords[i][1];
    if(x < minx)
      minx = x;
    if(x > maxx)
      maxx = x;
    if(y < miny)
      miny = y;
    if(y > maxy)
      maxy = y;
  }

  if(E_ABS(maxx - minx) < ZERO)
  {
    e_GeneralWarning("Zero extension in X direction, no mesh is generated !");
    return(FALSE);
  }

  if(E_ABS(maxy - miny) < ZERO)
  {
    e_GeneralWarning("Zero extension in Y direction, no mesh is generated !");
    return(FALSE);
  }

  d = (maxx - minx > maxy - miny) ? (maxx - minx) : (maxy - miny);
  d *= 0.2;

  /* store it */
  pm->Maxx = maxx + d;
  pm->Maxy = maxy + d;
  pm->Minx = minx - d;
  pm->Miny = miny - d;

  return(TRUE);
}


/*
 * Delete smoothing bit from all loops.
 */
static void ClearSmoothBit(PAVEDMESH *mesh)
{
  LOOP *curr, *start;

  curr = start = mesh->Fronts;
  do
  {
    DEL_BIT(curr->Type, MUST_SMOOTH);
    curr = curr->Next;
  }
  while(curr != start);
}


static void RowAdjustment(PAVEDMESH *mesh, LOOP *loop)
{
  int seamed, adjusted;
  int i, output;

  MeshSmoothing(mesh, loop, N_SMOOTH_ITERATION, FALSE);

  do
  {
    seamed   = FALSE;
    adjusted = FALSE;
    
    do
    {
      output = FALSE;

      /* check every node for seaming */
      for(i = 0; i < loop->NNodes; i++)
      {
        /* try to seam at loop head */
        if(SeamingAtLoopHead(mesh, loop, FALSE) == P_SEAM_DONE)
        {
          MeshSmoothing(mesh, loop, N_SMOOTH_ITERATION, FALSE);
          /* if we seamed, remember it */
          seamed = TRUE;
          output = TRUE;
        }
        /* go to next node */
        StepForward(loop, FALSE);
      }
    }
    while(output == TRUE);
 
    do
    {
      output = FALSE;
      /* classify all nodes in the loop */
      ClassifyALoop(mesh->NodeCoords, loop);
      /* check whether wedge or tuck should be inserted */
      output = WedgeTuckInsert(mesh, loop);
      /* if something was done smooth */
      if(output == P_YES)
      {
        MeshSmoothing(mesh, loop, N_SMOOTH_ITERATION, FALSE);
        adjusted = TRUE;
      }
    }
    while(output == P_YES);
  }
  while(seamed == TRUE || adjusted == TRUE);

}


/*
 * Pave the given mesh
 */
void pav_main(MGNI *mgni)
{
  PAVEDMESH *pm;
  LOOP      *loop;
  int        output, end;
  LOOP      *curr, *start;
  BG_DATA   *bgd;

  /* check that the angle ranges are valid */
  if(AngleRangeValidity() == FALSE)
  {
    e_GeneralWarning("PaveMesh: invalid angle range(s), no mesh is generated !");
    return;
  }

  /* initialise the paving mesh, 
     no undo for these operations  */
  pm = InitPavedMesh();

  /* check mesh dimensions */
  if(MeshExtension(mgni->mesh, pm) == FALSE)
  {
    e_free(pm);
    return;
  }

  pm->ctrl = mgni->ctrl;
  /* if backgorund or sources control */
  if(mgni->ctrl == CTRL_BACKGROUND || mgni->ctrl == CTRL_SOURCES)
  {
    get_bg_data(&(pm->bgd), mgni->bgmesh, mgni->pcs);
  }
  else
  {
    bgd = NULL;
  }

  if(pm->ctrl == CTRL_UNIFORM)
  {
    pm->uniform_h = mgni->h;
  }

  /* generate the paving boundaries, external, internal */
  InitFronts(mgni->mesh, mgni->geomset, pm);

  /* pave the mesh */
  while(1)
  {
    ClearSmoothBit(pm);

    /* take the actual front loop */
    loop = pm->Fronts;
    ADD_BIT(loop->Type, MUST_SMOOTH);

#ifdef PAVING_DEBUG
    e_GeneralWarning("----- No nodes in loop: %d -----", loop->NNodes);
#endif
    
    /* if there are more than 6 nodes in the list */
    if(loop->NNodes > 6)
    {
      /* start up the row */
      output = InitRow(pm, loop);
      
      /* while everything is fine */
      while(output == P_ROW_CONT)
      {
        /* check whether this is the end of the node */
        end = (CHK_BIT(loop->Nodes->Flag, NODE_SIDE)) ? 0 : 1;
/*        end = (CHK_BIT(loop->Nodes->Flag, NODE_END)) ? 1 : 0; */

        /* generate nodes, elements */
        output = FaceFactory(pm, loop);
        /* if everything is fine */
        if((output == P_FACE_ACCEPT) && !end)
        {
          /* go to the next node */
          StepForward(loop, FALSE);
          /* delete undo information */
          EmptyUndoQueue();
          output = P_ROW_CONT;
        }
        /* the row must be finished now */
        else
        {
          output = P_ROW_FINISHED;
        }
      }
      /* if there is less than 6 nodes */
      if(loop->NNodes < 6)
      {
#ifdef PAVING_DEBUG
        e_GeneralWarning("Resolution is required");
        e_GeneralWarning("No of remainder nodes: %d", loop->NNodes);
#endif
        /* close the loop appropriately */
        LoopClosure(pm, loop);
        /* if there is no more front, THE END */
        if(pm->Fronts == NULL)
          break;
      }
      else
      {
        /* row adjustment */
        curr = start = pm->Fronts;
        do
        {
          if(CHK_BIT(curr->Type, MUST_SMOOTH))
          {
            RowAdjustment(pm, curr);
          }
          curr = curr->Next;
        }
        while(curr != start);

      }
    }
    else
    {
#ifdef PAVING_DEBUG
      e_GeneralWarning("Resolution is required");
      e_GeneralWarning("No of remainder nodes: %d", loop->NNodes);
#endif
      /* close the loop appropriately */
      LoopClosure(pm, loop);
      /* if there is no more front, THE END */
      if(pm->Fronts == NULL)
        break;
    }
    /* take the next loop */
    NextLoop(pm, FALSE);
  }

  /* finally smooth the whole mesh */
  SmoothInterior(pm, NULL, mgni->smiter, FALSE);

  MeshClosure(pm);

  /* convert it to MDF mesh */
  mgni->outputmesh = ConvertToMesh(pm);

  if(bgd != NULL)
  {
    free_bg_data(bgd);
  }

}




