#ifndef PAVING_H
#define PAVING_H

#include "e_lib.h"
#include "mgni.h"
#include "bgdata.h"

/* ######################################################################### *\
 *                          Paving related constants                         *
\* ######################################################################### */

/* element size is controlled globally */
#define GLOBAL_SIZE_CONTROL


#define PAVING_DIMENSION    2

/* 
 * Maximum number of generated nodes at a node.
 * Number of generated nodes at:
 *     NODE_END:      0
 *     NODE_SIDE:     1
 *     NODE_CORNER:   3
 *     NODE_REVERSAL: 5
 */
#define MAX_GEN_NODE        5

/* there are 4 classes of nodes */
#define NO_NODE_CLASSES     4

/* node flags */
#define NODE_INTERNAL       1
#define NODE_FRONT          2
  /* classes of nodes */
#define NODE_END            4
#define NODE_SIDE           8
#define NODE_CORNER         16
#define NODE_REVERSAL       32
  /* other node flags */
#define NODE_NEW            64
#define NODE_BND            128 

#define ALL_NODE_TYPE       (NODE_END+NODE_SIDE+NODE_CORNER+NODE_REVERSAL)

/* loop types */
#define LOOP_EXTERNAL       1
#define LOOP_INTERNAL       2
/* warning for loop that it has to start off */
#define MUST_RESOLV         4
/* smoothing required on the loop */
#define MUST_SMOOTH         8


/* maximum number of end nodes that can be stored to determine some shape */
#define MAX_END_NODE        100

/* angle definitions in human form :-) */
#define ANGLE_10            (M_PI / 18.0)
#define ANGLE_45            (M_PI / 4.0)
#define ANGLE_90            (M_PI / 2.0)
#define ANGLE_180           (M_PI)
#define ANGLE_270           (M_PI * 3.0 / 2.0)
#define ANGLE_360           (M_PI * 2.0)

/* expansion and reduction angle thresholds */
#define EXPAND_ANGLE        (M_PI + (M_PI / 60.0))
#define REDUCE_ANGLE        (M_PI - (M_PI / 60.0))

/* maximum number of segments in proximity */
#define P_MAX_SEGMENT       3

/* maximum number of candidates considered for side merge */
#define P_MAX_CAND          3

/* angle tolerances for classification */
#define S1                  (M_PI/4.5)  /* 40 deg */
#define S2                  (M_PI/6.0)  /* 30 deg */
#define S3                  (M_PI/4.0)  /* 45 deg */
#define S4                  (M_PI/4.0)  /* 45 deg */
#define S5                  (M_PI/18.0) /* 10 deg */
#define S6                  (M_PI/2.5)  /* 72 deg */

/* shape definitions */
#define SHAPE_RECTANGLE     4
#define SHAPE_TRIANGLE      3
#define SHAPE_SEMICIRCLE    2

/* number of iterations for smoothing */
#define N_SMOOTH_ITERATION  10

/* number of internal levels for smoothing */
#define N_SMOOTH_LEVEL      2

/* angles for seaming */
#define EPS1_ANGLE          (M_PI/5.1528571)  /* 35 deg with regular node */
#define EPS2_ANGLE          (M_PI/18.0)       /* 10 deg with irregular node */

/* the allowable ratio between to adjacent side lengths in a quad */
#define MAX_SIDE_RATIO      2.0

/* 
 * In case of seaming if the ratio of the side langth is grater
 * than this value then a wedge is inserted 
 */
#define WEDGE_SIDE_RATIO    2.5

/* ######################################################################### */

/* used as a zero threshold */
#define ZERO                1e-6

/* allocation units */
#define ALLOC_NODEPOOL_SIZE 200 /* allocation unit for the dynamic nodes */
#define ALLOC_NODE_SIZE     16  /* allocation unit for new nodes */
#define ALLOC_ELEM_SIZE     16  /* allocation unit for new elements */
#define ALLOC_NE_SIZE       4   /* allocation unit for NodeElem */
#define ALLOC_SMOOTH_SIZE   200 /* allocation unit for smoothing nodes */

/* function return values */
#define P_ERROR             0
#define P_SUCCESS           1

#define P_NO                0
#define P_YES               1

#define P_BAD               0
#define P_GOOD              1


#define P_ROW_FINISHED      1
#define P_ROW_CONT          2

#define P_FACE_ACCEPT       1
#define P_FACE_RESOLV       2
#define P_FACE_UNDO         4

#define P_NO_SEAMING        0
#define P_SEAM_DONE         1
#define P_DELAY_SEAMING     2

/* bit operations */
#define CHK_BIT(v, b)       ((v) & (b))
#define ADD_BIT(v, b)       (v) = (v) | (b)
#define DEL_BIT(v, b)       (v) = (v) & (~(b))

#define CHK_BND(v)          ((v < 0) ? 1 : 0)
#define CHK_FRONT(v)        ((v > 0) ? 1 : 0)


/* ######################################################################### *\
 *                              Type definition                              *
\* ######################################################################### */

/* nodes in a loop */
typedef struct node_s
{
  struct node_s *Next;       /* next node in the loop */
  struct node_s *Prev;       /* previous node in the loop */
  int            Flag;       /* node properties */
  int            Node;       /* node number, reference to coordinates */
  double         Angle;      /* angle at the node */
  double         ReqLength;  /* projected length */
  int            Stage;      /* no. of stage which was completed */
  int            RefPrev;    /* through the stages stores the original */
  int            RefNext;       /* previous and next nodes */
  int            ProjNode;   /* the origin of the projection, */
                                /* mainly used for smoothing */
}
NODE;


/* circular loop, the last element points back to the first one */
typedef struct loop_s
{
  int            Type;      /* type of the loop + must resolv bit*/
  int            Primitive; /* if it is a primitive shape */
  struct loop_s *Next;      /* points to the next front loop */
  struct loop_s *Prev;      /* points to the previous front loop */
  int            NNodes;    /* no of nodes in the list */
  NODE          *Nodes;     /* loop head, all nodes accessed through this */
}
LOOP;


typedef struct
{
  double     Maxx, Maxy;    /* coordinates of the bounding box of the mesh */
  double     Minx, Miny;

  int        ctrl;
  double     uniform_h;
  BG_DATA   *bgd;

#ifdef GLOBAL_SIZE_CONTROL
  int        NBnd;          /* number of boundary nodes */
  double    *BndLenParam;   /* required length parameter at the boundary */
#endif

  int        NNodes;        /* number of nodes in the mesh */
  double   **NodeCoords;    /* coordinates of nodes */
  int       *NodeFlag;      /* node types, mesh bnd, on front, inside */

  int        NQuads;        /* number of quad elements */
  int      **QuadNodes;     /* the nodes of the quad elements */

  int       *NNodeElem;     /* node-element connectivity */
  int      **NodeElem;

  LOOP      *Fronts;        /* the list of fronts */
}
PAVEDMESH;

/* atomic operations */
#include "pav_operation.h"

/* graphical interface functions */
#define   PAVING_GRAPHICS

/* control.c */
void pav_main(MGNI *mgn);

/* preproc.c */
void InitFronts(MESH *mesh, int geomset, PAVEDMESH *pm);

/* postproc.c */
MESH * ConvertToMesh(PAVEDMESH *pm);

/* rowchoice.c */
int ClassifyANode(double **coords, int prev, int curr, int next, 
                  int *state, double *angle);
int ClassifyALoop(double **coords, LOOP *loop);
void RemoveTypeAmbiguouity(NODE *curr, int NNodeElem);
void FinalNodeTypeInLoop(PAVEDMESH *mesh, LOOP *loop);
int InitRow(PAVEDMESH *mesh, LOOP *loop);


/* paving.c */
int FindQuadOnSide(PAVEDMESH *mesh, int n1, int n2);
int FindNodeInQuad(int *Nodes, int n, int incr);
double AngleCalc(double xn1, double yn1, double xn2, double yn2,
                 double xn3, double yn3);
double AngleDivider(double xn1, double yn1, double xn2, double yn2,
                    double div, double angle);
#ifdef GLOBAL_SIZE_CONTROL
double ProjLenCalc(PAVEDMESH*mesh, int ncurr);
#else
double ProjLenCalc(PAVEDMESH*mesh, int nprev, int ncurr, int nnext,
                   double angle, int type);
#endif

int CreateEntities(PAVEDMESH *mesh, LOOP *loop, int genelem);
int FaceFactory(PAVEDMESH *mesh, LOOP *loop);
int InitElement(PAVEDMESH *mesh, LOOP *loop);

int GetNodeType(int nodeclass);

double QuadShapeQuality(double **Coords, int n1, int n2, int n3, int n4);
int QuadSideLengthQuality(double **Coords, int n1, int n2, int n3, int n4);

/* closure.c */
void LoopClosure(PAVEDMESH *mesh, LOOP *loop);
void MeshClosure(PAVEDMESH *mesh);

/* smooth.c */
int SmoothInterior(PAVEDMESH *mesh, LOOP *loop, int nsmoothiter, int undo);
int MeshSmoothing(PAVEDMESH *mesh, LOOP *loop, int nsmoothiter, int undo);

/* seaming */
int SeamingAtLoopHead(PAVEDMESH *mesh, LOOP *loop, int undo);

/* adjust.c */
int WedgeTuckInsert(PAVEDMESH *mesh, LOOP *loop);


#endif






