#include "pav_operation.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#ifdef PAVING_GRAPHICS
#include "pav_graphics.h"
#endif

/* always redraw the whole mesh, 
   if undefined only inremental drawing is done */
#define DRAW_ALL 

/* ############################################################################ */
/*                            Nothing to change below                           */
/* ############################################################################ */

#define UNDO_CREATE_NODE           1
#define UNDO_DELETE_NODE           2
#define UNDO_CHANGE_NODE_COORD     3

#define UNDO_CREATE_ELEMENT        4
#define UNDO_DELETE_ELEMENT        5
#define UNDO_CHANGE_ELEMENT_NODES  6

#define UNDO_SET_LOOP_HEAD         7
#define UNDO_STEP_FORWARD          8
#define UNDO_STEP_BACKWARD         9
#define UNDO_REMOVE_LOOP_NODE      10
#define UNDO_INSERT_LOOP_NODE      11
#define UNDO_NEXT_LOOP             12
#define UNDO_PREV_LOOP             13
#define UNDO_ADD_LOOP              14
#define UNDO_REMOVE_LOOP           15
#define UNDO_MERGE_CUT_LOOP        16

#define UNDO_STAGE_INCR            17
#define UNDO_STAGE_DECR            18

#define UNDO_MARKER                100

typedef struct undo_node
{
  struct undo_node *Next;
  int               Operation;
  int               Entity;
  double            X, Y;
  int               N[4];
  NODE             *Node;
  NODE             *RefNode;
  LOOP             *Loop;
}
UNDO;

/* list of undo nodes */
static UNDO *Undo = NULL;

/* pool of available nodes */
static UNDO *UndoPool = NULL;

/* pool of nodes */
static NODE *NodePool = NULL;


/* local undo functions */
static UNDO *GetNewUndo();
static void  AppendUndo(int operation, ...);


static int AddElemToNodeElem(PAVEDMESH *mesh, int n, int elem);
static int RemoveElemFromNodeElem(PAVEDMESH *mesh, int n, int elem);

/* ############################################################################# */
/*                             Undo functions                                    */
/* ############################################################################# */


static UNDO *GetNewUndo()
{
  UNDO *newnode;
  int   i;

  if(UndoPool == NULL)
  {
    for(i = 0; i < ALLOC_UNDOPOOL_SIZE; i++)
    {
      newnode = (UNDO *)e_calloc(1, sizeof(UNDO) );

      newnode->Next = UndoPool;
      UndoPool = newnode;
    }
  }

  newnode  = UndoPool;
  UndoPool = UndoPool->Next;

  return(newnode);
}


static void AppendUndo(int operation, ...)
{
  UNDO *node;
  va_list ap;

  /* get a new node */
  node = GetNewUndo();
  /* connect it to the list */
  node->Next      = Undo;
  Undo            = node;

  node->Operation = operation;
  
  va_start(ap, operation);
  switch(operation)
  {
    case UNDO_CREATE_NODE:
    case UNDO_CREATE_ELEMENT:
      node->Entity  = va_arg(ap, int);
      break;
    case UNDO_DELETE_NODE:
      node->Entity  = va_arg(ap, int);
      node->X       = va_arg(ap, double);
      node->Y       = va_arg(ap, double);
      node->N[0]    = va_arg(ap, int);
      break;
    case UNDO_CHANGE_NODE_COORD:
      node->Entity  = va_arg(ap, int);
      node->X       = va_arg(ap, double);
      node->Y       = va_arg(ap, double);
      break;

    case UNDO_DELETE_ELEMENT:
      node->Entity  = va_arg(ap, int);
      node->N[0]    = va_arg(ap, int);
      node->N[1]    = va_arg(ap, int);
      node->N[2]    = va_arg(ap, int);
      node->N[3]    = va_arg(ap, int);
      break;
    case UNDO_CHANGE_ELEMENT_NODES:
      node->Entity  = va_arg(ap, int);
      node->N[0]    = va_arg(ap, int);
      node->N[1]    = va_arg(ap, int);
      break;

    case UNDO_SET_LOOP_HEAD:
      node->Loop    = va_arg(ap, LOOP *);
      node->Node    = va_arg(ap, NODE *);
      break;
    case UNDO_STEP_FORWARD:
    case UNDO_STEP_BACKWARD:
      node->Loop    = va_arg(ap, LOOP *);
      break;
    case UNDO_REMOVE_LOOP_NODE:
      node->Node    = va_arg(ap, NODE *);
      /* fall through */
    case UNDO_INSERT_LOOP_NODE:
      node->Loop    = va_arg(ap, LOOP *);
      break;
    case UNDO_NEXT_LOOP:
    case UNDO_PREV_LOOP:
      break;
    case UNDO_REMOVE_LOOP:
    case UNDO_ADD_LOOP:
      node->Loop    = va_arg(ap, LOOP *);
      break;
    case UNDO_MERGE_CUT_LOOP:
      node->Node    = va_arg(ap, NODE *);
      node->RefNode = va_arg(ap, NODE *);
      break;
    case UNDO_STAGE_INCR:
    case UNDO_STAGE_DECR:
      node->Node    = va_arg(ap, NODE *);
      break;
    case UNDO_MARKER:
      break;
    default:
      e_GeneralError("AppendUndo: unknown undo operation !");
      break;
  }

  va_end(ap);
}


/* Public Undo functions */
/* ----------------------------------------------------------------------------- */

/* put down a marker */
void UndoMarker(void)
{
  AppendUndo(UNDO_MARKER);
}

/* Undo all operations until a marker */
void UndoUntilMarker(PAVEDMESH *mesh)
{
  UNDO *u;
  int   marker = FALSE;

  if(Undo == NULL)
    return;

  do
  {
    switch(Undo->Operation)
    {
      case UNDO_CREATE_NODE:
        DeleteNode(mesh, Undo->Entity, FALSE);
        break;
      case UNDO_DELETE_NODE:
        CreateNode(Undo->X, Undo->Y, Undo->N[0], mesh, FALSE);
        break;
      case UNDO_CHANGE_NODE_COORD:
        ChangeNodeCoord(Undo->Entity, Undo->X, Undo->Y, mesh, FALSE);
        break;
      case UNDO_CREATE_ELEMENT:
        DeleteElement(Undo->Entity, mesh, FALSE);
        break;
      case UNDO_DELETE_ELEMENT:
        CreateElement(Undo->N[0], Undo->N[1], 
                      Undo->N[2], Undo->N[3], mesh, FALSE);
        break;
      case UNDO_CHANGE_ELEMENT_NODES:
        ChangeElementNode(Undo->Entity, Undo->N[0], Undo->N[1], mesh, FALSE);
        break;

      case UNDO_SET_LOOP_HEAD:
        SetLoopHead(Undo->Loop, Undo->Node, FALSE);
        break;
      case UNDO_STEP_FORWARD:
        StepBackward(Undo->Loop, FALSE);
        break;
      case UNDO_STEP_BACKWARD:
        StepForward(Undo->Loop, FALSE);
        break;
      case UNDO_REMOVE_LOOP_NODE:
        InsertBehindFrontNode(mesh, Undo->Node, Undo->Loop, FALSE);
        break;
      case UNDO_INSERT_LOOP_NODE:
        RemoveBehindFrontNode(mesh, Undo->Loop, FALSE);
        break;
      case UNDO_NEXT_LOOP:
        PrevLoop(mesh, FALSE);
        break;
      case UNDO_PREV_LOOP:
        NextLoop(mesh, FALSE);
        break;
      case UNDO_ADD_LOOP:
        RemoveLoop(mesh, Undo->Loop, FALSE);
        break;
      case UNDO_REMOVE_LOOP:
        AddLoop(mesh, Undo->Loop, FALSE);
        break;
      case UNDO_MERGE_CUT_LOOP:
        MergeCutLoops(mesh, Undo->Node, Undo->RefNode, FALSE);
        break;
      case UNDO_STAGE_INCR:
        StageDecr(Undo->Node, FALSE);
        break;
      case UNDO_STAGE_DECR:
        StageIncr(Undo->Node, FALSE);
        break;
      case UNDO_MARKER:
        marker = TRUE;
        break;
      default:
        e_GeneralWarning("UndoUntilMarker: invalid undo operation !");
        break;
    }
    /* take next undo operation */
    u        = Undo;
    Undo     = u->Next;
    
    /* return it to the pool of memory */
    u->Next  = UndoPool;
    UndoPool = u;

  }
  while((Undo != NULL) && (marker == FALSE));

}


void InitUndo(void)
{
  UNDO *newnode;
  int   i;

  Undo = NULL;
  UndoPool = NULL;

  for(i = 0; i < ALLOC_UNDOPOOL_SIZE; i++)
  {
    newnode = (UNDO *)e_calloc(1, sizeof(UNDO) );

    newnode->Next = UndoPool;
    UndoPool = newnode;
  }
}

void EmptyUndoQueue(void)
{
  UNDO *u;

  while(Undo != NULL)
  {
    u = Undo;
    Undo = u->Next;
    /* return it to the pool of memory */
    u->Next  = UndoPool;
    UndoPool = u;
  }
}


/* ############################################################################# */
/*                       Mesh node manipulation functions                        */
/* ############################################################################# */

/* 
 * Create a node/meshpoint.
 */
int CreateNode(double x, double y, int flag, PAVEDMESH *mesh, int undo)
{
  int n;

  /* check coordinate validity */
  if(   x < mesh->Minx || x > mesh->Maxx 
     || y < mesh->Miny || y > mesh->Maxy)
  {
    e_GeneralError("CreateNode: Node coordinates are out of bounding box !");
  }

  n = mesh->NNodes;

  if(undo)
  {
    AppendUndo(UNDO_CREATE_NODE, n);
  }

  mesh->NodeCoords = e_incr_cont_d_matr(n, PAVING_DIMENSION, mesh->NodeCoords, ALLOC_NODE_SIZE);
  mesh->NNodeElem  = e_incr_i_vect(n, mesh->NNodeElem, ALLOC_NODE_SIZE);
  mesh->NodeElem   = e_incr_frag_i_matr(n, mesh->NodeElem, ALLOC_NODE_SIZE);
  mesh->NodeFlag   = e_incr_i_vect(n, mesh->NodeFlag, ALLOC_NODE_SIZE);

  mesh->NodeCoords[n][0] = x;
  mesh->NodeCoords[n][1] = y;
 
  mesh->NNodeElem[n]     = 0;
  mesh->NodeElem[n]      = NULL;


  if(flag & NODE_BND)
    mesh->NodeFlag[n] = -1;
  
  mesh->NNodes++;

#ifdef PAVING_GRAPHICS
  if(ShowGraphics)
  {
#ifdef DRAW_ALL
    mgnDrawAll(mesh->NNodes, mesh->NodeCoords,
               0, NULL, 
               mesh->NQuads, mesh->QuadNodes, 0,
               mgnDrawAllLoops, (void *)mesh->Fronts,NULL); 
#else
    mgnDrawNode(x, y, mesh->NNodes-1, G2_DRAW_NODE);
#endif
  }
#endif

  return P_SUCCESS;
}


/* 
 * Delete a node entity from the mesh 
 */
int DeleteNode(PAVEDMESH *mesh, int n, int undo)
{
  int    f;
  double x, y;  

  if(CHK_BND(mesh->NodeFlag[n]))
  {
    e_GeneralError("DeleteNode: Tried to delete a mesh boundary node !");
  }

  if(undo)
  {
    x = mesh->NodeCoords[n][0];
    y = mesh->NodeCoords[n][1];
    f = mesh->NodeFlag[n];

    AppendUndo(UNDO_DELETE_NODE, n, x, y, f);
  }

  e_free(mesh->NodeElem[n]);
  mesh->NodeElem[n]      = NULL;
  /* only nullify, do not decrease */
  mesh->NNodeElem[n]     = 0;
  mesh->NodeCoords[n][0] = 0.0;
  mesh->NodeCoords[n][1] = 0.0;
  mesh->NodeFlag[n]      = 0;

  if(n == mesh->NNodes-1)
    mesh->NNodes--;

#ifdef PAVING_GRAPHICS
  if(ShowGraphics)
  {
#ifdef DRAW_ALL
    mgnDrawAll(mesh->NNodes, mesh->NodeCoords,
               0, NULL, 
               mesh->NQuads, mesh->QuadNodes, 0,
               mgnDrawAllLoops, (void *)mesh->Fronts,NULL); 

#endif
  }
#endif
  
  return P_SUCCESS;
}

/* 
 * Change the coordinates of a node 
 */
int ChangeNodeCoord(int n, double dx, double dy, PAVEDMESH *mesh, int undo)
{
  double ox, oy;

  ox = mesh->NodeCoords[n][0];
  oy = mesh->NodeCoords[n][1];

  /* check coordinate validity */
  if(   ox+dx < mesh->Minx || ox+dx > mesh->Maxx 
     || oy+dy < mesh->Miny || oy+dy > mesh->Maxy)
  {
    e_GeneralError("ChangenodeCoord: Node coordinates are out of bounding box !");
  }

  if(undo)
  {
    AppendUndo(UNDO_CHANGE_NODE_COORD, n, -dx, -dy);
  }
  
  mesh->NodeCoords[n][0] += dx;
  mesh->NodeCoords[n][1] += dy;

  return(P_SUCCESS);
}

/* ############################################################################# */
/*                     Mesh element manipulation functions                       */
/* ############################################################################# */

/*
 * Add the element to the node-element connectivity list
 */
static int AddElemToNodeElem(PAVEDMESH *mesh, int n, int elem)
{
  mesh->NodeElem[n] = 
    e_incr_i_vect(mesh->NNodeElem[n], mesh->NodeElem[n], ALLOC_NE_SIZE);

  mesh->NodeElem[n][mesh->NNodeElem[n]] = elem;
  mesh->NNodeElem[n]++;
  
  return(P_SUCCESS);
}


/*
 * Remove the element from the node-element connectivity list
 */
static int RemoveElemFromNodeElem(PAVEDMESH *mesh, int n, int elem)
{
  int j, length;

  /* search in the list of connected elements, 
     backwards, because most of the time it is the last element */
  for(j = mesh->NNodeElem[n]-1; j >=0 ; j--)
  {
    /* did we find it ? */
    if(mesh->NodeElem[n][j] == elem)
    {
      length = mesh->NNodeElem[n]-j-1;
      /* is there any data after this one, then copy */
      if(length > 0)
      {
        memmove((void *)(mesh->NodeElem[n]+j),
                (void *)(mesh->NodeElem[n]+j+1),
                length * sizeof(int) );
      }
      break;
    }
  }
    
  if((j != 0) && (j == mesh->NNodeElem[n]))
  {
    e_GeneralError("DeleteElement: Internal Error 1");
  }

  /* reduce the length of reference list */
  mesh->NNodeElem[n]--;
  
  /* if no more, free it up */
  if(mesh->NNodeElem[n] == 0)
  {
    e_free(mesh->NodeElem[n]);
    mesh->NodeElem[n] = NULL;
  }
  
  return(P_SUCCESS);
}


/* 
 * Create a quad element from 4 points
 */
int CreateElement(int n1, int n2, int n3, int n4, PAVEDMESH *mesh, int undo)
{

  int elem, m, i;

  elem = mesh->NQuads;

  if(undo)
  {
    AppendUndo(UNDO_CREATE_ELEMENT, elem);
  }

  mesh->QuadNodes     = 
    e_incr_cont_i_matr(mesh->NQuads, 4, mesh->QuadNodes, ALLOC_ELEM_SIZE);

  mesh->QuadNodes[mesh->NQuads][0] = n1;
  mesh->QuadNodes[mesh->NQuads][1] = n2;
  mesh->QuadNodes[mesh->NQuads][2] = n3;
  mesh->QuadNodes[mesh->NQuads][3] = n4;

  for(i = 0; i < 4; i++)
  {
    m = mesh->QuadNodes[elem][i];

    AddElemToNodeElem(mesh, m, elem);
  }

  mesh->NQuads++;

#ifdef PAVING_GRAPHICS
  if(ShowGraphics)
  {
#ifdef DRAW_ALL
    mgnDrawAll(mesh->NNodes, mesh->NodeCoords,
               0, NULL, 
               mesh->NQuads, mesh->QuadNodes, 0,
               mgnDrawAllLoops, (void *)mesh->Fronts,NULL); 

#else
    mgnDrawQuad(mesh->NodeCoords, mesh->QuadNodes[mesh->NQuads-1],
                mesh->NQuads-1, 0);
#endif
  }
#endif

  return P_SUCCESS;
}


/* 
 * Delete an element from the mesh, 
 * remove references from the node-element connectivity data
 */
int DeleteElement(int elem, PAVEDMESH *mesh, int undo)
{
  int i, n, old, new;
  int node[4];


  for(i = 0; i < 4; i++)
  {
    node[i] = mesh->QuadNodes[elem][i];
  }

  if(undo)
  {
    n  = mesh->NQuads;
    AppendUndo(UNDO_DELETE_ELEMENT, n, node[0], node[1], node[2], node[3]);
  }

  mesh->NQuads--;
  
  if(elem != mesh->NQuads)
  {
    for(i = 0; i < 4; i++)
    {
      old = node[i];
      new = mesh->QuadNodes[mesh->NQuads][i];
      ChangeElementNode(elem, old, new, mesh, undo);
      node[i] = new;
    }

  }

  for(i = 0; i < 4; i++)
  {
    RemoveElemFromNodeElem(mesh, node[i], mesh->NQuads);
  }

#ifdef PAVING_GRAPHICS
  if(ShowGraphics)
  {
#ifdef DRAW_ALL
    mgnDrawAll(mesh->NNodes, mesh->NodeCoords,
               0, NULL, 
               mesh->NQuads, mesh->QuadNodes, 0,
               mgnDrawAllLoops, (void *)mesh->Fronts,NULL); 
#endif
  }
#endif

  return P_SUCCESS;
}

/*
 * In a quad element change the node 'old' into 'new' and
 * update node-element connectivity list
 */
int ChangeElementNode(int elem, int old, int new, PAVEDMESH *mesh, int undo)
{
  int i, n;

  if(undo)
  {
    AppendUndo(UNDO_CHANGE_ELEMENT_NODES, elem, new, old);
  }

  /* check all nodes in the quad element */
  for(i = 0; i < 4; i++)
  {
    n = mesh->QuadNodes[elem][i];
    if(n == old)
    {
      RemoveElemFromNodeElem(mesh, n, elem);
      AddElemToNodeElem(mesh, new, elem);

      mesh->QuadNodes[elem][i] = new;
      break;
    }
  }
  if(i == 4)
  {
    e_GeneralError("ChangeElementNode: node not found !");
  }

#ifdef PAVING_GRAPHICS
  if(ShowGraphics)
  {
#ifdef DRAW_ALL
    mgnDrawAll(mesh->NNodes, mesh->NodeCoords,
               0, NULL, 
               mesh->NQuads, mesh->QuadNodes, 0,
               mgnDrawAllLoops, (void *)mesh->Fronts,NULL); 
#endif
  }
#endif

  return(P_SUCCESS);
}


/* ############################################################################# */
/*                          Loop manipulation functions                          */
/* ############################################################################# */

/*
 * Set the head of the loop to the specified node,
 * there is no check that the node is really part of the loop !!!!!
 */
void SetLoopHead(LOOP *loop, NODE *node, int undo)
{
  if(undo)
  {
    AppendUndo(UNDO_SET_LOOP_HEAD, loop, loop->Nodes);
  }

  loop->Nodes = node;
}

/* 
 * Move the head of the loop one node forward.
 */
void StepForward(LOOP *loop, int undo)
{
  if(undo)
  {
    AppendUndo(UNDO_STEP_FORWARD, loop);
  }

  loop->Nodes = loop->Nodes->Next;

}


/* 
 * Move the head of the loop one node backward.
 */
void StepBackward(LOOP *loop, int undo)
{
  if(undo)
  {
    AppendUndo(UNDO_STEP_BACKWARD, loop);
  }

  loop->Nodes = loop->Nodes->Prev;

}


/*
 * Insert a NODE before the head of the LOOP.
 * It "pushes" all previous nodes back.
 */
void InsertBehindFrontNode(PAVEDMESH *mesh, NODE *node, LOOP *loop, int undo)
{
  NODE *prev;
  
/*  ADD_BIT(mesh->NodeFlag[node->Node], NODE_FIXED); */
  if(mesh->NodeFlag[node->Node] >= 0)
    mesh->NodeFlag[node->Node]++;

  if(undo)
  {
    AppendUndo(UNDO_INSERT_LOOP_NODE, loop);
  }
  
  prev = loop->Nodes->Prev;
  
  /* connect node with previous node */
  prev->Next        = node;
  node->Prev        = prev;
  
  /* connect node with loop head */
  loop->Nodes->Prev = node;
  node->Next        = loop->Nodes;

  loop->NNodes++;

#ifdef PAVING_GRAPHICS
  if(ShowGraphics)
  {
#ifdef DRAW_ALL
    mgnDrawAll(mesh->NNodes, mesh->NodeCoords,
               0, NULL, 
               mesh->NQuads, mesh->QuadNodes, 0,
               mgnDrawAllLoops, (void *)mesh->Fronts,NULL); 
#endif
  }
#endif

}


/*
 * Remove a NODE from behind the head of the LOOP.
 * The NODE memory is not freed-up and it is returned.
 */
void RemoveBehindFrontNode(PAVEDMESH *mesh, LOOP *loop, int undo)
{
  NODE *prev, *next, *node;
  
  node = loop->Nodes->Prev;

  /* mark the node, that it is not on the front any more */
  if(mesh->NodeFlag[node->Node] >= 1)
    mesh->NodeFlag[node->Node]--;

/*  DEL_BIT(mesh->NodeFlag[node->Node], NODE_FIXED); */

  if(undo)
  {
    AppendUndo(UNDO_REMOVE_LOOP_NODE, node, loop);
  }

  /* store references */
  prev = node->Prev;
  next = node->Next;

  /* can be deleted */
  node->Prev = NULL;
  node->Next = NULL;

  /* reconnect the loop */
  prev->Next = next;
  next->Prev = prev;
  
  loop->NNodes--;

  if(loop->NNodes == 0)
  {
    loop->Nodes = NULL;
  }

  /* if we do not store it in the undo list, recycle it */
  if(!undo)
  {
    RecycleNode(node, mesh);
  }

#ifdef PAVING_GRAPHICS
  if(ShowGraphics)
  {
#ifdef DRAW_ALL
    mgnDrawAll(mesh->NNodes, mesh->NodeCoords,
               0, NULL, 
               mesh->NQuads, mesh->QuadNodes, 0,
               mgnDrawAllLoops, (void *)mesh->Fronts,NULL); 
#endif
  }
#endif

}


/*
 * The next loop is taken as the front.
 */
void NextLoop(PAVEDMESH *mesh, int undo)
{
  if(undo)
  {
    AppendUndo(UNDO_NEXT_LOOP);
  }

  mesh->Fronts = mesh->Fronts->Next;
}


/*
 * The previous loop is taken as the front.
 */
void PrevLoop(PAVEDMESH *mesh, int undo)
{
  if(undo)
  {
    AppendUndo(UNDO_PREV_LOOP);
  }

  mesh->Fronts = mesh->Fronts->Prev;
}


/*
 * Add a loop to the mesh.
 * If no other fronts exist, add it simple,
 * if loop has valid 'Next' and 'Prev' components, use them to insert the loop,
 * in all other cases insert behind current front.
 */
void AddLoop(PAVEDMESH *mesh, LOOP *loop, int undo)
{
  LOOP *next, *prev;

  if(undo)
  {
    AppendUndo(UNDO_ADD_LOOP, loop);
  }

  /* no fronts at all */
  if(mesh->Fronts == NULL)
  {
    loop->Next = loop;
    loop->Prev = loop;
    mesh->Fronts = loop;
  }
  /* loop already knows its place */
  else if(   (loop->Next != NULL)
          && (loop->Prev != NULL) )
  {
    prev = loop->Prev;
    next = loop->Next;
    if(prev->Next != next)
      e_GeneralError("AddLoop: Internal error !");

    prev->Next = loop;
    loop->Prev = prev;
  
    loop->Next = next;
    next->Prev = loop;

  }
  /* insert before front */
  else
  {
    prev = mesh->Fronts->Prev;

    prev->Next = loop;
    loop->Prev = prev;
  
    loop->Next         = mesh->Fronts;
    mesh->Fronts->Prev = loop;

  }

#ifdef PAVING_GRAPHICS
  if(ShowGraphics)
  {
#ifdef DRAW_ALL
    mgnDrawAll(mesh->NNodes, mesh->NodeCoords,
               0, NULL, 
               mesh->NQuads, mesh->QuadNodes, 0,
               mgnDrawAllLoops, (void *)mesh->Fronts,NULL); 
#endif
  }
#endif

}


/*
 * Remove a loop from the front list.
 * If loop is the current front
 *   if that is the only loop, NULLifiy the front
 *   otherwise move fronts forward, delete loop references
 * otherwise remove loop, but preserve references
 */
void RemoveLoop(PAVEDMESH *mesh, LOOP *loop, int undo)
{
  if(mesh->Fronts == NULL)
  {
    e_GeneralWarning("RemoveLoop: Empty front list !");
    return;
  }
  
  /* the loop is on the front */
  if(mesh->Fronts == loop)
  {
    /* there is only one loop in the front */
    if(loop->Next == loop)
    {
      mesh->Fronts = NULL;
    }
    /* relocate the front */
    else
    {
      /* unlink the loop */
      loop->Next->Prev = loop->Prev;
      loop->Prev->Next = loop->Next;

      mesh->Fronts = loop->Next;
      loop->Next = NULL;
      loop->Prev = NULL;
    }
  }
  else
  {
    /* unlink the loop */
    loop->Next->Prev = loop->Prev;
    loop->Prev->Next = loop->Next;
  }

  if(undo)
  {
    AppendUndo(UNDO_REMOVE_LOOP, loop);
  }
  else
  {
    e_free(loop);
  }

#ifdef PAVING_GRAPHICS
  if(ShowGraphics)
  {
#ifdef DRAW_ALL
    mgnDrawAll(mesh->NNodes, mesh->NodeCoords,
               0, NULL, 
               mesh->NQuads, mesh->QuadNodes, 0,
               mgnDrawAllLoops, (void *)mesh->Fronts,NULL); 
#endif
  }
#endif

}


/*
 * Merge or cut two loops.
 */
void MergeCutLoops(PAVEDMESH *mesh, NODE *new, NODE *old, int undo)
{
  NODE *nnext, *onext;

  if(undo)
  {
    AppendUndo(UNDO_MERGE_CUT_LOOP, new, old);
  }
  
  nnext = new->Next;
  onext = old->Next;
  
  old->Next   = nnext;
  nnext->Prev = old;
  
  new->Next   = onext;
  onext->Prev = new;

#ifdef PAVING_GRAPHICS
  if(ShowGraphics)
  {
#ifdef DRAW_ALL
    mgnDrawAll(mesh->NNodes, mesh->NodeCoords,
               0, NULL, 
               mesh->NQuads, mesh->QuadNodes, 0,
               mgnDrawAllLoops, (void *)mesh->Fronts,NULL); 
#endif
  }
#endif

}


/*
 * Increment the stage for a node.
 */
int StageIncr(NODE *node, int undo)
{
  if(undo)
  {
    AppendUndo(UNDO_STAGE_INCR, node);
  }
  
  node->Stage++;

  return(P_SUCCESS);
}


/*
 * Decrement the stage for a node
 */
int StageDecr(NODE *node, int undo)
{
  if(undo)
  {
    AppendUndo(UNDO_STAGE_DECR, node);
  }
  
  node->Stage--;

  return(P_SUCCESS);
}


/* ############################################################################# */
/*                           Basic Node functions                                */
/* ############################################################################# */
/*  NO UNDO action */
/* ############### */

PAVEDMESH *InitPavedMesh(void)
{
  PAVEDMESH *mesh;
  
  mesh = (PAVEDMESH *)e_calloc(1, sizeof(PAVEDMESH) );
#ifdef WIN32
  srand(932865);
#else
  srand48(932865);
#endif
  
  return(mesh);
}


LOOP *InitLoop(int type, NODE *node)
{
  LOOP *loop;

  loop = (LOOP *)e_calloc(1, sizeof(LOOP) );

  loop->Type      = type;
  loop->Primitive = 0;
  loop->Nodes     = node;

  /* determine the number of nodes in loop */
  CountLoopNodes(loop);

  return(loop);
}


void CountLoopNodes(LOOP *loop)
{
  int   n;
  NODE *node;

  /* determine the number of nodes in loop */
  n = 0;
  node = loop->Nodes;
  do
  {
    n++;
    node = node->Next;
  }
  while(node != loop->Nodes);

  loop->NNodes = n;

}

/* 
 * Allocate or get new NODE from a pool.
 */
NODE *GetNewNode(PAVEDMESH *mesh)
{
  NODE *newnode;
  int   i;

  if(NodePool == NULL)
  {
    for(i = 0; i < ALLOC_NODEPOOL_SIZE; i++)
    {
      newnode = (NODE *)e_calloc(1, sizeof(NODE) );

      newnode->Next = NodePool;
      NodePool = newnode;
    }
  }

  newnode = NodePool;
  NodePool = NodePool->Next;

  ClearNode(newnode);

  return(newnode);
}


/* 
 * Clear all fields of a NODE.
 */
void ClearNode(NODE *node)
{
  node->Next      = NULL;
  node->Prev      = NULL;
  node->Flag      = 0;
  node->Node      = 0;
  node->Angle     = 0.0;
  node->ReqLength = 0.0;
  node->Stage     = 0;
  node->RefPrev   = 0;
  node->RefNext   = 0;
  node->ProjNode  = 0;
}


/*
 * Return the NODE back to the pool as an available node.
 * Clearing will happen when it is used again.
 */
void RecycleNode(NODE *node, PAVEDMESH *mesh)
{
  node->Prev = NULL;
  node->Next = NodePool;
  NodePool = node;
}


/*
 * Duplicates the given NODE.
 */
NODE *DuplicateNode(NODE *node, PAVEDMESH *mesh)
{
  NODE *newnode;

  newnode            = GetNewNode(mesh);
  newnode->Next      = node->Next;
  newnode->Prev      = node->Prev;
  newnode->Flag      = node->Flag;
  newnode->Node      = node->Node;
  newnode->Angle     = node->Angle;
  newnode->ReqLength = node->ReqLength;
  newnode->Stage     = node->Stage;
  newnode->RefPrev   = node->RefPrev;
  newnode->RefNext   = node->RefNext;
  newnode->ProjNode  = node->ProjNode;

  return(newnode);
}





