#include "pav_main.h"
#include "boundary.h"
#include "nodelem.h"
#include "bgdata.h"

/* functions in crvref.c */
static void BndLength(PAVEDMESH *mesh);

static void RefineAllBoundary(MESH *mesh, BND BList, int geomset, PAVEDMESH *pm);
static void RefineBoundary(MESH *mesh, int nlist, int *list, 
                           int geomset, PAVEDMESH *pm);
static int RefineSegment(MESH *mesh, int n1, int n2, int geomset, PAVEDMESH *pm,
                         LOOP **loop, int type);

/* ############################################################################ */

/* 
 * From a coarse mesh, extract the boundary, refine them according to the
 * specified mesh parameters and insert the generated points into the
 * paving mesh.
 */
void InitFronts(MESH *mesh, int geomset, PAVEDMESH *pm)
{
  int    *NoNodeElem;
  int   **NodeElem;
  int     NBList;
  BND    *BList;

  /* generate boundary of the coarse mesh */
  NoNodeElem = e_alloc_i_vect(mesh->NMeshPoints);
  NodeElem = e_init_frag_i_matr(mesh->NMeshPoints);
  e_InitMeshPointElemConnection(mesh, NoNodeElem, NodeElem);

  Gen2DBoundary(mesh, NoNodeElem, NodeElem, &NBList, &BList);

  if(NBList != 1)
  {
    e_GeneralError("More than one external boundary !");
  }

  /* refine boundary according to meshparameters */
  RefineAllBoundary(mesh, BList[0], geomset, pm);

  e_free_frag_i_matr(mesh->NMeshPoints, NodeElem);
  e_free(NoNodeElem);
  
  Free2DBoundary(NBList, BList);
  
  /* initial length on the boundary nodes */
  BndLength(pm);
}


/* 
 * Determine the length parameters for the mesh boundary nodes.
 */
static void BndLength(PAVEDMESH *mesh)
{
  LOOP *currloop, *startloop;
  NODE *currnode, *startnode;
  double dx1, dy1, dx2, dy2, l1, l2, length;
  int    curr, next, prev;

#ifdef GLOBAL_SIZE_CONTROL
  /* allocate boundary information */
  mesh->NBnd = mesh->NNodes;
  mesh->BndLenParam = e_alloc_d_vect(mesh->NBnd);
#endif
  
  /* go through all loops */
  currloop = startloop = mesh->Fronts;
  do
  {
    /* go through all nodes on the loop */
    currnode = startnode = currloop->Nodes;
    do
    {
      curr = currnode->Node;
      next = currnode->Next->Node;
      prev = currnode->Prev->Node;
      
      /* distance from next node */
      dx1 = mesh->NodeCoords[next][0] - mesh->NodeCoords[curr][0];
      dy1 = mesh->NodeCoords[next][1] - mesh->NodeCoords[curr][1];
      l1  = sqrt(dx1*dx1 + dy1*dy1);
      
      /* distance from previous node */
      dx2 = mesh->NodeCoords[prev][0] - mesh->NodeCoords[curr][0];
      dy2 = mesh->NodeCoords[prev][1] - mesh->NodeCoords[curr][1];
      l2  = sqrt(dx2*dx2 + dy2*dy2);
      
      /* average the length */
      length = (l1+l2)/2.0;
      
#ifdef GLOBAL_SIZE_CONTROL
      /* store in the array and in the front node */
      mesh->BndLenParam[curr] = length;
#endif
      currnode->ReqLength = length;
      
      currnode = currnode->Next;
    }
    while(currnode != startnode);
  
    currloop = currloop->Next;
  }
  while(currloop != startloop);
}

/*
 * Refine the external and internal boundary of a region.
 */
static void RefineAllBoundary(MESH *mesh, BND BList, int geomset, PAVEDMESH *pm)
{
  int i;

  RefineBoundary(mesh, BList.oe.nen, BList.oe.en, geomset, pm);

  if(BList.nie)
  {
    for(i = 0; i < BList.nie; i++)
    {
      RefineBoundary(mesh, BList.ie[i].nen, BList.ie[i].en, geomset, pm);
    }
  }

}

/* 
 * Refine a boundary segment. Find longest segment in boundary, for all other
 * segments generate nodes without any restriction, for the longest segment 
 * generate either even or odd number of nodes to get finally even
 * number of nodes.
 */
static void RefineBoundary(MESH *mesh, int nlist, int *list, 
                           int geomset, PAVEDMESH *pm)
{
  int      i, pre, cur, nnode, type;
  double   x, y, length;
  LOOP    *loop = NULL;
  double   maxlength;
  int      maxindex, index;

  /* find the longest segment */
  maxlength = 0.0;
  maxindex  = -1;
  pre = list[nlist-1]-1;
  for(i = 0; i < nlist; i++)
  {
    cur = list[i]-1;
    x = mesh->MeshPointCoords[pre][0] -
        mesh->MeshPointCoords[cur][0];
    y = mesh->MeshPointCoords[pre][1] -
        mesh->MeshPointCoords[cur][1];
    length = sqrt((x*x) + (y*y));
    if(maxlength < length)
    {
      maxlength = length;
      maxindex  = i;
    }    
    
    pre = cur;
  }
  
  if(maxindex == -1)
    e_GeneralError("RefineBoundary: Internal error !");
  
  /* refine all segments, except longest */
  nnode = 0;
  index = maxindex % nlist;
  pre = list[index];
  for(i = 1; i < nlist; i++)
  {
    index = (maxindex+i) % nlist;
    cur = list[index];
    nnode += RefineSegment(mesh, pre, cur, geomset, pm, &loop, E_SEGREF_NONE);
    pre = cur;
  }

  /* refine the longest segment, so there will be even number of nodes */
  index = (maxindex - 1 + nlist) % nlist;
  pre = list[index];
  index = maxindex;
  cur = list[index];
  if(nnode % 2)
  {
    type = E_SEGREF_EVEN;
  }
  else
  {
    type = E_SEGREF_ODD;
  }
  nnode += RefineSegment(mesh, pre, cur, geomset, pm, &loop, type);
  
  if(loop->NNodes % 2)
    e_GeneralError("RefineBoundary: odd number of nodes is generated !");
  
  /* add loop to front list */
  AddLoop(pm, loop, FALSE);

}


/*
 * Refine one segment.
 * Returns the number of inserted nodes.
 */
static int RefineSegment(MESH *mesh, int n1, int n2, int geomset, PAVEDMESH *pm,
                         LOOP **loop, int type)
{
  int          j;
  int          nurbs, n, looptype;
  double     **coords;
  double      *mpr, *par_t;
  double       x, y;
  NODE        *newnode;
  NURBSCURVE  *curve;
  double       t1, t2, d1, d2;
  int          ins_start, ins_end;
  MTRF         mtrf;

  coords = NULL;
  mpr    = NULL;
  par_t  = NULL;

  /* check whether we have a NURBS for it */
  nurbs = FALSE;
  if(geomset)
  {
    nurbs = e_GetCurve(mesh, n1, n2, &curve, &t1, &t2);
  }
  
  n1--;
  n2--;
  
  if(pm->ctrl == CTRL_MPR)
  {
    d1 = mesh->MeshParam.NodalMeshParams[n1];
    d2 = mesh->MeshParam.NodalMeshParams[n2];
  }
  else if(pm->ctrl == CTRL_BACKGROUND || pm->ctrl == CTRL_SOURCES)
  {
    mtrf.used = FALSE;
    d1 = get_bg_mpr(pm->bgd, mesh->MeshPointCoords[n1], pm->ctrl, &mtrf);
    d2 = get_bg_mpr(pm->bgd, mesh->MeshPointCoords[n2], pm->ctrl, &mtrf);
  }
  else if(pm->ctrl == CTRL_UNIFORM)
  {
    d1 = pm->uniform_h;
    d2 = pm->uniform_h;
  }
  
  if(nurbs)
  {
    /* curved boundary */
    n = e_CurveRefine(curve, t1, d1, t2, d2,
                      &coords, &mpr, &par_t, TRUE, type);
    ins_start = 1;
    ins_end   = n-1;
    
  }
  else
  {
    /* straight line */
    n = e_SegmentRefine(mesh->MeshPointCoords[n1], d1,
                        mesh->MeshPointCoords[n2], d2,
                        &coords, &mpr, TRUE, type);
    ins_start = 0;
    ins_end   = n;
  }
    
  /* insert the starting point */
  x = mesh->MeshPointCoords[n1][0];
  y = mesh->MeshPointCoords[n1][1];

  CreateNode(x, y, NODE_BND, pm, FALSE);

  newnode            = GetNewNode(pm);
  newnode->Node      = pm->NNodes-1;
    
  /* if required initialise the loop */
  if((*loop) == NULL)
  {
    newnode->Next = newnode;
    newnode->Prev = newnode;
    /* if no other loops, then this is external */
    if(pm->Fronts == NULL)
      looptype = LOOP_EXTERNAL;
    else
      looptype = LOOP_INTERNAL;
    (*loop) = InitLoop(looptype, newnode);
  }
  /* insert into the loop */
  else
  {
    InsertBehindFrontNode(pm, newnode, *loop, FALSE);
  }

  /* insert the generated nodes into the mesh and loop */
  for(j = ins_start; j < ins_end; j++)
  {
    x = coords[j][0];
    y = coords[j][1];
    
    CreateNode(x, y, NODE_BND, pm, FALSE);
    newnode            = GetNewNode(pm);
    newnode->Node      = pm->NNodes-1;
  
    InsertBehindFrontNode(pm, newnode, *loop, FALSE);
  }
  
  /* free up allocated memory */
  e_free_cont_d_matr(n, coords);
  e_free(mpr);
  if(nurbs)
  {
    e_free(par_t);
  }
  
  /* return the number of inserted nodes */
  return(n+1);
}



