#include "pav_main.h"
#include "cmb.h"


static NODE *SelectEnd(PAVEDMESH *mesh, LOOP *loop);
static int ValidRectangle(int *sides);
static int ValidTriangle(int *sides);
static int ValidSemiCircle(int *sides);
static void ShapeCheck(PAVEDMESH *mesh, LOOP *loop, int nshape, NODE **qn,
                       int *validity, int *nirregular, double *angledev);
static int ShapeCombinations(PAVEDMESH *mesh, LOOP *loop, int nshape,
                             int nrealend, NODE *realend[MAX_END_NODE],
                             int ncandend, NODE *candend[MAX_END_NODE],
                             int *bestshape, NODE *shape[4], 
                             int *bestirr, double *bestangle);
static void SetEndNodesForShape(int nshape, NODE *shape[4],
                                int ncandend, NODE *candend[MAX_END_NODE]);
static void SelectOneNodeType(NODE *curr, int NNodeElem, int type1, int type2);

static NODE *SelectRowStart(PAVEDMESH *mesh, LOOP *loop);

static int RowChoice(PAVEDMESH *mesh, LOOP *loop);

/* ######################################################################### */
/*                       Classification functions                            */
/* ######################################################################### */

/* 
 * Calculates the internal angle and classifies the nodes by the angle.
 * The function assumes, that external boundary counter-clockwise
 * and internal boundary clockwise.
 * Ambiguous types are possible !
 */
int ClassifyANode(double **coords, int prev, int curr, int next, 
                  int *state, double *angle)
{
  (*angle) = AngleCalc(coords[prev][0], coords[prev][1], 
                       coords[curr][0], coords[curr][1], 
                       coords[next][0], coords[next][1]  );

#ifdef PAVING_DEBUG
  if((E_ABS(*angle) < ZERO) || (E_ABS((*angle) - 2*M_PI) < ZERO) )
  {
    e_GeneralWarning("ClassifyNode : Internal angle near ZERO : %lf !", *angle);
  }
#endif

  /* Classify the node by the internal angle. A node can belong
     to two different classes, specified by the tolerances. */
  DEL_BIT((*state), ALL_NODE_TYPE);
  if((*angle) < (M_PI-S2))
  {
    ADD_BIT((*state), NODE_END);
  }

  if(   (*angle) > ((M_PI/2.0)+S1) 
     && (*angle) < ((3.0*M_PI/2.0)-S4) )
  {
    ADD_BIT((*state), NODE_SIDE);
  }

  if(   (*angle) > (M_PI+S3) 
     && (*angle) < ((2.0*M_PI)-S6) )
  {
    ADD_BIT((*state), NODE_CORNER);
  }
  
  if((*angle) >= ((3.0*M_PI/2.0)+S5))
  {
    ADD_BIT((*state), NODE_REVERSAL);
  }

  return(P_SUCCESS);
}

/*
 * Classify all nodes in a loop.
 */
int ClassifyALoop(double **coords, LOOP *loop)
{
  NODE *prev, *curr, *next, *start;

  curr = start = loop->Nodes;
  prev = curr->Prev;
  next = curr->Next;
  do
  {
    curr->Flag      = 0;
    curr->Angle     = 0.0;
    curr->Stage     = 0;
    /* 
     * Should not clear ReqLength and ProjNode since they preserve information for
     * smoothing. Though if we want to correct (and a little bit slower) then we
     * can set ProjNode to -1. In this case ReqLength will be recalculate when
     * it is possible and required.
     */
    ClassifyANode(coords, prev->Node, curr->Node, next->Node, 
                  &(curr->Flag), &(curr->Angle));

    curr->RefPrev = prev->Node;
    curr->RefNext = next->Node;

    prev = curr;
    curr = next;
    next = curr->Next;
  }
  while(curr != start);
  
  return 1;
}

/*
 * Select the next end node on the loop.
 */
static NODE *SelectEnd(PAVEDMESH *mesh, LOOP *loop)
{
  NODE  *start, *curr;
  int    started = FALSE;

  curr = start = loop->Nodes;
  do
  {
    if(CHK_BIT(curr->Flag, NODE_END))
    {
      break;
    }
    curr = curr->Next;
    started = TRUE;
  }
  while(start != curr);
  
  /* check whether we went around and not found a single end node,
     in this function it would be an error !! */
  if(started && start == curr)
  {
    e_GeneralError("SelectEnd: internal error !");
  }
  
  /* make it only end node */
  DEL_BIT(curr->Flag, ALL_NODE_TYPE);
  ADD_BIT(curr->Flag, NODE_END);

  return(curr);
}

/*
 * Check the validity of a rectangle shape to subdivide
 */
static int ValidRectangle(int *sides)
{
  /* sum of intervals is equal or greater than 4 */
  if(sides[0] + sides[1] + sides[2] + sides[3] <= 4)
    return(FALSE);

  return(TRUE);
}

/*
 * Check the validity of a triangle shape to subdivide
 */
static int ValidTriangle(int *sides)
{
  int i;

  /* each side should have at least two intervals */
  for(i = 0; i < 3; i++)
  {
    if(sides[i] < 2)
      return(FALSE);
  }

  /* the sum of the intervals for any two sides is two greater
     than the remaining side */
  if(sides[0] + sides[1] < sides[2] + 2)
    return(FALSE);
  if(sides[1] + sides[2] < sides[0] + 2)
    return(FALSE);
  if(sides[2] + sides[0] < sides[1] + 2)
    return(FALSE);

  return(TRUE);
}


/*
 * Check the validity of a semicircle shape to subdivide
 */
static int ValidSemiCircle(int *sides)
{
  int max;

  max = sides[0] > sides[1] ? 0 : 1;
  
  /* the intervals of the max side should be equal to or greater
     than the intervals on the other side which must be greater
     than 4 */
  if(   (sides[max] >= sides[(max+1) % 2])
     && (sides[(max+1) % 2] > 4) )
  {
    return(TRUE);
  }

  return(FALSE);
}

/*
 * Knowing the shape and the end nodes (all other nodes are side nodes)
 * we calculate the number of irregular nodes to generate and angle
 * deviation from perfect side and end node
 */
static void ShapeCheck(PAVEDMESH *mesh, LOOP *loop, int nshape, NODE **qn,
                       int *validity, int *nirregular, double *angledev)
{
  NODE *curr, *start;
  int   index, i, j, found;
  int   sides[4];

  (*validity)   = FALSE;
  (*nirregular) = 0;
  (*angledev)   = 0.0;
  sides[0]      = 0;
  sides[1]      = 0;
  sides[2]      = 0;
  sides[3]      = 0;
  
  /* go to a node after any of the end nodes */
  found = FALSE;
  start = curr = loop->Nodes;
  do
  {
    for(i = 0; i < nshape; i++)
    {
      if(curr == qn[i])
      {
        found = TRUE;
        break;
      }
    }
    curr = curr->Next;
  }
  while((curr != start) && (found == FALSE));

  if(!found)
  {
    e_GeneralError("ShapeCheck: internal error 1 !");
  }

  /* go around the loop */
  start = curr;
  index = 0;
  do
  {
    /* count the intervals between end nodes */
    sides[index]++;

    found = FALSE;
    /* check whether it is an end node */
    for(j = 0; j < nshape; j++)
    {
      if(qn[j] == curr)
      {
        /* check whether it produces irregular node */
        if(mesh->NNodeElem[qn[j]->Node]+1 != 4)
        {
          (*nirregular)++;
        }
        /* angle deviation */
        (*angledev) += E_ABS(ANGLE_90 - curr->Angle);
        /* count the intervals for the new sides */
        index++;
        if(index > 4)
        {
          e_GeneralError("ShapeCheck: internal error 3 !");
        }
        found = TRUE;
        break;
      }
    }
    /* it is a side node */
    if(!found)
    {
      /* angle deviation */
      (*angledev) += E_ABS(ANGLE_180 - curr->Angle);
      /* check whether it produces irregular node */
      if(mesh->NNodeElem[curr->Node]+2 != 4)
      {
        (*nirregular)++;
      }
    }
        
    curr = curr->Next;
  }
  while(curr != start);

  /* check the validity of the shape */
  if(nshape == SHAPE_RECTANGLE)
  {
    (*validity) = ValidRectangle(sides);
  }
  else if(nshape == SHAPE_TRIANGLE)
  {
    (*validity) = ValidTriangle(sides);
  }
  else if(nshape == SHAPE_SEMICIRCLE)
  {
    (*validity) = ValidSemiCircle(sides);
  }
  else
  {
    e_GeneralError("ShapeCheck: internal error 2 !");
  }
}


/*
 * Try all combinations of the end nodes to form the given shape
 */
static int ShapeCombinations(PAVEDMESH *mesh, LOOP *loop, int nshape,
                             int nrealend, NODE *realend[MAX_END_NODE],
                             int ncandend, NODE *candend[MAX_END_NODE],
                             int *bestshape, NODE *shape[4], 
                             int *bestirr, double *bestangle)
{
  int    i, j, next, shapefound;
  NODE  *qn[4];
  int    validity, nirregular;
  double angledev;
  int    combindex[4];

  shapefound = FALSE;
  nirregular = 0;
  angledev   = 0.0;
  /* 
   * no. of real ends is less than or equal to the no. of required end nodes
   * AND there is enough end nodes to select from
   */
  if((nrealend <= nshape) && (ncandend + nrealend >= nshape))
  {
    next = 0;
    do
    {
      for(i = 0; i < nrealend; i++)
      {
        qn[i] = realend[i];
      }
      if(i != nshape)
      {
        /* produce the combination from the candidates */
        next = cmb_NextKSubsetLex(ncandend, nshape - i, next, combindex);
        for(j = 0; i < nshape; i++, j++)
        {
          qn[i] = candend[combindex[j]];
        }
      }
    
      /* check the validity and quality of the shape */
      ShapeCheck(mesh, loop, nshape, qn, &validity, &nirregular, &angledev);
      /* it is a valid shape */
      if(validity == TRUE)
      {
        /* it is better than the best or 
           no irregular node is produced and angle deviation is smaller than 45 degree */
        if(   (*bestirr > nirregular && *bestangle > angledev)
           || (nirregular == 0  && angledev < ANGLE_45) )
        {
          *bestirr = nirregular;
          *bestangle = angledev;
          *bestshape = nshape;
          for(i = 0; i < nshape; i++)
          {
            shape[i] = qn[i];
          }
          shapefound = TRUE;
        }
      }
    }
    while(next != 0);
  }

  return(shapefound);
}


/*
 * Set the end nodes to end nodes and all other nodes to side nodes 
 */
static void SetEndNodesForShape(int nshape, NODE *shape[4],
                                int ncandend, NODE *candend[MAX_END_NODE])
{
  int  i, j, found;

  /* set the node types corresponding to the shape */
  for(i = 0; i < ncandend; i++)
  {
    DEL_BIT(candend[i]->Flag, ALL_NODE_TYPE);
    found = FALSE;
    for(j = 0; j < nshape; j++)
    {
      if(shape[j] == candend[i])
      {
        ADD_BIT(candend[i]->Flag, NODE_END);
        found = TRUE;
        break;
      }
    }
    if(!found)
    {
      ADD_BIT(candend[i]->Flag, NODE_SIDE);
    }
  }
}


/*
 * Remove any ambiguous type considering 
 * whether it produces irregular nodes or not
 */
static void SelectOneNodeType(NODE *curr, int NNodeElem, int type1, int type2)
{
  int nquad[NO_NODE_CLASSES] = {1, 2, 3, 4};
  int n1, n2;
  double prob;

  if((curr->Flag & ALL_NODE_TYPE) == (type1 + type2))
  {
    n1 = nquad[GetNodeType(type1)];
    n2 = nquad[GetNodeType(type2)];
    
    /* type 1 would not produce irregular node */
    if(NNodeElem+n1 == 4)
    {
      DEL_BIT(curr->Flag, ALL_NODE_TYPE);
      ADD_BIT(curr->Flag, type1);
    }
    /* type 2 would not produce irregular node */
    else if(NNodeElem+n2 == 4)
    {
      DEL_BIT(curr->Flag, ALL_NODE_TYPE);
      ADD_BIT(curr->Flag, type2);
    }
    /* random selection */
    else
    {
#ifdef WIN32
      prob = (double)rand()/(double)RAND_MAX;
#else
      prob = drand48();
#endif
      if(prob <= 0.5)
      {
        DEL_BIT(curr->Flag, ALL_NODE_TYPE);
        ADD_BIT(curr->Flag, type1);
      }
      else
      {
        DEL_BIT(curr->Flag, ALL_NODE_TYPE);
        ADD_BIT(curr->Flag, type2);
      }
    }
  }
}


void RemoveTypeAmbiguouity(NODE *curr, int NNodeElem)
{
  SelectOneNodeType(curr, NNodeElem, NODE_END,    NODE_SIDE);
  SelectOneNodeType(curr, NNodeElem, NODE_SIDE,   NODE_CORNER);
  SelectOneNodeType(curr, NNodeElem, NODE_CORNER, NODE_REVERSAL);
}


/*
 * Finalise the node types, only one node type should remain
 */
void FinalNodeTypeInLoop(PAVEDMESH *mesh, LOOP *loop)
{
  NODE *curr, *start;

  curr = start = loop->Nodes;
  do
  {
    /* overlaping node definitions are clarified */
    RemoveTypeAmbiguouity(curr, mesh->NNodeElem[curr->Node]);

    curr = curr->Next;
  }
  while(curr != start);
}


/* ######################################################################### */
/*                          Select row starting                              */
/* ######################################################################### */

/* identify, where the element generation should start */
static NODE *SelectRowStart(PAVEDMESH *mesh, LOOP *loop)
{
  int   nrealend, ncandend;
  NODE *realend[MAX_END_NODE];
  NODE *candend[MAX_END_NODE];
  NODE *start, *curr, *revsmall, *corsmall;
  int   nrevnode, ncornode, ncoronly;
  int   overload;
  NODE *shape[4];
  int   bestnirreg, bestshape;
  double bestangdev;

  /* go around the front loop and count all relevant information */
  overload = FALSE;
  nrealend = 0;
  ncandend = 0;
  nrevnode = 0;
  ncornode = 0;
  ncoronly = 0;
  revsmall = NULL;
  corsmall = NULL;
  start = curr = loop->Nodes;
  do
  {
    /* count end nodes look-alikes */
    if(CHK_BIT(curr->Flag, NODE_END))
    {
      /* it is only end node */
      if((curr->Flag & ALL_NODE_TYPE) == NODE_END)
      {
        if(nrealend == MAX_END_NODE)
        {
          if(overload == FALSE)
          {
            e_GeneralWarning("SelectRowStart: increase MAX_END_NODE in paving.h");
          }
          overload = TRUE;
        }
        else
        {
          realend[nrealend] = curr;
          nrealend++;
        }
      }
      else
      {
        if(ncandend == MAX_END_NODE)
        {
          if(overload == FALSE)
          {
            e_GeneralWarning("SelectRowStart: increase MAX_END_NODE in paving.h");
          }
          overload = TRUE;
        }
        else
        {
          candend[ncandend] = curr;
          ncandend++;
        }
      }
    }
    /* it is a reversal node */
    if(CHK_BIT(curr->Flag, NODE_REVERSAL))
    {
      nrevnode++;
      if(   (revsmall == NULL)
         || (revsmall->Angle > curr->Angle) )
      {
        revsmall = curr;
      }
    }
    /* corner node */
    if(CHK_BIT(curr->Flag, NODE_CORNER))
    {
      /* only corner node, no ambiguouity */
      if((curr->Flag & ALL_NODE_TYPE) == NODE_CORNER)
      {
        ncoronly++;
      }
    
      ncornode++;
      if(   (corsmall == NULL)
         || (corsmall->Angle > curr->Angle) )
      {
        corsmall = curr;
      }
    }
    
    curr = curr->Next;
  }
  while(curr != start);

  /* -------------------------------------------------------------------------- */
  /*                      Decision about the row start                          */
  /* -------------------------------------------------------------------------- */
  start = NULL;
  /* if there is/are reversal node(s) */
  if((nrevnode != 0) || (ncoronly != 0))
  {
    /* there is at least one real or candidate end node, 
       use the one which has the angle closest to 90 degree */
    if((nrealend != 0) || (ncandend != 0))
    {
      start = SelectEnd(mesh, loop);
    }
    /* if there are no end nodes (real or candidate) */
    else
    {
      /* if there are corner nodes as well */
      if(ncornode != 0)
      {
        start = corsmall;
        DEL_BIT(start->Flag, ALL_NODE_TYPE);
        ADD_BIT(start->Flag, NODE_CORNER);
      }
      /* only reversal nodes, select node with the smallest angle */
      else
      {
        start = revsmall;
        DEL_BIT(start->Flag, ALL_NODE_TYPE);
        ADD_BIT(start->Flag, NODE_REVERSAL);
      }
    }
  }
  /* there is no reversal node */
  else
  {
    /* no end nodes whatsoever */
    if(nrealend == 0 && ncandend == 0)
    {
      /* there is a corner node at least */
      if(ncornode != 0)
      {
        start = corsmall;
        DEL_BIT(start->Flag, ALL_NODE_TYPE);
        ADD_BIT(start->Flag, NODE_CORNER);
      }
      /* only side nodes */
      else
      {
        start = loop->Nodes;
        DEL_BIT(start->Flag, ALL_NODE_TYPE);
        ADD_BIT(start->Flag, NODE_SIDE);
      }
    }
    /* external loop with possible primitive shape */
    else if(CHK_BIT(loop->Type, LOOP_EXTERNAL))
    {
      if(!overload)
      {
        bestshape  = -1;
        bestnirreg = 100;
        bestangdev = ANGLE_90;
        if(   ShapeCombinations(mesh, loop, SHAPE_RECTANGLE, nrealend, realend,
                                ncandend, candend, 
                                &bestshape, shape, &bestnirreg, &bestangdev)
           || ShapeCombinations(mesh, loop, SHAPE_TRIANGLE, nrealend, realend,
                                ncandend, candend, 
                                &bestshape, shape, &bestnirreg, &bestangdev)
           || ShapeCombinations(mesh, loop, SHAPE_SEMICIRCLE, nrealend, realend,
                                ncandend, candend, 
                                &bestshape, shape, &bestnirreg, &bestangdev) )
        {
#ifdef PAVING_DEBUG
          e_GeneralWarning("Shape %d is recognised !", bestshape);
#endif
          SetEndNodesForShape(bestshape, shape, ncandend, candend);
          start = shape[0];
          loop->Primitive = bestshape;
        }
      }
    }

    /* if there was an overload, or could not find a shape 
       or internal loop or etc ... the start node is still NULL */
    if(start == NULL)
    {
      if((nrealend != 0) || (ncandend != 0))
      {
        start = SelectEnd(mesh, loop);
      }
    }
    
  }

  /* finalise the node types on the loop */
  FinalNodeTypeInLoop(mesh, loop);

  return(start);
}


static int RowChoice(PAVEDMESH *mesh, LOOP *loop)
{
  NODE *start;

  /* there is no primitive */
  if(loop->Primitive == 0)
  {
#ifdef PAVING_DEBUG
    e_GeneralWarning("Classifying !");
#endif

    /* classify all nodes in the loop */
    ClassifyALoop(mesh->NodeCoords, loop);

    /* no more adjustment to the row is required */
    start = SelectRowStart(mesh, loop);
    SetLoopHead(loop, start, FALSE);
  }
  /* there is still something to generate for the primitive */
  else
  {
#ifdef PAVING_DEBUG
    e_GeneralWarning("Processing primitive !");
#endif
    /* do not forget to reclassify this node to end node */
    DEL_BIT(loop->Nodes->Flag, ALL_NODE_TYPE);
    ADD_BIT(loop->Nodes->Flag, NODE_END);
  }

  /* if there is a primitive then it should be decreased at every access */
  if(loop->Primitive)
    loop->Primitive--;  

  return P_SUCCESS;
}


/*
 * Start the row generation, create the very first element
 */
int InitRow(PAVEDMESH *mesh, LOOP *loop)
{
  int trylevel, output, first, firsttry;

  trylevel = 0;
  firsttry = -1;
  do
  {
    /* previously something went wrong, 
       try harder to find something */
    if(CHK_BIT(loop->Type, MUST_RESOLV))
      trylevel++;
  
    first = TRUE;
    do
    {
      /* normal try */
      if(trylevel < 2)
      {
        RowChoice(mesh, loop);
      }
      /* we tried everything, lets go around try every single node */
      else
      {
        /* the undo will do the StepForward as well */
        ;
      }
      
      /* create the initial element */
      output = InitElement(mesh, loop);
      
      /* no success */
      if(output & P_FACE_UNDO)
      {
        /* for the first time store the node */
        if(first)
        {
          firsttry = loop->Nodes->Node;
          first = FALSE;
        }
        /* if we went around notify loop that it must resolv next time */
        else if(firsttry == loop->Nodes->Node)
        {
          ADD_BIT(loop->Type, MUST_RESOLV);
          break;
        }
      }
    }
    while(output & P_FACE_UNDO);

  }
  while((output & P_FACE_UNDO) && (trylevel == 1));

  /* nothing works */
  if(output & P_FACE_UNDO && trylevel == 2)
  {
#ifdef PAVING_DEBUG
    MESH *m;
    
    m = ConvertToMesh(mesh);
    e_PutMeshData(m, "out");
#endif

    e_GeneralError("MESH ERROR: The loop cannot start !");
  }

  /* if first element is created, we delete the MUST_RESOLV sign */
  if(!(output & P_FACE_UNDO))
  {
    DEL_BIT(loop->Type, MUST_RESOLV);
  }

  if(   (output & P_FACE_UNDO)
     || (output & P_FACE_RESOLV) )
  {
    return(P_ROW_FINISHED);
  }
  else
  {
    return(P_ROW_CONT);
  }
}




