#include "pav_main.h"

static int CheckSeamingCond(PAVEDMESH *mesh, NODE *node);
static int SeamingOnBoundary(PAVEDMESH *mesh, int node, int other);
static int NewWedgeNode(PAVEDMESH *mesh, LOOP *loop, double x, double y, 
                        double l1, double l2, int undo);

/* ######################################################################### */

/* 
 * Checks whether there should be seaming at the specified node.
 * Returns: P_YES -> if the node should be seamed
 *          P_NO  -> in all other cases
 */
static int CheckSeamingCond(PAVEDMESH *mesh, NODE *node)
{
  int     next, curr, prev;
  double  x1, x2, x3, y1, y2, y3, aa;
  int     nelem;

  next = node->Next->Node;
  curr = node->Node;
  prev = node->Prev->Node;

  x1 = mesh->NodeCoords[prev][0];
  y1 = mesh->NodeCoords[prev][1];
  x2 = mesh->NodeCoords[curr][0];
  y2 = mesh->NodeCoords[curr][1];
  x3 = mesh->NodeCoords[next][0];
  y3 = mesh->NodeCoords[next][1];

  aa = AngleCalc(x1, y1, x2, y2, x3, y3);

  nelem = mesh->NNodeElem[curr];

  /* in the case of a temporary element inversion it might to seam
     the element itself, well, do not allow it */
  if(nelem <= 1)
  {
    return(P_NO);
  }

  /* negative angle */
  if((nelem > 2) && (aa > 3.0/2.0*M_PI))
  {
    return(P_YES);
  }
  
  /* small angle, but with how many elements */
  if(nelem >= 4)
  {
    if(aa < EPS1_ANGLE)
    {
      return(P_YES);
    }
  }
  else
  {
    if(aa < EPS2_ANGLE)
    {
      return(P_YES);
    }
  }

  return(P_NO);
}


/*
 * If we would seam with a mesh boundary, then this function checks
 * whether the quad elements are good quality after seaming. If
 * everything is fine, then seaming is allowed.
 */
static int SeamingOnBoundary(PAVEDMESH *mesh, int node, int other)
{
  int i, j, quad;
  int qnode[4];

  /* check quad quality, we may allow it if they are good */
  for(i = 0; i < mesh->NNodeElem[node]; i++)
  {
    quad = mesh->NodeElem[node][i];
    for(j = 0; j < 4; j++)
    {
      if(mesh->QuadNodes[quad][j] == node)
        qnode[j] = other;
      else
        qnode[j] = mesh->QuadNodes[quad][j];
    }
    if(QuadShapeQuality(mesh->NodeCoords, qnode[0], qnode[1], 
                                          qnode[2], qnode[3]  ) < 0.5)
    {
      break;
    }
  }

  if(i == mesh->NNodeElem[node])
  {
    return(P_YES);
  }
  else
  {
    return(P_NO);
  }

}

/*
 * Creates a new wedge node.
 */
static int NewWedgeNode(PAVEDMESH *mesh, LOOP *loop, double x, double y, 
                        double l1, double l2, int undo)
{
  int    n;
  NODE  *newnode;

  CreateNode(x, y, NODE_FRONT, mesh, undo);
  n             = mesh->NNodes-1;
  newnode       = GetNewNode(mesh);
  newnode->Node = n;
  InsertBehindFrontNode(mesh, newnode, loop, undo);
  newnode->ProjNode = -1;

#ifdef GLOBAL_SIZE_CONTROL
  newnode->ReqLength = ProjLenCalc(mesh, newnode->Node);
#else
  newnode->ReqLength = (l1 + l2) / 2.0;
#endif
  
  return(n);
}


/*
 * Staring from the loop head, it does seaming until it is necessary
 * Returns: P_SEAM_DONE     -> if it performed seaming
 *          P_NO_SEAMING    -> if no seaming was necessary 
 *          P_DELAY_SEAMING -> if the seaming should be delayed until next
 *                             element creation
 */
int SeamingAtLoopHead(PAVEDMESH *mesh, LOOP *loop, int undo)
{
  int curr, new, old, quad, node;
  double x1, y1, x2, y2, x3, y3, x, y, l1, l2;
  int wn, wn1, wn2, wn3;
  double wx, wy;


  /* do seaming until it is necessary 
   * if the loop vanishes, stop
   */
  if((CheckSeamingCond(mesh, loop->Nodes) == P_YES) && loop->NNodes > 4)
  {
    new  = loop->Nodes->Prev->Node;
    old  = loop->Nodes->Next->Node;
    curr = loop->Nodes->Node;

    /* if we are on a real boundary */
    if(CHK_BND(mesh->NodeFlag[curr]))
    {
      if(   (SeamingOnBoundary(mesh, new, old) == P_NO)
         || (SeamingOnBoundary(mesh, old, new) == P_NO) )
      {
        return(P_DELAY_SEAMING);
      }
    }

#ifdef PAVING_DEBUG
    e_GeneralWarning("Seaming at node: %d", curr+1);
#endif

    /* check the difference in side length, if necessary insert wedge */
    x1 = mesh->NodeCoords[new][0];
    y1 = mesh->NodeCoords[new][1];
    x2 = mesh->NodeCoords[curr][0];
    y2 = mesh->NodeCoords[curr][1];
    x3 = mesh->NodeCoords[old][0];
    y3 = mesh->NodeCoords[old][1];
    
    x = x1 - x2;
    y = y1 - y2;
    l1 = sqrt((x * x) + (y * y));
    x = x2 - x3;
    y = y2 - y3;
    l2 = sqrt((x * x) + (y * y));
 
    if(   ((l1 / l2) > WEDGE_SIDE_RATIO)
       || ((l2 / l1) > WEDGE_SIDE_RATIO) )
    {
#ifdef PAVING_DEBUG
      e_GeneralWarning("Wedge inserted");
#endif
      
      /* decide into which side to insert the wedge and
         check that the side is not a fixed boundary */
      if((l1 > l2) && !CHK_BND(mesh->NodeFlag[new]))
      {
        wn = new;

        x  = x1;
        y  = y1;

        /* insert two nodes */
        wx  = (2*x+x2) / 3.0;
        wy  = (2*y+y2) / 3.0;
        wn1 = NewWedgeNode(mesh, loop, wx, wy, l1, l2, undo);

        wx  = (x+2*x2) / 3.0;
        wy  = (y+2*y2) / 3.0;
        wn2 = NewWedgeNode(mesh, loop, wx, wy, l1, l2, undo);
      
        /* find the quad and */
        quad = FindQuadOnSide(mesh, curr, wn);

        /* change the node number */
        ChangeElementNode(quad, curr, wn1, mesh, undo);

        /* the fourth node of the new quad wedge and */
        wn3 = FindNodeInQuad(mesh->QuadNodes[quad], wn1, 3);
        if(wn3 != new)
        {
          /* create the wedge element */
          CreateElement(curr, wn2, wn1, wn3, mesh, undo);
        }
        else
        {
          wn3 = FindNodeInQuad(mesh->QuadNodes[quad], curr, 1);
          if(wn3 == new)
          {
            e_GeneralError("SeamingAtLoopHead: inernal error 1 !");
          }
          /* create the wedge element */
          CreateElement(curr, wn3, wn1, wn2, mesh, undo);
        }
      
        new = wn2;

      }
      else if(!CHK_BND(mesh->NodeFlag[old]))
      {
        wn = old;

        x  = x3;
        y  = y3;
        
        StepForward(loop, undo);
        
        /* insert two nodes */
        wx = (x+2*x2) / 3.0;
        wy = (y+2*y2) / 3.0;
        wn1 = NewWedgeNode(mesh, loop, wx, wy, l1, l2, undo);

        wx = (2*x+x2) / 3.0;
        wy = (2*y+y2) / 3.0;
        wn2 = NewWedgeNode(mesh, loop, wx, wy, l1, l2, undo);
        
        /* find the quad and */
        quad = FindQuadOnSide(mesh, curr, wn);
        
        /* change the node number */
        ChangeElementNode(quad, curr, wn2, mesh, undo);
        
        /* the fourth node of the new quad wedge and */
        wn3 = FindNodeInQuad(mesh->QuadNodes[quad], wn2, 1);
        if(wn3 != old)
        {
          /* create the wedge element */
          CreateElement(curr, wn3, wn2, wn1, mesh, undo);
        }
        else
        {
          wn3 = FindNodeInQuad(mesh->QuadNodes[quad], wn2, 3);
          if(wn3 == old)
          {
            e_GeneralError("SeamingAtLoopHead: inernal error 2 !");
          }
          /* create the wedge element */
          CreateElement(curr, wn1, wn2, wn3, mesh, undo);
        }
        
        StepBackward(loop, undo);
        StepBackward(loop, undo);
        StepBackward(loop, undo);

        old = wn1;
      }
    }
    
    /* be sure that we do not touch mesh boundary nodes */
    if(!CHK_BND(mesh->NodeFlag[old]))
    {
      /* change the node numbers for all quads */
      while(mesh->NNodeElem[old] > 0)
      {
        quad = mesh->NodeElem[old][0];
        ChangeElementNode(quad, old, new, mesh, undo);
      }
  
      /* remove the nodes from the front */
      StepForward(loop, undo);
      RemoveBehindFrontNode(mesh, loop, undo);
      node = loop->Nodes->Node;
      StepForward(loop, undo);
      RemoveBehindFrontNode(mesh, loop, undo);
      /* delete this node, since it is not used any more */
      if(!CHK_BND(mesh->NodeFlag[node]))
      {
        DeleteNode(mesh, node, undo);
      }
    }
    else if(    CHK_BND(mesh->NodeFlag[old])
            && !CHK_BND(mesh->NodeFlag[new]) )
    {
      /* change the node numbers for all quads */
      while(mesh->NNodeElem[new] > 0)
      {
        quad = mesh->NodeElem[new][0];
        ChangeElementNode(quad, new, old, mesh, undo);
      }
  
      /* remove the nodes from the front */
      StepForward(loop, undo);
      RemoveBehindFrontNode(mesh, loop, undo);
      node = loop->Nodes->Node;
      RemoveBehindFrontNode(mesh, loop, undo);
      /* delete this node, since it is not used any more */
      if(!CHK_BND(mesh->NodeFlag[node]))
      {
        DeleteNode(mesh, node, undo);
      }
    }
    /* tried to merge two mesh boundary nodes */
    else
    {
      e_GeneralError("SeamingAtLoopHead: inernal error 3 !");
    }

    /* smooth after seaming */
    MeshSmoothing(mesh, loop, N_SMOOTH_ITERATION, undo);

    return(P_SEAM_DONE);
  }
  else
  {
    return(P_NO_SEAMING);
  }

}



