#include "pav_main.h"
#include "pav_operation.h"
#include "eg_lib.h"

static int IsoparametricSmooth(PAVEDMESH *mesh, int node, double *dax, double *day);
static int LengthSmooth(PAVEDMESH *mesh, int inode, int jnode, double ld, 
                        double dax, double day, double *dbx, double *dby);
static int AngleSmooth(PAVEDMESH *mesh, int nprev, int inode, int nnext, int jnode,
                       double ld, double *dcx, double *dcy);
/*static void CheckProjNode(PAVEDMESH *mesh, NODE *CurrNode); */
static int SmoothFront(PAVEDMESH *mesh, LOOP *loop, int nsmoothiter, int undo);

static void CollectInternalNodes(PAVEDMESH *mesh, int jnode,
                                 int *nlist, int **list);
static void SmoothAroundNode(PAVEDMESH *mesh, int inode, 
                             int *nlist, int **list, int undo);

/* ############################################################################ */

/* 
 * isoparametric smoothing of nodes on the active boundary
 */
static int IsoparametricSmooth(PAVEDMESH *mesh, int node, double *dax, double *day)
{
  int i, k, quad, diag, n;

  (*dax) = 0.0;
  (*day) = 0.0;
  for(i = 0; i < mesh->NNodeElem[node]; i++)
  {
    quad = mesh->NodeElem[node][i];
    diag = FindNodeInQuad(mesh->QuadNodes[quad], node, 2);

    for(k = 0; k < 4; k++)
    {
      n = mesh->QuadNodes[quad][k];
      /* node itself */
      if(n == node)
        continue;
      /* diagonally opposite node */
      if(n == diag)
      {
        (*dax) -= mesh->NodeCoords[n][0];
        (*day) -= mesh->NodeCoords[n][1];
      }
      else
      {
        (*dax) += mesh->NodeCoords[n][0];
        (*day) += mesh->NodeCoords[n][1];
      }
    }
  }

  (*dax) /= mesh->NNodeElem[node];
  (*day) /= mesh->NNodeElem[node];

  /* da = Vi' - Vi */
  (*dax) -= mesh->NodeCoords[node][0];
  (*day) -= mesh->NodeCoords[node][1];

  return(P_SUCCESS);
}


/* 
 * desired length smoothing of the nodes on the active boundary 
 */
static int LengthSmooth(PAVEDMESH *mesh, int inode, int jnode, double ld, 
                        double dax, double day, double *dbx, double *dby)
{
  double la;
  double x1, y1;

  /* the length of the vector Vij from Nj to the new location of Ni
     determined from isoparametric smooth */
  x1 = mesh->NodeCoords[inode][0] + dax - mesh->NodeCoords[jnode][0];
  y1 = mesh->NodeCoords[inode][1] + day - mesh->NodeCoords[jnode][1];
  la = sqrt(x1*x1+y1*y1);

  /* modified position */
  (*dbx) = mesh->NodeCoords[jnode][0] - mesh->NodeCoords[inode][0] + 
           (dax + mesh->NodeCoords[inode][0] - mesh->NodeCoords[jnode][0]) * ld / la;
  (*dby) = mesh->NodeCoords[jnode][1] - mesh->NodeCoords[inode][1] + 
           (day + mesh->NodeCoords[inode][1] - mesh->NodeCoords[jnode][1]) * ld / la;
  
  return(P_SUCCESS);
}

/* 
 * desired angular smoothing
 */
static int AngleSmooth(PAVEDMESH *mesh, int nprev, int inode, int nnext, int jnode,
                       double ld, double *dcx, double *dcy)
{
  double lq, lpb2;
  double x1, y1, x2, y2;
  int    status;
  double angle, pb1_angle, pb2_angle;
  double pb1x, pb1y, pb2[2], pe2[2], sign;
  double xi, yi, mu1, mu2;

  /* the angle between (Nj , Ni-1) and (Nj , Ni+1) */
  angle = AngleCalc(mesh->NodeCoords[nprev][0], mesh->NodeCoords[nprev][1], 
                    mesh->NodeCoords[jnode][0], mesh->NodeCoords[jnode][1], 
                    mesh->NodeCoords[nnext][0], mesh->NodeCoords[nnext][1]  );
  
  /* half the angle */
  pb1_angle = AngleDivider(mesh->NodeCoords[nprev][0], mesh->NodeCoords[nprev][1], 
                           mesh->NodeCoords[jnode][0], mesh->NodeCoords[jnode][1], 
                           0.5, angle);
  
  /* the length does not matter, we need only the direction */
  pb1x = mesh->NodeCoords[jnode][0] + cos(pb1_angle);
  pb1y = mesh->NodeCoords[jnode][1] + sin(pb1_angle);

  /* half the angle between Pb1 and (Nj , Ni) */
  x1 = mesh->NodeCoords[jnode][0]-pb1x;
  y1 = mesh->NodeCoords[jnode][1]-pb1y;
  x2 = mesh->NodeCoords[inode][0]-mesh->NodeCoords[jnode][0];
  y2 = mesh->NodeCoords[inode][1]-mesh->NodeCoords[jnode][1];

  sign = x1 * y2 - y1 * x2;

  /* we can calculate the division of the internal angle, so we have
     to make sure on which side of (Ni , Nj) the vector Pb1 will fall */
  /* it is zero, so the angle is the same */
  if(E_ABS(sign) < ZERO)
  {
    pb2_angle = pb1_angle;
  }
  else if(sign > 0)
  {
    pb2_angle = AngleCalc(pb1x, pb1y, 
                          mesh->NodeCoords[jnode][0], mesh->NodeCoords[jnode][1], 
                          mesh->NodeCoords[inode][0], mesh->NodeCoords[inode][1] );
    pb2_angle = AngleDivider(pb1x, pb1y, 
                             mesh->NodeCoords[jnode][0], mesh->NodeCoords[jnode][1], 
                             0.5, pb2_angle);
  }
  else
  {
    pb2_angle = AngleCalc(mesh->NodeCoords[inode][0], mesh->NodeCoords[inode][1],
                          mesh->NodeCoords[jnode][0], mesh->NodeCoords[jnode][1], 
                          pb1x, pb1y );
    pb2_angle = AngleDivider(mesh->NodeCoords[inode][0], mesh->NodeCoords[inode][1],
                             mesh->NodeCoords[jnode][0], mesh->NodeCoords[jnode][1], 
                             0.5, pb2_angle);
  }

  /* temporarily the end of the vector Pb2 is projected out 100 times
     of the generated length in both directions */
  pb2[0] = mesh->NodeCoords[jnode][0] + cos(pb2_angle) * 100.0 * ld;
  pb2[1] = mesh->NodeCoords[jnode][1] + sin(pb2_angle) * 100.0 * ld;

  pe2[0] = mesh->NodeCoords[jnode][0] - cos(pb2_angle) * 100.0 * ld;
  pe2[1] = mesh->NodeCoords[jnode][1] - sin(pb2_angle) * 100.0 * ld;
 
  /* intersection of Pb2 and (Ni-1 , Ni+1) */
  status = eg_LineLineIntersect2D(pb2, pe2, 
                                  mesh->NodeCoords[nprev], 
                                  mesh->NodeCoords[nnext],
                                  &mu1, &mu2);

  if(status == EG_INVALID)
  {
    e_GeneralWarning("AngleSmooth: geometry error !");
    (*dcx) = 0.0;
    (*dcy) = 0.0;

    return(P_ERROR);
  }

  if(   (status == EG_INTERSECT)
     && (mu1 >= 0.0 && mu1 <= 1.0)
     && (mu2 >= 0.0 && mu2 <= 1.0) )
  {
    /* intersection point */
    xi = pb2[0] + mu1 * (pe2[0] - pb2[0]);
    yi = pb2[1] + mu1 * (pe2[1] - pb2[1]);

    /* distance from intersection to Nj */
    x1 = xi - mesh->NodeCoords[jnode][0];
    y1 = yi - mesh->NodeCoords[jnode][1];
    lq = sqrt(x1*x1+y1*y1);
 
    /* final decision about the length of Pb2 */
    if(ld > lq)
      lpb2 = (ld + lq) * 0.5;
    else
      lpb2 = ld;
 
    /* final Pb2 vector */
    pb2[0] = mesh->NodeCoords[jnode][0] + cos(pb2_angle) * lpb2;
    pb2[1] = mesh->NodeCoords[jnode][1] + sin(pb2_angle) * lpb2;
 
    /* modification by this smoothing */
    (*dcx) = pb2[0] - mesh->NodeCoords[inode][0];
    (*dcy) = pb2[1] - mesh->NodeCoords[inode][1];
  }
  else
  {
    e_GeneralWarning("AngleSmooth: Angular smoother vector does not intersect new boundary !");
    (*dcx) = 0.0;
    (*dcy) = 0.0;

    return(P_ERROR);
  }
  
  return(P_SUCCESS);
}

/*
 * Determine the projecting node, it may change with time.
 */
static void CheckProjNode(PAVEDMESH *mesh, NODE *CurrNode)
{
  int     quad, c1, c2, lq, rq, refnode, inode;
  int     prev, next;
  
  inode = CurrNode->Node;
  prev  = CurrNode->Prev->Node;
  next  = CurrNode->Next->Node;

  quad  = mesh->NodeElem[inode][0];
  c1 = FindNodeInQuad(mesh->QuadNodes[quad], inode, 1);
  quad  = mesh->NodeElem[inode][1];
  c2 = FindNodeInQuad(mesh->QuadNodes[quad], inode, 1);

  if(   ((c1 == prev) || (c1 == next))
     && ((c2 != prev) || (c2 != next)) )
  {
    lq = mesh->NodeElem[inode][0];
    rq = mesh->NodeElem[inode][1];
    refnode = c2;
  }
  else if(   ((c1 != prev) || (c1 != next))
          && ((c2 == prev) || (c2 == next)) )
  {
    lq = mesh->NodeElem[inode][1];
    rq = mesh->NodeElem[inode][0];
    refnode = c1;
  }
  /* situation is unknown/invalid */
  else
  {
    refnode = -1;
  }

  /* set new reference node */
  if(refnode != CurrNode->ProjNode)
  {
    CurrNode->ProjNode = refnode;
  }
}


/* 
 * smooth the paving front
 */
static int SmoothFront(PAVEDMESH *mesh, LOOP *loop, int nsmoothiter, int undo)
{
  int     iter;
  double  dax, day, dbx, dby, dcx, dcy, dix, diy;
  double  ld;
  NODE   *CurrNode, *StartNode;
  int     inode, jnode, nprev, nnext;

#ifdef PAVING_DEBUG
  e_GeneralWarning("!!!!! Beginning of front smoothing !!!!!");
#endif

  /* number of row smoothing iterations */
  StartNode = loop->Nodes;
  for(iter = 0; iter < nsmoothiter; iter++)
  {
    CurrNode = StartNode;
    do
    {
      /* smooth all non-boundary nodes */
      if(!CHK_BND(mesh->NodeFlag[CurrNode->Node]))
      {
        inode = CurrNode->Node;

        /* isoparametric smooth */
        IsoparametricSmooth(mesh, inode, &dax, &day);

        /* in first iteration check reference node */
        if((iter == 0) && (mesh->NNodeElem[inode] == 2))
        {
          CheckProjNode(mesh, CurrNode);
        }

        /* only for nodes with two quads attached and 
           proper data of projected length */      
        if((mesh->NNodeElem[inode] == 2) && (CurrNode->ProjNode != -1))
        {
          nprev = CurrNode->Prev->Node;
          nnext = CurrNode->Next->Node;
          
          jnode = CurrNode->ProjNode;
          /* the length of the projection vector 
             used for the generation of the current node */
          ld = CurrNode->ReqLength;

          if(   LengthSmooth(mesh, inode, jnode, ld, dax, day, &dbx, &dby)    == P_ERROR
             || AngleSmooth(mesh, nprev, inode, nnext, jnode, ld, &dcx, &dcy) == P_ERROR )
          {
            /* if any error then simply isoparametric */
            dix = dax;
            diy = day;
          }
          else
          {
            /* final modification */
            dix = (dbx + dcx) * 0.5;
            diy = (dby + dcy) * 0.5;
          }
        }
        else
        /* for nodes to which more than two quads are attached, 
           only isparametric smoothing */
        {
          dix = dax;
          diy = day;
        }

        /* check whether any change should be done ? */
        if(E_ABS(dix) > ZERO || E_ABS(diy) > ZERO)
        {
          ChangeNodeCoord(inode, dix, diy, mesh, undo);
        }
      }
      
      CurrNode = CurrNode->Next;
    }
    while(CurrNode != StartNode);

  } /* no. of iterations */

#ifdef PAVING_DEBUG
  e_GeneralWarning("!!!!! End of front smoothing !!!!!");
#endif

  return(P_SUCCESS);
}


/*
 * Collect internal nodes to smooth
 */
static void CollectInternalNodes(PAVEDMESH *mesh, int jnode,
                                 int *nlist, int **list)
{
  int k;

  for(k = 0; k < (*nlist); k++)
  {
    if((*list)[k] == jnode)
      break;
  }
  if(   (k == (*nlist))
     && !CHK_FRONT(mesh->NodeFlag[jnode])
     && !CHK_BND(mesh->NodeFlag[jnode]) )
  {
    if(mesh->NNodeElem[jnode] == 0)
    {
      e_GeneralError("CollectInternalNodes: no quad attached !");
    }
  
    (*list) = e_incr_i_vect(*nlist, *list, ALLOC_SMOOTH_SIZE);
    (*list)[*nlist] = jnode;
    (*nlist)++;
  }
}

/*
 * Smooth around an internal node
 */
static void SmoothAroundNode(PAVEDMESH *mesh, int inode, 
                             int *nlist, int **list, int undo)
{
  int j, quad, jnode;
  double dx, dy, dd, l , cx, cy;

  /* only floating points are smoothed */
  if(   CHK_FRONT(mesh->NodeFlag[inode])
     || CHK_BND(mesh->NodeFlag[inode]) )
    return;

  dx = 0.0;
  dy = 0.0;
  dd = 0.0;
  for(j = 0; j < mesh->NNodeElem[inode]; j++)
  {
    quad  = mesh->NodeElem[inode][j];
    jnode = FindNodeInQuad(mesh->QuadNodes[quad], inode, 1);
    
    CollectInternalNodes(mesh, jnode, nlist, list);
    
    cx = mesh->NodeCoords[jnode][0] - mesh->NodeCoords[inode][0];
    cy = mesh->NodeCoords[jnode][1] - mesh->NodeCoords[inode][1];
    /* length of C vector */
    l = sqrt((cx*cx)+(cy*cy));
    dx += cx * l;
    dy += cy * l;
    dd += l;
  }

  dx = dx / dd;
  dy = dy / dd;

  if(E_ABS(dx) > ZERO || E_ABS(dy) > ZERO)
  {
    /* adjust the node's position */
    ChangeNodeCoord(inode, dx, dy, mesh, undo);
  }
}


/* 
 * smooth the interior nodes modified length-weighted Laplacian smoother 
 */
int SmoothInterior(PAVEDMESH *mesh, LOOP *loop, int nsmoothiter, int undo)
{
  NODE   *start, *curr;
  int     level, iter, j, inode, quad, jnode;
  int    *list[N_SMOOTH_LEVEL+1];
  int     nlist[N_SMOOTH_LEVEL+1];

  nlist[0] = 0;
  list[0]  = NULL;

#ifdef PAVING_DEBUG
  e_GeneralWarning("!!!!! Beginning of interior smoothing !!!!!");
#endif

  /* smooth the loop */
  if(loop != NULL)
  {
    /* collect non-fixed nodes */
    curr = start = loop->Nodes;
    do
    {
      inode = curr->Node;
      
      for(j = 0; j < mesh->NNodeElem[inode]; j++)
      {
        quad  = mesh->NodeElem[inode][j];
        jnode = FindNodeInQuad(mesh->QuadNodes[quad], inode, 1);

        CollectInternalNodes(mesh, jnode, &(nlist[0]), &(list[0]));
      }
 
      curr = curr->Next;
    }
    while(curr != start);
  }
  /* smooth the whole mesh */
  else
  {
    list[0]  = e_alloc_i_vect(mesh->NNodes);
    for(j = 0; j < mesh->NNodes; j++)
    {
      if(mesh->NNodeElem[j] != 0 && !CHK_BND(mesh->NodeFlag[j]))
      {
        list[0][nlist[0]] = j;
        nlist[0]++;
      }
    }
  }


  for(level = 0; level < N_SMOOTH_LEVEL; level++)
  {
#ifdef PAVING_DEBUG
    e_GeneralWarning("No of nodes to smooth: %d", nlist[level]);
#endif
    nlist[level+1] = 0;
    list[level+1]  = NULL;

    for(iter = 0; iter < nsmoothiter; iter++)
    {
      /* forward smoothing */
      for(j = 0; j < nlist[level]; j++)
      {
        inode = list[level][j];
        SmoothAroundNode(mesh, inode, &(nlist[level+1]), &(list[level+1]), undo);
      }
      
      /* backward smoothing */
      for(j = nlist[level]-1; j >= 0; j--)
      {
        inode = list[level][j];
        SmoothAroundNode(mesh, inode, &(nlist[level+1]), &(list[level+1]), undo);
      }

    } /* no of smoothing iteration */
  } /* no of level */

  /* free-up memory */
  for(j = 0; j < N_SMOOTH_LEVEL+1; j++)
  {
    e_free(list[j]);
  }

#ifdef PAVING_DEBUG
  e_GeneralWarning("!!!!! End of interior smoothing !!!!!");
#endif

  return(P_SUCCESS);
}


int MeshSmoothing(PAVEDMESH *mesh, LOOP *loop, int nsmoothiter, int undo)
{

  SmoothFront(mesh, loop, nsmoothiter, undo);
  SmoothInterior(mesh, loop, nsmoothiter, undo);

  return(P_SUCCESS);
}




