#include "mgni.h"
#include "rgr.h"
#include "eg_lib.h"

/*****************************************************************************/

static double **crd;
static int offset;

static int new_al_edge_nodes(EDGE *edg, double ***coords, int *np, int *tp,
                             double **mpr, int ***tpnl, double gprm,
                             double umpr, int geomset, MESH *mesh, int ngn,
                             int *sortx, int *sorty, int *rankx, int *ranky);
static int handle_curve_refine(EDGE *edg, int *b, NURBSCURVE * curve,
                               double tfn, double tsn, double ***coords,
                               int *np, int *tp, double **mpr, int ***tpnl,
                               double gprm, double umpr);
static int handle_segment_refine(EDGE *edg, int *b, double ***coords, int *np,
                                 int *tp, double **mpr, int ***tpnl,
                                 double gprm, double umpr);
static int compare_sortx(const void *i, const void *j);
static int compare_sorty(const void *i, const void *j);
static void register_node(double ***coords, int fn, int sn, double current_t,
                          int *np, int *tp, double **mpr, int ***tpnl,
                          double umpr, double t_diff, int *segment_done,
                          double zval);
static void list_bubble(double **coords, int fn, int sn, double t1, double t2,
                        double radius, int ngn, int *sortx, int *sorty,
                        int *rankx, int *ranky, int *bubble, int *nbn);
static int obtain_interval(double current_t, double t_diff, double *t1,
                           double *t2);
static int check_bubble(double **coords, int fn, int sn, double t1, double t2,
                        int *bubble, int nbn, double umpr, double gprm,
                        double *current_t);
static int proj_bubble_check(double **coords, int fn, int sn, double t1,
                             double t2, int *bubble, int nbn, double umpr,
                             double gprm, double *current_t);
static void get_pos(double **coords, int fn, int sn, double t, double *pos);
static double perpdist(double *pos1, double *pos2, double *pos);
static void solve_quadratic(double a, double b, double c, double *s1,
                            double *s2);

int new_algn_bnd_nodes(double ***coords, int *np, int *tp, int *stnp,
                       double **mpr, BND *bnd, int ***tpnl, double gprm,
                       double umpr, int geomset, MESH *mesh, int ngn)
{
  /* make carefully placed boundary refinement nodes adjusted to
     a regular grid already present as the last ngnn nodes in coords
   */

  int nnp;                      /* return value = number of new nodes */
  int *sortx, *sorty, *rankx, *ranky; /* sort and rank vectors  for searching */
  int i, e;

  (*stnp) = (*np);
  nnp = 0;

  /* global search init */
  sortx = e_alloc_i_vect(ngn);
  sorty = e_alloc_i_vect(ngn);
  rankx = e_alloc_i_vect(ngn);
  ranky = e_alloc_i_vect(ngn);
  for (i = 0; i < ngn; i++)
  {
    sortx[i] = i;
    sorty[i] = i;
    rankx[i] = i;
    ranky[i] = i;
  }
  crd = *coords;                /* make coords accessible for sorting */
  offset = (*np - ngn);         /* make offset accessible for sorting */
  qsort(sortx, ngn, sizeof(int), compare_sortx);
  qsort(sorty, ngn, sizeof(int), compare_sorty);
  for (i = 0; i < ngn; i++)
  {
    rankx[sortx[i]] = i;
    ranky[sorty[i]] = i;
  }

  nnp += new_al_edge_nodes(&(bnd->oe), coords, np, tp, mpr, tpnl, gprm, umpr,
                           geomset, mesh, ngn, sortx, sorty, rankx, ranky);
  for (e = 0; e < bnd->nie; e++)
    nnp +=
      new_al_edge_nodes(&(bnd->ie[e]), coords, np, tp, mpr, tpnl, gprm, umpr,
                        geomset, mesh, ngn, sortx, sorty, rankx, ranky);

  e_free(sortx);
  e_free(sorty);
  e_free(rankx);
  e_free(ranky);
  return nnp;
}

static int new_al_edge_nodes(EDGE *edg, double ***coords, int *np, int *tp,
                             double **mpr, int ***tpnl, double gprm,
                             double umpr, int geomset, MESH *mesh, int ngn,
                             int *sortx, int *sorty, int *rankx, int *ranky)
{
  int b;                        /* index on the edge */
  int fn, sn;                   /* bnd node numbers (first, second) */
  NURBSCURVE *curve;            /* nurbs curve for curve refine */
  double tfn;                   /* curve param for fn */
  double tsn;                   /* curve param for sn */
  int nsp;                      /* number of segment points */
  int nnp;                      /* the return value = number of new edge nodes */
  int s;                        /* counter */

  int *bubble;                  /* the node indices of the grid nodes that belong to the bubble */
  int nbn;                      /* number of nodes in the bubble */

  double length;                /* length of a segment */
  int segment_done;
  double current_t;
  double t_diff;
  double t1, t2;

  nnp = 0;
  for (b = 0; b < edg->nen; b++)
  {
    fn = edg->en[b];
    sn = edg->en[(b + 1) % edg->nen];
    if (geomset)
      if (e_GetCurve(mesh, fn + 1, sn + 1, &curve, &tfn, &tsn))
      {
        nnp +=
          handle_curve_refine(edg, &b, curve, tfn, tsn, coords, np, tp, mpr,
                              tpnl, gprm, umpr);
        continue;
      }
    /* OK we know we can handle the segment as straight */

    /* first preliminary checks */
    length = eg_PointDist2D((*coords)[fn], (*coords)[sn]);
    if (length / umpr < 2.0)
    {
      nnp +=
        handle_segment_refine(edg, &b, coords, np, tp, mpr, tpnl, gprm, umpr);
      continue;
    }

    /* then start making the new nodes */
    segment_done = FALSE;
    current_t = 0.0;
    t_diff = umpr / length;
    bubble = e_alloc_i_vect(50);
    nbn = 0;
    nsp = 0;
    while (!segment_done)
    {
      if (!obtain_interval(current_t, t_diff, &t1, &t2))
      {
        register_node(coords, fn, sn, t1, np, tp, mpr, tpnl, umpr, t_diff,
                      &segment_done, -5.0);
        nsp++;
        continue;
      }
      list_bubble((*coords), fn, sn, t1, t2, umpr * gprm * 2, ngn, sortx,
                  sorty, rankx, ranky, bubble, &nbn);
      if (nbn == 0)
      {
        list_bubble((*coords), fn, sn, t1, t2, umpr * (1.0 + gprm), ngn,
                    sortx, sorty, rankx, ranky, bubble, &nbn);
        if (!proj_bubble_check
            ((*coords), fn, sn, t1, t2, bubble, nbn, umpr, gprm, &current_t))
          register_node(coords, fn, sn, current_t, np, tp, mpr, tpnl, umpr,
                        t_diff, &segment_done, -20.0);
        else
          register_node(coords, fn, sn, current_t, np, tp, mpr, tpnl, umpr,
                        t_diff, &segment_done, 0.0);
      }
      else
      {
        if (!check_bubble
            ((*coords), fn, sn, t1, t2, bubble, nbn, umpr, gprm, &current_t))
          register_node(coords, fn, sn, current_t, np, tp, mpr, tpnl, umpr,
                        t_diff, &segment_done, -5.0);
        else
          register_node(coords, fn, sn, current_t, np, tp, mpr, tpnl, umpr,
                        t_diff, &segment_done, 0.0);
      }
      nsp++;
    }
    e_free(bubble);
    nnp += nsp;
    for (s = 0; s < nsp; s++)
      edg->en = e_incr_i_vect(edg->nen + s, edg->en, EDGE_INCR);
    for (s = edg->nen-1; s > b; s--)
      edg->en[nsp + s] = edg->en[s];
    for (s = 0; s < nsp; s++)
      edg->en[b + 1 + s] = (*np) - nsp + s;
    edg->nen += nsp;
    b += nsp;
  }
  return nnp;
}

static int proj_bubble_check(double **coords, int fn, int sn, double t1,
                             double t2, int *bubble, int nbn, double umpr,
                             double gprm, double *current_t)
{
  /* so the first bubble with R=2*0.7 was empty
     we got a new bubble with R=1+0.7
     now project every bubblenode onto the interval-line
     if the projection falls inside the interval than take
     that projective node.
     if all projections fall outside then treat it as a failed node. */

  int b;
  int bnode;
  double diff1[2], diff2[2];
  double t;
  int s;
  double t_diff = 1.0;

  if (nbn == 0)
  {
    *current_t = t2;
    return FALSE;
  }

  for (b = 0; b < nbn; b++)
  {
    bnode = bubble[b];
    for (s = 0; s < 2; s++)
    {
      diff1[s] = coords[sn][s] - coords[fn][s];
      diff2[s] = coords[bnode][s] - coords[fn][s];
    }
    t = e_VectDotProd(2, diff1, diff2) / e_VectDotProd(2, diff1, diff1);
    if ((t >= t1) && (t <= t2))
    {
      *current_t = t;
      return TRUE;
    }
    else
    {
      if (t < t1)
      {
        if (t1 - t < t_diff)
        {
          *current_t = t1;
          t_diff = t1 - t;
        }
      }
      else if (t > t2)
      {
        if (t - t2 < t_diff)
        {
          *current_t = t2;
          t_diff = t - t2;
        }
      }
    }
  }
  return FALSE;
}

static int check_bubble(double **coords, int fn, int sn, double t1, double t2,
                        int *bubble, int nbn, double umpr, double gprm,
                        double *current_t)
{
  int b;
  int bnode;
  double dmax, dmin;
  int failed_nodes;
  double posmin[2], posmax[2];
  double init_t2 = t2;
  double pd;
  double a, bd, c, ts1, ts2;

  get_pos(coords, fn, sn, t1, posmin);
  get_pos(coords, fn, sn, t2, posmax);

  failed_nodes = 0;

  for (b = 0; b < nbn; b++)
  {
    bnode = bubble[b];
    dmin = eg_PointDist2D(coords[bnode], posmin);
    dmax = eg_PointDist2D(coords[bnode], posmax);
    if (E_MAX(dmin, dmax) < umpr)
      continue;                 /* unconditional pass */
    pd = perpdist(posmin, posmax, coords[bnode]);
    if (pd > umpr)
    {
      failed_nodes++;
      continue;
    }
    a =
      pow((coords[sn][0] - coords[fn][0]),
          2) + pow((coords[sn][1] - coords[fn][1]), 2);
    bd =
      2.0 * (coords[sn][0] - coords[fn][0]) * (coords[fn][0] -
                                               coords[bnode][0]) +
      2.0 * (coords[sn][1] - coords[fn][1]) * (coords[fn][1] -
                                               coords[bnode][1]);
    c =
      pow((coords[fn][0] - coords[bnode][0]),
          2) + pow((coords[fn][1] - coords[bnode][1]), 2) - umpr * umpr;
    solve_quadratic(a, bd, c, &ts1, &ts2); /* ts1 must be <=ts2 */
    if (ts1 <= t1)
    {
      if (ts2 >= t2)
        continue;               /* case J, F, G */
      if ((ts2 < t2) && (ts2 >= t1))
        t2 = ts2;               /* case A, H, K */
      else if (ts2 < t1)
        failed_nodes++;         /* case C */
    }
    else if (ts1 > t1)
    {
      if (ts2 >= t2)
        if (ts1 > t2)
          failed_nodes++;       /* case D */
        else
          t1 = ts1;             /* case B, I, L */
      else
      {
        t1 = ts1;               /* case E, N */
        t2 = ts2;
      }
    }
  }
  if (failed_nodes > 1)
  {
    *current_t = init_t2;
    return FALSE;
  }
  *current_t = t2;              /* also takes care of nbn==0 */
  return TRUE;

}

static void solve_quadratic(double a, double b, double c, double *s1,
                            double *s2)
{
  double d;
  double t1, t2;

  if (E_ABS(a) < 1e-10)
  {
    if (E_ABS(b) < 1e-10)
      e_GeneralError("solve_quadratic: you supplied zero coefficients");
    *s1 = -1.0 * c / b;
    *s2 = *s1;
    return;
  }

  d = b * b - 4.0 * a * c;
  if (d < -1e-6)
    e_GeneralError("solve_quadratic: no solutions");
  if (d < 0.0)                  /* for those case when d=-1e-8) */
    d = 0.0;

  t1 = (-1.0 * b + sqrt(d)) / (2.0 * a);
  t2 = (-1.0 * b - sqrt(d)) / (2.0 * a);
  *s1 = E_MIN(t1, t2);
  *s2 = E_MAX(t1, t2);

}

static void get_pos(double **coords, int fn, int sn, double t, double *pos)
{
  int s;

  for (s = 0; s < 2; s++)
    pos[s] = coords[fn][s] + t * (coords[sn][s] - coords[fn][s]);
}

static void list_bubble(double **coords, int fn, int sn, double t1, double t2,
                        double radius, int ngn, int *sortx, int *sorty,
                        int *rankx, int *ranky, int *bubble, int *nbn)
{
  int sxmin, symin, sxmax, symax; /* begin-index end-index of search region */
  int x;                        /* index inside the search region */
  double yposmin, yposmax;
  double xposmin, xposmax;
  double pos1[2], pos2[2];
  double d1, d2, diag;
  int include;


  get_pos(coords, fn, sn, t1, pos1);
  get_pos(coords, fn, sn, t2, pos2);

  diag = sqrt(pow(eg_PointDist2D(pos1, pos2), 2) + pow(radius, 2));

  yposmin = E_MIN(pos1[1], pos2[1]);
  yposmax = E_MAX(pos1[1], pos2[1]);
  xposmin = E_MIN(pos1[0], pos2[0]);
  xposmax = E_MAX(pos1[0], pos2[0]);

  symin = 0;
  symax = 0;
  sxmin = 0;
  sxmax = 0;
  while (coords[offset + sorty[symin]][1] < yposmin - radius)
  {
    symin++;
    if (symin == (ngn))
      break;
  }
  while (coords[offset + sorty[symax]][1] < yposmax + radius)
  {
    symax++;
    if (symax == (ngn))
      break;
  }
  symax--;
  while (coords[offset + sortx[sxmin]][0] < xposmin - radius)
  {
    sxmin++;
    if (sxmin == (ngn))
      break;
  }
  while (coords[offset + sortx[sxmax]][0] < xposmax + radius)
  {
    sxmax++;
    if (sxmax == (ngn))
      break;
  }
  sxmax--;

  *nbn = 0;
  for (x = sxmin; x <= sxmax; x++)
    if ((ranky[sortx[x]] >= symin) && (ranky[sortx[x]] <= symax))
    {
      include = FALSE;
      d1 = eg_PointDist2D(pos1, coords[offset + sortx[x]]);
      d2 = eg_PointDist2D(pos2, coords[offset + sortx[x]]);
      if (d1 < radius)
        include = TRUE;
      else if (d2 < radius)
        include = TRUE;
      else if (perpdist(pos1, pos2, coords[offset + sortx[x]]) < radius)
        if (E_MAX(d1, d2) < diag)
          include = TRUE;
      if (include)
      {
        bubble[*nbn] = offset + sortx[x];
        (*nbn)++;
      }
    }
}

static double perpdist(double *pos1, double *pos2, double *pos)
{
  /* the perpendicular distance from pos to
     the line made by pos1 and pos2 */

  double diff1[2];
  double diff2[2];
  double pos3[2];
  double t;
  int s;

  for (s = 0; s < 2; s++)
  {
    diff1[s] = pos2[s] - pos1[s];
    diff2[s] = pos[s] - pos1[s];
  }
  t = e_VectDotProd(2, diff1, diff2) / e_VectDotProd(2, diff1, diff1);
  for (s = 0; s < 2; s++)
    pos3[s] = pos1[s] + t * diff1[s];

  return eg_PointDist2D(pos, pos3);
}


static int obtain_interval(double current_t, double t_diff, double *t1,
                           double *t2)
{

  *t2 = E_MIN(current_t + t_diff, 1.0 - 0.5 * t_diff);
  *t1 = current_t + 0.5 * t_diff;

  if ((*t2 - *t1) < 1e-6)
  {
    *t1 = (current_t + 1.0) / 2.0;
    return FALSE;
  }
  return TRUE;

}

static void register_node(double ***coords, int fn, int sn, double current_t,
                          int *np, int *tp, double **mpr, int ***tpnl,
                          double umpr, double t_diff, int *segment_done,
                          double zval)
{
  if ((1.0 - current_t) < t_diff)
    *segment_done = TRUE;

  if ((*tp) - (*np) < 1)
  {
    (*coords) = e_ext_cont_d_matr((*tp), 3, (*coords), 1 + (*np) - (*tp));
    (*tpnl) = e_ext_frag_i_matr((*tp), (*tpnl), 1 + (*np) - (*tp));
    (*mpr) = e_ext_d_vect((*tp), (*mpr), 1 + (*np) - (*tp));
    (*tp) += 1 + (*np) - (*tp);
  }

  get_pos(*coords, fn, sn, current_t, (*coords)[*np]);
  (*coords)[*np][2] = zval;
  (*tpnl)[(*np)] = e_alloc_i_vect(2);
  (*tpnl)[(*np)][0] = 0;
  (*tpnl)[(*np)][1] = 0;
  (*mpr)[(*np)] = umpr;
  (*np)++;
}

static int handle_curve_refine(EDGE *edg, int *b, NURBSCURVE * curve,
                               double tfn, double tsn, double ***coords,
                               int *np, int *tp, double **mpr, int ***tpnl,
                               double gprm, double umpr)
{
  int fn, sn;                   /* bnd node numbers (first, second) */
  double **seg_crd;             /* coordinates of segment nodes */
  double *seg_mpr;              /* and their mpr's */
  double *seg_t;                /* and their t-params in case of curves */
  int nsp;                      /* number of segment points */
  int s;                        /* counter */

  fn = edg->en[*b];
  sn = edg->en[(*b + 1) % edg->nen];
  nsp = e_CurveRefine(curve, tfn, umpr, tsn, umpr,
                      &seg_crd, &seg_mpr, &seg_t, TRUE, E_SEGREF_NONE);
  adjust_curve_refine_output(&nsp, seg_crd);

  if (nsp == 0)
    return 0;

  if ((*tp) - (*np) < nsp)
  {
    (*coords) = e_ext_cont_d_matr((*tp), 3, (*coords), nsp + (*np) - (*tp));
    (*tpnl) = e_ext_frag_i_matr((*tp), (*tpnl), nsp + (*np) - (*tp));
    (*mpr) = e_ext_d_vect((*tp), (*mpr), nsp + (*np) - (*tp));
    (*tp) += nsp + (*np) - (*tp);
  }

  memcpy(&((*coords)[(*np)][0]), &(seg_crd[0][0]), 3 * nsp * sizeof(double));
  for (s = 0; s < nsp; s++)
    edg->en = e_incr_i_vect(edg->nen + s, edg->en, EDGE_INCR);
  for (s = edg->nen-1; s > *b; s--)
    edg->en[nsp + s] = edg->en[s];
  for (s = 0; s < nsp; s++)
  {
    edg->en[*b + 1 + s] = (*np);
    (*tpnl)[(*np)] = e_alloc_i_vect(2);
    (*tpnl)[(*np)][0] = 0;
    (*tpnl)[(*np)][1] = 0;
    (*mpr)[(*np)] = umpr;
    (*np)++;
  }

  edg->nen += nsp;
  *b += nsp;
  e_free(seg_mpr);
  e_free_cont_d_matr(nsp + 2, seg_crd);
  e_free(seg_t);

  return nsp;
}

static int handle_segment_refine(EDGE *edg, int *b,
                                 double ***coords, int *np, int *tp,
                                 double **mpr, int ***tpnl, double gprm,
                                 double umpr)
{
  int fn, sn;                   /* bnd node numbers (first, second) */
  double **seg_crd;             /* coordinates of segment nodes */
  double *seg_mpr;              /* and their mpr's */
  int nsp;                      /* number of segment points */
  int s;                        /* counter */

  fn = edg->en[*b];
  sn = edg->en[(*b + 1) % edg->nen];
  nsp = e_SegmentRefine((*coords)[fn], umpr, (*coords)[sn], umpr,
                        &seg_crd, &seg_mpr, TRUE, E_SEGREF_NONE);

  if (nsp == 0)
    return 0;

  if ((*tp) - (*np) < nsp)
  {
    (*coords) = e_ext_cont_d_matr((*tp), 3, (*coords), nsp + (*np) - (*tp));
    (*tpnl) = e_ext_frag_i_matr((*tp), (*tpnl), nsp + (*np) - (*tp));
    (*mpr) = e_ext_d_vect((*tp), (*mpr), nsp + (*np) - (*tp));
    (*tp) += nsp + (*np) - (*tp);
  }

  memcpy(&((*coords)[(*np)][0]), &(seg_crd[0][0]), 3 * nsp * sizeof(double));
  for (s = 0; s < nsp; s++)
    edg->en = e_incr_i_vect(edg->nen + s, edg->en, EDGE_INCR);
  for (s = edg->nen-1; s > *b; s--)
    edg->en[nsp + s] = edg->en[s];
  for (s = 0; s < nsp; s++)
  {
    edg->en[*b + 1 + s] = (*np);
    (*tpnl)[(*np)] = e_alloc_i_vect(2);
    (*tpnl)[(*np)][0] = 0;
    (*tpnl)[(*np)][1] = 0;
    (*mpr)[(*np)] = umpr;
    (*np)++;
  }

  edg->nen += nsp;
  *b += nsp;
  e_free(seg_mpr);
  e_free_cont_d_matr(nsp, seg_crd);

  return nsp;
}

static int compare_sortx(const void *i, const void *j)
{
  int *pi, *pj;

  pi = (int *) i;
  pj = (int *) j;

  if (crd[*pi + offset][0] > crd[*pj + offset][0])
    return +1;
  if (crd[*pi + offset][0] < crd[*pj + offset][0])
    return -1;
  return 0;
}

static int compare_sorty(const void *i, const void *j)
{
  int *pi, *pj;

  pi = (int *) i;
  pj = (int *) j;

  if (crd[*pi + offset][1] > crd[*pj + offset][1])
    return 1;
  if (crd[*pi + offset][1] < crd[*pj + offset][1])
    return -1;
  return 0;
}

/* end of file */
